#!/usr/bin/perl

# (c) Franz Rueschendorf  1999-2006
# gh_modview.pl  : Prepare Genehunter output for displaying with gnuplot
# Syntax     : perl [path]\gh_modview.pl [-options] <infile>
# Infile     : <gh.out>  (default)
# Outfiles   : mymodv.gp and a lot of *.lod
# View result: gnuplot mymodv.gp
# 
# Modified   : 04-May-2006, read outfiles from GH-Modscore's sex specific analysis   


use Cwd;
use strict;

################################  V A R I A B L E S ###########################

my $maxlines = 0; 	# number of lines (input-file)
my $pfad = getcwd();# working directory
my $deltmp=1;		# delete temporary files
my $label=1;
my $legend=1;		# print legend (default)
my $showtitel=1;	# show title (default)
my $keybelow=0;		# 1 - set key below
my $sr=-1;			# region of modscore analysis to plot, 0 = initial model
my $ps=0;			# Print postscript file
my $step=0.3;		# stepwide, helps to remove needles from plot
my @data=();		# Genehunter outfile data
my $version="1.2";	# version

my $xmin = 0;
my $xmax = 10;
my $ymin = -2;
my $ymax = 6;

my $infile="";


####################################  M A I N  ################################

print "\n\n";
print " * * * * * * * * * * * *  GH_MODVIEW v$version  * * * * * * * * * * * *\n";
print " * * * * * * *  View Genehunter results with GNUPLOT * * * * * * *\n";


for(my $i=0;$i<=$#ARGV;$i++) {
	if ($ARGV[$i] =~ "-h" ) { &Help;}
	elsif ($ARGV[$i] =~ "-clean" ) { &Clean;}
	elsif ($ARGV[$i] =~ "-nd" ) { $deltmp=0;}
	elsif ($ARGV[$i] =~ "-kb" ) { $keybelow=1;}
	elsif ($ARGV[$i] eq "-nl" ) { $legend=0;}
	elsif ($ARGV[$i] =~ "-nk" ) { $legend=0;}
	elsif ($ARGV[$i] =~ "-nt" ) { $showtitel=0;}
	elsif ($ARGV[$i] eq "-nlabel" ) {$label=0;}
	elsif ($ARGV[$i] eq "-i" ) { $sr=0;}
	elsif ($ARGV[$i] =~ "-r" ) { $sr=$ARGV[$i]; $sr =~ s/-r//;}
	elsif ($ARGV[$i] =~ "-ps" ) { $ps=1;}
	else {$infile = $ARGV[$i];}
}


if (-f "$infile") {

	unlink glob ("*.lod");
	unlink "mymodv.gp";
	print "\nRead: $infile \n";
	&ReadFile;
	&Check;
	&ReadLods; 
	system "gnuplot mymodv.gp";
	print "\nto create a Postscript file of the plot, use option -ps\n\n";
	if ($deltmp==1) {&Clean;}	
}                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           
else {
	&Help;
	print "\n Need infile!\n";
}

###########################  R E A D  F I L E   ################################

sub ReadFile { 
	open (IN, "$infile") || die "could not open infile $infile :$!\n";
	while (<IN>){
		push @data,$_;
	}
	close (IN);
	$maxlines=$#data;
}

#################################  C H E C K  ##################################

sub Check {   # gets first and last position of genetic map

	my $cr=0;	# counter for rounds of MODscore calculations (result by pedigree)
 	foreach (@data){          
		if ($_ =~ /^Results by pedigree/) {$cr++;}
	}
	print "found $cr round(s) of modscore analysis\n";
	if ($sr > $cr) {
		print "selected round of modscore analysis $sr is greater than the last one found: $cr\n\n";
		exit 1;
	}
	elsif ( $sr == 0) { print "\tinitial model selected\n";}
	elsif ($sr == -1) { $sr=$cr; print "last round selected ($sr)\n";}
	else { print "round selected $sr\n";}

	my $count=0; # line counter
	my $pedc=0; # pedigree counter
	my $jetzt = $maxlines;
	my $lastpos=0;
	my $firstpos=0;

	$cr=0;
	foreach (@data){
		$count++;
		if ($_ =~ /^Results by pedigree/) {$cr++;}

		if ($sr == 0) {
			if ($_ =~ /^analyzing/) {$pedc++;}
		}
		else{
			if (($_ =~ /^Scores for pedigree/) && ($sr == $cr)) {$pedc++;}
		}
		if ($pedc > 1){last;}
		if (($_ =~ /^$/) || ($_ =~ /^OBS/)) {$jetzt = $maxlines;}
		my @feld =	split(' ',$_);
		if (($feld[0] =~ /position/) && ($feld[2] =~ /score/)){ $jetzt=$count;}
		if (($feld[0] =~ /pos\.\(avg/) && ($feld[4] =~ /score/)){ $jetzt=$count;}
		if (($pedc ==1) && ($count > $jetzt)) {
			if ($count == ($jetzt+1)) {$firstpos=$feld[0];}  
			$lastpos=$feld[0];
		}	
	}
	my $diff=$lastpos-$firstpos;
	$step=$diff/1000;	# stepwide depends on screen resolution
	# if you see needles in the plot, decrease the $step value
	
	if (($pedc == 0)&& ($sr > 0)) {print "\nNo Modscore analysis done in $infile\n";exit 1;}
#	else {
#		print "Firstposition: $firstpos, Lastposition: $lastpos\n";
#		print "Diff/Step : $diff\t$step\n";
#		}	
}

#############################  R E A D   L O D S  ##############################

sub ReadLods {   # Lods by familiy

	if ($sr == 0) { print "\nContribution of families to LOD score  \n"; }
	else { print "\nContribution of families to MOD score  \n"; }
	
	my $jetzt = $maxlines;
	my $schalt = 0;
	my $pedc = 0;		# pedigree counter
	my $pedopen = 0;
	my $pedlodf="";
	my ($lastposy,$lastnegy,$lastposx,$lastlod,$i);
	my @array=();	# command line for gnuplot
	my @pos=();		# add up positive scores
	my @neg=();		# add ip negative scores
	my @ped=();
	my $cr=0;
	my $count=0; # line counter
	my $cl=1;		# column of LOD/MOD score (normal - 1; sex specific - 3)

	open (CNTRL, "> mymodv.gp") || die "could not open mymodv.gp :$!\n";

	foreach (@data){
		if ($_ =~ /^Results by pedigree/) {$cr++;}
		my @feld =	split(' ',$_);
		$count++;
  
		if (($feld[0] =~ /analyzing/) && ($sr == 0)) { 
			my $pedfeld = $feld[2];
			@ped=split('\.',$pedfeld);
			print "Read pedigree: $ped[0] \n";
			$pedlodf = "$ped[0].lod";
			open (out1,"> $pedlodf") || die "could not open $pedlodf :$!\n";
			$pedopen = 1;
			$pedc++;
		}

		if (($feld[0] =~ /Scores/) && ($feld[2] =~ /pedigree/) && ($cr == $sr)) { 
			my $pedfeld = $feld[3];
			@ped=split('\:',$pedfeld);
			print "Read pedigree: $ped[0] \n";
			$pedlodf = "$ped[0].lod";
			open (out1,"> $pedlodf") || die "could not open $pedlodf :$!\n";
			$pedopen = 1;
			$pedc++;
		}

		if (($feld[0] =~ /position/) && ($feld[2] =~ /score/) && ($pedopen == 1)) {
			$pedopen = 2;
			$jetzt = $count;
			$i=0;
			$cl=1;
		}

		if (($feld[0] =~ /pos\.\(avg/) && ($feld[4] =~ /score/) && ($pedopen == 1)) {
			$pedopen = 2;
			$jetzt = $count;
			$i=0;
			$cl=3;
		}
	
		if (($feld[0] =~ /OBSERVED/) || ($#feld < 1)) { 
			$jetzt = $maxlines;
			if ($pedc > 0) {
				if ($pedopen == 2) {
					close (out1);
					$pedopen = 0;
					if ($schalt == 0) { 
						@array = " \'$pedlodf\' with lines";
						$schalt = 1; 
					}
					else { 
						push (@array, "\, \'$pedlodf\' with lines");
					}
				}
			}
		}

		if ($count > $jetzt) {
			$i++;
			if ($pedc == 1) {
				$pos[$i] = 0.0;
				$neg[$i] = 0.0;
			}
			if ($feld[0] > $xmax) {$xmax = $feld[0];}
			if ($feld[0] < $xmin) {$xmin = $feld[0];}
	
			########## fill up regions with low dense positions ####
			if ($i > 1) {
				my $difflod=$feld[$cl]-$lastlod;
				my $diffx=$feld[0]-$lastposx;
				my $n=$diffx/$step;
				my ($y1,$y2);
				for (my $k=1;$k<$n;$k++) { 
					my $delta=($k*$step/$diffx);

					my $x1=$lastposx+($k*$step);
					my $deltalod=$lastlod+($delta*$difflod);
	
					if ($deltalod > 0) {
						my $diffy=$pos[$i]-$lastposy;
						$y1=$lastposy+($delta*$diffy);
						$y2=$y1+$deltalod;
					}
					else {
						my $diffy=$neg[$i]-$lastnegy;
						$y1=$lastnegy+($delta*$diffy);
						$y2=$y1+$deltalod;
					}

					print out1 "$x1 $y1\n";
					print out1 "$x1 $y2\n\n"
				}
			}
			$lastposy=$pos[$i];
			$lastnegy=$neg[$i];
			$lastposx=$feld[0];
			$lastlod=$feld[$cl];

			###########################	
			if ($feld[$cl] > 0) {
				print out1 "$feld[0] $pos[$i]\n";
				$pos[$i] = $pos[$i]+$feld[$cl];
				print out1 "$feld[0] $pos[$i]\n\n";
				if ($pos[$i] > $ymax) { $ymax = $pos[$i];}
			}
			elsif ($feld[$cl] < 0) {
				print out1 "$feld[0] $neg[$i]\n";
				$neg[$i] = $neg[$i] + $feld[$cl];
				print out1 "$feld[0] $neg[$i]\n\n";
				if ($neg[$i] < $ymin) { $ymin = $neg[$i];}
			}
		}
	} # end while

	$ymax += 1;						# y-achse auf ganze LOD-Zahl bringen
	my @buff =	split('\.',$ymax);		# round to integer
	$ymax=$buff[0];

	if ($ymin < -10) { $ymin = -10;}

	my $titel="";
	if ($sr == 0) { $titel = "$pfad/$infile - Contribution of families to LOD score "; }
	else { $titel = "$pfad/$infile - Contribution of families to MOD score ";}
	
	print CNTRL "set autoscale\n";
	if ($showtitel == 1) { print CNTRL "set title \"$titel\"\n";}

	if ($label == 1) {
		print CNTRL "set xlabel \"cM\"\n";
		if ($sr == 0) {print CNTRL "set ylabel \"LOD score\"\n";}
		else {print CNTRL "set ylabel \"MOD score\"\n";}
	}

	if ($legend == 0) {print CNTRL "set nokey\n";}
	elsif ($keybelow == 1) {print CNTRL "set key below\n";}
	elsif ($pedc > 14) {;}	 
	else {$xmax = $xmax + ($xmax/6);}

	print CNTRL "plot \[$xmin:$xmax\] \[$ymin:$ymax\] ";
	print CNTRL "@array";
	print CNTRL "\n";
	print CNTRL "\n pause -1 \"Hit return to continue\"\n\n";
	if ($ps == 1) {	
		print CNTRL "set terminal postscript color solid\n";
		print CNTRL "set output \"gh_modview.ps\"\n";
		print CNTRL "replot\n";
	}

	close (CNTRL);

} # end read_lods


###-----------------------------  H E L P  ---------------------------------###
sub Help {
	print " * * * * * * *  (c) Franz Rueschendorf  1999-2006  * * * * * * * * \n";
	print " * * * * * * * * * Max-Delbrueck-Center, Berlin  * * * * * * * * * \n";
	print " * * * * * * * * * email: fruesch\@mdc-berlin.de  * * * * * * * * * \n\n";
	print "Usage        : perl [path]\\gh_modview.pl [-options] <infile>\n";
	print "Infile       : <gh.out>\n";
	print "Outfiles     : mymodv.gp, *.lod \n" ;
	print "View results : gnuplot mymodv.gp\n\n";
	print "Options      :\n";
	print "\t-clean\tdelete temporary files only\n";
	print "\t-h\thelp \(this\)\n";
	print "\t-i\tplot results from initial genetic model\n";
	print "\t  \t \(default: last MOD score analysis\)\n";
	print "\t-kb\tset key \(legend\) below\n";
	print "\t-nd\tdon\'t delete temporary files\n";
	print "\t-nl\tdon\'t show legend\n";
	print "\t-nlabel\tdon\'t show label\n";
	print "\t-nt\tdon\'t show title\n";
	print "\t-ps\tprint postscript file\n";
	print "\t-r<n>\tplot results from <n-th> round of MOD score analysis,\n";
	print "\t  \t if the 'modscore' command has been executed more than once\n";
	print "\t  \t \(default: last MOD score analysis\)\n";
	
	exit 0;
}

###----------------------------  C L E A N  -------------------------------###
sub Clean { 
	print "delete temporary files\n";
	unlink glob("*.lod");
	unlink glob("*.gp");
	exit 0;
}

###----------------------------    E N D    -------------------------------###
## end gh_modview.pl
