#ifndef FAMILY_H
#define FAMILY_H

#include "Person.h"
#include "Vector.h"
#include "InheritanceVector.h"
#include "PhasedGenotypes.h"
#include "IntSet.h"
#include "Matrix.h"
#include "PQTerm.h"

typedef struct FamilyStruct *Family;

Family Family_Create(int id);
void Family_Destroy(Family);
int Family_GetId(Family);
int Family_GetNumberOfMembers(Family);
extern int (*Family_GetNumberOfFounders)(Family);
extern int (*Family_GetNumberOfNonfounders)(Family);
int Family_AddFounder(Family, Person);
int Family_AddNonfounder(Family, Person);
Person Family_GetMember(Family, int);
Person Family_GetFounder(Family, int);
Person Family_GetNonfounder(Family, int);
Vector Family_GetNonfounders(Family);
Vector Family_GetFounders(Family);
Vector Family_GetMembers(Family);
void Family_SetParents(Family);
void Family_ClearPhase(Family);
void Family_SetAlleleOffsetsForFounders(Vector);
void Family_SetAlleleOriginsForNonfounders(Vector, InheritanceVector);
void Family_SetAlleleOffsetsForNonfounders(Vector);
void Family_SetPhase(Family, InheritanceVector);
PhasedGenotypes Family_GetPhasedGenotypesForNonfounders(Family, PhasedGenotypes phased_founder_genotypes);
extern IntSet (*Family_FindCompatiblePhasedFounderGenotypes)(Family, int marker);
Matrix Family_GetPenetranceMatrix(Family);
void Family_ComputePenetrances(Family);
Vector Family_GetPQTerms(Family);
void Family_ComputePQTerms(Family);
int Family_GetNumberOfDiseaseGenotypes(Family);
int Family_GetNumberOfInheritanceVectors(Family);
void Family_Show(Family family);
void Family_SetSingleton(Family family);
int Family_GetSingleton(Family family);
int Family_AddSortedNonfounder(Family, Person);
Vector Family_GetSortedNonfounders(Family);
void Family_SetAlleleOriginsForNonfounders_old(Vector, InheritanceVector);
void Family_SetPhase_old(Family, InheritanceVector);

#endif
