#ifndef LINKAGEMARKERFILE_H
#define LINKAGEMARKERFILE_H

#include "IntVector.h"
#include "Marker.h"
#include "UserLinkageMarkerFile.h"

typedef struct LinkageMarkerFileStruct *LinkageMarkerFile;

#include "PedigreeFile.h"

LinkageMarkerFile LinkageMarkerFile_Create(const char *filename, PedigreeFile);
LinkageMarkerFile LinkageMarkerFile_Create_old(const char *filename);
void LinkageMarkerFile_Destroy(LinkageMarkerFile);
int LinkageMarkerFile_GetNumberOfMarkers(LinkageMarkerFile);
IntVector LinkageMarkerFile_GetMapOrder(LinkageMarkerFile);
double LinkageMarkerFile_GetDiseaseLocusAlleleFrequency1(LinkageMarkerFile);
double LinkageMarkerFile_GetDiseaseLocusAlleleFrequency2(LinkageMarkerFile);
int LinkageMarkerFile_GetNumberOfLiabilityClasses(LinkageMarkerFile);
int LinkageMarkerFile_GetNumberOfPenetrances(LinkageMarkerFile);
double LinkageMarkerFile_GetPenetrance(LinkageMarkerFile, int which);
int LinkageMarkerFile_GetNumberOfAlleles(LinkageMarkerFile, int marker);
const char *LinkageMarkerFile_GetMarkerName(LinkageMarkerFile, int marker);
double LinkageMarkerFile_GetAlleleFrequency(LinkageMarkerFile, int marker, int allele);
int LinkageMarkerFile_IsSexSpecific(LinkageMarkerFile);
double LinkageMarkerFile_GetDistanceBetween(LinkageMarkerFile, int marker_1, int marker_2);
double LinkageMarkerFile_GetMaleDistanceBetween(LinkageMarkerFile, int marker_1, int marker_2);
double LinkageMarkerFile_GetFemaleDistanceBetween(LinkageMarkerFile, int marker_1, int marker_2);
int LinkageMarkerFile_GetNumberOfPhenotypes(void);
int LinkageMarkerFile_GetNumberOfCovariates(void);
int LinkageMarkerFile_GetMarkerIndex(LinkageMarkerFile, const char *marker_name);
int LinkageMarkerFile_IsReady(LinkageMarkerFile);
Marker LinkageMarkerFile_GetMarkerByIndex(LinkageMarkerFile, int marker_index);
Marker LinkageMarkerFile_GetMarkerByName(LinkageMarkerFile, const char *marker_name);
int LinkageMarkerFile_WithinMaxDistance(LinkageMarkerFile, Vector markers, double maximum_marker_distance);
void print_distances(LinkageMarkerFile);
Vector LinkageMarkerFile_GetMarkers(LinkageMarkerFile);
void LinkageMarkerFile_WriteWithNewLinkageFreqs(LinkageMarkerFile, UserLinkageMarkerFile, Vector freqs, const char *filename);
int LinkageMarkerFile_IsCompatibleWith(LinkageMarkerFile, UserLinkageMarkerFile);
int LinkageMarkerFile_IsCompatibleWithTestMarkers(LinkageMarkerFile, Vector test_marker_names);

#endif
