#ifndef MODSCORER_H
#define MODSCORER_H

#include "AnnotatedFamily.h"
#include "DoubleVector.h"
#include "HaplotypeMatrix.h"
#include "LinkageMarkerFile.h"
#include "ModelParamsFile.h"
#include "ModScore.h"
#include "ModScoreList.h"
#include "PenetranceMatrix.h"
#include "UserLinkageMarkerFile.h"
#include "Vector.h"

typedef ModScore *(*ModScoreFunction)(
    Vector test_markers,
    Vector annotated_families,
    DoubleVector em_hapfreqs,
    LinkageMarkerFile linkage_file,
    ModelParams user_model,
    int number_of_test_marker_sets,
    ModScore *grid_mod_score);

typedef struct ModScorerStruct {
    Vector families;
    Vector inheritance_distros;
    int test_markers_per_set;
    double max_marker_distance;
    LinkageMarkerFile linkage_file;
    UserLinkageMarkerFile user_linkage_file;
    ModScoreList mod_score_list;
    ModelParamsFile model_params_file;
} *ModScorer;

typedef struct MatrixCacheStruct {
    double **matrix;
    int nrow;
    int ncol;
} *MatrixCache;

ModScorer ModScorer_Create(
    Vector families,
    Vector inheritance_distros,
    int test_markers_per_set,
    double max_marker_distance,
    LinkageMarkerFile linkage_file,
    UserLinkageMarkerFile user_linkage_file,
    ModelParamsFile model_params_file,
    ModScoreList mod_score_list);

double ModScorer_ComputeModScore(
    ModScorer mod_scorer,
    Vector annotated_families,
    Vector test_markers,
    double disease_allele_frequency,
    DoubleVector hapfreqs,
    DoubleVector penetrances,
    DoubleVector em_hapfreqs);

void ModScorer_ComputeModScoresForRegions(ModScorer, Vector regions);
void ModScorer_ComputeModScoreForTestMarkers(ModScorer mod_scorer, Vector test_marker_names);
void ModScorer_ComputeModScoreForTestMarkersAndUserModels(ModScorer mod_scorer, Vector test_marker_names);

double ModScorer_ScoreUnderH0(
    DoubleVector penetrances,
    double disease_allele_frequency,
    DoubleVector weights,
    DoubleVector weights_h1,
    DoubleVector em_hapfreqs,
    PenetranceMatrix penetrance_matrix,
    HaplotypeMatrix haplotype_matrix_without_disease_locus,
    Vector pqterms,
    int number_of_founders,
    double **cache_pen_matrix);

double ModScorer_ScoreUnderH1(
    DoubleVector penetrances,
    DoubleVector haplotype_frequencies,
    DoubleVector weights,
    PenetranceMatrix penetrance_matrix,
    HaplotypeMatrix haplotype_matrix,
    int number_of_founders,
    double **cache_pen_matrix);

ModScoreList ModScorer_GetModScoreList(ModScorer);

#endif
