#ifndef PERSON_H
#define PERSON_H

#include "Genotype.h"
#include "PhasedGenotypes.h"
#include "AlleleOrigin.h"
#include "Vector.h"

typedef struct PersonStruct *Person;

enum { MISSING_ALLELE = -1 };

Person Person_Create(const char *id);
void Person_Destroy(Person);

const char *Person_GetId(Person);
void Person_SetFatherId(Person, const char *);
void Person_SetMotherId(Person, const char *);
const char *Person_GetFatherId(Person);
const char *Person_GetMotherId(Person);
void Person_SetFather(Person, Person father);
void Person_SetMother(Person, Person mother);
Person Person_GetFather(Person);
Person Person_GetMother(Person);

void Person_SetMarkerGenotypes(Person, int *genotypes, int nmarkers);
void Person_SetMarkerAlleles2(Person person, const int *alleles, int number_of_markers);
int Person_GetMarkerAllele1(Person, int marker);
int Person_GetMarkerAllele2(Person, int marker);
Genotype Person_GetMarkerGenotype(Person, int marker);
int Person_GetNumberOfMarkers(Person);
int Person_IsActualGenotypeCompatibleWithGenotype(Person, int marker, Genotype);
int Person_AreActualGenotypesCompatibleWithGenotypes(Vector persons, int marker, PhasedGenotypes);

void Person_SetAffected(Person);
void Person_SetUnaffected(Person);
int Person_IsAffected(Person);
int Person_IsAffectionStatusKnown(Person);

void Person_SetMale(Person);
int Person_IsMale(Person);
void Person_SetFemale(Person);
int Person_IsSexKnown(Person);

void Person_SetPaternalAlleleOffset(Person, int);
int Person_GetPaternalAlleleOffset(Person);
int Person_HasPaternalAlleleOffset(Person);
void Person_SetMaternalAlleleOffset(Person, int);
int Person_GetMaternalAlleleOffset(Person);
int Person_HasMaternalAlleleOffset(Person);

void Person_SetMaternalAlleleOrigin(Person, AlleleOrigin);
void Person_SetPaternalAlleleOrigin(Person, AlleleOrigin);
AlleleOrigin Person_GetMaternalAlleleOrigin(Person);
AlleleOrigin Person_GetPaternalAlleleOrigin(Person);

int Person_HasPaternalBitFixed(Person);
int Person_HasMaternalBitFixed(Person);

void Person_FixPaternalBit(Person);
void Person_FixMaternalBit(Person);

const char *Person_GetDescription(Person);
void Person_Show(Person);

void Person_SetSimulatedMarkerAllele1(Person person, int marker_index, int allele);
void Person_SetSimulatedMarkerAllele2(Person person, int marker_index, int allele);
int Person_IsSimulated(Person person);
int Person_SimulatedParents(Person person);
void Person_SetSimulated(Person person);
void Person_SetNotSimulated(Person person);
void Person_MarkGenotypeAsMissing(Person person, int marker_index);
int Person_IsGenotypeMissing(Person person, int marker_index);

#endif
