#ifndef HASHTABLETEMPLATEIMPL_H
#define HASHTABLETEMPLATEIMPL_H

#include <stdlib.h>

#include "List.h"
#include "Memory.h"

#define HashTableTemplateImpl(T) \
struct T ## HashTableStruct { \
    int size; \
    List *buckets; \
}; \
\
typedef struct ElementStruct *Element; \
struct ElementStruct { \
    T key; \
    void *value; \
}; \
\
static Element create_element(T key, void *value); \
static int key_equals(void *element, void *key); \
\
T ## HashTable T ## HashTable_Create(int size) \
{ \
    T ## HashTable table; \
    int i; \
\
    if (size == 0) \
        return NULL; \
    table = (T ## HashTable) Memory_Malloc(sizeof *table); \
    table->size = 0; \
\
    table->buckets = (List *) Memory_Malloc(size * sizeof(List)); \
    table->size = size; \
    for (i = 0; i < table->size; i++) \
        table->buckets[i] = List_Create(); \
\
    return table; \
} \
\
void T ## HashTable_Destroy(T ## HashTable table) \
{ \
    int i; \
\
    if (table == NULL) \
        return; \
\
    for (i = 0; i < table->size; i++) \
        Memory_Free(table->buckets[i]); \
    Memory_Free(table->buckets); \
    Memory_Free(table); \
} \
\
void T ## HashTable_Put(T ## HashTable table, T key, void *value) \
{ \
    int i = HASH(key, table); \
    Element element = create_element(key, value); \
    List old_bucket = table->buckets[i]; \
\
    table->buckets[i] = List_AddItem(List_RemoveIf(key_equals, old_bucket, &key), element); \
    List_Destroy(old_bucket); \
} \
\
static int key_equals(void *element, void *key) \
{ \
    Element e = element; \
    T *k = key; \
\
    return IS_SAME_KEY(e->key, *k); \
} \
\
static Element create_element(T key, void *value) \
{ \
    Element element = (Element) Memory_Malloc(sizeof *element); \
\
    element->key = key; \
    element->value = value; \
\
    return element; \
} \
\
int T ## HashTable_Get(T ## HashTable table, T key, void **value) \
{ \
    int i = HASH(key, table); \
    List list; \
\
    for (list = table->buckets[i]; !List_IsEmpty(list); list = List_GetNext(list)) { \
        Element element = List_GetItem(list); \
        if (IS_SAME_KEY(element->key, key)) { \
            *value = element->value; \
            return 1; \
        } \
    } \
    return 0; \
} \
\
int T ## HashTable_GetSize(T ## HashTable table) \
{ \
    int i, size = 0; \
\
    for (i = 0; i < table->size; i++) \
        size += List_GetSize(table->buckets[i]); \
\
    return size; \
} \
\
int T ## HashTable_GetLength(T ## HashTable table) \
{ \
    return table->size; \
} \
\
T *T ## HashTable_SortKeys(T ## HashTable table, T ## compare_func compare) { \
    int nelems = T ## HashTable_GetSize(table); \
    T *sorted = Memory_Malloc(nelems * sizeof(T)); \
    int i = 0; \
    for (int j = 0; j < table->size; j++) { \
        List list = table->buckets[j]; \
        while (!List_IsEmpty(list)) { \
            Element element = List_GetItem(list); \
            sorted[i++] = element->key; \
            list = List_GetNext(list); \
        } \
    } \
    qsort(sorted, nelems, sizeof(T), compare); \
    return sorted; \
}

#endif
