#ifndef TRIPLET_H
#define TRIPLET_H

// We introduced the Triplet because we wanted to use a combination of
// up to 3 phased genotypes as a key in a hash table.  Since hash keys
// have to be single objects we had to bundle the phased genotypes in
// a struct.  With every combination of phased genotypes we associate
// an array of exponents for haplotype frequencies.  The last element
// of the exponent array is not an exponent but represents the number
// of times that the given exponent combination occurred for different
// inheritance vectors in a pedigree.
//
// The following example with 2 test markers and 1 founder illustrates
// that 2 unique triplets may lead to the same exponent combination:
//
// Triplet A = {2, 1, 0} where 2 = 1|0 from TM1
//                         and 1 = 0|1 from TM2
//
//     => haplotypes 01 and 10 (TM2 at MSB, TM1 at LSB),
//        i.e., ++exponents[1] and ++exponents[2]
//
// Triplet B = {1, 2, 0} where 1 = 0|1 from TM1
//                         and 2 = 1|0 from TM2
//
//     => haplotypes 10 and 01 (TM2 at MSB, TM1 at LSB),
//        i.e., ++exponents[2] and ++exponents[1]

typedef struct TripletStruct *Triplet;

Triplet Triplet_Create(int i, int j, int k);
int Triplet_Equal(Triplet, Triplet);
unsigned Triplet_Hash(Triplet);
void Triplet_Print(Triplet);

#endif
