#ifndef VECTOR_H
#define VECTOR_H

#include "NELEMS.h"

typedef struct VectorStruct *Vector;
typedef int (*EqualsFunction)(void *, void *);
typedef void *(*Vector_CollectFunction)(void *, void *);
typedef void (*Vector_ReduceFunction)(void *result_so_far, void *next_element);

Vector Vector_CreateEmpty(void);
Vector Vector_Create(int initial_capacity);
// Destroy storage, but leave elements untouched.
void Vector_Destroy(Vector);
int Vector_GetSize(Vector);
int Vector_GetCapacity(Vector);
void Vector_PushBack(Vector, void *);
void Vector_Set(Vector, int, void *);
void *Vector_Get(Vector, int);
void **Vector_GetInternalArray(Vector);
void Vector_Reverse(Vector);
int Vector_Equals(Vector, Vector, EqualsFunction);
Vector Vector_Copy(Vector);
Vector Vector_FindCombinations(Vector vectors);
Vector Vector_With(int number_of_elements, ...);
Vector Vector_FromArray(void *array, int number_of_elements, int size_of_elements);
Vector Vector_FromCopyOfArray(void *array, int number_of_elements, int size_of_elements);
#define Vector_FromStaticArray(array) Vector_FromArray((array), NELEMS(array), sizeof((array)[0]))
#define Vector_FromCopyOfStaticArray(array) Vector_FromCopyOfArray((array), NELEMS(array), sizeof((array)[0]))
Vector Vector_Collect(Vector, Vector_CollectFunction, void *additional_arguments);
Vector Vector_SafeCollect(Vector, Vector_CollectFunction, void *additional_arguments);
void Vector_Reduce(Vector, Vector_ReduceFunction, void *result);
double Vector_ProductOfDoubles(Vector);

int Vector_IncrementIndices(int *indices, Vector vectors);

#endif
