#ifndef VECTORTEMPLATE_H
#define VECTORTEMPLATE_H

// Given a simple or typedef'd type T, this header file creates a
// vector type with elements of type T and the below interface.

// TVector TVector_Create(size_t capacity);
// void TVector_Destroy(TVector);
// int TVector_PushBack(TVector, T value);
// T *TVector_GetInternalArray(TVector);
// T TVector_Get(TVector, int index);
// int TVector_GetSize(TVector);
// int TVector_GetCapacity(TVector, size_t *capacity);
// int TVector_Set(TVector, int index, T value);
// int TVector_Resize(TVector, size_t *capacity);
// TVector TVector_Copy(TVector);
// TVector TVector_With(int number_of_elements, ...);
// TVector TVector_AsView(T *beg, T *end);

// The header file of the new vector type must include the header file
// "VectorTemplate.h" and call 'VectorTemplate(T)'.  Note the missing
// final semicolon.
//
// The implementation file must include "VectorTemplateImpl.h", the
// header of the new vector type, say "TVector.h", and the call
// 'VectorTemplateImpl(T)'.  Note the missing final semicolon.

#include <stddef.h>

#define C(A,B) A ## B

#define VectorTemplate(T) \
typedef struct C(T,VectorStruct) *C(T,Vector); \
C(T,Vector) C(T,Vector_Create)(size_t capacity); \
void C(T,Vector_Destroy)(C(T,Vector)); \
int C(T,Vector_PushBack)(C(T,Vector), T value); \
T *C(T,Vector_GetInternalArray)(C(T,Vector)); \
T C(T,Vector_Get)(C(T,Vector), int index); \
int C(T,Vector_GetSize)(C(T,Vector)); \
int C(T,Vector_GetCapacity)(C(T,Vector), size_t *capacity); \
int C(T,Vector_Set)(C(T,Vector), int index, T value); \
int C(T,Vector_Resize)(C(T,Vector), size_t capacity); \
C(T,Vector) C(T,Vector_Copy)(C(T,Vector)); \
C(T,Vector) C(T,Vector_With)(int number_of_elements, ...); \
C(T,Vector) C(T,Vector_AsView)(T *beg, T *end);

#endif
