#ifndef VECTORTEMPLATEIMPL_H
#define VECTORTEMPLATEIMPL_H

#include "Memory.h"
#include "FailWith.h"
#include <stddef.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#define VectorTemplateImpl(T) \
struct T ## VectorStruct { \
    T *k; \
    size_t size; \
    size_t capacity; \
}; \
\
static int double_capacity(T ## Vector); \
static int initialize_storage(T ## Vector); \
static void set_to_zero(T *, int n); \
\
static void set_to_zero(T *k, int n) \
{ \
    for (int i = 0; i < n; i++) \
        k[i] = 0; \
} \
\
static int double_capacity(T ## Vector v) \
{ \
    T *k; \
    size_t nbytes = 2 * v->capacity * sizeof(T); \
    if ((k = Memory_Realloc(v->k, nbytes)) == NULL) \
        return 0; \
    v->k = k; \
    set_to_zero(v->k + v->capacity, v->capacity); \
    v->capacity *= 2; \
    return 1; \
} \
\
static int initialize_storage(T ## Vector v) \
{ \
    T *k; \
    if ((k = Memory_Malloc(sizeof *k)) == NULL) \
        return 0; \
    v->k = k; \
    set_to_zero(v->k, 1); \
    v->capacity = 1; \
    return 1; \
} \
\
T ## Vector T ## Vector_Create(size_t capacity) \
{ \
    T ## Vector v; \
    T *k; \
    if ((v = Memory_Malloc(sizeof *v)) == NULL) \
        return NULL; \
    if ((k = Memory_Malloc(capacity * sizeof(T))) == NULL) { \
        Memory_Free(v); \
        return NULL; \
    } \
    v->k = k; \
    set_to_zero(v->k, capacity); \
    v->size = 0; \
    v->capacity = capacity; \
    return v; \
} \
\
int T ## Vector_PushBack(T ## Vector v, T value) \
{ \
    if (v->capacity == 0) \
        initialize_storage(v); \
    if (v->size == v->capacity && !double_capacity(v)) \
        return 0; \
    v->k[v->size++] = value; \
    return 1; \
} \
\
T * T ## Vector_GetInternalArray(T ## Vector v) \
{ \
    return v->k; \
} \
\
inline T T ## Vector_Get(T ## Vector v, int index) \
{ \
    return v->k[index]; \
} \
\
int T ## Vector_GetSize(T ## Vector v) \
{ \
    if (v == NULL) \
        FailWith("Vector_GetSize: vector is NULL"); \
    return v->size; \
} \
\
int T ## Vector_GetCapacity(T ## Vector v, size_t *capacity) \
{ \
    *capacity = v->capacity; \
    return 1; \
} \
\
int T ## Vector_Set(T ## Vector v, int index, T value) \
{ \
    if (v->capacity == 0) \
        initialize_storage(v); \
    while ((unsigned long) index >= v->capacity) \
        if (!double_capacity(v)) \
            return 0; \
    v->k[index] = value; \
    if (v->size < (size_t) index + 1) \
        v->size = index + 1; \
    return 1; \
} \
\
void T ## Vector_Destroy(T ## Vector v) \
{ \
    if (v == NULL) \
        return; \
    if (v->k != NULL) { \
        Memory_Free(v->k); \
        v->k = NULL; \
    } \
    Memory_Free(v); \
} \
\
int T ## Vector_Resize(T ## Vector v, size_t capacity) \
{ \
    T *p; \
    if (capacity == 0) \
        v->k = NULL; \
    else if ((p = Memory_Realloc(v->k, capacity * sizeof(T))) == NULL) { \
        FailWith("Vector_Resize: failed to allocate memory for %lu elements", \
            (unsigned long) capacity); \
    } \
    else { \
        for (size_t i = v->capacity; i < capacity; i++) \
            p[i] = 0; \
        v->k = p; \
    } \
    v->capacity = capacity; \
    if (capacity < v->size) \
        v->size = capacity; \
\
    return 1; \
} \
\
T ## Vector T ## Vector_Copy(T ## Vector vector) \
{ \
    T ## Vector copy = T ## Vector_Create(T ## Vector_GetSize(vector)); \
    for (int i = 0; i < T ## Vector_GetSize(vector); i++) \
        T ## Vector_PushBack(copy, T ## Vector_Get(vector, i)); \
    return copy; \
} \
\
T ## Vector T ## Vector_With(int number_of_elements, ...) \
{ \
    T ## Vector vector = T ## Vector_Create(number_of_elements); \
    va_list ap; \
    va_start(ap, number_of_elements); \
    while (number_of_elements--) \
        T ## Vector_PushBack(vector, va_arg(ap, T)); \
    va_end(ap); \
    return vector; \
}

#endif
