#!/bin/bash
set -e

function usage {
    echo Usage:
    echo
    echo "    bash $0 COMMANDS_FILE OUTPUT_DIRECTORY PERMUTATIONS MAX_PROCESSES"
    echo
    echo 'COMMANDS_FILE      ghm commands file; input files have to be specified'
    echo '                   using absolute (!) paths and the output file has to'
    echo '                   be called output.txt'
    echo
    echo 'OUTPUT_DIRECTORY   directory in which permutations will be created'
    echo
    echo 'PERMUTATIONS       number of permutations to run'
    echo
    echo 'MAX_PROCESSES      maximum number of parallel processes to use'
    exit 1
}

COMMANDS_FILE="$1"
OUTPUT_DIRECTORY="$2"
PERMUTATIONS="$3"
MAX_PROCESSES="$4"

if [ -z "$1" ] || [ -z "$2" ] || [ -z "$3" ] || [ -z "$4" ]; then
    usage
fi

if ! which ghm &>/dev/null; then
    echo 'Cannot find "ghm" executable'
    exit 1
fi

if [ ! -f "${COMMANDS_FILE}" ]; then
    echo "COMMANDS_FILE does not exist: ${COMMANDS_FILE}"
    usage
fi

if ! grep -q 'set random seed' "${COMMANDS_FILE}"; then
    echo 'COMMANDS_FILE must contain "set random seed" command'
    usage
fi

if ! [ "$PERMUTATIONS" -eq "$PERMUTATIONS" ] &>/dev/null; then
    echo "PERMUTATIONS must be a number: $PERMUTATIONS"
    usage
fi

if ! [ "${MAX_PROCESSES}" -eq "${MAX_PROCESSES}" ] &>/dev/null; then
    echo "MAX_PROCESSES must be a number: ${MAX_PROCESSES}"
    usage
fi

nprocs=0  # current number of parallel processes
perm=0    # number of current permutation

while [ "$perm" -lt "$PERMUTATIONS" ]; do
    if [ "$nprocs" -eq "$MAX_PROCESSES" ]; then
        wait
        nprocs=$(( nprocs - 1 ))
    fi
    perm=$(( perm + 1 ))
    echo -n "$perm "
    d="${OUTPUT_DIRECTORY}/$perm"
    if [ -d "$d" ]; then
        echo "Permutation directory exists already: $d"
        echo "Please, delete it manually before proceeding."
        exit 1
    fi
    mkdir -p "$d"
    cp "${COMMANDS_FILE}" "$d"
    sed -i "s/set random seed.*/set random seed $RANDOM/" "$d/$( basename ${COMMANDS_FILE} )"
    ( cd "$d" && ghm < "$( basename ${COMMANDS_FILE} )" &>/dev/null ) &
    nprocs=$(( nprocs + 1 ))
done

echo

while [ "$nprocs" -gt 0 ]; do
    wait
    nprocs=$(( nprocs - 1 ))
done

echo done
