#include "AnnotatedFamily.h"
#include "Memory.h"
#include <stddef.h>

struct AnnotatedFamilyStruct {
    Family family;
    HaplotypeMatrix haplotype_matrix;
    HaplotypeMatrix haplotype_matrix_without_disease_locus;
    InheritanceDistribution inheritance_distribution;
    int singleton;
    FounderHaplotypes founder_haplotypes;
};

AnnotatedFamily AnnotatedFamily_Create(Family family, Vector test_marker_set)
{
    AnnotatedFamily annotated_family = (AnnotatedFamily) Memory_Malloc(sizeof(struct AnnotatedFamilyStruct));
    annotated_family->family = family;
    annotated_family->haplotype_matrix = HaplotypeMatrix_Create(family, test_marker_set);
    annotated_family->haplotype_matrix_without_disease_locus = HaplotypeMatrix_RemoveDiseaseLocus(annotated_family->haplotype_matrix);
    annotated_family->inheritance_distribution = NULL;
    annotated_family->singleton = Family_GetSingleton(family);
    annotated_family->founder_haplotypes = FounderHaplotypes_Create(family, test_marker_set);

    return annotated_family;
}

AnnotatedFamily AnnotatedFamily_CreateWithInheritanceDistro(Family family, Vector test_marker_set, InheritanceDistribution distro)
{
    AnnotatedFamily annotated_family = AnnotatedFamily_Create(family, test_marker_set);
    AnnotatedFamily_SetInheritanceDistribution(annotated_family, distro);
    return annotated_family;
}

void AnnotatedFamily_Destroy(AnnotatedFamily annotated_family)
{
    if (!annotated_family)
        return;

    HaplotypeMatrix_Destroy(annotated_family->haplotype_matrix);
    HaplotypeMatrix_Destroy(annotated_family->haplotype_matrix_without_disease_locus);
    FounderHaplotypes_Destroy(annotated_family->founder_haplotypes);
    Memory_Free(annotated_family);
}

Family AnnotatedFamily_GetFamily(AnnotatedFamily annotated_family)
{
    return annotated_family->family;
}

PenetranceMatrix AnnotatedFamily_GetPenetranceMatrix(AnnotatedFamily annotated_family)
{
    return Family_GetPenetranceMatrix(annotated_family->family);
}

HaplotypeMatrix AnnotatedFamily_GetHaplotypeMatrix(AnnotatedFamily annotated_family)
{
    return annotated_family->haplotype_matrix;
}

HaplotypeMatrix AnnotatedFamily_GetHaplotypeMatrixWithoutDiseaseLocus(AnnotatedFamily annotated_family)
{
    return annotated_family->haplotype_matrix_without_disease_locus;
}

void AnnotatedFamily_SetInheritanceDistribution(AnnotatedFamily annotated_family, InheritanceDistribution distro)
{
    annotated_family->inheritance_distribution = distro;
}

InheritanceDistribution AnnotatedFamily_GetInheritanceDistribution(AnnotatedFamily annotated_family)
{
    return annotated_family->inheritance_distribution;
}

Vector AnnotatedFamily_GetPQTerms(AnnotatedFamily annotated_family)
{
    return Family_GetPQTerms(annotated_family->family);
}

int AnnotatedFamily_GetSingleton(AnnotatedFamily annotated_family)
{
    return annotated_family->singleton;
}

FounderHaplotypes AnnotatedFamily_GetFounderHaplotypes(AnnotatedFamily annotated_family)
{
    return annotated_family->founder_haplotypes;
}

int AnnotatedFamily_GetNumberOfInheritanceVectors(AnnotatedFamily annotated_family)
{
    return Family_GetNumberOfInheritanceVectors(annotated_family->family);
}
