#include "FounderHaplotypes.h"
#include "Marker.h"
#include "Memory.h"

#include <stdio.h>

#define MAX_NUMBER_OF_TEST_MARKERS 3

struct FounderHaplotypesStruct {
    int number_of_founders;
    int number_of_test_markers;
    int number_of_haplotypes;
    TripletHashTable hash_table;
    TripletHashTable hash_table_1;
    TripletHashTable hash_table_2;
    TripletHashTable hash_table_3;
};

// Given phased test marker genotypes [ab, cd, ef] where a-f either 0
// or 1, return the integer with bits fdb if left, otherwise eca.
static int em_get_index_of_haplotype(int left, Genotype *test_marker_genotype, int number_of_test_markers)
{
    int shift_width = left ? 1 : 0;
    int result = 0x0;
    for (int test_marker = number_of_test_markers - 1; test_marker >= 0; test_marker--) {
        result <<= 1;
        result |= (test_marker_genotype[test_marker] >> shift_width) & 0x1;
    }
    return result;
}

static int em_build_haplotype_index(int left, int index_of_person, PhasedGenotypes *test_marker_genotypes, int number_of_test_markers)
{
    Genotype genotype_of_test_markers[MAX_NUMBER_OF_TEST_MARKERS] = { 0 };
    for (int i = 0; i < number_of_test_markers; i++)
        genotype_of_test_markers[i] = PhasedGenotypes_Get(test_marker_genotypes[i], index_of_person);
    return em_get_index_of_haplotype(left, genotype_of_test_markers, number_of_test_markers);
}

static void populate_hash_table_1_test_marker(
    TripletHashTable hash_table,
    int number_of_phased_genotypes,
    int number_of_haplotypes,
    int number_of_founders,
    IntSet *compatible_phased_test_marker_genotypes,
    int which_marker
)
{
    PhasedGenotypes test_marker_genotypes[1];
    for (int i = 0; i < number_of_phased_genotypes; i++) {
        if (IntSet_Contains(compatible_phased_test_marker_genotypes[which_marker], i)) {
            test_marker_genotypes[0] = i;
            Triplet triplet = Triplet_Create(i, 0, 0);
            int *exponents;
            if (!TripletHashTable_Get(hash_table, triplet, (void **) &exponents)) {
                exponents = Memory_Calloc(number_of_haplotypes + 1, sizeof(int));
                for (int f = 0; f < number_of_founders; f++) {
                    ++exponents[em_build_haplotype_index(0, f, test_marker_genotypes, 1)];
                    ++exponents[em_build_haplotype_index(1, f, test_marker_genotypes, 1)];
                }
                exponents[number_of_haplotypes] = 1;  // coefficient = 1
                TripletHashTable_Put(hash_table, triplet, exponents);
            } else {
                exponents = TripletHashTable_GetExponents(hash_table, triplet);
                ++exponents[number_of_haplotypes];  // increment coefficient
                TripletHashTable_Put(hash_table, triplet, exponents);
            }
        }
    }
}

static void populate_hash_table_2_test_markers(
    TripletHashTable hash_table,
    int number_of_phased_genotypes,
    int number_of_haplotypes,
    int number_of_founders,
    IntSet *compatible_phased_test_marker_genotypes
)
{
    PhasedGenotypes test_marker_genotypes[2];
    for (int i = 0; i < number_of_phased_genotypes; i++) {
        if (IntSet_Contains(compatible_phased_test_marker_genotypes[0], i)) {
            for (int j = 0; j < number_of_phased_genotypes; j++) {
                if (IntSet_Contains(compatible_phased_test_marker_genotypes[1], j)) {
                    test_marker_genotypes[0] = i;
                    test_marker_genotypes[1] = j;
                    Triplet triplet = Triplet_Create(i, j, 0);
                    int *exponents;
                    if (!TripletHashTable_Get(hash_table, triplet, (void **) &exponents)) {
                        exponents = Memory_Calloc(number_of_haplotypes + 1, sizeof(int));
                        for (int f = 0; f < number_of_founders; f++) {
                            ++exponents[em_build_haplotype_index(0, f, test_marker_genotypes, 2)];
                            ++exponents[em_build_haplotype_index(1, f, test_marker_genotypes, 2)];
                        }
                        exponents[number_of_haplotypes] = 1;
                        TripletHashTable_Put(hash_table, triplet, exponents);
                    } else {
                        exponents = TripletHashTable_GetExponents(hash_table, triplet);
                        ++exponents[number_of_haplotypes];
                        TripletHashTable_Put(hash_table, triplet, exponents);
                    }
                }
            }
        }
    }
}

static void populate_hash_table_3_test_markers(
        TripletHashTable hash_table,
        int number_of_phased_genotypes,
        int number_of_haplotypes,
        int number_of_founders,
        IntSet *compatible_phased_test_marker_genotypes
        )
{
    PhasedGenotypes test_marker_genotypes[3];
    for (int i = 0; i < number_of_phased_genotypes; i++) {
        if (IntSet_Contains(compatible_phased_test_marker_genotypes[0], i)) {
            for (int j = 0; j < number_of_phased_genotypes; j++) {
                if (IntSet_Contains(compatible_phased_test_marker_genotypes[1], j)) {
                    for (int k = 0; k < number_of_phased_genotypes; k++) {
                        if (IntSet_Contains(compatible_phased_test_marker_genotypes[2], k)) {
                            test_marker_genotypes[0] = i;
                            test_marker_genotypes[1] = j;
                            test_marker_genotypes[2] = k;
                            Triplet triplet = Triplet_Create(i, j, k);
                            int *exponents;
                            if (!TripletHashTable_Get(hash_table, triplet, (void **) &exponents)) {
                                exponents = Memory_Calloc(number_of_haplotypes + 1, sizeof(int));
                                for (int f = 0; f < number_of_founders; f++) {
                                    ++exponents[em_build_haplotype_index(0, f, test_marker_genotypes, 3)];
                                    ++exponents[em_build_haplotype_index(1, f, test_marker_genotypes, 3)];
                                }
                                exponents[number_of_haplotypes] = 1;
                                TripletHashTable_Put(hash_table, triplet, exponents);
                            } else {
                                exponents = TripletHashTable_GetExponents(hash_table, triplet);
                                ++exponents[number_of_haplotypes];
                                TripletHashTable_Put(hash_table, triplet, exponents);
                            }
                        }
                    }
                }
            }
        }
    }
}

FounderHaplotypes FounderHaplotypes_Create(Family family, Vector test_marker_set)
{
    FounderHaplotypes founder_haplotypes = Memory_Malloc(sizeof(*founder_haplotypes));
    founder_haplotypes->number_of_founders = Family_GetNumberOfFounders(family);
    founder_haplotypes->number_of_test_markers = Vector_GetSize(test_marker_set);
    founder_haplotypes->number_of_haplotypes = 1 << (Vector_GetSize(test_marker_set));
    IntSet *compatible_phased_test_marker_genotypes = Memory_Malloc(Vector_GetSize(test_marker_set) * sizeof(IntSet));
    int number_of_phased_genotypes = 1 << (2 * Family_GetNumberOfFounders(family));
    int max_buckets = 4096;
    TripletHashTable hash_table = NULL;
    TripletHashTable hash_table_1 = NULL;
    TripletHashTable hash_table_2 = NULL;
    TripletHashTable hash_table_3 = NULL;
    int number_of_test_markers = Vector_GetSize(test_marker_set);
    int number_of_buckets = number_of_phased_genotypes;
    int number_of_buckets_123 = 0;
    if (number_of_test_markers == 1) {
        number_of_buckets = number_of_phased_genotypes;
        number_of_buckets_123 = number_of_phased_genotypes;
    } else if (number_of_test_markers == 2) {
        number_of_buckets = number_of_phased_genotypes * number_of_phased_genotypes;
        number_of_buckets_123 = number_of_phased_genotypes;
    } else if (number_of_test_markers == 3) {
        number_of_buckets = number_of_phased_genotypes * number_of_phased_genotypes * number_of_phased_genotypes;
        number_of_buckets_123 = number_of_phased_genotypes;
    }
    if (number_of_buckets > max_buckets) {
        number_of_buckets = max_buckets;
    }
    if (number_of_buckets_123 > max_buckets) {
        number_of_buckets_123 = max_buckets;
    }
    hash_table = TripletHashTable_Create(number_of_buckets);
    hash_table_1 = TripletHashTable_Create(number_of_buckets_123);
    hash_table_2 = TripletHashTable_Create(number_of_buckets_123);
    hash_table_3 = TripletHashTable_Create(number_of_buckets_123);
    founder_haplotypes->hash_table = hash_table;
    for (int inheritance_vector = 0; inheritance_vector < Family_GetNumberOfInheritanceVectors(family); inheritance_vector++) {
        Family_SetPhase(family, inheritance_vector);
        for (int i = 0; i < Vector_GetSize(test_marker_set); i++) {
            compatible_phased_test_marker_genotypes[i] = Family_FindCompatiblePhasedFounderGenotypes(family, Marker_GetIndex(Vector_Get(test_marker_set, i)));
        }
        int number_of_haplotypes = 1 << number_of_test_markers;
        if (number_of_test_markers == 1) {
            populate_hash_table_1_test_marker(
                hash_table,
                number_of_phased_genotypes,
                number_of_haplotypes,
                founder_haplotypes->number_of_founders,
                compatible_phased_test_marker_genotypes,
                0
            );
            founder_haplotypes->hash_table_1 = founder_haplotypes->hash_table;
        } else if (number_of_test_markers == 2) {
            populate_hash_table_2_test_markers(
                hash_table,
                number_of_phased_genotypes,
                number_of_haplotypes,
                founder_haplotypes->number_of_founders,
                compatible_phased_test_marker_genotypes
            );
            founder_haplotypes->hash_table_1 = hash_table_1;
            populate_hash_table_1_test_marker(
                hash_table_1,
                number_of_phased_genotypes,
                2,
                founder_haplotypes->number_of_founders,
                compatible_phased_test_marker_genotypes,
                0
            );
            founder_haplotypes->hash_table_2 = hash_table_2;
            populate_hash_table_1_test_marker(
                hash_table_2,
                number_of_phased_genotypes,
                2,
                founder_haplotypes->number_of_founders,
                compatible_phased_test_marker_genotypes,
                1
            );
        } else if (number_of_test_markers == 3) {
            populate_hash_table_3_test_markers(
                hash_table,
                number_of_phased_genotypes,
                number_of_haplotypes,
                founder_haplotypes->number_of_founders,
                compatible_phased_test_marker_genotypes
            );
            founder_haplotypes->hash_table_1 = hash_table_1;
            populate_hash_table_1_test_marker(
                hash_table_1,
                number_of_phased_genotypes,
                2,
                founder_haplotypes->number_of_founders,
                compatible_phased_test_marker_genotypes,
                0
            );
            founder_haplotypes->hash_table_2 = hash_table_2;
            populate_hash_table_1_test_marker(
                hash_table_2,
                number_of_phased_genotypes,
                2,
                founder_haplotypes->number_of_founders,
                compatible_phased_test_marker_genotypes,
                1
            );
            founder_haplotypes->hash_table_3 = hash_table_3;
            populate_hash_table_1_test_marker(
                hash_table_3,
                number_of_phased_genotypes,
                2,
                founder_haplotypes->number_of_founders,
                compatible_phased_test_marker_genotypes,
                2
            );
        }
    }
    return founder_haplotypes;
}

int FounderHaplotypes_WithRecombination(FounderHaplotypes founder_haplotypes)
{
    // A recombination event occurred if no compatible phased founder
    // test marker genotypes were found.  (Compatible phased founder
    // genotypes are coded as Triplets irrespective of whether we have
    // 1, 2, or 3 test markers.)
    return TripletHashTable_GetSize(founder_haplotypes->hash_table) == 0;
}

void FounderHaplotypes_Destroy(FounderHaplotypes founder_haplotypes)
{
    TripletHashTable_Destroy(founder_haplotypes->hash_table);
    Memory_Free(founder_haplotypes);
}

TripletHashTableIterator FounderHaplotypes_GetIterator(FounderHaplotypes founder_haplotypes) {
    return TripletHashTableIterator_Create(founder_haplotypes->hash_table);
}

TripletHashTableIterator FounderHaplotypes_GetIterator_1(FounderHaplotypes founder_haplotypes) {
    return TripletHashTableIterator_Create(founder_haplotypes->hash_table_1);
}

TripletHashTableIterator FounderHaplotypes_GetIterator_2(FounderHaplotypes founder_haplotypes) {
    return TripletHashTableIterator_Create(founder_haplotypes->hash_table_2);
}

TripletHashTableIterator FounderHaplotypes_GetIterator_3(FounderHaplotypes founder_haplotypes) {
    return TripletHashTableIterator_Create(founder_haplotypes->hash_table_3);
}

int FounderHaplotypes_GetNumberOfHaplotypes(FounderHaplotypes founder_haplotypes) {
    return founder_haplotypes->number_of_haplotypes;
}

int FounderHaplotypes_GetNumberOfFounders(FounderHaplotypes founder_haplotypes) {
    return founder_haplotypes->number_of_founders;
}

int FounderHaplotypes_GetNumberOfTestMarkers(FounderHaplotypes founder_haplotypes) {
    return founder_haplotypes->number_of_test_markers;
}
