#include "HaplotypeFrequencyGrid.h"
#include "Haplotypes.h"
#include "DoubleVector.h"

static DoubleVector change_pivot_by_delta(int direction, DoubleVector expected, int pivot);
static double sum_of_non_pivot_elements(int pivot, DoubleVector frequencies);
static double pivot_element(int direction, double frequency);
static double non_pivot_element(int direction, double frequency, DoubleVector frequencies, int pivot);

Vector HaplotypeFrequencyGrid_Create(Vector markers)
{
    Vector result = Vector_CreateEmpty();
    DoubleVector expected = Haplotypes_GetExpectedFrequencies(markers);
    Vector_PushBack(result, expected);
    for (int i = 0; i < DoubleVector_GetSize(expected); i++) {
        Vector_PushBack(result, change_pivot_by_delta(+1, expected, i));
        Vector_PushBack(result, change_pivot_by_delta(-1, expected, i));
    }
    return result;
}

static DoubleVector change_pivot_by_delta(int direction, DoubleVector frequencies, int pivot)
{
    DoubleVector result = DoubleVector_Copy(frequencies);
    for (int i = 0; i < DoubleVector_GetSize(frequencies); i++) {
        double freq = DoubleVector_Get(frequencies, i);
        if (i == pivot)
            DoubleVector_Set(result, i, pivot_element(direction, freq));
        else
            DoubleVector_Set(result, i, non_pivot_element(direction, freq, frequencies, pivot));
    }
    return result;
}

static double pivot_element(int direction, double frequency)
{
    double amount = .5 * frequency;
    if (direction == 1 && frequency + amount > 1.0)
        return 1.0;
    return frequency + direction * amount;
}

static double non_pivot_element(int direction, double frequency, DoubleVector frequencies, int pivot)
{
    int opposite_direction = -1 * direction;
    double total_weight = sum_of_non_pivot_elements(pivot, frequencies);
    double pivot_frequency = DoubleVector_Get(frequencies, pivot);
    double amount = .5 * pivot_frequency;
    if (direction == 1 && pivot_frequency + amount > 1.0)
        // New frequency of pivot element is 1.0 so nothing left to
        // distribute over non-pivot elements
        return 0.0;
    return frequency + opposite_direction * amount * (frequency / total_weight);
}

static double sum_of_non_pivot_elements(int pivot, DoubleVector frequencies)
{
    double result = 0;
    for (int i = 0; i < DoubleVector_GetSize(frequencies); i++)
        if (i != pivot)
            result += DoubleVector_Get(frequencies, i);
    return result;
}
