#include "InheritanceDistribution.h"
#include "Memory.h"
#include "npl.h"

struct InheritanceDistributionStruct {
    double **matrix;
    int number_of_mod_positions;
    int number_of_inheritance_vectors;
};

InheritanceDistribution InheritanceDistribution_CreateFromMatrix(double **matrix, int nrow, int ncol)
{
    InheritanceDistribution distro = Memory_Malloc(sizeof(*distro));
    distro->number_of_mod_positions = nrow;
    distro->number_of_inheritance_vectors = ncol;
    distro->matrix = Memory_Malloc(nrow * sizeof(double *));
    for (int i = 0; i < nrow; i++)
        distro->matrix[i] = Memory_Malloc(ncol * sizeof(double));
    for (int i = 0; i < nrow; i++)
        for (int j = 0; j < ncol; j++)
            distro->matrix[i][j] = matrix[i][j];
    return distro;
}

void InheritanceDistribution_Destroy(InheritanceDistribution distro)
{
    if (!distro) {
        return;
    }
    for (int i = 0; i < distro->number_of_mod_positions; i++) {
        Memory_Free(distro->matrix[i]);
    }
    Memory_Free(distro->matrix);
    Memory_Free(distro);
}

double InheritanceDistribution_Get(InheritanceDistribution distro, int position, int inheritance_vector)
{
    return distro->matrix[position][inheritance_vector];
}

int InheritanceDistribution_GetNumberOfInheritanceVectors(InheritanceDistribution distro)
{
    if (distro == NULL && !g_jla_p_value) {  // singleton case, no simulation master run
        return 1;
    }
    return distro->number_of_inheritance_vectors;
}
