#include "Marker.h"
#include "Memory.h"
#include "ModScoreList.h"

#include <float.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define MAX(x, y) ((x) > (y) ? (x) : (y))

struct ModScoreListStruct {
    Vector list;
};

ModScoreList ModScoreList_Create(void)
{
    ModScoreList mod_score_list = Memory_Malloc(sizeof *mod_score_list);
    mod_score_list->list = Vector_Create(1);
    return mod_score_list;;
}

void ModScoreList_Add(ModScoreList mod_score_list, ModScore *mod_score)
{
    Vector_PushBack(mod_score_list->list, mod_score);
}

static void print_float(FILE *fp, double x)
{
    char s[100];
    ModScoreList_Format(s, x);
    fprintf(fp, "%s", s);
}

static double squared(double x)
{
    return x * x;
}

static void print_ld_2x2(FILE *fp, ModScore *mod_score)
{
    // a b
    // c d
    double a = DoubleVector_Get(mod_score->hapfreqs, 0);
    double b = DoubleVector_Get(mod_score->hapfreqs, 1);
    double c = DoubleVector_Get(mod_score->hapfreqs, 2);
    double d = DoubleVector_Get(mod_score->hapfreqs, 3);
    double ab = a+b, cd = c+d, ac = a+c, bd = b+d, N = a+b+c+d;
    double chisq = squared(a - ab*ac/N) / (ab*ac/N)
                 + squared(b - ab*bd/N) / (ab*bd/N)
                 + squared(c - cd*ac/N) / (cd*ac/N)
                 + squared(d - cd*bd/N) / (cd*bd/N);
    const int min_tab_dimen = 2;
    double cramer_v = sqrt(chisq / N / (min_tab_dimen - 1));
    fprintf(fp, " %.3f", cramer_v);
}

static void print_ld_4x2(FILE *fp, ModScore *mod_score)
{
    // a b
    // c d
    // e f
    // g h
    double a = DoubleVector_Get(mod_score->hapfreqs, 0);
    double b = DoubleVector_Get(mod_score->hapfreqs, 1);
    double c = DoubleVector_Get(mod_score->hapfreqs, 2);
    double d = DoubleVector_Get(mod_score->hapfreqs, 3);
    double e = DoubleVector_Get(mod_score->hapfreqs, 4);
    double f = DoubleVector_Get(mod_score->hapfreqs, 5);
    double g = DoubleVector_Get(mod_score->hapfreqs, 6);
    double h = DoubleVector_Get(mod_score->hapfreqs, 7);
    double ab = a+b, cd = c+d, ef = e+f, gh = g+h,
           aceg = a+c+e+g, bdfh = b+d+f+h,
           N = a+b+c+d+e+f+g+h;
    double chisq = squared(a - ab*aceg/N) / (ab*aceg/N)
                 + squared(b - ab*bdfh/N) / (ab*bdfh/N)
                 + squared(c - cd*aceg/N) / (cd*aceg/N)
                 + squared(d - cd*bdfh/N) / (cd*bdfh/N)
                 + squared(e - ef*aceg/N) / (ef*aceg/N)
                 + squared(f - ef*bdfh/N) / (ef*bdfh/N)
                 + squared(g - gh*aceg/N) / (gh*aceg/N)
                 + squared(h - gh*bdfh/N) / (gh*bdfh/N);
    const int min_tab_dimen = 2;
    double cramer_v = sqrt(chisq / N / (min_tab_dimen - 1));
    fprintf(fp, " %.3f", cramer_v);
}

static void print_ld_8x2(FILE *fp, ModScore *mod_score)
{
    // a b
    // c d
    // e f
    // g h
    // i j
    // k l
    // m n
    // o p
    double a = DoubleVector_Get(mod_score->hapfreqs, 0);
    double b = DoubleVector_Get(mod_score->hapfreqs, 1);
    double c = DoubleVector_Get(mod_score->hapfreqs, 2);
    double d = DoubleVector_Get(mod_score->hapfreqs, 3);
    double e = DoubleVector_Get(mod_score->hapfreqs, 4);
    double f = DoubleVector_Get(mod_score->hapfreqs, 5);
    double g = DoubleVector_Get(mod_score->hapfreqs, 6);
    double h = DoubleVector_Get(mod_score->hapfreqs, 7);
    double i = DoubleVector_Get(mod_score->hapfreqs, 8);
    double j = DoubleVector_Get(mod_score->hapfreqs, 9);
    double k = DoubleVector_Get(mod_score->hapfreqs, 10);
    double l = DoubleVector_Get(mod_score->hapfreqs, 11);
    double m = DoubleVector_Get(mod_score->hapfreqs, 12);
    double n = DoubleVector_Get(mod_score->hapfreqs, 13);
    double o = DoubleVector_Get(mod_score->hapfreqs, 14);
    double p = DoubleVector_Get(mod_score->hapfreqs, 15);
    double ab = a+b, cd = c+d, ef = e+f, gh = g+h,
           ij = i+j, kl = k+l, mn = m+n, op = o+p,
           acegikmo = a+c+e+g+i+k+m+o,
           bdfhjlnp = b+d+f+h+j+l+n+p,
           N = a+b+c+d+e+f+g+h+i+j+k+l+m+n+o+p;
    double chisq = squared(a - ab*acegikmo/N) / (ab*acegikmo/N)
                 + squared(b - ab*bdfhjlnp/N) / (ab*bdfhjlnp/N)
                 + squared(c - cd*acegikmo/N) / (cd*acegikmo/N)
                 + squared(d - cd*bdfhjlnp/N) / (cd*bdfhjlnp/N)
                 + squared(e - ef*acegikmo/N) / (ef*acegikmo/N)
                 + squared(f - ef*bdfhjlnp/N) / (ef*bdfhjlnp/N)
                 + squared(g - gh*acegikmo/N) / (gh*acegikmo/N)
                 + squared(h - gh*bdfhjlnp/N) / (gh*bdfhjlnp/N)
                 + squared(i - ij*acegikmo/N) / (ij*acegikmo/N)
                 + squared(j - ij*bdfhjlnp/N) / (ij*bdfhjlnp/N)
                 + squared(k - kl*acegikmo/N) / (kl*acegikmo/N)
                 + squared(l - kl*bdfhjlnp/N) / (kl*bdfhjlnp/N)
                 + squared(m - mn*acegikmo/N) / (mn*acegikmo/N)
                 + squared(n - mn*bdfhjlnp/N) / (mn*bdfhjlnp/N)
                 + squared(o - op*acegikmo/N) / (op*acegikmo/N)
                 + squared(p - op*bdfhjlnp/N) / (op*bdfhjlnp/N);
    const int min_tab_dimen = 2;
    double cramer_v = sqrt(chisq / N / (min_tab_dimen - 1));
    fprintf(fp, " %.3f", cramer_v);
}

static void print_ld(FILE *fp, ModScore *mod_score, int nhapfreqs)
{
    if (nhapfreqs == 2*2) {
        print_ld_2x2(fp, mod_score);
    } else if (nhapfreqs == 4*2) {
        print_ld_4x2(fp, mod_score);
    } else {
        print_ld_8x2(fp, mod_score);
    }
}

void ModScoreList_Write(ModScoreList mod_score_list, FILE *fp, int number_of_test_markers)
{
    Vector rows = mod_score_list->list;
    int nrows = Vector_GetSize(rows);
    if (nrows == 0) {
        fprintf(fp, "No results were produced\n");
        return;
    }
    int has_deltas = ((ModScore *) Vector_Get(rows, 0))->deltas != NULL;

    // Determine width of pos, mod, and marker name columns
    double max_pos = 0, max_mod = 0;
    unsigned *marker_width = Memory_Malloc(number_of_test_markers * sizeof(unsigned));
    for (int i = 0; i < number_of_test_markers; i++) {
        marker_width[i] = strlen("markerX");
    }
    for (int i = 0; i < nrows; i++) {
        ModScore *mod_score = Vector_Get(rows, i);
        max_pos = MAX(max_pos, mod_score->position);
        max_mod = MAX(max_mod, fabs(mod_score->mod_score));
        for (int j = 0; j < number_of_test_markers; j++) {
            const char *marker_name = Marker_GetName(Vector_Get(mod_score->test_markers, j));
            marker_width[j] = MAX(marker_width[j], strlen(marker_name));
        }
    }
    const char *pos_fmt = max_pos <   10 ? "%4.2f"
                        : max_pos <  100 ? "%5.2f"
                        : max_pos < 1000 ? "%6.2f"
                        :                  "%7.2f";
    int pos_width = max_pos <   10 ? strlen("9.99")
                  : max_pos <  100 ? strlen("99.99")
                  : max_pos < 1000 ? strlen("999.99")
                  :                  strlen("9999.99");
    const char *mod_fmt = max_mod <    10 ? "%5.2f"
                        : max_mod <   100 ? "%6.2f"
                        : max_mod <  1000 ? "%7.2f"
                        : max_mod < 10000 ? "%8.2f"
                        :                   "%9.2f"; // extra space for (potential) minus sign
    int mod_width = max_mod <    10 ? strlen(" 9.99")
                  : max_mod <   100 ? strlen(" 99.99")
                  : max_mod <  1000 ? strlen(" 999.99")
                  : max_mod < 10000 ? strlen(" 9999.99")
                  :                   strlen(" 99999.99");

    // Print header: pos marker1..M mod p F0 F1_P F1_M F2 H0..N EM0..N/2
    fprintf(fp, "%*s", pos_width, "pos");
    for (int i = 0; i < number_of_test_markers; i++) {
        fprintf(fp, " %*s%d", marker_width[i]-1, "marker", i+1);
    }
    fprintf(fp, " %*s", mod_width, "mod");
    fprintf(fp, "   %6s", "p");
    fprintf(fp, "     %4s     %4s     %4s     %4s", "F0", "F1_P", "F1_M", "F2");
    int nhapfreqs = 1 << (number_of_test_markers + 1);
    for (int i = 0; i < nhapfreqs; i++) {
        fprintf(fp, "      %sH%d", (i < 10 ? " " : ""), i);
    }
    for (int i = 0; i < nhapfreqs/2; i++) {
        fprintf(fp, "     EMH%d", i);
    }
    fprintf(fp, "    LD");
    if (has_deltas) {
        for (int i = 0; i < nhapfreqs/2; i++) {
            fprintf(fp, "        d%d", i + 1);
        }
    }
    fprintf(fp, "\n");

    // Print rows
    for (int i = 0; i < nrows; i++) {
        ModScore *mod_score = Vector_Get(rows, i);
        fprintf(fp, pos_fmt, mod_score->position);
        for (int j = 0; j < number_of_test_markers; j++) {
            Marker m = Vector_Get(mod_score->test_markers, j);
            fprintf(fp, " %*s", marker_width[j], Marker_GetName(m));
        }
        fprintf(fp, " ");
        fprintf(fp, mod_fmt, mod_score->mod_score);
        fprintf(fp, " ");
        print_float(fp, mod_score->disease_allele_frequency);
        fprintf(fp, " ");
        print_float(fp, MAX(0, DoubleVector_Get(mod_score->penetrances, 0)));
        fprintf(fp, " ");
        print_float(fp, MAX(0, DoubleVector_Get(mod_score->penetrances, 1)));
        fprintf(fp, " ");
        print_float(fp, MAX(0, DoubleVector_Get(mod_score->penetrances, 2)));
        fprintf(fp, " ");
        print_float(fp, MAX(0, DoubleVector_Get(mod_score->penetrances, 3)));
        for (int j = 0; j < nhapfreqs; j++) {
            fprintf(fp, " ");
            print_float(fp, MAX(0, DoubleVector_Get(mod_score->hapfreqs, j)));
        }
        for (int j = 0; j < nhapfreqs/2; j++) {
            fprintf(fp, " ");
            print_float(fp, MAX(0, DoubleVector_Get(mod_score->em_hapfreqs, j)));
        }
        print_ld(fp, mod_score, nhapfreqs);
        if (has_deltas) {
            for (int j = 0; j < nhapfreqs/2; j++) {
                fprintf(fp, " %s", DoubleVector_Get(mod_score->deltas, j) < 0.0 ? "" : " ");
                print_float(fp, DoubleVector_Get(mod_score->deltas, j));
            }
        }
        fprintf(fp, "\n");
    }
}

void ModScoreList_Format(char *dest, double x)
{
    // x       0 <--> eps <--> 1e-6 <--> 1 - 1e-6 <--> 1 - eps <--> 1
    // format    0.00     %8.2e     %.6f          %8.2e        1.00
    const double eps = DBL_EPSILON;
    const char *fmt = "%.6f";
    const char *scientific = "%8.2e";
    if (eps < x && x < 1e-6) {
        fmt = scientific;
    } else if (1 - 1e-6 < x && x < 1 - eps) {
        fmt = scientific;
    }
    sprintf(dest, fmt, x);
}
