#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "IO.h"
#include "Memory.h"
#include "ModelParamsFile.h"

// Legacy includes for print messages to log file
#define INC_LIB
#define INC_SHELL
#include "npl.h"


struct ModelParamsFileStruct {
    ModelParams *models;
    int number_of_models;
};

void check_penetrances_between_0_and_1(ModelParamsFile);
void check_model_parameters_for_validity(ModelParamsFile);

ModelParamsFile ModelParamsFile_Create(const char *filename)
{
    ModelParamsFile f = Memory_Malloc(sizeof *f);
    int max_models = 2;
    f->models = Memory_Malloc(max_models * sizeof(ModelParams));
    FILE *file = fopen(filename, "rb");
    if (!file) {
        sf(ps, "Failed to open file: \"%s\"\n", filename);
        pr();
        exit(1);
    }
    int n = 0;
    char *line;
    while ((line = IO_GetLine(file)) != NULL) {
        if (n >= max_models) {
            max_models *= 2;
            f->models = Memory_Realloc(f->models, max_models * sizeof(ModelParams));
        }
        ModelParams p = ModelParams_Create(line);
        if (p == NULL) {
            sf(ps, "Wrong number of model parameters in line %d of your model file.\nPlease correct the model file!", n);
            pr();
            fclose(file);
            exit(1);
        }
        f->models[n++] = p;
    }
    fclose(file);
    f->number_of_models = n;
    check_model_parameters_for_validity(f);
    return f;
}

void check_model_parameters_for_validity(ModelParamsFile f) {
    check_penetrances_between_0_and_1(f);
}

void check_penetrances_between_0_and_1(ModelParamsFile f) {
    for (int i = 0; i < f->number_of_models; i++) {
        ModelParams p = ModelParamsFile_GetModel(f, i);
        if (!ModelParams_PenetrancesBetween0And1(p)) {
            sf(ps, "Penetrances are not between 0.0 and 1.0 in line %d of your model file!\nPlease correct the model file!", i + 1);
            pr();
            exit(1);
        }
    }
}

int ModelParams_PenetrancesBetween0And1(ModelParams p)
{
    int penetrances_between_0_and_1 = 1;
    if ((p->f0 < 0.0) || (p->f0 > 1.0)) {
        penetrances_between_0_and_1 = 0;
    }
    if (penetrances_between_0_and_1) {
        if ((p->f1_pat < 0.0) || (p->f1_pat > 1.0)) {
            penetrances_between_0_and_1 = 0;
        }
    }
    if (penetrances_between_0_and_1) {
        if ((p->f1_mat < 0.0) || (p->f1_mat > 1.0)) {
            penetrances_between_0_and_1 = 0;
        }
    }
    if (penetrances_between_0_and_1) {
        if ((p->f2 < 0.0) || (p->f2 > 1.0)) {
            penetrances_between_0_and_1 = 0;
        }
    }
    return penetrances_between_0_and_1;
}

int ModelParamsFile_GetNumberOfModels(ModelParamsFile f)
{
    return f->number_of_models;
}

ModelParams ModelParamsFile_GetModel(ModelParamsFile f, int model)
{
    return f->models[model];
}

ModelParams ModelParams_Create(const char *line)
{
    ModelParams p = Memory_Malloc(sizeof *p);
    const char *s = line;
    char *t;
    p->disease_allele_frequency = strtod(s, &t);
    if (s == t) goto FAILED;
    s = t;
    p->f0 = strtod(s, &t);
    if (s == t) goto FAILED;
    s = t;
    p->f1_pat = strtod(s, &t);
    if (s == t) goto FAILED;
    s = t;
    p->f1_mat = strtod(s, &t);
    if (s == t) goto FAILED;
    s = t;
    p->f2 = strtod(s, &t);
    if (s == t) goto FAILED;
    s = t;
    // Fail if there's another number.
    strtod(s, &t);
    if (s != t) goto FAILED;
    return p;
  FAILED:
    Memory_Free(p);
    return NULL;
}
