#include "PenetranceMatrix.h"
#include "Family.h"
#include "InheritanceVector.h"
#include "PhasedGenotypes.h"

PenetranceMatrix PenetranceMatrix_Create(Family family)
{
    PenetranceMatrix penetrance_matrix;
    InheritanceVector inheritance_vector;
    PhasedGenotypes phased_founder_genotypes;
    int number_of_inheritance_vectors = Family_GetNumberOfInheritanceVectors(family);
    int number_of_phased_founder_genotypes = Family_GetNumberOfDiseaseGenotypes(family);

    penetrance_matrix = PenetranceMatrix_CreateEmpty(number_of_inheritance_vectors, number_of_phased_founder_genotypes);
    for (inheritance_vector = 0; (int) inheritance_vector < number_of_inheritance_vectors; inheritance_vector++) {
        Family_SetPhase(family, inheritance_vector);
        for (phased_founder_genotypes = 0; (int) phased_founder_genotypes < number_of_phased_founder_genotypes; phased_founder_genotypes++) {
            PhasedGenotypes phased_nonfounder_genotypes = Family_GetPhasedGenotypesForNonfounders(family, phased_founder_genotypes);
            Penetrances penetrances = Penetrances_Create();
            Vector founders = Family_GetFounders(family);
            Vector nonfounders = Family_GetNonfounders(family);
            Penetrances_GetForFounders(penetrances, founders, phased_founder_genotypes);
            Penetrances_GetForNonfounders(penetrances, nonfounders, phased_nonfounder_genotypes);
            PenetranceMatrix_Set(penetrance_matrix, inheritance_vector, phased_founder_genotypes, penetrances);
        }
    }
    return penetrance_matrix;
}

PenetranceMatrix PenetranceMatrix_CreateWithZero(int number_of_rows, int number_of_columns)
{
    PenetranceMatrix penetrance_matrix = PenetranceMatrix_CreateEmpty(number_of_rows, number_of_columns);
    for (int i = 0; i < number_of_rows; i++)
        for (int j = 0; j < number_of_columns; j++)
            PenetranceMatrix_Set(penetrance_matrix, i, j, Penetrances_Create());
    return penetrance_matrix;
}
