#include "Penetrances.h"
#include "Memory.h"

typedef PenetranceIndex (*Penetrances_GetForPerson)(Person, Genotype);
typedef PenetranceIndex (*Penetrances_GetPenetranceIndexForPerson)(Person, Genotype);

static void Penetrances_GetForPersons(Penetrances_GetPenetranceIndexForPerson, Penetrances, Vector, PhasedGenotypes);

Penetrances Penetrances_Create(void)
{
    Penetrances penetrances = (Penetrances) Memory_Malloc(NUMBER_OF_PENETRANCES * sizeof(*penetrances));
    int i;

    for (i = 0; i < NUMBER_OF_PENETRANCES; i++)
        penetrances[i] = 0;

    return penetrances;
}

void Penetrances_Destroy(Penetrances penetrances)
{
    Memory_Free(penetrances);
}

PenetranceIndex Penetrances_GetForFounder(Person p, Genotype disease_genotype)
{
    int penetrance_index[][4] = {
        // OO     OI        IO     II
        { UF0, UFF1_AVG, UFF1_AVG, UF2 },  // UNAFFECTED
        {  F0,  FF1_AVG,  FF1_AVG,  F2 },  // AFFECTED
    };
    int affection_status = Person_IsAffected(p) ? 1 : 0;

    return penetrance_index[affection_status][disease_genotype];
}

PenetranceIndex Penetrances_GetForNonfounder(Person p, Genotype disease_genotype)
{
    int penetrance_index[][4] = {
        // OO    OI       IO     II
        { UF0, UF1_MAT, UF1_PAT, UF2 },  // UNAFFECTED
        {  F0,  F1_MAT,  F1_PAT,  F2 },  // AFFECTED
    };
    int affection_status = Person_IsAffected(p) ? 1 : 0;

    return penetrance_index[affection_status][disease_genotype];
}

static void Penetrances_GetForPersons(Penetrances_GetPenetranceIndexForPerson get_penetrance_index_for_person,
    Penetrances penetrances, Vector persons, PhasedGenotypes phased_genotypes)
{
    int number_of_persons = Vector_GetSize(persons);
    int i, penetrance_index;
    Genotype disease_genotype;

    for (i = 0; i < number_of_persons; i++) {
        Person person = Vector_Get(persons, i);
        if (!Person_IsAffectionStatusKnown(person))
            continue;
        disease_genotype = PhasedGenotypes_Get(phased_genotypes, i);
        penetrance_index = get_penetrance_index_for_person(person, disease_genotype);
        ++penetrances[penetrance_index];
    }
}

void Penetrances_GetForNonfounders(Penetrances penetrances, Vector nonfounders, PhasedGenotypes phased_nonfounder_genotypes)
{
    Penetrances_GetForPersons(Penetrances_GetForNonfounder, penetrances, nonfounders, phased_nonfounder_genotypes);
}

void Penetrances_GetForFounders(Penetrances penetrances, Vector founders, PhasedGenotypes phased_founder_genotypes)
{
    Penetrances_GetForPersons(Penetrances_GetForFounder, penetrances, founders, phased_founder_genotypes);
}
