#include "Region.h"
#include "Memory.h"
#include "LinkageMarkerFile.h"
#include "UserLinkageMarkerFile.h"
// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"

struct RegionStruct {
    Vector markers;
    Vector linkage_markers;
};

static Region create_region(const char *first_marker_name, const char *last_marker_name, LinkageMarkerFile linkage_file, UserLinkageMarkerFile user_linkage_file);

Region Region_Create(const char *from, const char *to, LinkageMarkerFile linkage_file, UserLinkageMarkerFile user_linkage_file)
{
    if (linkage_file == NULL || user_linkage_file == NULL) {
        return NULL;
    }
    return create_region(from, to, linkage_file, user_linkage_file);
}

void Region_Destroy(Region region)
{
    Vector_Destroy(region->markers);
    Vector_Destroy(region->linkage_markers);
    Memory_Free(region);
}

static Region create_region(const char *from, const char *to, LinkageMarkerFile linkage_file, UserLinkageMarkerFile user_linkage_file)
{
    int first = LinkageMarkerFile_GetMarkerIndex(linkage_file, from);
    int last = LinkageMarkerFile_GetMarkerIndex(linkage_file, to);
    if (first == -1 || last == -1)
        return NULL;
    if (first > last) {
        int tmp = first;
        first = last;
        last = tmp;
    }
    Region region = Memory_Malloc(sizeof(*region));
    region->markers = Vector_CreateEmpty();
    for (int marker_index = first; marker_index <= last; marker_index++) {
        Marker marker = LinkageMarkerFile_GetMarkerByIndex(linkage_file, marker_index);
        if (LinkageMarkerFile_GetNumberOfAlleles(linkage_file, marker_index) != 2) {
            sf(ps, "\nMarker %s is not a SNP and will therefore be omitted in JLA analysis.\n\n", LinkageMarkerFile_GetMarkerName(linkage_file, marker_index));
            pr();
            continue;
        } else
            Vector_PushBack(region->markers, marker);
    }
    region->linkage_markers = Vector_CreateEmpty();
    for (int i = 0; i < Vector_GetSize(region->markers); i++) {
        Marker marker = Vector_Get(region->markers, i);
        if (UserLinkageMarkerFile_Contains(user_linkage_file, Marker_GetName(marker)))
            Vector_PushBack(region->linkage_markers, marker);
    }
    return region;
}

Vector Region_GetMarkers(Region region)
{
    return region->markers;
}

Vector Region_GetLinkageMarkers(Region region)
{
    return region->linkage_markers;
}
