#include "TestMarkerSetGenerator.h"
#include "Memory.h"
#include "LinkageMarkerFile.h"
#include "OrderedSubsetGenerator.h"
#include "UserLinkageMarkerFile.h"
#include <stdlib.h>

// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"

struct TestMarkerSetGeneratorStruct {
    OrderedSubsetGenerator generator;
    double max_marker_distance;
};

TestMarkerSetGenerator TestMarkerSetGenerator_Create(Region region, int number_of_markers_per_set, double max_marker_distance)
{
    TestMarkerSetGenerator generator = Memory_Malloc(sizeof(*generator));
    Vector markers = Region_GetMarkers(region);
    if (Vector_GetSize(markers) < number_of_markers_per_set) {
        sf(ps, "Not enough markers in association region: got %d, need at least %d\n",
            Vector_GetSize(markers), number_of_markers_per_set);
        pr();
        exit(1);
    }
    generator->generator = OrderedSubsetGenerator_Create(markers, number_of_markers_per_set);
    generator->max_marker_distance = max_marker_distance;
    return generator;
}

Vector TestMarkerSetGenerator_Next(TestMarkerSetGenerator generator, LinkageMarkerFile linkage_file, UserLinkageMarkerFile user_linkage_file)
{
    Vector markers;
    while ((markers = OrderedSubsetGenerator_Next(generator->generator)) != NULL) {
        if (LinkageMarkerFile_WithinMaxDistance(linkage_file, markers, generator->max_marker_distance)
                && UserLinkageMarkerFile_CountLinkageMarkers(user_linkage_file, markers) == 0)
        {
            return markers;
        }
    }
    return NULL;
}

int TestMarkerSetGenerator_GetNumberOfSets(TestMarkerSetGenerator generator)
{
    return OrderedSubsetGenerator_GetNumberOfSets(generator->generator);
}
