#include "UserAssociationRegionsFile.h"

#include "IO.h"
#include "LinkageMarkerFile.h"
#include "ParseObject.h"
#include "Region.h"
#include "UserLinkageMarkerFile.h"

#include <stdio.h>
#include <stdlib.h>

// Legacy includes.
#define INC_LIB
#define INC_SHELL
#include "npl.h"


Vector UserAssociationRegionsFile_ReadRegions(
    const char *filename,
    LinkageMarkerFile linkage_file,
    UserLinkageMarkerFile user_linkage_file)
{
    if (linkage_file == NULL || user_linkage_file == NULL) {
        return NULL;
    }
    FILE *fp = fopen(filename, "rb");
    if (!fp) {
        sf(ps, "Failed to open file: \"%s\"\n", filename);
        pr();
        exit(1);
    }
    Vector regions = Vector_CreateEmpty();
    const char *line = NULL;
    int line_number = 0;
    while ((line = IO_GetLine(fp)) != NULL) {
        line_number++;
        ParseObject parser = ParseObject_Create(line);
        Vector marker_names = ParseObject_SplitIntoWords(parser);
        int ok = 1;
        if (Vector_GetSize(marker_names) < 2) {
            ok = 0;
        } else if (Vector_GetSize(marker_names) > 2) {
            // Trailing #-comment after the region is allowed.
            ok = ((const char *) Vector_Get(marker_names, 2))[0] == '#';
        }
        if (!ok) {
            sf(ps, "Error: need exactly 2 markers to denote a region in line %d of %s: %s",
                line_number, filename, line);
            pr();
            exit(1);
        }
        Region region = Region_Create(Vector_Get(marker_names, 0), Vector_Get(marker_names, 1), linkage_file, user_linkage_file);
        if (region == NULL) {
            return NULL;
        }
        Vector_PushBack(regions, region);
        ParseObject_Destroy(parser);
        Vector_Destroy(marker_names);
    }
    fclose(fp);
    free((void *) line);
    return regions;
}
