/* GENEHUNTER - Kruglyak, Daly, Reeve-Daly, and Lander
   Whitehead Institute for Biomedical Research, Copyright 1996
   modified by Konstantin Strauch, Johannes Dietter, Manuel Mattheisen
   and Markus Brugger.    Copyright (c) 1999-2019 */

#include <search.h>
#define INC_LIB
#define INC_SHELL
#define real_eq(x,y) (fabs((x)-(y))<1e-10)
#include "npl.h"
#include "dSFMT.h"
#include <strings.h>
#include <time.h>

void init_disease_info (void); /* in locprep.c - called by the
				  join_liability_classes command */
void fill_marker_structs (char*,int*);        /* in locprep.c */
void check_user_marker_build_distances (void); /* in locprep.c */

/* externs */
int map_function, units, scan_steps;
double scan_increment,scan_increment_m, scan_increment_f, off_end_dist, 
       off_end_dist_m, off_end_dist_f;
char* map_name;
char* map_file;
int pedigrees_calculated;
int discard_unaffected;
int score_function, analysis_type;
int pval_precision;
int show_expected_recs;
int postscript_output;
int haplotype_output;
int haplo_method;
int dump_requirements;
double haplo_ratio;
char *current_filename;
unsigned long *m_pow2;
int ps_scaling;
double constant_ps_scale;
int single_point_mode;
int max_bits_to_analyze;
int skip_large;
bool scan_done; /*** new ***/
bool display_scores; /*** new ***/
int use_letters;
int compute_sharing; /*** new 11/99 ***/
int imprinting;               /*** for imprinting version ***/
int include_untyped;          /*** new in MOD version ***/
int modcalc;                  /*** for MOD version ***/
double lowpos, highpos;       /*** for MOD version ***/
int modliab;                  /*** for MOD version ***/
int penetrance_restriction;   /*** for MOD version ***/
int allfreq_restriction;      /*** for MOD version ***/
double highest_allfreq;       /*** for MOD version ***/
int dimensions;               /*** for MOD version ***/
int saved_models;             /*** for MOD version ***/
int modscore_done;            /*** for MOD version ***/
int changed_modscore_options; /*** for PMOD extension ***/
int long_mod_output;          /*** for MOD version ***/
int maxim_type;               /*** for MOD versions >= 3.0 ***/
bool replicates_error_flag;   /*** for PMOD extension ***/
int sequential_sim;           /*** for PMOD extension ***/
bool replicates_store_result; /*** for PMOD extension ***/
bool replicates_store_pre;    /*** for PMOD extension ***/
int num_replicates;           /*** for PMOD extension ***/
int untyped_founders;          /*** for PMOD extension ***/
int full_marker_inf;          /*** for PMOD extension ***/
int simulate_untyped;         /*** for PMOD extension ***/
int best_position;             /*** for PMOD extension ***/
bool debug_out_pv;            /*** for PMOD extension ***/
int g_random_seed;            /*** for PMOD extension ***/
bool calc_pval;               /*** for PMOD extension ***/
bool show_distribution;       /*** for PMOD extension ***/
int norm_allfreq;             /*** for PMOD extension ***/
char *user_title;   /*** for PostScript output, new 04/2004 (K. Strauch) ***/
int user_title_set;
bool algebra; /* for algebraic calculation mode 10/2013 */
bool mobit_simulate;
bool g_scan_steps_changed;
bool g_use_map;

void init_variables(void)
{
    int i;

    num_markers = 0;
    num_loci = 0;       /*** new ***/
    num_phenotypes = 0; /*** new ***/
    num_covs = 0;       /*added (K. Strauch 04/2005) */
    scan_steps = 5;
    off_end_dist = 0.0;
    off_end_dist_m = 0.0;
    off_end_dist_f = 0.0;
    scan_increment = 0.0;
    scan_increment_m = 0.0;
    scan_increment_f = 0.0; 
    map_function = HALDANE;
    units = CENTIMORGANS;
    array(map_name, 1000, char);
    array(map_file, 1000, char);
    strcpy(map_name,"");

    pedigrees_calculated=0;
    discard_unaffected=FALSE;
    score_function = SCORE_ALL;
    pval_precision = DEFAULT_PRECISION;
    analysis_type = BOTH_ANALYSES;
    show_expected_recs = FALSE;
    postscript_output = FALSE;
    haplotype_output = FALSE;
    dump_requirements = FALSE;
 /* 17 April 2003, haplo method default is now maxprob (as says in paper) */
	 haplo_method = 1;   /* 0=viterbi,  1=max prob. , initialze to 1, max prob*/
    haplo_ratio=1.0;
    array(current_filename, 1000, char);
    strcpy(current_filename,"");

    /*** new ***/
    scan_done = FALSE;
    pair_setting = FIRST_PAIR;
    start_values = CONSTANT;
    means_by_sex = TRUE;
    display_scores = TRUE;
    /*** new ***/

    array(m_pow2, AVAILABLE_BITS, unsigned long);

    m_pow2[0]=1;
    for (i=1; i<AVAILABLE_BITS; i++) m_pow2[i]=2*m_pow2[i-1];

    ps_scaling = FIT_TO_PAGE;
    constant_ps_scale = 5.0;
    single_point_mode = FALSE;
    max_bits_to_analyze = AVAILABLE_BITS - 10;
    skip_large=FALSE;
    use_letters=FALSE;

    /* external TDT variables */
    TDTcount=0;
    one_parent_TDT=FALSE;
    missing_TDT=FALSE;
    dhskip=0;
    qtdt_on=0;
    compute_sharing=FALSE; /*** new 11/99 ***/ /* changed to FALSE for ghi21 */

    ind_sp = 0; /* default for sex specific analysis: none*/
    ind_use_map_marker = 0; /*usage of map markers*/
    imprinting = FALSE;
    include_untyped = TRUE; /* new in MOD ver: use untyped w/known phenotype */
    modcalc = FALSE;        /* save a lot of memory if no MOD is calculated */
    lowpos = -999.0;  /* initialize maximization range of genetic positions */
    highpos = 999.0;  /*   for MOD-score analysis to be the entire map      */
    modliab = 1;      /*   default: optimize liab class 1 with MOD analysis */
    penetrance_restriction = TRUE; /** for MOD: heteroz pen between homoz **/
    allfreq_restriction = TRUE;  /** for MOD: disease allele freq <= value **/
    highest_allfreq = 0.5;  /* MOD: upper bound for the disease allele freq */
    dimensions = 2;   /* for MOD: number of parms optimized at the same time */
    saved_models = -1; /* for MOD: number of models ('single') to be stored */
                       /* default is all initial models (-1) */
    modscore_done = FALSE;   /* for MOD version */
    changed_modscore_options = FALSE; /*** for PMOD extension ***/
    long_mod_output = FALSE; /* for MOD version */
    maxim_type = 1;          /* for MOD version: maximization option, default
                                is 1 = 'standard' */

    pmod_calc_flag = FALSE;      /* PMOD: is pmod function running? */
    replicates_count = 0;        /* PMOD: no replicates simulated yet */
    num_replicates = 0;          /* PMOD: number of replicates for simulations */
    sequential_sim = 0;          /* PMOD: calculate p-value without a sequential simulation method */
    replicates_store_result = FALSE; /* PMOD: store pedigree results for each replicate is off */
    replicates_store_pre = FALSE;    /* PMOD: store pedigree information for each replicate is off */
    replicates_error_flag=FALSE; /* PMOD: no error occured while reading number of replicates */
    full_marker_inf = FALSE;     /* PMOD: randomly simulated alleles for founders */     
    simulate_untyped= FALSE;     /* PMOD: unytped persons in real data will stay untyped in replicates */
    best_position= FALSE;        /* PMOD: p-value calculation only at best position? */
    debug_out_pv = FALSE;        /* PMOD: Debug will not be printed to replicate files */
    g_random_seed = -1;          /* PMOD: Default value for random seed: current system time */
    show_distribution = FALSE;   /* PMOD: Information on the distribution of simulated scores */
    norm_allfreq = FALSE;        /* PMOD: Normalize allfreqs when sum != 1.0 */
    array(user_title, 1000, char);
    strcpy(user_title,"");  /*** for PS output, new 04/2004 (K. Strauch) ***/
    user_title_set = FALSE;
    algebra = TRUE;
    mobit_simulate = FALSE;
    g_scan_steps_changed = FALSE;
    g_use_map = FALSE;
    g_association = 0;
    g_frequencies = 0;
}


command set_scan_increment(void)
{
  double rnum, old_increment;
  char *word;
  int inum, old_steps;

  g_scan_steps_changed = TRUE;

  if (g_association) {
      print("The command to change the scan positions will be ignored for JLA analysis.\n");
      return;
  }

  use_uncrunched_args();
  word = get_temp_string();

  old_increment=scan_increment;
  old_steps=scan_steps;
  if (!stoken(&args,sREQUIRED,word)) {
    if (scan_steps <= 0) {
      sf(ps,"Scanning is currently being done in constant increments of %.1lf cM",
	 scan_increment); pr(); nl();
    } else {
      sf(ps,"Scanning is currently being done in %d equal steps per map interval",
	 scan_steps); pr(); nl();
    }
    
  } else {
    if (modcalc && (scan_done || modscore_done)) {
      print("Pedigrees have already been analyzed.\n");
      print("This command will therefore be ignored.\n");
      return;
    }
    if (!strncasecmp(word,"DISTANCE",4)) {
      if (!rtoken(&args,rREQUIRED,&rnum)) {
	print("error: a cM distance is required");
      } else if (rnum < 0.00) {
	print("error: the distance must be a positive number");
      } else {
	scan_increment = rnum;
	scan_steps = 0;
	sf(ps,
	   "Scanning will now be done in constant increments of %.1lf cM",
	   scan_increment); pr();
	nl();
      }
    } else if (!strncasecmp(word,"STEP",3)) {
      if (!itoken(&args,iREQUIRED,&inum)) {
	print("error: an integer number of intervals is required");
      } else if (inum < 1) {
	print("error: the number of intervals must be a positive number");
      } else {
	scan_steps = inum;
	scan_increment = 1.0;
	sf(ps,
	   "Scanning will be done in %d steps per map interval",
	   scan_steps); pr();
	nl();
      }
    }
    else {
      print("error: the first argument must be 'distance' or 'steps'\n");
    }
    if ((scan_steps != old_steps || scan_increment != old_increment) && scan_done) {
      scan_done = FALSE;
      sf(ps, "You will need to scan pedigrees again for this to take effect."); pr(); 
      nl();
    }
  }
}


command set_scan_increment_m(void)
{ 
  double rnum_m, old_increment;
  char *word;
  int inum, old_steps;

  use_uncrunched_args();
  word = get_temp_string();
  
  old_increment=scan_increment;
  old_steps=scan_steps;
  if (!stoken(&args,sREQUIRED,word)) {
    if (scan_steps <= 0) {
      sf(ps,"Scanning is currently being done in constant increments of %.1lf cM",
	 scan_increment_m); pr(); nl();
    } else {
      sf(ps,"Scanning is currently being done in %d equal steps per map interval",
	 scan_steps); pr(); nl();
    }
    
  } else {
    if (modcalc && (scan_done || modscore_done)) {
      print("Pedigrees have already been analyzed.\n");
      print("This command will therefore be ignored.\n");
      return;
    }
    if (!strncasecmp(word,"DISTANCE",4)) {
      if (!rtoken(&args, rREQUIRED, &rnum_m)) {
	print("error: a cM distance is required");
      } else if (rnum_m < 0.00) {
	print("error: the distance must be a positive number");
      } else {
	scan_increment_m = rnum_m;
	scan_steps = 0;
	sf(ps,
	   "Scanning will now be done in constant increments of %.1lf cM",
	   scan_increment_m); pr();
	nl();
      }
    } else if (!strncasecmp(word,"STEP",3)) {
      if (!itoken(&args,iREQUIRED,&inum)) {
	print("error: an integer number of intervals is required");
      } else if (inum < 1) {
	print("error: the number of intervals must be a positive number");
      } else {
	scan_steps = inum;
	scan_increment = 1.0;
	sf(ps,
	   "Scanning will be done in %d steps per map interval",
	   scan_steps); pr();
	nl();
      }
    }
    else {
      print("error: the first argument must be 'distance' or 'steps'\n");
    }
    if ((scan_steps != old_steps || scan_increment != old_increment) && scan_done) {
      scan_done = FALSE;
      sf(ps, "You will need to scan pedigrees again for this to take effect."); pr(); 
      nl();
    }
  }
}

command set_scan_increment_f(void)
{ 
  double rnum_f, old_increment;
  char *word;
  int inum, old_steps;

  use_uncrunched_args();
  word = get_temp_string();
  
  old_increment=scan_increment;
  old_steps=scan_steps;
  if (!stoken(&args,sREQUIRED,word)) {
    if (scan_steps <= 0) {
      sf(ps,"Scanning is currently being done in constant increments of %.1lf cM",
	 scan_increment_f); pr(); nl();
    } else {
      sf(ps,"Scanning is currently being done in %d equal steps per map interval",
	 scan_steps); pr(); nl();
    }
    
  } else {
    if (modcalc && (scan_done || modscore_done)) {
      print("Pedigrees have already been analyzed.\n");
      print("This command will therefore be ignored.\n");
      return;
    }
    if (!strncasecmp(word,"DISTANCE",4)) {
      if (!rtoken(&args, rREQUIRED, &rnum_f)) {
	print("error: a cM distance is required");
      } else if (rnum_f < 0.00) {
	print("error: the distance must be a positive number");
      } else {
	scan_increment_f = rnum_f;
	scan_steps = 0;
	sf(ps,
	   "Scanning will now be done in constant increments of %.1lf cM",
	   scan_increment_f); pr();
	nl();
      }
    } else if (!strncasecmp(word,"STEP",3)) {
      if (!itoken(&args,iREQUIRED,&inum)) {
	print("error: an integer number of intervals is required");
      } else if (inum < 1) {
	print("error: the number of intervals must be a positive number");
      } else {
	scan_steps = inum;
	scan_increment = 1.0;
	sf(ps,
	   "Scanning will be done in %d steps per map interval",
	   scan_steps); pr();
	nl();
      }
    }
    else {
      print("error: the first argument must be 'distance' or 'steps'\n");
    }
    if ((scan_steps != old_steps || scan_increment != old_increment) && scan_done) {
      scan_done = FALSE;
      sf(ps, "You will need to scan pedigrees again for this to take effect."); pr(); 
      nl();
    }
  }
}


/* map_order[] must be fewer than MAX_LOCI_IN_MAP (100) */

void print_map_order(void);
int lookup_locus(char*);

command set_use_map()
{
    char *loc1, *word;
    int new_order[MAXLOCI], num_new_order;
    int i, this_one, use_cM;
    volatile int changed=0;
    double rnum, new_distance_m[MAXLOCI], new_distance_f[MAXLOCI];
    MARKER *p, *q;

    if (g_association) {
        print("The 'use' command will be ignored in the case of JLA analysis.\n");
        return;
    }

    g_use_map = TRUE;

    use_uncrunched_args();

    strcpy(self_delimiting,"-");

    run {
      loc1 = get_temp_string();
      word = get_temp_string();

      if (!stoken(&args,sREQUIRED,loc1)) { /* no args */ }
      else {
        if (modcalc && (scan_done || modscore_done)) {
	  error("cannot change map with 'use' if pedigrees have already been\n       analyzed with 'modcalc' activated");
        }
        if (ind_use_map_marker != 0){
          sf(ps,"cannot change map with 'use' if a map file has been read");
          error(ps);
          /* ind_use_map_marker = 0; */
          /*sf(ps,"switching from map defined distances to user defined distances.\n"),pr();
          sf(ps,"In order to use map defined distances, restart program.\n"),pr(); */

        }

	new_order[0] = lookup_locus(loc1);
	num_new_order = 1;

	if (!stoken(&args,sREQUIRED,word)) {
	  /* only one marker listed, two-point analysis */
	  /* skip to the end */

	} else {
	  /* we will either get markers or '-' indicating ranges,
	     loop over all args until done */
	  while (1) {

            if (ind_sp==1){
              if (!rtoken(&word,rREQUIRED,&rnum)) {
                  error("non-number where distance was expected");
              } else {
                  new_distance_f[num_new_order-1]= rnum;
              }

              if (!rtoken(&args,rREQUIRED,&rnum)) {
                  error("non-number where distance was expected");
              } else {
                  new_distance_m[num_new_order-1]= rnum;
              }
            }else if(ind_sp==0){
              if (!rtoken(&word,rREQUIRED,&rnum)) {
                error("non-number where distance was expected");
              }else{
                new_distance_m[num_new_order-1]= rnum;
                new_distance_f[num_new_order-1]= rnum;
              }
            }

            if(!stoken(&args,sREQUIRED,word)) {
	        error("map must end with a marker");
	    } else {
	        this_one = lookup_locus(word);
		new_order[num_new_order] = this_one;
		num_new_order++;
	    }
	    if (!stoken(&args,sREQUIRED,word)) break;
	  }
	}
	
	/* check whether rec-fracs or cM are used */
	use_cM=FALSE;
	for (i=0; i<num_new_order-1; i++) 
	  if (new_distance_m[i] >= 0.50 || new_distance_f[i] >= 0.50)
	    { use_cM=TRUE; break; }

	if (!use_cM) {
           for (i=0; i<num_new_order-1; i++){ 
	      new_distance_m[i] = rec_to_dist(new_distance_m[i]);
	      new_distance_f[i] = rec_to_dist(new_distance_f[i]);
           } 
	}

	/* update the global map order */
	for (i=0; i<num_new_order-1; i++) {
	  map_order[i] = new_order[i];
	  map_distance_m[i] = new_distance_m[i];
	  map_distance_f[i] = new_distance_f[i];
          map_distance[i] = 0.5*(map_distance_f[i] + map_distance_m[i]);
	}
	map_order[i] = new_order[i];
	num_in_map_order = num_new_order;

	if (num_in_map_order > 1) {
	  if (!use_cM) {
	    print("Map distances have been interpreted as recombination fractions.\n");
	    print("In the following, genetic distances will be given in cM units.\n");
	  } else print("Map distances have been interpreted as genetic distances in cM units.\n");
	}

	hdestroy();   /* delete hash table */
	if (first_marker) for (p = first_marker; p != NULL; p = q) {
	  q = p->next;
	  free(p);
	}
	first_marker = NULL;  last_marker = NULL;

	fill_marker_structs(map_name, &nr_marker);
	check_user_marker_build_distances();

	changed = 1;
      }
	
      if (num_markers > 0) {
	if (num_in_map_order > 0) {
	  print_map_order();
	  if (changed) print("The selected markers and their positions can be found in the file\n'user_markers_in_map'.\n");
	} else {
	  print("No marker set has been selected for the analysis yet.\n");
	  print("Please specify markers and distances with the 'use' command.\n");
	}
      } else print
	     ("no data is currently loaded - use 'load markers <filename>'\n");
      if (changed && scan_done) {
	scan_done = FALSE;
	print("You will need to scan pedigrees again for this to take effect.\n");
      }

      /* reset this to previous value */
      strcpy(self_delimiting,"");

    } on_error {
        /* interrupt sent when lookup_locus fails */
        print("map not changed\n");
	strcpy(self_delimiting,"");
	if (msg != INTERRUPT && msg != SOFTABORT) relay_messages;
    }
}

int lookup_locus(char *str)
{
    int i, this_marker;
    
    this_marker = -1;

    for (i=0; i<num_markers; i++) {
        if (!strcasecmp(locus[i].name,str)) { this_marker=i; break; }
    }

    if (this_marker == -1) {
        if (!itoken(&str,iREQUIRED,&this_marker)) {
	    sf(ps,"marker %s not found in data set",str);
	    error(ps);
	}
	if ((this_marker < 1) || (this_marker > num_markers)) {
	  sf(ps,"marker %d not found in data set",this_marker);
	  error(ps);
	}
	this_marker--; /* 1 is locus[0] */
    }
    return(this_marker);
}


void print_map_order(void)
{
    int i;

    sf(ps,"Current map (%d markers):\n",num_in_map_order); pr();
    if (ind_sp==0){
      for (i=0; i<num_in_map_order-1; i++) {
          sf(ps,"%s %.4lf ",locus[map_order[i]].name,map_distance[i]); pr();
      }
      sf(ps,"%s\n",locus[map_order[i]].name); pr();
    }else if (ind_sp==1){
      for (i=0; i<num_in_map_order-1; i++) {
        sf(ps,"%s %.4lf %.4lf ",locus[map_order[i]].name,map_distance_f[i],map_distance_m[i]); pr();
      }
      sf(ps,"%s\n",locus[map_order[i]].name); pr();
    }  
}
    

command set_units(void)
{
    char type[TOKLEN+1];
    int old_units;

    if (g_association) {
        print("The 'units' command will be ignored in the case of JLA analysis.\n"
              "Test marker positions are printed in cM units.\n");
        return;
    }

    get_one_arg(stoken,"",type);

    old_units=units;
    if (nullstr(type)) {
	if (units==RECFRACS)
	  print("the 'units' are currently recombination-fractions\n");
	else if (map_function==HALDANE)
	  print("the 'units' are currently (Haldane) centimorgans\n");
	else if (map_function==KOSAMBI)
	  print("the 'units' are currently (Kosambi) centimorgans\n");
	else send(CRASH);

    } else { /* have args */
      if (modcalc && (scan_done || modscore_done)) {
        print("Pedigrees have already been analyzed.\n");
        print("This command will therefore be ignored.\n");
        return;
      }
      if (matches(type,"recombination fractions") || 
	    matches(type,"rec-fracs") ||
	    matches(type,"recombination-fractions") ||
	    matches(type,"rec fracs") ||
	    matches(type,"rf")) units=RECFRACS;
	else if (matches(type,"centimorgans") || 
		 matches(type,"cm")) units=CENTIMORGANS;
	else usage_error(1);

	if (units==RECFRACS)
	  print("the 'units' are now recombination-fractions.\n");
	else if (map_function==HALDANE)
	  print("the 'units' are now set to (Haldane) centimorgans.\n");
	else if (map_function==KOSAMBI)
	  print("the 'units' are now set to (Kosambi) centimorgans.\n");
	else send(CRASH);
	if (units != old_units && scan_done) {
	  scan_done = FALSE;
	  print("You will need to scan pedigrees again for this to take effect.\n");
	}
      }
}

command set_map_function(void)
{
    char type[TOKLEN+1];
    int old_map;

    get_one_arg(stoken,"",type);
    
    old_map=map_function;
    if (nullstr(type)) {
        sf(ps,"The %s map function is currently in use.\n",
	   (map_function==HALDANE)?"Haldane":"Kosambi"); pr();
    } else {
      if (modcalc && (scan_done || modscore_done)) {
        print("Pedigrees have already been analyzed.\n");
        print("This command will therefore be ignored.\n");
        return;
      }
      if (matches(type,"haldane")) map_function=HALDANE;
	else if(matches(type,"kosambi")) map_function=KOSAMBI;
	else usage_error(1);
      
	sf(ps,"The %s map function is now in use.\n",
	   (map_function==HALDANE)?"Haldane":"Kosambi"); pr();
	if (map_function != old_map && scan_done) {
	  scan_done=FALSE;
	  print("You will need to scan pedigrees again for this to take effect.\n");
	}
    }
}

command set_off_end_distance(void)
{
    double new_val, old_val;

    g_scan_steps_changed = TRUE;

    if (g_association) {
        print("The command to change the off end positions will be ignored in the case of JLA analysis.\n");
        return;
    }

    get_one_arg(rtoken,-999.0,&new_val);

    old_val=off_end_dist;
    if (new_val < 0.0) {
        sf(ps,"Scanning will be done %.1lf cM beyond the ends of the map\n",
	   off_end_dist); pr();
	   off_end_dist_m = off_end_dist;
	   off_end_dist_f = off_end_dist;
    } else {
      if (modcalc && (scan_done || modscore_done)) {
        print("Pedigrees have already been analyzed.\n");
        print("This command will therefore be ignored.\n");
        return;
      }
      if (new_val < 0.50) off_end_dist = rec_to_dist(new_val);
        else off_end_dist = new_val;
	sf(ps,"Scanning will now be done %.1lf cM beyond the ends of the map",
	   off_end_dist); pr(); nl();
	   off_end_dist_m = off_end_dist;
	   off_end_dist_f = off_end_dist;
	if (off_end_dist != old_val && scan_done) {
	  scan_done=FALSE;
	  print("You will need to scan pedigrees again for this to take effect.\n");
	}
    }
}

command set_off_end_distance_m(void)
{
    double new_val, old_val;

    get_one_arg(rtoken,-999.0,&new_val);

    old_val=off_end_dist_m;
    if (new_val < 0.0) {
        sf(ps,"Scanning will be done %.1lf cM beyond the ends of the map\n",
	   off_end_dist_m); pr();
    } else {
      if (modcalc && (scan_done || modscore_done)) {
        print("Pedigrees have already been analyzed.\n");
        print("This command will therefore be ignored.\n");
        return;
      }
      if (new_val < 0.50) off_end_dist_m = rec_to_dist(new_val);
        else off_end_dist_m = new_val;
	sf(ps,"Scanning will now be done %.1lf cM beyond the ends of the map",
	   off_end_dist); pr(); nl();
	if (off_end_dist_m != old_val && scan_done) {
	  scan_done=FALSE;
	  print("You will need to scan pedigrees again for this to take effect.\n");
	}
    }
}

command set_off_end_distance_f(void)
{
    double new_val, old_val;

    get_one_arg(rtoken,-999.0,&new_val);

    old_val=off_end_dist_f;
    if (new_val < 0.0) {
        sf(ps,"Scanning will be done %.1lf cM beyond the ends of the map\n",
	   off_end_dist_f); pr();
    } else {
      if (modcalc && (scan_done || modscore_done)) {
        print("Pedigrees have already been analyzed.\n");
        print("This command will therefore be ignored.\n");
        return;
      }
      if (new_val < 0.50) off_end_dist_f = rec_to_dist(new_val);
        else off_end_dist_f = new_val;
	sf(ps,"Scanning will now be done %.1lf cM beyond the ends of the map",
	   off_end_dist); pr(); nl();
	if (off_end_dist_f != old_val && scan_done) {
	  scan_done=FALSE;
	  print("You will need to scan pedigrees again for this to take effect.\n");
	}
    }
}



command show_total_stat(void)
{
    int i,j,total_pval_max,this_score,max_num_scores;
    double total,rawtotal,infototal,entropytotal,pval_total,pvent_total;
    double *apmstat,*infosum,*varinfo,*var_scale;
    /* double alpha_min, varsum; */
    double h_alpha,h_lod,*poslod,*new_score_list;
    double *prior_dist, this_max, total_max;
    /* double *post_dist; */
    double chisq_prob(double,double), fixed_alpha;
    FILE *fp;
    char *lod_str, *pos_str, type[TOKLEN+1], *filename;
    void linkhet(double*,int,double*,double*,int);

    if (pedigrees_calculated == 0)
      error("no pedigree file has been analyzed with the 'scan' command yet");

    if (replicates_count > 0) {
      print("Since p-values have already been calculated, this command will be ignored.\n");
      return;
    }

    get_arg(stoken,"hom",type);
    get_arg(rtoken,-999.0,&fixed_alpha);
    if (fixed_alpha > 1.0) error("fixed alpha must be between 0.0 and 1.0");

    array(apmstat, num_w_pos, double);
    array(infosum, num_w_pos, double);
    array(varinfo, num_w_pos, double);
    array(poslod, pedigrees_calculated, double);
    array(var_scale, pedigrees_calculated, double);

    lod_str = get_temp_string();
    pos_str = get_temp_string();

    /* New variance scales for better p-value calculation */
    /* varsum = 0.0; */
    for (i=0; i<pedigrees_calculated; i++) {
      if (null_variance[i] < .0001) null_variance[i]=.0001;
      var_scale[i] = 1.0/sqrt(null_variance[i]);
      /* varsum += var_scale[i]; */
    }
    /* for (i=0; i<pedigrees_calculated; i++)  var_scale[i] /= varsum; */

    /* get maximum raw score for allocation/convolution */
    total_max=0.0; max_num_scores=1;
    for (i=0; i<pedigrees_calculated; i++) {
      this_max=0.0;
      for (j=0; j<num_pscores[i]; j++) {
	  if (pscore_list[i][j] > this_max) this_max=pscore_list[i][j];
      }
      total_max += (this_max*var_scale[i]);
      if (num_pscores[i] > max_num_scores) max_num_scores=num_pscores[i];
    }
    array(new_score_list, max_num_scores, double);
    init_convolution(total_max);
    total_pval_max=(int)(total_max*pval_precision)+1000;
    array(prior_dist,total_pval_max,double);

    if (analysis_type != LOD_ANALYSIS) {
      print("\nTotalling pedigrees"); flush();
      for (j=0; j<total_pval_max; j++) prior_dist[j]=0.0;
      for (i=0; i<num_pscores[0]; i++) {
        this_score = (int) ((pscore_list[0][i]*pval_precision*var_scale[0])+.5);
	prior_dist[this_score] = prior_pscore[0][i];
      }
      for (i=1; i<pedigrees_calculated; i++) {
	print(":"); flush();
	for (j=0; j<num_pscores[i]; j++) new_score_list[j]=(pscore_list[i][j]*var_scale[i]);
        convolve(prior_dist, prior_pscore[i], new_score_list, num_pscores[i]);
      }
      print("\n");
    }
    
    if (ind_sp==1){

    if (analysis_type == NPL_ANALYSIS) 
      print("pos.(avg. female  male)   NPL_score   p-value     information\n");
    else if (analysis_type == LOD_ANALYSIS) {
      if (modscore_done && modcalc==2) {   /* only with modcalc single */
	if (!strncasecmp(type,"het",3))
	  print("pos.(avg. female  male)  MOD_score (alpha,  HMOD)     information\n");
	else
	  print("pos.(avg. female  male)  MOD_score   information\n");
      } else {
	if (!strncasecmp(type,"het",3))
	  print("pos.(avg. female  male)  LOD_score (alpha,  HLOD)     information\n");
	else
	  print("pos.(avg. female  male)  LOD_score   information\n");
      }
    }
    else { /* BOTH */
      if (modscore_done && modcalc==2) {   /* only with modcalc single */
	if (!strncasecmp(type,"het",3))
          print("pos.(avg. female  male)  MOD_score (alpha,  HMOD)    NPL_score  p-value     information\n");
	else
          print("pos.(avg. female  male)  MOD_score    NPL_score  p-value     information\n");
      } else {
	if (!strncasecmp(type,"het",3)) 
          print("pos.(avg. female  male)  LOD_score  (alpha,  HLOD)   NPL_score p-value     information\n");
	else 
          print("pos.(avg. female  male)  LOD_score   NPL_score p-value     information\n");
      }
    }
    
    }else if (ind_sp==0){
    
    if (analysis_type == NPL_ANALYSIS) 
      print("position  NPL_score   p-value     information\n");
    else if (analysis_type == LOD_ANALYSIS) {
      if (modscore_done && modcalc==2) {   /* only with modcalc single */
	if (!strncasecmp(type,"het",3))
	  print("position MOD_score  (alpha, HMOD)     information\n");
	else
	  print("position MOD_score   information\n");
      } else {
	if (!strncasecmp(type,"het",3))
	  print("position LOD_score  (alpha, HLOD)     information\n");
	else
	  print("position LOD_score   information\n");
      }
    }
    else { /* BOTH */
      if (modscore_done && modcalc==2) {   /* only with modcalc single */
	if (!strncasecmp(type,"het",3))
	  print("position MOD_score  (alpha, HMOD)    NPL_score  p-value     information\n");
	else
	  print("position MOD_score   NPL_score  p-value     information\n");
      } else {
	if (!strncasecmp(type,"het",3)) 
	  print("position LOD_score  (alpha, HLOD)    NPL_score  p-value     information\n");
	else 
	  print("position LOD_score   NPL_score  p-value     information\n");
      }
    }   
    
    }

    for (i=0; i<num_w_pos; i++) {

        /* post_dist stuff can be removed if we decide just to go with 
	   examining the mean vs. the prior_dist for our significance */

        total=0.0; rawtotal=0.0; infototal=0.0; 
	entropytotal=0.0; pval_total=1.0; pvent_total=0.0;
	wpos[i]=0.5*(wpos_m[i]+wpos_f[i]);
	for (j=0; j<pedigrees_calculated; j++) {
	    total += apm_stat[j][i];
	    rawtotal += (raw_stat[j][i]*var_scale[j]);
	    infototal += apm_info[j][i];
	    entropytotal += apm_entropy[j][i];
	    pvent_total += pvec_entropy[j][i];
	    poslod[j] = sub_lods[j][i];
	}

	if (!strncasecmp(type,"het",3)) {
	  if (fixed_alpha > 0.0) {
	    h_alpha = fixed_alpha;
	    linkhet(poslod,pedigrees_calculated,&h_alpha,&h_lod,TRUE);
	  } else {
	    linkhet(poslod,pedigrees_calculated,&h_alpha,&h_lod,FALSE);
	  }
	}

	pval_total=0.0;
	for (j=(int)(rawtotal*pval_precision); j<=total_pval_max; j++)
	  pval_total += prior_dist[j];

#ifdef DEBUG_DIST
	FILE *fpdebug = open_file("debug.dists","w");
	for (j=0; j<=total_pval_max; j++) {
	  if (prior_dist[j] > 0.0) {
	    sf(ps,"%8d   %.10lf\n",j,prior_dist[j]); 
	    fpr(fpdebug);
	  }
	}
	close_file(fpdebug);
#endif

	apmstat[i] = total/sqrt((double)pedigrees_calculated);
	varinfo[i] = infototal/(double)pedigrees_calculated;
	if (varinfo[i] < 0.00) varinfo[i]=0.00;
	infosum[i] = pvent_total/(double)pedigrees_calculated;

	if (total_lod[i] < -9000.0) strcpy(lod_str," -INFINITY");
	else sf(lod_str,"%10.6lf",total_lod[i]); 
        
	
	if (ind_sp==1){ 
	  if (!single_point_mode) sf(pos_str,"%7.2lf %7.2lf %7.2lf" ,0.5*(wpos_m[i] + wpos_f[i]), wpos_f[i], wpos_m[i]);  
	  else sf(pos_str,"%s",locus[i].name); 
	}else if(ind_sp==0){
	  if (!single_point_mode) sf(pos_str,"%7.2lf" ,wpos[i]);  
	  else sf(pos_str,"%s",locus[i].name); 
	}
	
	/* else sf(pos_str,"loc%-4d",i+1); */

	if (analysis_type == NPL_ANALYSIS) {
	  if (pval_total < .00001) {
	    sf(ps,"%s %9.5lf      %8.4g   %9.6lf\n",
	       pos_str, apmstat[i], pval_total, 
	       pvent_total/((double)pedigrees_calculated));
	    pr();
	  } else {
	    sf(ps,"%s %9.5lf     %8.6lf   %9.6lf\n",
	       pos_str, apmstat[i], pval_total, 
	       pvent_total/((double)pedigrees_calculated));
	    pr();
	  }
	} else if (analysis_type==LOD_ANALYSIS) {
	    if (!strncasecmp(type,"het",3))
	      sf(ps,"%s %s (%6.4lf,%7.4lf)    %9.6lf\n",
		 pos_str, lod_str, h_alpha, h_lod,  
		 pvent_total/((double)pedigrees_calculated));
	    else
	      sf(ps,"%s %s   %9.6lf\n",
		 pos_str, lod_str,  
		 pvent_total/((double)pedigrees_calculated));
	    pr(); 
	} else { /* BOTH */
	  if (pval_total < .00001) {
	    if (!strncasecmp(type,"het",3))
	      sf(ps,"%s %s (%6.4lf,%7.4lf)  %9.5lf %8.4g     %9.6lf\n",
		 pos_str, lod_str, h_alpha, h_lod, apmstat[i], 
		 pval_total, pvent_total/((double)pedigrees_calculated));
	    else
	      sf(ps,"%s %s   %9.5lf %8.4g     %9.6lf\n",
		 pos_str, lod_str, apmstat[i], 
		 pval_total, pvent_total/((double)pedigrees_calculated));
	    pr(); 
	  } else {
	    if (!strncasecmp(type,"het",3))
	      sf(ps,"%s %s (%6.4lf,%7.4lf)  %9.5lf %8.6lf     %9.6lf\n",
		 pos_str, lod_str, h_alpha, h_lod, apmstat[i], 
		 pval_total, pvent_total/((double)pedigrees_calculated));
	    else
	      sf(ps,"%s %s   %9.5lf %8.6lf     %9.6lf\n",
		 pos_str, lod_str, apmstat[i], 
		 pval_total, pvent_total/((double)pedigrees_calculated));
	    pr(); 
	  }
	}

	/**********
	if (pval_total < .00001 || alpha_min < .00001) {
	  sf(ps,"%7.2lf %s %9.6lf %9.6lf %9.6lf %9.6lf %9.4g %9.4g\n",
	     wpos[i], lod_str, apmstat[i], 
	     infototal/((double)pedigrees_calculated), 
	     entropytotal/((double)pedigrees_calculated), 
	     pvent_total/((double)pedigrees_calculated),
	     pval_total, alpha_min); pr();
	} else {
	  sf(ps,"%7.2lf %s %9.6lf %9.6lf %9.6lf %9.6lf %9.6lf %9.6lf\n",
	     wpos[i], lod_str, apmstat[i],
	     infototal/((double)pedigrees_calculated), 
	     entropytotal/((double)pedigrees_calculated), 
	     pvent_total/((double)pedigrees_calculated),
	     pval_total, alpha_min); pr();
	}
	**********/
    }

    if (num_w_pos > 1 && postscript_output && !single_point_mode) {

      if (analysis_type != LOD_ANALYSIS) {
	filename = get_temp_string();
	strcpy(filename,"npl_plot.ps");
	fp = open_out_file(filename,"postscript plot");
	if (score_function == SCORE_PAIRS)
	  draw_graph(fp,wpos,apmstat,num_w_pos,"NPL plot","Z-pairs");
	else if (score_function == SCORE_ALL)
	  draw_graph(fp,wpos,apmstat,num_w_pos,"NPL plot","Z-all");
	else if (score_function == SCORE_ONE)
	  draw_graph(fp,wpos,apmstat,num_w_pos,"NPL plot","Z-one");
	else if (score_function == SCORE_HOM)
	  draw_graph(fp,wpos,apmstat,num_w_pos,"NPL plot","Z-hom");
	else error("undefined NPL score function selected");
      }

      if (analysis_type != NPL_ANALYSIS) {
	filename = get_temp_string();
	if (modscore_done && modcalc==2) strcpy(filename,"mod_plot.ps");
	else strcpy(filename,"lod_plot.ps");
	fp = open_out_file(filename,"postscript plot");
	if (modscore_done && modcalc==2)   /* only with modcalc single */
	  draw_graph(fp,wpos,total_lod,num_w_pos,"MOD plot","MOD");
	else draw_graph(fp,wpos,total_lod,num_w_pos,"LOD plot","LOD");
      }

      filename = get_temp_string();
      strcpy(filename,"info_content.ps");
      fp = open_out_file(filename,"postscript plot");
      draw_graph(fp,wpos,infosum,num_w_pos,"Information content","Information Content");

      /*****
      fp = open_out_file("varinfo.ps","postscript plot");
      draw_graph(fp,wpos,varinfo,num_w_pos,"variance info","variance info");
      *****/
    }

    free_convolution();

    /** unarray(post_dist, double); **/
    unarray(prior_dist, double);

    unarray(poslod, double);
    unarray(varinfo, double);
    unarray(infosum, double);
    unarray(apmstat, double);
}


command set_discard(void)
{
    char type[TOKLEN+1];
    bool old_discarded;

    get_one_arg(stoken,"",type);

    old_discarded=discard_unaffected;
    if (nullstr(type)) {
        sf(ps,"Unaffected children are currently %s.\n",
	   (discard_unaffected)?"discarded":"used"); pr();
    } else {
      if (modcalc && (scan_done || modscore_done)) {
        print("Pedigrees have already been analyzed.\n");
        print("This command will therefore be ignored.\n");
        return;
      }
      if (matches(type,"on")) discard_unaffected=TRUE;
	else if(matches(type,"off")) discard_unaffected=FALSE;
	else usage_error(1);
      
        sf(ps,"Unaffected children are now %s.\n",
	   (discard_unaffected)?"discarded":"used"); pr();
	if (discard_unaffected != old_discarded && scan_done) {
	  scan_done=FALSE;
	  print("You will need to scan pedigrees again for this to take effect.\n");	  
	}
    }
}

command set_skip_large(void)
{
    char type[TOKLEN+1];
    bool old_val;

    get_one_arg(stoken,"",type);

    old_val=skip_large;
    if (nullstr(type)) {
        sf(ps,"Large pedigrees are currently %s.\n",
	   (skip_large)?"skipped":"used but trimmed"); pr();
    } else {
      if (modcalc && (scan_done || modscore_done)) {
        print("Pedigrees have already been analyzed.\n");
        print("This command will therefore be ignored.\n");
        return;
      }
      if (matches(type,"on")) skip_large=TRUE;
	else if(matches(type,"off")) skip_large=FALSE;
	else usage_error(1);
      
        sf(ps,"Large pedigrees are now %s.\n",
	   (skip_large)?"skipped":"used but trimmed"); pr();
	if (skip_large != old_val && scan_done) {
	  scan_done=FALSE;
	  print("You will need to scan pedigrees again for this to take effect.\n");
	}  
    }
}

command set_letters(void)
{
    char type[TOKLEN+1];

    get_one_arg(stoken,"",type);

    if (nullstr(type)) {
        sf(ps,"Haplotypes are currently displayed as %s.\n",
	   (use_letters)?"inferred letters":"raw numeric alleles"); pr();
    } else {
        if (matches(type,"on")) use_letters=TRUE;
	else if(matches(type,"off")) use_letters=FALSE;
	else usage_error(1);
      
        sf(ps,"Haplotpes are now displayed as %s.\n",
	   (use_letters)?"inferred letters":"raw numeric alleles"); pr();
    }
}

command set_score_function(void)
{
    char type[TOKLEN+1];
    get_one_arg(stoken,"",type);

    if (nullstr(type)) {
        sf(ps,"The current score function is '%s'\n",
	   (score_function==SCORE_PAIRS)?"pairs":
	   ((score_function==SCORE_ALL)?"all":
	    ((score_function==SCORE_ONE)?"one":"hom"))); pr();
    } else {
        if (matches(type,"pairs")) { 
	    score_function=SCORE_PAIRS; pval_precision=DEFAULT_PRECISION;
	    print("The current score function is now 'pairs'\n");
	} else if (matches(type,"all")) {
	    score_function=SCORE_ALL; pval_precision=DEFAULT_PRECISION/100;
	    print("The current score function is now 'all'\n");
	} else if (matches(type,"one")) {
	    score_function=SCORE_ONE; pval_precision=DEFAULT_PRECISION;
	    print("The current score function is now 'one'\n");
	} else if (matches(type,"hom")) {
	  score_function=SCORE_HOM; pval_precision=DEFAULT_PRECISION;
	  print("The current score function is now 'hom'\n");
	} else usage_error(1);
      
    }
}

command set_analysis_type(void)
{
    char type[TOKLEN+1];
    int old_val;

    get_one_arg(stoken,"",type);

    old_val = analysis_type;
    if (nullstr(type)) {
        sf(ps,"The current analysis type is '%s'\n",
	   (analysis_type==NPL_ANALYSIS)?"NPL":
	   ((analysis_type==LOD_ANALYSIS)?"LOD":"BOTH")); pr();
    } else {
      if (modcalc && (scan_done || modscore_done)) {
        print("Pedigrees have already been analyzed.\n");
        print("This command will therefore be ignored.\n");
        return;
      }
      if (matches(type,"NPL")) {
	    analysis_type=NPL_ANALYSIS;
	    if (g_frequencies) {
	        sf(ps, "The 'allele frequency estimation' command only works with analysis types 'BOTH' or 'LOD'.\n");
	        pr();
	        exit(1);
	    }
	  }
	else if(matches(type,"LOD") || matches(type,"LINKAGE"))
	    analysis_type=LOD_ANALYSIS;
	else if (matches(type,"BOTH") || matches(type,"both"))
	    analysis_type=BOTH_ANALYSES;
	else usage_error(1);
      
        sf(ps,"The analysis type is now '%s'\n",
	   (analysis_type==NPL_ANALYSIS)?"NPL":
	   ((analysis_type==LOD_ANALYSIS)?"LOD":"BOTH")); pr();
	if (analysis_type != old_val && scan_done) {
	  scan_done=FALSE;
	  print("You will need to scan pedigrees again for this to take effect.\n");
	}
    }
}

command set_count_recs(void)
{
    char type[TOKLEN+1];
    get_one_arg(stoken,"",type);

    if (nullstr(type)) {
        sf(ps,"Count recs is '%s'\n",
	   show_expected_recs?"on":"off"); pr();
    } else {
        if (matches(type,"ON") || matches(type,"TRUE")) 
	    show_expected_recs=TRUE;
	else if(matches(type,"OFF") || matches(type,"FALSE"))
	    show_expected_recs=FALSE;
	else usage_error(1);
      
        sf(ps,"Count recs is now '%s'\n",
	   show_expected_recs?"on":"off"); pr();
    }
}

command set_postscript_output(void)
{
    char type[TOKLEN+1];
    get_one_arg(stoken,"",type);

    if (nullstr(type)) {
        sf(ps,"Postscript output is '%s'\n",
	   postscript_output?"on":"off"); pr();
    } else {
        if (matches(type,"ON") || matches(type,"TRUE")) 
	    postscript_output=TRUE;
	else if(matches(type,"OFF") || matches(type,"FALSE"))
	    postscript_output=FALSE;
	else usage_error(1);
      
        sf(ps,"Postscript output is now '%s'\n",
	   postscript_output?"on":"off"); pr();
    }
}

command set_haplotype(void)
{
    char type[TOKLEN+1];
    get_one_arg(stoken,"",type);

    if (nullstr(type)) {
        sf(ps,"Haplotype output is '%s'\n",
	   haplotype_output?"on":"off"); pr();
    } else {
        if (matches(type,"ON") || matches(type,"TRUE")) 
	    haplotype_output=TRUE;
	else if(matches(type,"OFF") || matches(type,"FALSE"))
	    haplotype_output=FALSE;
	else usage_error(1);
      
        sf(ps,"Haplotype output is now '%s'\n",
	   haplotype_output?"on":"off"); pr();
    }
}

command set_haplomethod(void)
{
    char type[TOKLEN+1];
    get_one_arg(stoken,"",type);

    if (nullstr(type)) {
        sf(ps,"Haplotype method is '%s'\n",
			  (haplo_method == 0)? "Viterbi" : "MaxProb"); 
		  pr();
    } else {
		if (matches(type,"VITERBI") || matches (type, "VIT"))
		  haplo_method = 0;
		else if (matches(type,"MAXPROB") || matches(type,"MAX PROB"))
		  haplo_method = 1;
	else usage_error(1);
        sf(ps,"Haplotype method is now '%s'\n",
			  (haplo_method == 0)? "Viterbi" : "MaxProb"); 
		  pr();
    }
}

command set_dump_req(void)
{
    char type[TOKLEN+1];
    get_one_arg(stoken,"",type);

    if (nullstr(type)) {
        sf(ps,"Dump Requirements is '%s'\n",
	   dump_requirements?"on":"off"); pr();
		  if(dump_requirements) {
			 sf(ps,"  Program will only calculate and dump memory requirements\n");
			 pr();
		  }
    } else {
        if (matches(type,"ON") || matches(type,"TRUE")) 
	    dump_requirements=TRUE;
	else if(matches(type,"OFF") || matches(type,"FALSE"))
	    dump_requirements=FALSE;
	else usage_error(1);
        sf(ps,"Dump Requirements is '%s'\n",
	   dump_requirements?"on":"off"); pr();
		  if(dump_requirements) {
			 sf(ps,"  Program will now only calculate and dump memory requirements.\n");
			 pr();
		  }
    }
}

command set_haplo_ratio(void)
{
    double new_val;

    get_one_arg(rtoken,haplo_ratio,&new_val);

    if (new_val <= 0.0 || new_val > 1.0) {
        sf(ps,"Haplo ratio must be between 0 and 1");
	error(ps);
    } else {
        haplo_ratio = new_val;
    }
    sf(ps,"Haplo ratio is currently %.6lf\n",haplo_ratio); pr();
}


command dump_peak_info(void)
{
    int i,j,index;
    double pos;
    char *filename,*tempstr;
    FILE *fp;

    get_one_arg(rtoken,-9999.9,&pos);

    if (pedigrees_calculated == 0)
      error("no pedigree file has been analyzed with the 'scan' command yet");

    if (pos < -9999.0) 
      error("this command requires an argument which is valid cM position in the most recent scan");

    index = -1;
    for (i=0; i<num_w_pos; i++)
      if (fabs(wpos[i]-pos) < .01) { index=i; break; }
    if (index == -1) {
      sf(ps,"%.2lf is not a valid scan position in the most recent scan",pos);
      error(ps);
    }

    filename = get_temp_string();
    tempstr = get_temp_string();
    strcpy(filename, current_filename);
    split_string(filename,&tempstr,'.');
    sf(ps,"%.2lf",pos);
    strcat(filename,ps);

    fp = open_out_file(filename,"peak info");

    sf(ps,"%d pedigrees  (data file: %s, position: %.2lf)\n",
       pedigrees_calculated, current_filename, pos);
    fpr(fp);

    for (i=0; i<pedigrees_calculated; i++) {
      sf(ps,"%.6lf  \t%d\n",raw_stat[i][index],num_pscores[i]); fpr(fp);
      for (j=0; j<num_pscores[i]; j++) { 
	sf(ps,"%.6lf ",pscore_list[i][j]); fpr(fp);
      }
      sf(ps,"\n"); fpr(fp);
      for (j=0; j<num_pscores[i]; j++) { 
	sf(ps,"%.6lf ",prior_pscore[i][j]); fpr(fp);
      }
      sf(ps,"\n"); fpr(fp);
    }
    close_file(fp);
}


command set_single_point_mode(void)
{
    char type[TOKLEN+1];
    bool old_val;

    get_one_arg(stoken,"",type);

    old_val=single_point_mode;
    if (nullstr(type)) {
        sf(ps,"Single point mode is '%s'\n",
	   single_point_mode?"on":"off"); pr();
    } else {
      if (modcalc && (scan_done || modscore_done)) {
        print("Pedigrees have already been analyzed.\n");
        print("This command will therefore be ignored.\n");
        return;
      }
      if (matches(type,"ON") || matches(type,"TRUE")) 
	    single_point_mode=TRUE;
	else if(matches(type,"OFF") || matches(type,"FALSE"))
	    single_point_mode=FALSE;
	else usage_error(1);
      
        sf(ps,"Single point mode is now '%s'\n",
	   single_point_mode?"on":"off"); pr();
	if (single_point_mode != old_val && scan_done) {
	  scan_done=FALSE;
	  print("You will need to scan pedigrees again for this to take effect.\n");
	}
    }
}

command set_compute_sharing(void)
{
    char type[TOKLEN+1];
    bool old_val;

    get_one_arg(stoken,"",type);

    old_val=compute_sharing;
    if (nullstr(type)) {
        sf(ps,"compute sharing is '%s'\n",
	   compute_sharing?"on":"off"); pr();
    } else {
      if (modcalc && (scan_done || modscore_done)) {
        print("Pedigrees have already been analyzed.\n");
        print("This command will therefore be ignored.\n");
        return;
      }
      if (matches(type,"ON") || matches(type,"TRUE")) 
	    compute_sharing=TRUE;
	else if(matches(type,"OFF") || matches(type,"FALSE"))
	    compute_sharing=FALSE;
	else usage_error(1);
      
        sf(ps,"compute sharing is now '%s'\n",
	   compute_sharing?"on":"off"); pr();
	if (compute_sharing != old_val && scan_done) {
	  scan_done=FALSE;
	  print("You will need to scan pedigrees again for this to take effect.\n");
	}
    }
}

command set_max_bits(void)
{
    int new_val;

    get_one_arg(itoken,max_bits_to_analyze,&new_val);

    if (new_val < 4 || new_val > 30) {
        sf(ps,"Max bits to analyze must be between 4 and 30");
	error(ps);
    } else {
        max_bits_to_analyze = new_val;
    }
    sf(ps,"Currently analyzing a maximum of %d bits per pedigree\n",
       max_bits_to_analyze); pr();
}

/* Not commands but interface stuff that has no other home */

void draw_graph(FILE *fp, double *xpos, double *yval, 
		int num_points, char *graph_title, char *y_title)
{
    int i, top_y, bottom_y, zero_to_one, num_labels;
    int last_one = 0;
    char **y_axis_label, **locus_name, *x_title, *new_y_title;
    double xscale, yscale, max_x, min_y, max_y, too_close, x_offset;
    double old_dist = 0.0;
    double *y_axis_val, *new_map_pos, total_pos;
    double define_x_scale(double);

    ps_file_start(fp);
    ps_page_start(fp,1);

    fprintf(fp,"31 319 moveto\n");
    fprintf(fp,"GS 90 rotate 0 0 rmoveto /Times-Roman FF 12 SF F (%s)S GR\n",
	  graph_title);
    if (user_title_set) fprintf(fp,
      "GS 90 rotate 0 -14 rmoveto /Times-Roman FF 10 SF F (%s)S GR\n",
			   user_title);
    else fprintf(fp,
	  "GS 90 rotate 0 -14 rmoveto /Times-Roman FF 10 SF F (Pedigree file: %s)S GR\n",
	  current_filename);

    /* Calculate appropriate scales, etc. */

    max_y = min_y = 0.0;
    for (i=0; i<num_points; i++) {
        if (yval[i] < -1000.0) continue; /* -INFINITY LOD score */
	
	if (yval[i] > max_y) max_y=yval[i];
	if (yval[i] < min_y) min_y=yval[i];
    }

    if (max_y <= 1.0 && min_y >= 0.0) {
        top_y = 1; bottom_y=0;
	yscale=400.0;
	zero_to_one = TRUE;

    } else {
        zero_to_one = FALSE;
        if (max_y < 3.0) top_y=3;
	else top_y= (int)(max_y+1.0);

	if (min_y > -3.0) bottom_y=-3;
	else bottom_y= (int) (min_y-1.0);

	/* hardwired MLS hack */
	/* top_y=6.0; */
	/* bottom_y=0.0; */

	yscale = 400.0/(top_y-bottom_y);
	if (yscale > 65.0) yscale=65.0;

    }

    /* define x_scale for the graph - looks up based on user settings */
    if (xpos[0] < 0.0) x_offset = (-1.0)*xpos[0];
    else          x_offset = 0.0;
    max_x = xpos[num_points-1] + x_offset;
    xscale = define_x_scale(max_x);


    /* eliminate any off chart points (e.g.,-INFINITY LODs) */

    for (i=0; i<num_points; i++) 
      if (yval[i] < bottom_y) yval[i]=bottom_y;

    /* create y-axis labels */

    num_labels = (top_y-bottom_y)+2;
    matrix(y_axis_label, max(11,num_labels), 10, char);
    array(y_axis_val, max(11,num_labels), double);

    if (zero_to_one) {
      for (i=0; i<=10; i++) {
	y_axis_val[i] = (double) i/10.0;
	sprintf(y_axis_label[i],"%.1lf",y_axis_val[i]);
      }
    } else {
      for (i=0; i<=(top_y-bottom_y); i++) {
        y_axis_val[i] = (double) i+bottom_y;
	sprintf(y_axis_label[i],"%.1lf",y_axis_val[i]);
      }
    }

    /* create x-axis labels */

    matrix(locus_name, num_in_map_order+1, MAX_NAME_SIZE*2, char);
    /* was MAX_NAME_SIZE*10 before; now MAX_NAME_SIZE*2 to save memory,
       important when using large numbers of SNPs  (K. Strauch 07/2007) */
    array(new_map_pos, num_in_map_order+1, double);

    too_close = 6.0/xscale;
    for (i=0; i<num_in_map_order; i++) {
        if (i>0) {
	    if (map_distance[i-1]+old_dist < too_close) {
		if (strlen(locus_name[last_one]) < MAX_NAME_SIZE-2) {
		    /* if strlen inserted to avoid array overflow
		       when using many SNPs  (K. Strauch 07/2007) */
		    strcat(locus_name[last_one]," ");
		    strcat(locus_name[last_one],locus[map_order[i]].name);
		}
		strcpy(locus_name[i],"");
		old_dist += map_distance[i-1];
	    } else {
	        strcpy(locus_name[i],locus[map_order[i]].name);
		last_one=i;
		old_dist=0.0;
	    }
	} else {
	  strcpy(locus_name[i],locus[map_order[i]].name);
	  last_one=i;
	  old_dist=0.0;
	}
    }

    total_pos=0;
    for (i=0; i<num_in_map_order-1; i++) {
      new_map_pos[i]=total_pos+x_offset;
      total_pos+=map_distance[i];
    }
    new_map_pos[i]=total_pos+x_offset;

    /* draw axes */
    x_title = get_temp_string();
    sf(x_title,"%d cM",(int) max_x);

    new_y_title = get_temp_string();
    strcpy(new_y_title,y_title);
    if (zero_to_one)
      draw_axes(fp, num_in_map_order, new_map_pos, locus_name,
		11, y_axis_val, y_axis_label, new_y_title,
		x_title, xscale, yscale, 0.0, TRUE);
    else
      draw_axes(fp, num_in_map_order, new_map_pos, locus_name,
		(top_y-bottom_y)+1, y_axis_val, y_axis_label, new_y_title,
		x_title, xscale, yscale, 0.0, TRUE);

    if (x_offset > 0.0)
      for (i=0; i<num_points; i++) xpos[i]+=x_offset;

    do_bezier(fp, xpos, yval, num_points, 0.0, 0.0,
	      SOLID_LINE, xscale, yscale);

    if (x_offset > 0.0)
      for (i=0; i<num_points; i++) xpos[i]-=x_offset;
	      
    ps_page_end(fp);
    ps_file_end(fp);

    unarray(new_map_pos, double);
    unmatrix(locus_name, num_in_map_order+1, char);
    unarray(y_axis_val, double);
    unmatrix(y_axis_label, max(11,num_labels), char);

    run {
      close_file(fp);
    } except_when(CANTCLOSE) { }

}


void draw_mult_y_plots(FILE *fpps, double *xpos, double **yval,
		       int num_points, int num_y_plots, char *title,
		       char **ylabel, char *y_title)
{
  int i, j, top_y, bottom_y, num_y_val;
  int last_one = 0;
  char **y_axis_label,**locus_name, *x_title;
  double xscale, yscale, *y_axis_val, max_x, max_y, min_y;
  double old_dist = 0.0;
  double x_offset, *new_map_pos, y_increment,tmp_y,too_close,total_pos;

  ps_file_start(fpps);
  ps_page_start(fpps,1);

  /***** Print general information and a key at the top *****/
  fprintf(fpps,"31 319 moveto\n");
  fprintf(fpps,"GS 90 rotate 0 0 rmoveto /Times-Roman FF 12 SF F (%s)S GR\n",
	  title);
  if (user_title_set) fprintf(fpps,
    "GS 90 rotate 0 -12 rmoveto /Times-Roman FF 10 SF F (%s)S GR\n",
			      user_title);
  else fprintf(fpps,
	  "GS 90 rotate 0 -12 rmoveto /Times-Roman FF 10 SF F (Pedigree file: %s)S GR\n",
	  current_filename);
  fprintf(fpps,"22 130 rmoveto\n");
  fprintf(fpps,"GS 90 rotate 0 0 rmoveto /Times-Roman FF 10 SF F (Key:)S GR\n");
  for (i=0; i<num_y_plots; i++) {
    fprintf(fpps,"GS 90 rotate %d %d rmoveto ",(i/3)*135,-10-((i%3)*9));
    fprintf(fpps,"/Times-Roman FF 9 SF F (%s)S GR\n",
	    ylabel[i]);
    fprintf(fpps,"GS 90 rotate %d %d rmoveto ",((i/3)*135)+63,-7-((i%3)*9));
    fprintf(fpps,"%s 20 0 rlineto stroke GR\n",line_choice(i));
  }

  
  /***** Calculate appropriate scales, etc. *****/

  /***** Y-axis stuff *****/

  /* Find the min & max out of all the y-plots */
  max_y = min_y = 0.0;
  for (j=0; j<num_y_plots; j++)
    for (i=0; i<num_points; i++) {
      if (yval[j][i] < -1000.0) continue;  /* -INFINITY LOD score */
      if (yval[j][i] > max_y) max_y=yval[j][i];
      if (yval[j][i] < min_y) min_y=yval[j][i];
    }

  /* Set the y-limits on the page according to the min & max */
  if (max_y < 0.0) {
    if (max_y >= -3.0) 
      top_y = 0;
    else top_y = (int)(max_y+1.5);
  } else if (max_y <= 1.0) {
    top_y=1;
  } else { top_y = (int)(max_y+1.5);}
  if (min_y >= 0.0) bottom_y=0;
  else bottom_y = (int)(min_y-1.5);

  yscale = 440.0/(double)(top_y-bottom_y);

  /* eliminate -INFINITY points */
  for (j=0; j<num_y_plots; j++) 
    for (i=0; i<num_points; i++) 
      if (yval[j][i] < bottom_y) 
	yval[j][i]=bottom_y;

  /* set the number of marks that will be shown along the axis */
  if (top_y-bottom_y < 5) {
    y_increment = (double)(top_y-bottom_y)/10.0;
    num_y_val = 10;
  } else {
    y_increment = 1.0;
    num_y_val = top_y-bottom_y+2;
  }

  /* Store the values to be printed on the y-axis as doubles
     and strings */
  matrix(y_axis_label, num_y_val+2, 10, char);
  array(y_axis_val, num_y_val+2, double);
  tmp_y = (double)(bottom_y);
  i=0;
  while(tmp_y < ((double)(top_y))) {
    y_axis_val[i] = tmp_y;
    sprintf(y_axis_label[i],"%.1lf",y_axis_val[i]);
    tmp_y += y_increment;
    i++;
  }
  num_y_val = i;


  /***** X-axis stuff ******/

  if (xpos[0] < 0.0) x_offset = (-1.0)*xpos[0];
  else               x_offset = 0.0;
  max_x = xpos[num_points-1]+x_offset;
  xscale = 600.0/max_x;

  /* Locus name labels */

  matrix(locus_name, num_in_map_order+1, MAX_NAME_SIZE*2, char);
  /* was MAX_NAME_SIZE*10 before; now MAX_NAME_SIZE*2 to save memory,
     important when using large numbers of SNPs  (K. Strauch 07/2007) */
  array(new_map_pos, num_in_map_order+1, double);

  too_close = 6.0/xscale;
  for (i=0; i<num_in_map_order; i++) {
    if (i>0) {
      if (map_distance[i-1]+old_dist < too_close) {
	if (strlen(locus_name[last_one]) < MAX_NAME_SIZE-2) {
	  /* if strlen inserted to avoid array overflow
	     when using many SNPs  (K. Strauch 07/2007) */
	  strcat(locus_name[last_one]," ");
	  strcat(locus_name[last_one],locus[map_order[i]].name);
	}
	strcpy(locus_name[i],"");
	old_dist += map_distance[i-1];
      } else {
	strcpy(locus_name[i],locus[map_order[i]].name);
	last_one=i;
	old_dist=0.0;
      }
    } else {
      strcpy(locus_name[i],locus[map_order[i]].name);
      last_one=i;
      old_dist=0.0;
    }
  }

  total_pos=0;
  for (i=0; i<num_in_map_order-1; i++) {
    new_map_pos[i]=total_pos+x_offset;
    total_pos+=map_distance[i];
  }
  new_map_pos[i]=total_pos+x_offset;

  x_title = get_temp_string();
  sf(x_title,"%.1lf cM",max_x);

  /****** Drawing stuff *******/
  /* Draw the axes */
  draw_axes(fpps, num_in_map_order, new_map_pos, locus_name, 
	    num_y_val, y_axis_val, y_axis_label, y_title, 
	    x_title, xscale, yscale, (-2.0), TRUE);

  /* Draw all the analytical curves */
  if (x_offset > 0.0)
    for (i=0; i<num_points; i++) xpos[i]+=x_offset;
  for (i=0; i<num_y_plots; i++) {
    do_bezier(fpps, xpos, yval[i], num_points, 0.0, 0.0,
	      line_choice(i), xscale, yscale);
  }
  

  /* Reset the values */
  if (x_offset > 0.0)
    for (i=0; i<num_points; i++) xpos[i]-=x_offset;

  /* Finish up the postscript file */
  ps_page_end(fpps);
  ps_file_end(fpps);

  unarray(new_map_pos, double);
  unarray(y_axis_val,double);
  unmatrix(locus_name, num_in_map_order+1, char);
  unmatrix(y_axis_label, num_y_val+2, char);

}


FILE *open_out_file(char *default_file, char *purpose)
{
  char *out_file;
  char * volatile tmp;
  bool valid;
  FILE *fp;

  array(out_file, PATH_LENGTH+1, char);
  valid = FALSE;
  while(!valid) {
    fp = NULL;
    tmp = get_temp_string();
    sf(ps,"\nfile to store %s [%s]: \n",purpose,default_file);
    input(ps,tmp,TEMP_STRING_LEN);
    if (!nullstr(tmp)) {
      run {
	nstrcpy(out_file,tmp,PATH_LENGTH);
	/* User has specified a file */
	fp = open_file(out_file,"r");
	sf(ps," file '%s' already exists, overwrite? y/n [n]: \n",
	   out_file);
	close_file(fp);
	tmp = get_temp_string();
	input(ps,tmp,TEMP_STRING_LEN);
	if (tmp[0] == 'y' || tmp[0] == 'Y') {
	  run {
	    fp = open_file(out_file,"w");
	    valid = TRUE;
	  } except_when(CANTOPEN) {
	    print("can't write to that file, try again\n");
	  }
	}
      } except_when(CANTOPEN) {
	run {
	  fp = open_file(out_file,"w");
	  valid = TRUE;
	} except_when(CANTOPEN) {
	  sf(ps," can't open file %s, try again\n",out_file); pr();
	}
      }
    } else {
      /* Use the default filename */
      run {
	fp = open_file(default_file,"r");
	sf(ps," file '%s' already exists, overwrite? y/n [n]: \n",
	   default_file);
	close_file(fp);
	tmp = get_temp_string();
	input(ps,tmp,TEMP_STRING_LEN);
	if (tmp[0] == 'y' || tmp[0] == 'Y') {
	  run {
	    fp = open_file(default_file,"w");
	    valid = TRUE;
	  } except_when(CANTOPEN) {
	    print("can't write to that file, try again\n");
	  }
	}
      } except_when (CANTOPEN) {
	run {
	  fp = open_file(default_file,"w");
	  valid=TRUE;
	} except_when (CANTOPEN) {
	  sf(ps,
	     " can't open file '%s' (maybe permission problem?) try again\n",
	     default_file); pr();
	}
      }
    }
  }
  unarray(out_file, char);
  return(fp);
}



double define_x_scale(double max_x)
{
    double xscale;

    if (ps_scaling == FIT_TO_PAGE) {

      xscale = 640.0/max_x;
      if (xscale > 25.0) xscale=25.0;

    } else { /* CONSTANT_SCALE */
      xscale = constant_ps_scale;
      if (xscale*max_x > 650.0) {
	  /* warning: picture won't fit */
	  xscale = 640.0/max_x;
	  sf(ps,"warning: constant scale factor too large, reducing to %.1lf",
	     xscale); pr(); nl();
      } else if (xscale*max_x < 100.0) {
	  /* warning: picture rather small and insignificant looking */
	  xscale = 100.0/max_x;
	  sf(ps,"warning: constant scale factor too small, raising to %.1lf",
	     xscale); pr(); nl();
      }

    }

    return(xscale);
}


command set_postscript_scaling(void)
{
    char *tempstr;
    int num;
    double rnum;

    print("Methods to scale drawings:\n");
    print("1) Scale chromosome drawings to fill the page\n");
    print("2) Set a fixed value for dots per cM\n");
    tempstr = get_temp_string();
    sf(ps,"Enter choice [%d]: ",ps_scaling);
    input(ps,tempstr,TEMP_STRING_LEN);
    itoken(&tempstr,ps_scaling,&num);
    if (num > 2 || num < 1) error("bad value entered");
    ps_scaling = num;

    if (ps_scaling == 2) {
      print(
"There are roughly 650 dots available for the longest map so your scaling\n");
      print(
"factor should be roughly 650 / length of your longest chromosome in cM\n");
      
      sf(ps,"Enter the factor [%.1lf]: ",constant_ps_scale);
      input(ps,tempstr,TEMP_STRING_LEN);
      rtoken(&tempstr,constant_ps_scale,&rnum);
      if (rnum <= 0.0) error("scale factor must be positive");
      constant_ps_scale = rnum;
      sf(ps,"Drawing will be done with a fixed scale of %.1lf dots per cM\n",
	 constant_ps_scale); pr();
    } else {
      print("Drawings will be scaled to full-page size\n");
    }
}


void set_display_scores(void)
{
  char type[TOKLEN+1];
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    sf(ps,"Screen display of NPL scores, LOD scores, and haplotypes is now '%s'\n",
       display_scores?"on":"off"); pr();
  } else {
    if (matches(type,"ON") || matches(type,"TRUE")) 
      display_scores=TRUE;
    else if(matches(type,"OFF") || matches(type,"FALSE"))
      display_scores=FALSE;
    else usage_error(1);
    
    sf(ps,"Screen display of NPL scores, LOD scores, and haplotypes is now '%s'\n",
       display_scores?"on":"off"); pr();
  }
}


command set_imprinting(void)
{
  char type[TOKLEN+1];
  int old_val;

  old_val = imprinting;
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    sf(ps,"Imprinting mode is '%s'.\n",
       imprinting?"on":"off"); pr();
  } else if ((num_markers != 0) && (imprinting==TRUE) &&
	     (matches(type,"OFF") || matches(type,"FALSE"))) {
      error("if markers have already been loaded, imprinting mode can only be\n       turned from 'off' to 'on' but not vice versa");
  } else {
    if (matches(type,"ON") || matches(type,"TRUE"))
      imprinting=TRUE;
    else if(matches(type,"OFF") || matches(type,"FALSE"))
      imprinting=FALSE;
    else usage_error(1);

    if (modscore_done && (imprinting != old_val)) {
	/* do not allow p-value calculation if 'imprinting'
	   is changed after executing 'modscore' */
	changed_modscore_options = TRUE;
    }
    sf(ps,"Imprinting mode is now '%s'.\n",
       imprinting?"on":"off"); pr();
  }
}


command set_include_untyped(void)
{
  char type[TOKLEN+1];
  bool old_include;

  get_one_arg(stoken,"",type);

  old_include=include_untyped;
  if (nullstr(type)) {
    sf(ps,"Non-genotyped but phenotyped individuals with no kids are %s.\n",
       (include_untyped)?"included":"excluded"); pr();
  } else {
    if (modcalc && (scan_done || modscore_done)) {
      print("Pedigrees have already been analyzed.\n");
      print("This command will therefore be ignored.\n");
      return;
    }
    if (matches(type,"ON") || matches(type,"TRUE"))
      include_untyped=TRUE;
    else if(matches(type,"OFF") || matches(type,"FALSE"))
      include_untyped=FALSE;
    else usage_error(1);

    sf(ps,"Non-genotyped but phenotyped individuals with no kids are now %s.\n",
       (include_untyped)?"included":"excluded"); pr();
    if (include_untyped != old_include && scan_done) {
      scan_done=FALSE;
      print("You will need to scan pedigrees again for this to take effect.\n");
    }
  }
}


command set_modcalc(void)
{
  char type[TOKLEN+1];

  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    sf(ps,"Storage of inheritance-vector probs for MOD-score analysis is %s.\n",
       (modcalc)?"activated":"deactivated"); pr();
    if (modcalc) {
      if ((lowpos<-990.0) && (highpos>990.0)) {
        sf(ps,"Maximization will be done %s.\n",
	   (modcalc==1) ? "globally over all genetic positions"
	                : "separately for each genetic position");
	pr();
      } else {
        sf(ps,"Maximization will be done %s between %.2lf and %.2lf cM.\n",
	   (modcalc==1) ? "globally over all pos"
	                : "separately for each pos",
	   lowpos, highpos);
        pr();
      }
    }
  } else if (scan_done || modscore_done) {
    print("Pedigrees have already been analyzed.\n");
    print("This command will therefore be ignored.\n");
  } else if(matches(type,"ON") || matches(type,"TRUE")) {
    print("Please use either 'global', 'single', or 'off' as an option for modcalc.\n");
  } else {
    if(matches(type,"GLOBAL")) modcalc=1;
    else if(matches(type,"SINGLE")) modcalc=2;
    else if(matches(type,"OFF") || matches(type,"FALSE")) modcalc=FALSE;
    else usage_error(1);

    sf(ps,"Storage of inheritance-vector probs for MOD-score analysis is now %s.\n",
       (modcalc)?"activated":"deactivated"); pr();
    if (modcalc) {
      if ((lowpos<-990.0) && (highpos>990.0)) {
	sf(ps,"Maximization will be done %s,\n",
	   (modcalc==1) ? "globally over all genetic positions"
	                : "separately for each genetic position");
	pr();
      } else {
	sf(ps,"Maximization will be done %s between %.2lf and %.2lf cM,\n",
	   (modcalc==1) ? "globally over all pos"
	                : "separately for each pos",
	   lowpos, highpos);
        pr();
      }
      sf(ps,"by varying penetrances of liability class %d and disease allele frequency.\n", modliab);
      pr();
      if ((num_markers != 0) && (modliab > num_liability_classes)) {
	sf(ps,"Warning: liability class to be optimized (%d) does not exist.\n", modliab);
	pr();
	print("Use the 'liability class' command in order to change this.\n");
      } else print("Use the 'liability class' command in case you want to change this.\n");
    }
  }
}


command set_positions(void)
{
  double newlowpos, newhighpos;
  int changed=FALSE;

  if (g_association) {
      print("The command to specify the lowest and highest positions for MOD score calculation\n"
            "will be ignored for JLA analysis.\n");
      return;
  }

  get_arg(rtoken,-999.0,&newlowpos);
  get_arg(rtoken,999.0,&newhighpos);

  if ((scan_done || modscore_done) &&
      ((!real_eq(newlowpos,-999.0)) || (!real_eq(newhighpos,999.0)))) {
    print("Pedigrees have already been analyzed.\n");
    print("This command will therefore be ignored.\n");
  } else {
    if ((!real_eq(newlowpos,-999.0)) || (!real_eq(newhighpos,999.0))) {
      /*** if either parameter is given ***/
      if((newlowpos>-990.0) && (newhighpos<990.0) &&
          ((newhighpos>=newlowpos) || real_eq(newlowpos, newhighpos))) {    
	lowpos=newlowpos;
	highpos=newhighpos;
	changed=TRUE;
	if (modscore_done) changed_modscore_options = TRUE;
	/* do not allow p-value calculation if the set of positions
	   to be analyzed is changed after executing 'modscore' */
      }
      else usage_error(1);
    }
    if (modcalc) {
      if ((lowpos<-990.0) && (highpos>990.0)) {
	sf(ps,"Maximization will be done %s.\n",
	   (modcalc==1) ? "globally over all genetic positions"
	                : "separately for each genetic position");
	pr();
      } else {
	sf(ps,"Maximization will be done %s between %.2lf and %.2lf cM.\n",
	   (modcalc==1) ? "globally over all pos"
	                : "separately for each pos",
	   lowpos, highpos);
	pr();
      }
    } else { /* modcalc=FALSE */
      if ((lowpos<-990.0) && (highpos>990.0))
	print("Maximization will be done for all genetic positions.\n");
      else {
        sf(ps,"Maximization will be done for all positions between %.2lf and %.2lf cM.\n", lowpos, highpos);
        pr();
      }
      if (changed) print("Warning: modcalc must be 'global' or 'single' for this to have an effect.\n");
    }
  }
}


command set_liability_class(void)
{
    int new_val;

    get_one_arg(itoken,modliab,&new_val);

    if ((num_markers != 0) &&
	((new_val < 1) || (new_val > num_liability_classes))) {
	if (num_liability_classes == 1)
	    error("liability class optimized for MOD calculation must be 1");
	else {
	    sf(ps,"liability class optimized for MOD calculation must be between 1 and %d", num_liability_classes);
	    error(ps);
	}
    } else if (new_val < 1 || new_val > MAX_LIABILITY_CLASSES) {
        sf(ps,"liability class optimized for MOD calculation must be between 1 and %d", MAX_LIABILITY_CLASSES);
	error(ps);
    } else {
	if (modscore_done && (new_val != modliab)) {
	    /* do not allow p-value calculation if the liability class
	       to be optimized is changed after executing 'modscore' */
	    changed_modscore_options = TRUE;
	}
        modliab = new_val;
    }
    sf(ps,"Currently optimizing penetrances of liability class %d for MOD calculation.\n",
       modliab); pr();
    if (!modcalc) print("Warning: modcalc must be 'global' or 'single' for this to have an effect.\n");
}


command join_liability_classes(void)
{
    int i, j, new_val, found=FALSE, done=FALSE, pendiff, count=0;
    int join[MAX_LIABILITY_CLASSES+1], newcode[MAX_LIABILITY_CLASSES+1];

    if (num_markers == 0 || scan_done)
	error("command must be executed between 'load markers' and 'scan pedigrees'");
    else if (num_liability_classes == 1)
	error("no joining possible since there is only one liability class");
    else {
	do {
	    get_arg(itoken,-999,&new_val);
	    if (new_val == -999) {
		done = TRUE;
	    } else {  /* Parameter entered */
		if (new_val < 1 || new_val > num_liability_classes) {
		    sf(ps,"liability classes to be joined must be between 1 and %d", num_liability_classes);
		    error(ps);
		} else {
		    /* first check if liab. class has been entered before */
		    for (i=0; i<count; i++) if (new_val == join[i])
			error("a liability class cannot be specified more than once");
		    join[count] = new_val;
		    count++;
		    if (count >= 2) found = TRUE;
		}
	    }
	} while (!done);
	if (!found) error("you need to enter at least two liability classes to be joined");
	else {  /* found; we need to join and recode the liability classes */
	    new_val = 999;  /* variable is reused now for joined liab. class */
	    for (j=0; j<count; j++)
		if (join[j] < new_val) new_val = join[j];
	    /* now new_val contains recoded number of the joined liab. class,
	       i.e., the entered liability class with the smallest number   */
	    pendiff = FALSE;
	    for (j=1; j<count; j++)   /* count is always >= 2 */
		if ((!real_eq(penetrance[join[j]][HOM_UNAFF],
			      penetrance[join[0]][HOM_UNAFF])) ||
		    (!real_eq(penetrance[join[j]][HET_AU],
                              penetrance[join[0]][HET_AU])) ||
		    (!real_eq(penetrance[join[j]][HET_UA],
                              penetrance[join[0]][HET_UA])) ||
		    (!real_eq(penetrance[join[j]][HOM_AFF],
                              penetrance[join[0]][HOM_AFF]))) pendiff = TRUE;
	    if (pendiff && (join[0] != new_val)) {
		/*** copy penetrances of join[0] (1st entered class
		     to be joined) to the recoded new joined class ***/
		penetrance[new_val][HOM_UNAFF] =
		    penetrance[join[0]][HOM_UNAFF];
		penetrance[new_val][HET_AU] =
		    penetrance[join[0]][HET_AU];
		penetrance[new_val][HET_UA] =
		    penetrance[join[0]][HET_UA];
		penetrance[new_val][HOM_AFF] =
		    penetrance[join[0]][HOM_AFF];
	    }

	    num_liability_classes = 0;  /* recount them all the way */
	    for (i=0; i<=MAX_LIABILITY_CLASSES; i++) newcode[i] = 0;
	    /* these variables are used to keep track of whether it is
	       the first time we have seen a particular liability class,
	       and of the new recoded number of that liability class   */
	    for (i=1; i<=old_num_liability_classes; i++) {
		for (j=0; j<count; j++) if (recode_liab_class[i]==join[j]) {
		    /* this is one of the classes to be joined */
		    if (!newcode[new_val]) {
			num_liability_classes++;
			/* only increment num_liability_classes the first time
			   we find a liability class to be joined - this will
			   necessarily be the one for which the old as well
			   as the new number is new_val - since the other
			   classes to be joined no longer count */
			newcode[new_val] = new_val;
		    }
		    recode_liab_class[i] = new_val;
		    /* no need to copy penetrance values for joined class;
		       this has already been done above if necessary     */
		    break;  /* i.e., the for j loop */
		}
		if (j==count) { /* not found, this liab.class is not joined */
		    if (!newcode[recode_liab_class[i]]) {
			/* we see this liability class for the first time */
			num_liability_classes++;
			penetrance[num_liability_classes][HOM_UNAFF] =
			    penetrance[recode_liab_class[i]][HOM_UNAFF];
			penetrance[num_liability_classes][HET_AU] =
			    penetrance[recode_liab_class[i]][HET_AU];
			penetrance[num_liability_classes][HET_UA] =
			    penetrance[recode_liab_class[i]][HET_UA];
			penetrance[num_liability_classes][HOM_AFF] =
			    penetrance[recode_liab_class[i]][HOM_AFF];
			newcode[recode_liab_class[i]] = num_liability_classes;
		    }
		    recode_liab_class[i] = newcode[recode_liab_class[i]];
		}
	    }
	    init_disease_info();
	    print("\nThe liability classes are now joined and recoded as follows:\n\n");
	    if (imprinting) {
		print("   Initial class        Recoded class              Penetrances\n");
		print("(as in dat/prefile)    (after joining)    f+/+    fm/+    f+/m    fm/m\n");
		for (i=1; i<=old_num_liability_classes; i++) {
		    sf(ps,"        %2d        --->       %2d          %6.4lf  %6.4lf  %6.4lf  %6.4lf\n",
		       i, recode_liab_class[i],
		       penetrance[recode_liab_class[i]][HOM_UNAFF],
		       penetrance[recode_liab_class[i]][HET_AU],
		       penetrance[recode_liab_class[i]][HET_UA],
		       penetrance[recode_liab_class[i]][HOM_AFF]);  pr();
		}
	    } else {  /* no imprinting */
		print("   Initial class        Recoded class           Penetrances\n");
		print("(as in dat/prefile)    (after joining)     f+/+    fHet    fm/m\n");
		for (i=1; i<=old_num_liability_classes; i++) {
		    sf(ps,"        %2d        --->       %2d           %6.4lf  %6.4lf  %6.4lf\n",
		       i, recode_liab_class[i],
		       penetrance[recode_liab_class[i]][HOM_UNAFF],
		       penetrance[recode_liab_class[i]][HET_AU],
		       penetrance[recode_liab_class[i]][HOM_AFF]);  pr();
		}
	    }
	    if (pendiff) {
		print("\nWarning: the penetrances specified in the locus datafile differ\n");
		print("between the liability classes to be joined.  The penetrances of the\n");
		sf(ps,"previous liability class %d will be used for the joined liability class.\n", join[0]);  pr();
	    }
	    print("\nPlease refer to the recoded numbering when you select a liability class\n");
	    print("for which penetrances should be varied in the MOD-score analysis\n");
	    print("or if you select a further set of liability classes to be joined.\n");
	    print("Note that the liability classes in the pedigree file still need to\n");
	    print("reflect the initial definition in the locus datafile.  The only purpose\n");
	    print("of this command is to allow for a MOD-score analysis in which the\n");
	    print("penetrances of several liability classes should be jointly varied.\n");
	    if (modcalc && modliab > num_liability_classes) {
		sf(ps,"\nWarning: liability class to be optimized (%d) does not exist.\n", modliab);  pr();
		print("Use the 'liability class' command in order to change this.\n");
	    }
	}
    }
}


command set_penetrance_restriction(void)
{
  char type[TOKLEN+1];
  int old_val;

  old_val = penetrance_restriction;
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    sf(ps,"Penetrance restriction \"f+/+ <= (fm/+,f+/m) <= fm/m\" is '%s'.\n",
       penetrance_restriction?"on":"off"); pr();
  } else {
    if (matches(type,"ON") || matches(type,"TRUE"))
      penetrance_restriction=TRUE;
    else if(matches(type,"OFF") || matches(type,"FALSE"))
      penetrance_restriction=FALSE;
    else usage_error(1);

    if (modscore_done && (penetrance_restriction != old_val)) {
	/* do not allow p-value calculation if the penetrance
	   restriction is changed after executing 'modscore' */
	changed_modscore_options = TRUE;
    }
    sf(ps,"Penetrance restriction \"f+/+ <= (fm/+,f+/m) <= fm/m\" is now '%s'.\n",
       penetrance_restriction?"on":"off"); pr();
    if (!modcalc) print("Warning: modcalc must be 'global' or 'single' for this to have an effect.\n");
  }
}


command set_allfreq_restriction(void)
{
  char type[TOKLEN+1];
  int old_val;

  old_val = allfreq_restriction;
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    sf(ps,"Disease allele frequency restriction (MOD) is '%s'.\n",
       allfreq_restriction?"on":"off"); pr();
    if (allfreq_restriction) {
      sf(ps,"Current upper bound for the disease allele frequency is %.2lf.\n",
         highest_allfreq); pr();
      print("Use the 'highest allfreq' command if you want to change the upper bound.\n");
    }
    if (!modcalc) print("Warning: modcalc must be 'global' or 'single' for this to have an effect.\n");
  } else {
    if (matches(type,"ON") || matches(type,"TRUE"))
      allfreq_restriction=TRUE;
    else if(matches(type,"OFF") || matches(type,"FALSE"))
      allfreq_restriction=FALSE;
    else usage_error(1);

    if (modscore_done && (allfreq_restriction != old_val)) {
	/* do not allow p-value calculation if the disease allele
	   frequency restriction is changed after executing 'modscore' */
	changed_modscore_options = TRUE;
    }
    sf(ps,"Disease allele frequency restriction (MOD) is now '%s'.\n",
       allfreq_restriction?"on":"off"); pr();
    if (allfreq_restriction) {
      sf(ps,"Current upper bound for the disease allele frequency is %.2lf.\n",
         highest_allfreq); pr();
      print("Use the 'highest allfreq' command if you want to change the upper bound.\n");
    }
    if (!modcalc) print("Warning: modcalc must be 'global' or 'single' for this to have an effect.\n");
  }
}


command set_highest_allfreq(void)
{
  double new_val, old_val;

  get_one_arg(rtoken,-999.0,&new_val);

  old_val = highest_allfreq;
  if (!real_eq(new_val,-999.0)) {
    if ((new_val < 0.01 || new_val > 0.99) && (!real_eq(new_val,0.01))
         && (!real_eq(new_val,0.99))) {
      sf(ps,"upper bound for disease allfreq (MOD) must be between 0.01 and 0.99");
      error(ps);
    } else {
      highest_allfreq = new_val;
      if (real_eq(new_val,0.01)) highest_allfreq = 0.01;
      if (real_eq(new_val,0.99)) highest_allfreq = 0.99;
      if (modscore_done && (!real_eq(highest_allfreq, old_val))) {
	  /* do not allow p-value calculation if the highest disease
	     allele frequency is changed after executing 'modscore' */
	  changed_modscore_options = TRUE;
      }
    }
  }
  sf(ps,"Current upper bound for the disease allele frequency (MOD) is %.2lf.\n", highest_allfreq); pr();
  if (!modcalc) print("Warning: modcalc must be 'global' or 'single' for this to have an effect.\n");
  if (!allfreq_restriction) print("Warning: 'allfreq restriction' must be 'on' for this to have an effect.\n");
}


command set_dimensions(void)
{
    int new_val;

    get_one_arg(itoken,dimensions,&new_val);

    if (new_val < 1 || new_val > 5) {
        sf(ps,"number of parameters jointly changed (MOD) must be between 1 and 5");
        error(ps);
    } else {
	if (modscore_done && (new_val != dimensions)) {
	    /* do not allow p-value calculation if dimensions
	       is changed after executing 'modscore' */
	    changed_modscore_options = TRUE;
	}
        dimensions = new_val;
    }
    if (dimensions == 1) 
        print("Currently only 1 parameter is changed at a time for MOD calculation.\n");
    else {
	sf(ps,"Currently up to %d parameters are jointly changed for MOD calculation.\n", dimensions); pr();
    }
}


command set_saved_models(void)
{
	int new_val;
	unsigned long int mod_memcount=0;

	get_one_arg(itoken,-999,&new_val);
	if (modscore_done && (new_val != -999)) {
		error("cannot change no. of saved models if MOD has already been calculated");
	} else if (scan_done && algebra && new_val != -999) {
		error("cannot change no. of saved models if pedigrees have already been\n       analyzed by algebraic calculation");
	} else {
		if (new_val != -999) {
			if (new_val < -1 || new_val > 32767) {
				sf(ps,"number of models saved for MOD calculation must be between 0 and 32767\n       - or specify '-1' to save all initial models");
				error(ps);
			} else {
				saved_models = new_val;
			}
		}
		if (saved_models == -1)
			print("Currently all initial models are saved in memory for MOD calculation.\n");
		else if (saved_models == 1) 
			print("Currently 1 model is saved in memory for MOD calculation.\n");
		else {
			sf(ps,"Currently %d models are saved in memory for MOD calculation.\n",
					saved_models); pr();
		}

		if (((modcalc==1 && num_replicates > 0) || modcalc == 2) && saved_models>0 && (scan_done || modscore_done)) {
			for (famnum=0; famnum<num_pedigrees; famnum++)
				mod_memcount += save_real_num_vecs[famnum_pointer[famnum]] * sizeof(double);
			mod_memcount *= saved_models;
			sf(ps,"With the current dataset, this will need %lu bytes of memory.\n", mod_memcount); pr();
		}
	}
	if (algebra && new_val != -999) {
		sf(ps, "Warning: Model saving is only relevant with algebraic calculation mode 'off'.\n"); pr();
	}
}


command set_long_mod_output(void)
{
  char type[TOKLEN+1];
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    sf(ps,"Long output with MOD-score analysis (modcalc single) is '%s'.\n",
       long_mod_output?"on":"off"); pr();
  } else {
    if (matches(type,"ON") || matches(type,"TRUE"))
      long_mod_output=TRUE;
    else if(matches(type,"OFF") || matches(type,"FALSE"))
      long_mod_output=FALSE;
    else usage_error(1);

    sf(ps,
       "Long output with MOD-score analysis (modcalc single) is now '%s'.\n",
       long_mod_output?"on":"off"); pr();
  }
}


command set_user_title(void)
{
    char *tmp;

    print("Please enter a title to be used for the PostScript plots\n");
    print("instead of the pedigree filename.\n");
    tmp = get_temp_string();
    input("",tmp,TEMP_STRING_LEN);
    if (!nullstr(tmp)) {
	strcpy (user_title, tmp);
	user_title_set = TRUE;
	sf(ps, "The title used for PostScript plots is now:\n%s\n",
	   user_title); pr();
    } else {
	print("The pedigree filename will be used as the title for the PostScript plots.\n");
	user_title_set = FALSE;
    }
}


command use_pub_map(void){
  char *file_name, *file_name_with_path;
  volatile int could_open=1;
  FILE * volatile fp;

  use_uncrunched_args();
  file_name = get_temp_string();
  file_name_with_path = get_temp_string();
  
  if (!stoken(&args, sREQUIRED, file_name)) {
    if (strcmp(map_name,"")!=0){
      sf(ps,"current marker map: %s", map_name); pr(); nl();    
    }else{
      sf(ps,"Please supply the name of a file containing the map you wish to use.\nChoose either 'duffy.txt', 'marshfield.txt', 'nievergelt.txt', 'rutgers.txt',\nor supply a different map file, but with the same format as in the\nfiles mentioned above.");pr();nl();
    }    
  }else{
    if (num_markers > 0) {
      sf(ps,"Genetic marker info is already loaded for %d markers.\n",
         num_markers); pr();
      print("This command will therefore be ignored.\n");
      return;
    }
    run{fp=open_file(file_name,READ);}except_when(CANTOPEN){
	could_open = 0;
    }
    if (!could_open) {
	could_open = 2;
	strcpy (file_name_with_path, file_name);
	if (map_make_filename_in_dir(file_name_with_path,
				     DEFAULT_DIR,CODE_DIR)) {
	    run { fp=open_file(file_name_with_path,READ); }
	    except_when(CANTOPEN) { could_open = 0; }
	} else could_open = 0;
    }
    /* sf(ps,"file_name = %s, file_name_with_path = %s, could_open = %d\n",
       file_name, file_name_with_path, could_open); pr(); */
    /* could_open now is 0 if neither file could be opened,
                         1 if the file could be opened in local dir,
			 2 if the file could be opened in GHM_DIR        */
    if (!could_open) {
	sf(ps,"can't open file %s (neither in local directory nor in GHM_DIR)",
	   file_name);
	error(ps);
    } else if (could_open == 1) {
	sf(ps,"Map file %s in local directory will be used.\n", file_name);
	pr();
    } else if (could_open == 2) {
        sf(ps,"Map file %s will be used.\n",
	   file_name_with_path);
        pr();
    }
    close_file(fp);

    if (!strncasecmp(file_name,"decode.txt",3)) {
      strcpy(map_name,"decode");
      strcpy(map_file, file_name);
      ind_use_map_marker = 1;  
    }else if(!strncasecmp(file_name,"marshfield.txt",3)) { 
      strcpy(map_name,"marshfield");
      strcpy(map_file, file_name);
      ind_use_map_marker = 1;
    }else if(!strncasecmp(file_name,"duffy.txt",3)) {
      strcpy(map_name,"duffy");
      strcpy(map_file, file_name);
      ind_use_map_marker = 1;
    }else if(!strncasecmp(file_name,"rutgers.txt",3)) {
      strcpy(map_name,"rutgers");
      strcpy(map_file, file_name);
      ind_use_map_marker = 1;
    }else if(!strncasecmp(file_name,"nievergelt.txt",3)) {
      strcpy(map_name,"nievergelt");
      strcpy(map_file, file_name);
      ind_use_map_marker = 1;
    }else{
      strcpy(map_name,"usermap");
      strcpy(map_file, file_name);
      ind_use_map_marker = 1; 
    }  
    if (could_open == 2) strcpy (map_file, file_name_with_path);
  }   
}


command set_maximization(void)
{
  char type[TOKLEN+1];
  int old_val;

  old_val = maxim_type;
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    if (maxim_type == 1) {
      sf(ps,"Maximization will be performed with the standard set of predefined\ntrait models, followed by a fine maximization.\n");
      pr();
    } else if (maxim_type == 2) {
      sf(ps,"Maximization will be performed with the larger set of predefined\ntrait models, followed by a fine maximization.\n");
      pr();
      if (allfreq_restriction) {
	sf(ps,"Please consider turning 'allfreq restriction off' in order to cover\na wider range of values for the disease allele frequency as well.\n");
	pr();
      }
    } else if (maxim_type == 4) {
      sf(ps,"Maximization will be done only with the larger set of predefined\ntrait models, without a fine maximization.\n");
      pr();
    } else if (maxim_type == 5) {
      sf(ps,"Maximization will be done only with the standard set of predefined\ntrait models, without a fine maximization.\n");
      pr();
    } else if (maxim_type == 6) {
      sf(ps,"Maximization will be done only with the set of user-defined trait models\n(those entered by the 'model' command), without a fine maximization.\n");
      pr();
    }
  } else if(matches(type,"ON") || matches(type,"TRUE")|| matches(type,"OFF")|| matches(type,"FALSE")) {
    print("Please use either 'standard', 'dense', 'grid', 'small', or 'user'\nas an option for 'maximization'.\n");
  } else {
    if(matches(type,"STANDARD")) maxim_type=1;
    else if(matches(type,"DENSE")) maxim_type=2;
    else if(matches(type,"GRID")) maxim_type=4;
    else if(matches(type,"SMALL")) maxim_type=5;
    else if(matches(type,"USER")) maxim_type=6;
    else usage_error(1);
    if (maxim_type == 1) {
      sf(ps,"Maximization will now be performed with the standard set of predefined\ntrait models, followed by a fine maximization.\n");
      pr();
    } else if (maxim_type == 2) {
      sf(ps,"Maximization will now be performed with the larger set of predefined\ntrait models, followed by a fine maximization.\n");
      pr();
      if (allfreq_restriction) {
	sf(ps,"Please consider turning 'allfreq restriction off' in order to cover\na wider range of values for the disease allele frequency as well.\n");
	pr();
      }
    } else if (maxim_type == 4) {
      sf(ps,"Maximization will now be done only with the larger set of predefined\ntrait models, without a fine maximization.\n");
      pr();
    } else if (maxim_type == 5) {
      sf(ps,"Maximization will now be done only with the standard set of predefined\ntrait models, without a fine maximization.\n");
      pr();
    } else if (maxim_type == 6) {
      sf(ps,"Maximization will now be done only with the set of user-defined trait models\n(those entered by the 'model' command), without a fine maximization.\n");
      pr();
    }
    if (modscore_done && (maxim_type != old_val)) {
	/* do not allow p-value calculation if the maximization
	   type is changed after executing 'modscore' */
	changed_modscore_options = TRUE;
    }
  }
}


command set_number_of_replicates(void)
{
  int new_val;
  char type[TOKLEN+1];
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    if (num_replicates == 0) {
      sf(ps,"P-value calculation is not active (number of replicates = 0).\n"); pr();
    } else { 
      sf(ps,"%d replicates are currently used to calculate p-values\n", num_replicates);
      pr();
    }
    return;
  }

  new_val = atoi(type);

  if (new_val < 0) {
    error("Number of replicates must be > 0 (or = 0 -> off).\n");
  }
  if (scan_done && new_val != num_replicates) {
    replicates_error_flag = TRUE;
    error("Number of replicates must be specified before 'scan pedigrees'.");
  }
  num_replicates = new_val;

  if (num_replicates == 0) {
    sf(ps,"P-value calculation is no longer active (number of replicates = 0).\n"); pr();
  } else if (num_replicates == 1) {
    sf(ps,"%d replicate will be used to calculate p-value.\n", num_replicates); pr();
  } else if (num_replicates > 1) {
    sf(ps,"%d replicates will be used to calculate p-value.\n", num_replicates); pr();
  } else usage_error(1);
}

command set_sequential_simulation(void)
{
  int new_val;

  get_one_arg(itoken,sequential_sim,&new_val);
  sequential_sim = new_val;

  if (sequential_sim == 0) {
    sf(ps,"Sequential simulation is deactivated. The full number of replicates\n"
          "provided with the 'number of replicates' command will be simulated and\n"
          "analyzed.\n"); pr();
  } else if (sequential_sim > 0) {
    if (sequential_sim == 1)
      sf(ps,"Sequential simulation is activated. %d replicate must at least\n"
          "provide a score equal to the real data score.\n", sequential_sim);
    else 
      sf(ps,"Sequential simulation is activated. %d replicates must at least\n"
          "provide a score equal to the real data score.\n", sequential_sim);
    pr();

    sf(ps,"\nPlease take in mind, that the overall number of replicates you\n"
          "have choosen will stay the maximum number of replicates that will\n"
          "be analyzed.\n"); pr();
  } else if (sequential_sim < 0) {
    sf(ps,"Number of replicates, which must at least provide a score equal to\n"
          "the real data score, must be > 0! Instead of your choice of \n"
          "'%d', now 50 replicates must provide a score equal or higher\n"
          "than the real data score.\n", sequential_sim);
    pr();
    sf(ps,"\nPlease keep in mind that the overall number of replicates is\n"
          "still bounded by the number that you have specified using the\n"
          "'number of replicates' option\n");
    pr();
    sequential_sim = 50;
  } else usage_error(1);
}

command set_store_replicates(void)
{
  char type[TOKLEN+1];
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    sf(ps,"Prefiles of all replicates will %s stored.\n",
       replicates_store_pre?"be":"not be"); pr();
    sf(ps,"Analysis results for all replicates will %s stored.\n",
       replicates_store_result?"be":"not be"); pr();
  } else {
    if (matches(type,"PREFILE") || matches(type,"PRE")) {
      replicates_store_pre=TRUE;
      replicates_store_result=FALSE;
    } else if(matches(type,"BOTH") || matches(type,"ALL")) {
      replicates_store_pre=TRUE;
      replicates_store_result=TRUE;
    } else if(matches(type,"OFF") || matches(type,"FALSE")) {
      replicates_store_pre=FALSE;
      replicates_store_result=FALSE;
    } else usage_error(1);
    sf(ps,"The prefiles of all replicates will %s stored.\n",
       replicates_store_pre?"now be":"no longer be"); pr();
    sf(ps,"The analysis results for all replicates will %s stored.\n",
       replicates_store_result?"now be":"no longer be"); pr();
  }
}

command set_untyped_founders(void)
{
  char type[TOKLEN+1];
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    sf(ps,"Setting all founders' simulated genotypes to unknown is '%s'.\n",
       untyped_founders?"on":"off"); pr();
  } else {
    if (matches(type,"ON") || matches(type,"TRUE")) 
      untyped_founders=TRUE;
    else if(matches(type,"OFF") || matches(type,"FALSE"))
      untyped_founders=FALSE;
    else usage_error(1);

    sf(ps,"Setting all founders' simulated genotypes to unknown is '%s'.\n",
       untyped_founders?"on":"off"); pr();
  }
}

command set_marker_information(void)
{
  char type[TOKLEN+1];
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    sf(ps,"Founder genotypes %s be simulated with distinguishable alleles.\n",
       full_marker_inf?"will":"will not"); pr();
  } else {
    if (matches(type,"ON") || matches(type,"TRUE")) 
      full_marker_inf=TRUE;
    else if(matches(type,"OFF") || matches(type,"FALSE"))
      full_marker_inf=FALSE;
    else usage_error(1);

    sf(ps,"Founder genotypes %s be simulated with distinguishable alleles.\n",
       full_marker_inf?"will now":"will no longer"); pr();
  }
}

command set_simulate_untyped(void)
{
  char type[TOKLEN+1];
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
     sf(ps,"Missing genotypes in the real data set %s be filled in the simulated data sets.\n",
       simulate_untyped?"will":"will not"); pr();
  } else {
    if (matches(type,"ON") || matches(type,"TRUE")) 
      simulate_untyped=TRUE;
    else if(matches(type,"OFF") || matches(type,"FALSE"))
      simulate_untyped=FALSE;
    else usage_error(1);

    sf(ps,"Missing genotypes in the real data set %s be filled in the simulated data sets.\n",
       simulate_untyped?"will":"will no longer"); pr();
  }
}

command set_best_position(void)
{
  char type[TOKLEN+1];
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    sf(ps,"P-value calculation only at the best position is '%s'.\n",
       best_position?"on":"off"); pr();
  } else {
    if (matches(type,"ON") || matches(type,"TRUE")) 
      best_position=TRUE;
    else if(matches(type,"OFF") || matches(type,"FALSE"))
      best_position=FALSE;
    else usage_error(1);
    
    sf(ps,"P-value calculation only at the best position is now '%s'.\n",
       best_position?"on":"off"); pr();
  }
}

command set_debug_out_pv(void)
{
  char type[TOKLEN+1];
  get_one_arg(stoken,"",type);

  if (!replicates_store_result) {
    sf(ps,"Storing of replicates must be activated to give a detailed output.\n");
    error(ps);
  }

  if (nullstr(type)) {
    sf(ps,"Detailed output for p-value calculation is '%s'.\n",
       debug_out_pv?"on":"off"); pr();
  } else {
    if (matches(type,"ON") || matches(type,"TRUE")) 
      debug_out_pv=TRUE;
    else if(matches(type,"OFF") || matches(type,"FALSE"))
      debug_out_pv=FALSE;
    else usage_error(1);

    sf(ps,"Detailed output for p-value calculation is now '%s'.\n",
       debug_out_pv?"on":"off"); pr();
  }
}

command set_random_seed(void)
{
    int seed;
    get_one_arg(itoken, -1, &seed);
    if (seed < 0) {
        sf(ps, "Invalid random seed.  Random seed must be in [0, 32767].  "
               "System time will be used instead.\n");
        pr();
        g_random_seed = time(NULL);
    } else {
        sf(ps, "Random seed will be %d.\n", seed);
        pr();
        g_random_seed = seed;
    }
}

command set_distribution(void)
{
  char type[TOKLEN+1];
  get_one_arg(stoken,"",type);

  if (nullstr(type)) {
    sf(ps,"Providing information on the distribution of simulated scores is '%s'\n",
           show_distribution?"on":"off"); pr();
  } else {
    if (matches(type,"ON") || matches(type,"TRUE")) 
      show_distribution=TRUE;
    else if(matches(type,"OFF") || matches(type,"FALSE"))
      show_distribution=FALSE;
    else usage_error(1);

    sf(ps,"Providing information on the distribution of simulated scores is now '%s'\n",
           show_distribution?"on":"off"); pr();
    }
}

command set_norm_allfreq(void)
{
  char type[TOKLEN+1];
  get_one_arg(stoken,"",type);

  if (num_markers > 0) {
    print("\nMarkers have already been loaded, this command will therefore be ignored.\n");
  } else {
    if (nullstr(type)) {
      if (norm_allfreq == 2)
        sf(ps,"Using normalized marker allele frequencies is off,\n"
              "and the analysis is performed even if marker allele\n"
              "frequencies sum to a value <= 0.9 or >= 1.1.\n");
      else if (norm_allfreq == 1)
        sf(ps,"Using normalized marker allele frequencies is on.\n");
      else if (norm_allfreq == 0)
        sf(ps,"Using normalized marker allele frequencies is off,\n"
              "but the analysis is stopped if marker allele frequencies\n"
              "sum to a value <= 0.9 or >= 1.1.\n");
      pr();
    } else {
      if (matches(type,"OLD") || matches(type,"old"))
	norm_allfreq = 2;
      else if (matches(type,"ON") || matches(type,"TRUE")) 
        norm_allfreq = 1;
      else if(matches(type,"OFF") || matches(type,"FALSE"))
        norm_allfreq = 0;
      else usage_error(1);

      if (norm_allfreq == 2)
	sf(ps,"Using normalized marker allele frequencies is now off,\n"
	      "and the analysis is performed even if marker allele\n"
              "frequencies sum to a value <= 0.9 or >= 1.1.\n");
      else if (norm_allfreq == 1)
        sf(ps,"Using normalized marker allele frequencies is now on.\n");
      else if (norm_allfreq == 0)
        sf(ps,"Using normalized marker allele frequencies is now off,\n"
              "but the analysis is stopped if marker allele frequencies\n"
              "sum to a value <= 0.9 or >= 1.1.\n");
      pr();
    }
  }
}

command algebraic_calculation(void)
{
	char type[TOKLEN+1];
	get_one_arg(stoken,"",type);

	if (nullstr(type)) {
		sf(ps,"Algebraic calculation mode is '%s'.\n",
				algebra?"on":"off"); pr();
	}    
	else {
		if (scan_done) {
			/* this is not possible! */
			sf(ps, "Calculation mode cannot be changed after pedigrees have been loaded.\n");
			pr();
		}
		else if (matches(type,"ON") || matches(type,"TRUE"))
		{
			if (analysis_type == NPL_ANALYSIS)
			{
				sf(ps, "Algebraic calculation mode is only useful for\n"
					   "analysis types 'LOD' and 'BOTH'.\n"
					   "Command is therefore ignored.\n");
				pr();
			}
			if(analysis_type == LOD_ANALYSIS || analysis_type == BOTH_ANALYSES) {
				algebra = TRUE;
				sf(ps,"Algebraic calculation mode is now '%s'.\n",
						algebra?"on":"off"); pr();
			}
		}
		else if(matches(type,"OFF") || matches(type,"FALSE"))
		{
			algebra=FALSE;
			sf(ps,"Algebraic calculation mode is now '%s'.\n",
					algebra?"on":"off"); pr();
		}
		else usage_error(1);
	}
}

command mobit_simulation(void)
{
	char type[TOKLEN+1];
	get_one_arg(stoken,"",type);

	if (nullstr(type)) {
		sf(ps,"MOBIT simulation mode is '%s'.\n",
				mobit_simulate?"on":"off"); pr();
	}    
	else {
		if (scan_done) {
			/* this is not possible! */
			sf(ps, "MOBIT simulation mode cannot be specified after pedigrees have been loaded.\n");
			pr();
			mobit_simulate = FALSE;
		}
		else if (pmod_calc_flag) {
			sf(ps, "Cannot execute 'mobit simulation' command after 'calculate p value' command.  These commands are mutually exclusive.\n");
			pr();
			mobit_simulate = FALSE;
		}
		else if (matches(type,"ON") || matches(type,"TRUE"))
		{
			if (analysis_type == NPL_ANALYSIS || algebra == FALSE)
			{
				sf(ps, "MOBIT simulation mode is only possible when algebraic calculation mode\n"
                       "is set to 'on' and analysis types are either\n"
					   "'LOD' or 'BOTH'.\n"
					   "Command is therefore ignored.\n");
				pr();
				mobit_simulate = FALSE;
			}
			if((analysis_type == LOD_ANALYSIS || analysis_type == BOTH_ANALYSES) && algebra == TRUE) {
				mobit_simulate = TRUE;
				sf(ps,"*** WARNING ***\n"
                      "MOBIT simulation mode is now '%s'.\n", mobit_simulate?"on":"off");
                pr();
                sf(ps, "Sample is analyzed with prior permutation of parental inheritance.\n"
                       "Several independent GHM runs using this program option can hence be combined to\n"
                       "give a permutation p value for the MOBIT imprinting test stastic, i.e.,\n"
                       "MOD score taking imprinting into account - MOD score.\n");
                pr();
			}
		}
		else if(matches(type,"OFF") || matches(type,"FALSE"))
		{
			mobit_simulate = FALSE;
			sf(ps,"MOBIT simulation mode is now '%s'.\n",
					mobit_simulate?"on":"off"); pr();
		}
		else usage_error(1);
	}
}
