/* GENEHUNTER - Kruglyak, Daly, Reeve-Daly, and Lander
   Whitehead Institute for Biomedical Research, Copyright 1996
   modified by Konstantin Strauch, Johannes Dietter and 
   Manuel Mattheisen, Copyright (c) 2004-2008 */
   
#define INC_LIB
#define INC_MISC
#define INC_SHELL
#include "npl.h"
#include <strings.h>

#define REMOVE_BITS

void set_dad_index(int,int);
void set_mom_index(int,int);
void check_for_unlinked(int);
void sort_pedigree(int);
void resort_pedigree(int non_originals);
void check_for_loop(int,int,int,int,int);
void fill_placeholder_alleles_0(int, int, int, PED_MEMBER *);
void fill_placeholder_alleles_1(int, int, PED_MEMBER *);

/* Internal methods */
void sim_linkage_pedigree(char*);
void simulate_family(int);
void add_rec(int bit, double pos);
void tally_score(double this_score, int sim, double this_pos);


int num_pedigrees;

int num_sims, num_chroms_to_sim;
double chrom_length;

typedef struct {
    double pos;
    int bit;
} RECOMBINATION;

RECOMBINATION rec[1000];
int num_recs;

typedef struct {
    double *score;
    double *pos;
    int num_regions;
} SIMCHROM;

SIMCHROM **sim_chrom;

double get_max_NPL(int*);

command simulation(void)
{

  /* Determine the type of pedigree we are dealing with (LINKAGE
     or our own internal format and call the appropriate method) */

  char *filename,s1[100],*tempstr;
  FILE *fpin;
  void prep_stanford_pedigree(char*),prep_linkage_pedigree(char*);
  int i,j,n,a,b,c,d,e;
  double chisq_prob(double,double);
  int *choice;
  double *NPLmax;

  run {
    filename = get_temp_string();
    use_uncrunched_args();
    get_one_arg(stoken,"",filename);
	  
    if (nullstr(filename)) print("need filename as argument\n");
    
    fpin = open_file(filename,"r");
    _fgetln(fpin);
    while(nullstr(ln)) _fgetln(fpin);

    n = sscanf(ln,"%s %d %d %d %d %d",s1,&a,&b,&c,&d,&e);

    tempstr = get_temp_string();
    input("number of simulations: ",tempstr,TEMP_STRING_LEN);
    itoken(&tempstr,100000,&num_sims);

    tempstr = get_temp_string();
    input("chromosome length: ",tempstr,TEMP_STRING_LEN);
    rtoken(&tempstr,150.0,&chrom_length);

    num_chroms_to_sim=1000;

    if (n == 6)  {
      matrix (sim_chrom, MAX_PEDIGREES, num_chroms_to_sim, SIMCHROM);
      
      sim_linkage_pedigree(filename); /* LINKAGE has 6 or more fields */

    } else { 
      sf(ps,"error: can't recognize first line of pedigree file (%s)\n",ln);
      pr();
    }

    /* Now tally the simulation results */
    array(NPLmax, num_sims, double);
    array(choice, num_pedigrees, int);

    for (i=0; i<num_sims; i++) {

      /* select a random simulated chromosome from each pedigree */
      for (j=0; j<num_pedigrees; j++) {
	choice[j] = num_chroms_to_sim*randnum();
      }
      NPLmax[i] = get_max_NPL(choice);
    }
    inv_rsort(NPLmax, num_sims);

    sf(ps,"Maximum NPL seen: %.2lf\n",NPLmax[0]); pr();
    sf(ps,".0001: %.2lf\n",NPLmax[num_sims/10000]); pr();
    sf(ps,".001: %.2lf\n",NPLmax[num_sims/1000]); pr();
    sf(ps,".002: %.2lf\n",NPLmax[num_sims/500]); pr();
    sf(ps,".0025: %.2lf\n",NPLmax[num_sims/400]); pr();
    sf(ps,".003: %.2lf\n",NPLmax[(num_sims*3)/1000]); pr();
    sf(ps,".004: %.2lf\n",NPLmax[(num_sims*4)/1000]); pr();
    sf(ps,".005: %.2lf\n",NPLmax[(num_sims*5)/1000]); pr();
    sf(ps,".006: %.2lf\n",NPLmax[(num_sims*6)/1000]); pr();
    sf(ps,".007: %.2lf\n",NPLmax[(num_sims*7)/1000]); pr();
    sf(ps,".008: %.2lf\n",NPLmax[(num_sims*8)/1000]); pr();
    sf(ps,".009: %.2lf\n",NPLmax[(num_sims*9)/1000]); pr();
    sf(ps,".01: %.2lf\n",NPLmax[num_sims/100]); pr();
    sf(ps,".02: %.2lf\n",NPLmax[num_sims/50]); pr();
    sf(ps,".03: %.2lf\n",NPLmax[(num_sims*3)/100]); pr();
    sf(ps,".04: %.2lf\n",NPLmax[(num_sims*4)/100]); pr();
    sf(ps,".05: %.2lf\n",NPLmax[num_sims/20]); pr();

    i=0;
    while (NPLmax[i] > 4.05) i++;
    sf(ps,"4.05 exceeded %.5lf of the time\n",(double)i/(double)num_sims);
    pr();
  } on_error {
    if (msg == CANTOPEN) {
      sf(ps,"Can't open file '%s'\n",filename); pr();
    } else {
      relay_messages;
    }
  }
}


/*********************************/
/* LINKAGE PEDIGREE READING CODE */
/*********************************/

#define MISSING (-1) /* Value to set uncalculated fields to */
#define MALE 1
#define FEMALE 2
static char current_ped[MAX_NAME_SIZE];
static int num_allele_data[MAX_INDIVIDUALS];
static int list_index;

void sim_linkage_pedigree(char *filename)
{

  FILE * volatile fpin;
  char ped_name[MAX_NAME_SIZE];
  char indiv_ID[MAX_NAME_SIZE];
  char father_ID[MAX_NAME_SIZE], mother_ID[MAX_NAME_SIZE];
  int sex, j, datum;

  /* Open the input file with the LINKAGE format pedigrees, and
     the output file where the inheritance vectors will be written */

  num_pedigrees=0;

  run {

    fpin = NULL; 
    fpin = open_file(filename,"r");
    strcpy(current_filename, filename);

  } except_when(CANTOPEN) {
    sf(ps,"Can't open file: %s\n",filename); pr();
    return;
  }

  matrix(famtree,1,MAX_INDIVIDUALS+5,PED_MEMBER);
  run {
    list_index=0;
    strcpy(current_ped,"");
    while(1) {

      _fgetln(fpin);
      if (!maxstoken(&ln,sREQUIRED,ped_name,MAX_NAME_SIZE)) {
	/* blank line, read the next line */
      }
      else {
	/* actual line of data */
	if (nullstr(current_ped)) {
	  /* Record the new pedigree information for the first pedigree */
	  strcpy(current_ped,ped_name);
	  num_pedigrees++;
	} else if (strcasecmp(current_ped,ped_name) != 0) {
	  /* This line is the start of a new pedigree,
	     dump out the old one and reset the list index to 0 */
	  simulate_family(list_index); /* scans the family */

	  strcpy(current_ped,ped_name);
	  list_index=0;
	  num_pedigrees++;
	}
	
	/* Read the individual's index within the pedigree */
	if (!maxstoken(&ln,sREQUIRED,indiv_ID,MAX_NAME_SIZE)) {
	  sf(ps,"bad input line - no indivdual ID"); error(ps);
	}
	/* Read the parents' indices in the pedigree */
	if (!maxstoken(&ln,sREQUIRED,father_ID,MAX_NAME_SIZE)) {
	  sf(ps,"bad input line - no paternal ID for indiv %s",
	     indiv_ID); error(ps);
	}
	if (!maxstoken(&ln,sREQUIRED,mother_ID,MAX_NAME_SIZE)) {
	  sf(ps,"bad input line - no maternal ID for indiv %s",
	     indiv_ID); error(ps);
	}
	/* Read the sex of the individual */
	if (!itoken(&ln,iREQUIRED,&sex)) {
	  sf(ps,"bad input line - no sex for individual %s",indiv_ID);
	  error(ps);
	}
	
	/* Record the information in the family tree struct */
	strcpy(famtree[0][list_index].indiv_ID,indiv_ID);
	strcpy(famtree[0][list_index].dad,father_ID);
	strcpy(famtree[0][list_index].mom,mother_ID);
	famtree[0][list_index].sex=sex;
	/* Initialize values to be filled in later */
	famtree[0][list_index].nkids=0;
	famtree[0][list_index].dad_index=MISSING;
	famtree[0][list_index].mom_index=MISSING;
	
	/* Check that this isn't a duplicate indiv_ID */
	for (j=0; j<list_index; j++)
	  if (!strcasecmp(famtree[0][j].indiv_ID,
			  famtree[0][list_index].indiv_ID)) {
	    sf(ps,"indiv %s appears more than once in pedigree %s",
	       famtree[0][list_index].indiv_ID,current_ped); error(ps);
	  }
	    
	/* Read in the genotype data -store in static structs */
	j=0;
	while (itoken(&ln,iREQUIRED,&datum)) {
	  allele_data[list_index][j]=datum;
	  j++;
	}
	num_allele_data[list_index]=j;
	list_index++;
      }
    }	
    close_file(fpin);
    unarray(famtree,PED_MEMBER);
  } m_except {
      when SOFTABORT: 
          num_pedigrees = 0;
          close_file(fpin); fpin=NULL;
          unarray(famtree,PED_MEMBER);
          relay;
          break;
      when ENDOFILE:
          if (list_index > 0) {
	    run { simulate_family(list_index); } on_error {
	    }
	  }
          close_file(fpin); fpin=NULL;
          unarray(famtree,PED_MEMBER);
          break;
      default:
          close_file(fpin); fpin=NULL;
          unarray(famtree,PED_MEMBER);
          relay;
    }
}


void simulate_family(int num_in_ped)
{
  int i;
  int originals,non_originals,real_non_originals;
  int num_vecs, vec, cur_place, num_bits;
  volatile int skip_ped=FALSE;
  double graph_assign_probs(int,int,int,int,int**,bool);
  double exp, exp_2, apm_score(int**,int), varnull;
  double ** volatile pvector=NULL;
  double * volatile score=NULL;
  int num_aff, parents_informative(int);
  int ** volatile affected_allele = NULL;
  int * volatile uninformative_marker=NULL;
  int * volatile num_real_vecs=NULL;
  int * volatile best_pvector=NULL;
  int has_data(int);
  int real_num_bits,real_num_vecs,new_vec;
  int sim, pvec, mask;
  double this_length, rnum, next_rec, this_score;
  
  /* For all individuals, look up the person's mom and dad
     and record it in the pedigree structure */
  run {
    sf(ps,"\nsimulating pedigree %s...\n",current_ped); pr();

    if (num_allele_data[0]%2==1) 
      allele_offset = 1;
    else 
      allele_offset = 2;

    for (i=0; i<num_in_ped; i++) {
      set_dad_index(i,num_in_ped);
      set_mom_index(i,num_in_ped);
      famtree[0][i].analyzed_parent=FALSE; /* for counting inh. vec complexity */
      famtree[0][i].dad_bit_uninf=FALSE;
      famtree[0][i].mom_bit_uninf=FALSE;
      famtree[0][i].kids_found=FALSE; /* used later in bit chopping section */
    }

    for (i=0; i<num_in_ped; i++) famtree[0][i].discard=FALSE;

    /* get rid of unlinked individuals - set discard to TRUE */
    check_for_unlinked(num_in_ped);

    /* at this point we can eliminate from consideration:
        b) individuals that are unaffected and have no kids
           (they add next to nothing to a perfect-data NPL simulation) */

    for (i=0; i<num_in_ped; i++) {
      if (famtree[0][i].discard) continue;
      
      /* if (discard_unaffected) { */
      /* discard_unaffected is a flag ("discard") set at command level */

      /* Since we are simulating perfect data here, unaffecteds can
	 have no affect on the resulting NPL score so we skip them */
      if ( famtree[0][i].nkids == 0 && 
	  allele_data[i][0] != 2 ) {
	  famtree[0][i].discard = TRUE;
      }
    }
  
    originals = non_originals = 0;
    cur_place=0;

    /* set placeholder alleles for original pedigree members */
    for (i=0; i<num_in_ped; i++) {
      if (famtree[0][i].dad_index == MISSING &&
	  famtree[0][i].mom_index == MISSING) {
	if (!famtree[0][i].discard) {
	  originals++;
	  famtree[0][i].place_allele[0]=cur_place; cur_place++;
	  famtree[0][i].place_allele[1]=cur_place; cur_place++;
	}
      }

      else if (famtree[0][i].dad_index == MISSING ||
	       famtree[0][i].mom_index == MISSING) {
        sf(ps,"person %s has ONE parent",famtree[0][i].indiv_ID); error(ps);
      }
      else {
	if (!famtree[0][i].discard) non_originals++;
	famtree[0][i].place_allele[0]=-1;
	famtree[0][i].place_allele[1]=-1;
      }
    }

    allele_offset=1;
    
    /* fill in affectation_status and liability_class in PED_MEMBERs */
    for (i=0; i<num_in_ped; i++) {
      famtree[0][i].affectation_status = allele_data[i][0];
      if (allele_offset == 2) famtree[0][i].liability_class = allele_data[i][1];
      else famtree[0][i].liability_class = 1;
    }

    /* now let's do a preliminary sort of non-originals so that
       we can accurately determine how much analysis needs to be done */

    /* new_sort_pedigree(); */
    sort_pedigree(num_in_ped);
    resort_pedigree(non_originals);

    /* now read through this sorted list of non_originals until we
       hit the limit of our analytic capabilities */

    num_bits = real_num_bits = (non_originals*2);
    real_non_originals = non_originals;

    if (non_originals < 1) {
      sf(ps,"no affected non-original members in pedigree %s...SKIPPING\n",
	 current_ped);  pr();
      skip_ped = TRUE; num_pedigrees--;
    } 

    if (!skip_ped) {
  
      /* Check for impossible loops */
      for (i=0; i<num_in_ped; i++) {
	/* We've already know that no one can be their own
	   direct parent so we'll start up a level */
	if (famtree[0][i].dad_index != MISSING) { 
	  /* If one parent is missing,both are */
	  check_for_loop(i,famtree[0][i].sex,famtree[0][i].dad_index,1,num_in_ped);
	  check_for_loop(i,famtree[0][i].sex,famtree[0][i].mom_index,1,num_in_ped);
	}
      }

      /* we have 2^2*non_originals possible pvectors */
    
      /* now connect to graphing code for determining inh. vector probs */
      num_vecs=1;
      for (i=0; i<num_bits; i++) num_vecs *= 2;
      real_num_vecs=1;
      for (i=0; i<real_num_bits; i++) real_num_vecs *= 2;

      matrix (pvector, num_in_map_order+1, real_num_vecs, double);
      array (score, real_num_vecs, double);
      matrix (affected_allele, num_in_ped, 2, int);
      array (uninformative_marker, num_in_map_order+1, int);
      array (num_real_vecs, num_in_map_order+1, int);
      array (best_pvector, num_in_map_order+1, int);
    
      for (i=0; i<num_in_map_order; i++) {
	uninformative_marker[i]=TRUE; num_real_vecs[i]=0;
      }
      exp = exp_2 = 0.0; 
      
      new_vec=0;
    
      for (vec=0; vec<num_vecs; vec++) {

	if (vec == 0)
	  fill_placeholder_alleles_0(vec,real_non_originals,0,famtree[famnum]);

	/* now every member of famtree has two placeholder alleles
	   filled in and we can advance to the prob assignment
	   for each of the markers */

	fill_placeholder_alleles_1(vec,real_non_originals,famtree[famnum]);

	/* now obtain NPL score for this assignment of placeholder
	   alleles for this pedigree */

	for (i=0,num_aff=0; i<num_in_ped; i++) {
	  if (famtree[0][i].discard || famtree[0][i].place_allele[0]==-1) continue;

	  if (allele_data[i][0] == 2) {
	      /* this individual is affected */
	      affected_allele[num_aff][0] = famtree[0][i].place_allele[0];
	      affected_allele[num_aff][1] = famtree[0][i].place_allele[1];
	      num_aff++;
	  }
	}
	score[new_vec] = apm_score(affected_allele, num_aff);
	exp += score[new_vec];
	exp_2 += (score[new_vec]*score[new_vec]);

	new_vec++;
      }

      if (new_vec != real_num_vecs) error("num_vecs don't match in hackprep");
      
      exp /= real_num_vecs;
      exp_2 /= real_num_vecs;
      varnull = exp_2 - (exp*exp);

      /* Now we've stored the score for each pvector, we now need to
	 simulate a chromosome by choosing one at random and then
	 having bits flip to represent crossovers */

      for (sim=0; sim<num_chroms_to_sim; sim++) {
	array(sim_chrom[num_pedigrees-1][sim].pos, num_bits*5, double);
	array(sim_chrom[num_pedigrees-1][sim].score, num_bits*5, double);
	sim_chrom[num_pedigrees-1][sim].num_regions=0;

	/* simulate recombinations */
	num_recs=0;
	for (i=0; i<num_bits; i++) {
	    this_length=0.0;
	    while (this_length < chrom_length) {
	        rnum = randnum();
		next_rec = -100.0 * log(rnum);
		this_length += next_rec;
		if (this_length < chrom_length)
		  add_rec(i,this_length);
	    }
	}

	/* pick a random starting pvector */
	rnum = randnum();
	pvec = (int)(rnum*num_vecs);
	this_score = (score[pvec]-exp)/sqrt(varnull);
	tally_score(this_score,sim,0.0);

	for (i=0; i<num_recs; i++) {
	  mask = m_pow2[rec[i].bit];
	  pvec = pvec ^ mask;
	  this_score = (score[pvec]-exp)/sqrt(varnull);
	  tally_score(this_score,sim,rec[i].pos);
	}
      }

    } /* if (!skip_ped) */
  
  } on_exit {
    /* free memory */
    unarray(best_pvector, int);
    unarray(num_real_vecs, int);
    unarray(uninformative_marker, int);
    unmatrix (affected_allele, num_in_ped, int);
    unarray (score, double);
    unmatrix (pvector, num_in_map_order+1, double);
    relay_messages;
  }
}


void add_rec(int bit, double pos)
{
    int i=0, j;

    while (i<num_recs) {
      if (rec[i].pos > pos) break;
      i++;
    }
    
    for (j=num_recs-1; j>=i; j--) {
      rec[j+1].pos=rec[j].pos;
      rec[j+1].bit=rec[j].bit;
    }
    rec[i].pos=pos;
    rec[i].bit=bit;
    num_recs++;
}


void tally_score(double this_score, int sim, double this_pos)
{
    int num;

    num = sim_chrom[num_pedigrees-1][sim].num_regions;
    sim_chrom[num_pedigrees-1][sim].score[num]=this_score;
    sim_chrom[num_pedigrees-1][sim].pos[num]=this_pos;
    num++;
    sim_chrom[num_pedigrees-1][sim].num_regions=num;
}

double get_max_NPL(int *choice)
{
    int i,position[MAX_PEDIGREES],minsim;
    double max, this_score, minpos, compute_score(int*,int*);

    for (i=0; i<num_pedigrees; i++) position[i]=0;

    max = compute_score(choice,position);

    while (TRUE) {
      minpos=9999.0; minsim=-1;
      for (i=0; i<num_pedigrees; i++) {
	if (position[i]+1 == sim_chrom[i][choice[i]].num_regions) continue;

	if (sim_chrom[i][choice[i]].pos[position[i]+1] < minpos) {
	  minpos=sim_chrom[i][choice[i]].pos[position[i]+1];
	  minsim=i;
	}
      }
      if (minsim < 0) break;

      position[minsim] = position[minsim]+1;
      this_score = compute_score(choice,position);
      if (this_score > max) max=this_score;
    }

    return(max);
}

double compute_score(int *choice, int *position)
{
    int i;
    double sum=0.0;

    for (i=0; i<num_pedigrees; i++) {
      sum += sim_chrom[i][choice[i]].score[position[i]];
    }
    return(sum/sqrt(num_pedigrees));
}

