#include "List.h"
#include "Memory.h"

struct ListStruct {
    void *item;
    List next;
};

static List create_empty_list(void);

List List_Create(void)
{
    return NULL;
}

void List_Destroy(List list)
{
    List current_element, next_element;

    current_element = list;
    while (current_element) {
        next_element = current_element->next;
        Memory_Free(current_element);
        current_element = next_element;
    }
}

List List_AddItem(List list, void *item)
{
    List new_list = create_empty_list();

    new_list->item = item;
    new_list->next = list;

    return new_list;
}

static List create_empty_list(void)
{
    List list = (List) Memory_Malloc(sizeof *list);

    return list;
}

List List_GetNext(List list)
{
    return list->next;
}

void *List_GetItem(List list)
{
    return list->item;
}

int List_GetSize(List list)
{
    List p;
    int size = 0;

    for (p = list; p; p = p->next)
        ++size;

    return size;
}

int List_HasNext(List list)
{
    return list && list->next;
}

int List_IsEmpty(List list)
{
    return list == NULL;
}

List List_RemoveIf(Predicate predicate, List list, void *additional_arguments)
{
    List p, new_list = List_Create();

    for (p = list; p; p = p->next)
        if (!predicate(p->item, additional_arguments))
            new_list = List_AddItem(new_list, p->item);

    return new_list;
}
