#include "MathFunctions.h"
#include <stddef.h>

static void combine(Vector items, int n, Vector vs, Vector v, int k);

long Math_Factorial(int n)
{
    if (n == 0)
        return 1;
    long factorial = n;
    while (--n)
        factorial *= n;
    return factorial;
}

long Math_Choose(int n, int k)
{
    if (k == 0)
        return 0;
    return Math_Factorial(n) / (Math_Factorial(k) * Math_Factorial(n - k));
}

// Return a vector of vectors containing all combinations of n
// elements from items.
Vector Math_Combine(Vector items, int n)
{
    Vector combinations = Vector_Create(Math_Choose(Vector_GetSize(items), n));
    combine(items, n, combinations, Vector_CreateEmpty(), 0);
    return combinations;
}

// items -- elements based on which to construct combinations
// n -- desired size of combinations of elements of items
// vs -- vector of vectors with combination of n elements from items
// v -- current combination (possibly of size < n, so far)
// k -- next element of items that is eligible for inclusion in the
//      current combination
static void combine(Vector items, int n, Vector vs, Vector v, int k)
{
    if (Vector_GetSize(v) == n) {
        Vector_PushBack(vs, v);
        return;
    }
    for (int i = k; i < Vector_GetSize(items); i++) {
        Vector copy = Vector_Copy(v);
        Vector_PushBack(copy, Vector_Get(items, i));
        combine(items, n, vs, copy, i + 1);
    }
    Vector_Destroy(v);
}

int Math_Log2(int n)
{
    if (n <= 0)
        return -1;
    int result = 0;
    while ((n >>= 1))
        ++result;
    return result;
}

double Math_IntegerPower(double base, int exponent)
{
    double result = 1.0;
    while (exponent > 0) {
        if (exponent % 2 == 0) {
            base *= base;
            exponent /= 2;
        } else {
            result *= base;
            --exponent;
        }
    }
    return result;
}
