#include "OrderedSubsetGenerator.h"
#include "Memory.h"

struct OrderedSubsetGeneratorStruct {
    Vector items;
    int *indices;
    int size_of_subsets;
    int issued_subsets;
};

static int find_next_indices(OrderedSubsetGenerator);
static int update_indices_to_the_right(int *indices, int i, int number_of_indices, int length);

OrderedSubsetGenerator OrderedSubsetGenerator_Create(Vector items, int n)
{
    OrderedSubsetGenerator generator = Memory_Malloc(sizeof(*generator));
    generator->items = items;
    generator->size_of_subsets = n;
    generator->indices = Memory_Malloc(n * sizeof(int));
    for (int i = 0; i < n; i++)
        generator->indices[i] = i;
    generator->issued_subsets = 0;
    return generator;
}

Vector OrderedSubsetGenerator_Next(OrderedSubsetGenerator generator)
{
    if (find_next_indices(generator)) {
        Vector subset = Vector_Create(generator->size_of_subsets);
        for (int i = 0; i < generator->size_of_subsets; i++)
            Vector_PushBack(subset, Vector_Get(generator->items, generator->indices[i]));
        ++generator->issued_subsets;
        return subset;
    } else
        return NULL;
}

static int find_next_indices(OrderedSubsetGenerator generator)
{
    return generator->issued_subsets == 0
        || OrderedSubsetGenerator_IncrementOrderedIndices(generator->indices, generator->size_of_subsets, generator->items);
}

int OrderedSubsetGenerator_IncrementOrderedIndices(int *indices, int number_of_indices, Vector vector)
{
    int length = Vector_GetSize(vector);
    for (int i = number_of_indices - 1; i >= 0; i--) {
        ++indices[i];
        if (indices[i] < Vector_GetSize(vector)
                && update_indices_to_the_right(indices, i, number_of_indices, length))
            return 1;
    }
    return 0;
}

static int update_indices_to_the_right(int *indices, int i, int number_of_indices, int length)
{
    for (int j = i + 1; j < number_of_indices; j++) {
        indices[j] = indices[j - 1] + 1;
        if (indices[j] >= length)
            return 0;
    }
    return 1;
}

int OrderedSubsetGenerator_GetNumberOfSets(OrderedSubsetGenerator generator)
{
    return generator->size_of_subsets * Vector_GetSize(generator->items);
}
