#include "HashTableTemplateImpl.h"
#include "List.h"
#include "Memory.h"
#include "TripletHashTable.h"

#define HASH(key, table) ((unsigned long) Triplet_Hash(key) % (table)->size)
#define IS_SAME_KEY(K1, K2) Triplet_Equal((K1), (K2))

HashTableTemplateImpl(Triplet)

struct TripletHashTableIteratorStruct {
    List bucket;
    int bucket_index;
    int items_seen;
    int total_items;
    TripletHashTable table;
};

TripletHashTableIterator TripletHashTableIterator_Create(TripletHashTable table) {
    TripletHashTableIterator it = Memory_Malloc(sizeof *it);
    it->table = table;
    it->bucket = 0;
    it->bucket_index = 0;
    it->items_seen = 0;
    it->total_items = TripletHashTable_GetSize(table);
    return it;
}

Triplet TripletHashTableIterator_GetTriplet(TripletHashTableIterator it) {
    Element element = List_GetItem(it->bucket);
    Triplet triplet = element->key;
    return triplet;
}

void *TripletHashTableIterator_GetValue(TripletHashTableIterator it) {
    Element element = List_GetItem(it->bucket);
    Triplet key = element->key;
    void *value;
    int ok = TripletHashTable_Get(it->table, key, (void **) &value);
    return ok ? value : 0;
}

void TripletHashTableIterator_Next(TripletHashTableIterator it) {
    if (it->bucket == 0) {
        for (int i = 0; i < it->table->size; i++) {
            List bucket = it->table->buckets[i];
            if (List_IsEmpty(bucket)) {
                continue;
            }
            it->bucket = bucket;
            it->bucket_index = i;
            break;
        }
        it->items_seen++;
    } else if (List_HasNext(it->bucket)) {
        it->bucket = List_GetNext(it->bucket);
        it->items_seen++;
    } else if (TripletHashTableIterator_HasNext(it)) {
        while (List_IsEmpty(it->table->buckets[++it->bucket_index])) {
            ;
        }
        it->bucket = it->table->buckets[it->bucket_index];
        it->items_seen++;
    }
}

int TripletHashTableIterator_HasNext(TripletHashTableIterator it) {
    return it->items_seen < it->total_items;
}

void *TripletHashTable_GetExponents(TripletHashTable table, Triplet key) {
    void *exponents;
    int ok = TripletHashTable_Get(table, key, (void **) &exponents);
    return ok ? exponents : 0;
}
