/* GENEHUNTER - Kruglyak, Daly, Reeve-Daly, and Lander
   Whitehead Institute for Biomedical Research, Copyright 1996
   modified by Konstantin Strauch, Johannes Dietter, Manuel Mattheisen
   and Markus Brugger.    Copyright (c) 1999-2019 */

#define INC_LIB
#define INC_SHELL
#include "dSFMT.h"
#include "nlopt.h"
#include "npl.h"
#include "time.h"
#include "limits.h"
#include "math.h"
#include "memory_realloc.h"
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#define REMOVE_BITS

#ifndef _XOPEN_SOURCE
#define _XOPEN_SOURCE 500   # Feature Test Macro for strdup(3)
#endif
#include <string.h>

#include "Association.h"
#include "EM.h"
#include "LinkageMarkerFile.h"
#include "Marker.h"
#include "Memory.h"
#include "PedigreeFile.h"
#include "UserAssociationRegionsFile.h"
#include "UserLinkageMarkerFile.h"
#include "Vector.h"

/* for FindIllegalBits() */
#define COMP_GENOTYPES(a,b,c,d) ((a==c && b==d) || (a==d && b==c))

/* For Debugging and Testing, define as necessary: */
#undef PR_00
#undef PR_01
#undef PR_02  /* not used */
#undef PR_03
#undef PR_04
#undef PR_05
#undef PR_06
#undef PR_07  /* not used */
#undef PR_08  /* not used */
#undef PR_09  /* not used */
#undef PR_10  /* not used */
#undef PR_11  /* not used */
/* warning: if use pr_12 or 13, must define both to either 0 or 1 (or non-0)*/
/*   define PR_12 1 and define pr_13 0  */
#undef PR_12 /* search result absolute restrictions*/
#undef PR_13 /* search result relative restrictions*/
#undef PR_14  /* not used */
#undef PR_15
#undef PR_16
#undef PR_17
#undef PR_18
#undef PR_19  /* Debugging for maskHomP */
#undef PR_20  /* not used */

/* partition prior computation */
/* set to 1 as necessary: */
#define SKIP_SHORTCUT 0
#define TEST_SHORTCUT 0
#define EXIT_EARLY 0
#define COMPUTE_SIZE_ONLY 0

#undef DUMP_SINGLE_P_INFOC
#undef DUMP_TIME

const char *g_pedigree_file = NULL;

int *g_is_singleton;
int *g_is_trio;
int g_first_pedigree;

void allocate_pair_analysis_storage(void);
void extract_warray(void);
void set_weights(int phenotype_index);
void make_member_list(char *ped_name, int total_num_members);
void scan_pedigree(int *nz_bits_s, int *nz_bits_s_m, int *nz_bits_s_f, unsigned long *nz_pointer_s, double *pvct_s, double *num_par_score_s, int f_num_in_map_order_s, int *f_uninformative_s, int *f_map_order_s, int real_num_bits, int real_num_bits_m_s, int real_num_bits_f_s, double *score, int score_mask_rdc_s, double exp, double exp2, int pednum, int *best_pv, int num_bits, int num_bits_m_s, int num_bits_f_s, int u_mask, int *f_mask, int **maskT, int this_mod_pos_s);

int ***save_maskTest=NULL;  /* make it 3dim and extern for MOD-score version */
static int *maskHomP;
static int inconsistent_genotypes[MAXLOCI];
static double  pr_time[200], pr_time_all=0;

/* Internal methods */
void prep_linkage_pedigree(char*);
void analyze_family(int);
void orig_haplotypes (int,int*,int**,int**,int**);
void draw_pedigree(int,int,int*,char*,int**,int*,int*,int,int**,int,int**);
void resort_pedigree(int);
int Define_family_map_order(int, int, int, int *, int *, int *, int, int);
void fill_placeholder_alleles_2(int, int, BITS, int, int);
int FindIllegalBits(int, int, int, int, int, int *, int *);
void Set_npl_score_masks(int *, int *, int*, int, int);
void print_requirements(int, int, int, int, int, int, int *, int, int *);
void fill_placeholder_alleles_0(int, int, int, PED_MEMBER *);
int Setup_local_pad_inhvec_2(int, int, int *, int *);
void fill_placeholder_alleles_1(int, int, PED_MEMBER *);

static int total_meioses;
static int total_meioses_m;
static int total_meioses_f;
static int total_affecteds[12];
int k_full[64];
FILE *haplofp;

int **sorted_ped;                       /* Store the indices of
					   famtree sorted so that all
					   ancestors are at the front
					   of the list */
                       /* sorted_ped: use a matrix for MOD-score version */
                       /* sorted_ped: extern for PMOD */

static bool placed[MAX_INDIVIDUALS]; /* Boolean list same length as famtree
					 indicating whether or not the 
					 corresponding indiv in famtree has 
					 been placed yet or not */
static int num_placed;
         /* Keep track of how many individuals have been placed */

double npl_pos, npl_max;

/* new variables for MOD-score analysis; need to store them for all peds */
/* all extern */
int famnum;
int num_pedigrees_allocated;
int *save_num_in_ped;
int *save_real_non_originals;
int *save_num_inbred;
int *save_old_mom;
int *save_num_bits;
int *save_num_bits_m;
int *save_num_bits_f;
int *save_real_num_bits;
int *save_real_num_bits_m;
int *save_real_num_bits_f;
int *save_real_num_vecs;
int *save_npadMS;
int **save_padMask, **save_padShift;
int **save_nz_bits, **save_nz_bits_m, **save_nz_bits_f;
unsigned long **save_nz_pointer;
int *save_f_num_in_map_order;
int **save_f_map_order, **save_f_uninformative;
int *save_uninformative_mask, **save_flip_mask;
double **save_pvct;
double *save_unlinked, **save_p_one_pos, ***save_p_all_pos;
char **pedigree_names;
int datafile_model_incompatible, scan_error_flag;


/* The following variables are used for the calculation of p-Values for the MODs: */
int skipped_peds;              /* count skipped peds */
char skipped_peds_char[1000];  /* print skipped peds */
int famnum_count;              /* unbiased famnum with skipped peds */
int *famnum_pointer;           /* points on pedigree based on famnum_count */
int **quick_model_famnum;      /* quick_model calculated for this famnum? */
bool *saved_model_allocated;   /* saved_model_vectors allocated for this famnum? */
bool save_famtree_allocated;   /* was save_famtree allocated? */
bool diff_discard;             /* discarded indiv changed over replicates? */
double **save_par_score;       /* store datfile model, needs to be implemented */
bool *real_num_vecs_calculated;    /* only when real_vecs are calculated 
                               saved_model_vectors can be allocated in modscore.c*/

int *save_bf_originals, *save_bf_non_originals;
int **save_bf_original_list, **save_bf_non_original_list;

/* The following variables are used by algebra_force; the matrices need to be set up only once but must be stored in a multidimensional structure */

unsigned int **algebra_array; /* stores unique power combinations derived from new_vecs (more specifically from their or_vecs) */
unsigned int **alg_coeff_combi_check; /* binary (0/1) matrix that checks if a certain combination of powers and coefficient has already been found; 0 means no - 1 means yes; needed in algebra_matrix() only */
int **address_array; /* stores positions of final doubles for each unique powers-coefficient combination */
int *list_num_terms; /* stores number of unique powers-coefficient combinations for each inheritance vector class */
double *pen_0, *pen_1, *pen_2, *pen_3, *pen_4, *pen_5, *pen_6, *pen_7, *pen_8, *pen_9, *alg_afreq_1, *alg_afreq_2; /* store doubles derived from algebra_array */
int **vec_class_pointer; /* stores positions of unique powers-coefficient combinations for each inh. vector class */
int unique_alg_terms, unique_combis, vec_classes; /* a) gives number of unique power combinations; b) gives unique powers-coefficient combinations; c) gives highest inh. vector class starting with class 0 */
int *unique_terms_line, *unique_coeff; /* a) and b) store unique powers-coefficient combination; in a) is information about powers, in b) about coefficients */ 
int mem_limit1 = 1000, mem_increment1 = 1000, mem_limit2 = 40, mem_increment2 = 20;
int mem_limit3 = 40000, mem_increment3 = 40000, mem_limit4 = 500, mem_increment4 = 100;
int mem_limit5 = 1000, mem_increment5 = 500, mem_limit6 = 500, mem_increment6 = 500;
int mem_limit7 = 100, mem_increment7 = 100;
int mem_limit8 = 20, mem_increment8 = 20, mem_increment9 = 50;
int mem_limit9;
double *pow_prep_vec;
double *pow_coeff_prep_vec;
double **par_score_address_vec;
double *alg_prob_vec; /* for cache optimization, stores likelihood contribution of each class in array */
int **new_vec_to_alg_prob; /* with dimensions [famnum] and [new_vec], with pointer to position in alg_prob_vec */
int *vec_class_members; /* represents number of members of each inh. vector class */
int *num_fam_in_vec_class; /* how many families are in an inheritance vector class */
int **new_vecs_of_vec_class, **vec_class_to_family; /* a) stores for each inh. vector class its members in terms of new_vecs; b) stores IDs of pedigrees contributing new_vecs to vec_class; a) and b) are complementary arrays, such that information on new_vec and corresponding family is given when filling the par_score_vec class-wise; as cache-optimzed version is made, these vectors are not in use anymore */
int max_indiv;
double **par_score;
bool *algebra_analyzed; /* indicates whether a family has already been analyzed; needed with pmod */
int *is_currently_in_data; /* which version of a family was analyzed the last run; needed with pmod */
int **version_pointer; /* which position a family version has in the consecutively enlarged algebra_pointer; needed with pmod */
int *version_counter; /* how many versions of a family are there?; needed with pmod */
double **datfile_par_score;
double **untotal_datfile_par_score;
/* Both datfile_par_score and untotal_datfile_par_score are needed for
   analyze_family() to check whether the datfile model is compatible
   with the pedigree.  From a programming point of view they are ragged arrays, that
   is, arrays of arrays where the member arrays can have different
   length.  They differ from other dynamically allocated structures in the
   program in that neither the top-level "spine" array nor the
   member arrays are ever reallocated.  Therefore the memory location of
   both structures and their subcomponents never changes and it is safe
   to store pointers that directly reference those memory locations.
*/
int peds_pmod;
int **family_to_vec_class, *num_classes_in_family; /* which vec_classes are in a pedigree and how many are they?; needed for pmod & saved models */
int *algebra_pointer;
bool *was_skipped_before;
ALGEBRA_PED_MEMBER ***algebra_famtree;
/* New for MOBIT simulations */
int ***permutation_units; /* with dims [famnum][unit][member]; allocate with number of nonfounders and initialze with 0 */
int ***permutation_units_indexes; /* with dims [famnum][unit][parental indexes]; initialize with 0 (?) */
int *number_of_permutation_units; /* with dims [famnum]; initialize with 0 */
int **number_of_members_in_permutation_unit; /* with dims [famnum][unit]; initialize with 0 */
int **number_of_parental_indexes_in_unit; /* with dims [famnum][unit], initialize with 0 */
int **permutation_unit_flipped; /* with dims [famnum][unit][0 or 1]; initialize with 0; holds coin flip result */
int mobit_random_seed = -1;
time_t seed;

command pedigree_prep(void)
{

    /* Determine the type of pedigree we are dealing with (LINKAGE
       or our own internal format and call the appropriate method) */

    char *filename,*hfilename;
    char s1[MAX_NAME_SIZE], s2[MAX_NAME_SIZE],
    s3[MAX_NAME_SIZE], s4[MAX_NAME_SIZE];
    FILE *fpin;
    void prep_stanford_pedigree(void),prep_linkage_pedigree(char*);
    int i,n,a,b;
    double theta, chisq_prob(double,double), total_total;
    volatile int temp_ser;
    volatile int temp_hap;
    volatile int temp_num_in_map_order;
    int temp_map[MAXLOCI];

    if (num_markers == 0) {
        print("Please load marker information using 'load markers' first.\n");
        replicates_error_flag = TRUE;
        return;
    }

    if (num_in_map_order == 0) {
        print("No marker set has been selected for the analysis yet.\n");
        print("Please specify markers and distances with the 'use' command.\n");
        replicates_error_flag = TRUE;
        return;
    }

    if (scan_done && modcalc) {
        sf(ps,"The scan pedigrees command has already been called with 'modcalc %s'.\n", (modcalc==1)?"global":"single");  pr();
        print("This command will therefore be ignored.\n");
        return;
    }

    if (replicates_count > 0) {
        print("Since p-values have already been calculated, this command will be ignored.\n");
        return;
    }

    run {
        filename = get_temp_string();
        use_uncrunched_args();
        get_one_arg(stoken,"",filename);

        if (nullstr(filename)) print("need filename as argument\n");
        g_pedigree_file = strdup(filename);
        if(g_association) {
            g_ped_file = PedigreeFile_Create(g_pedigree_file);
            g_linkage_file = LinkageMarkerFile_Create(g_linkage_marker_file, g_ped_file);
            if (!g_frequencies) {
                if (g_user_linkage_file == NULL || g_user_linkage_marker_file == NULL || g_user_association_regions_file == NULL) {
                    sf(ps, "One or several input files for JLA analysis are missing. If you wanted to compute\nallele frequencies only, please use the 'allele frequency estimation' command.\n.");
                    pr();
                    exit(1);
                }
            }
            if (!LinkageMarkerFile_IsCompatibleWith(g_linkage_file, g_user_linkage_file)) {
                sf(ps, "Markers in %s do not match with those in %s\n",
                    g_user_linkage_marker_file, g_linkage_marker_file);
                pr();
                exit(1);
            }
            if (!g_frequencies) {
                g_jla_regions = UserAssociationRegionsFile_ReadRegions(
                    g_user_association_regions_file,
                    g_linkage_file,
                    g_user_linkage_file);
                if (g_jla_regions == NULL) {
                    sf(ps, "Markers in %s do not match with those in %s\n",
                        g_user_association_regions_file, g_linkage_marker_file);
                    pr();
                    exit(1);
                }
            }
            sf(ps, "No. of families: %d\n", PedigreeFile_GetNumberOfFamilies(g_ped_file));
            pr();
            print_distances(g_linkage_file);
            if (g_random_seed < 0) {
                g_random_seed = time(NULL);
            }
            init_gen_rand(g_random_seed);
            nlopt_srand((unsigned long) g_random_seed);
            if (!g_frequencies) {
                sf(ps, "Random seed for JLA analysis: %d\n", g_random_seed);
                pr();
            }
        }

        fpin = open_file(filename,"r");
        _fgetln(fpin);
        while(nullstr(ln)) _fgetln(fpin);

        n = sscanf(ln,"%s %s %s %s %d %d",s1,s2,s3,s4,&a,&b);
        close_file(fpin); fpin=NULL;

        if (n > 0) {

            sf(ps,"scan pedigrees is running...\n"); pr();

            if (haplotype_output) {
                hfilename = get_temp_string();
                strcpy(hfilename,"haplo.dump");
                haplofp = open_file(hfilename,"w");
            }

            if (single_point_mode) {
                temp_ser = show_expected_recs; show_expected_recs=FALSE;
                temp_hap = haplotype_output;   haplotype_output=FALSE;
                temp_num_in_map_order = num_in_map_order;
                for (i=0; i < num_in_map_order; i++) temp_map[i] = map_order[i];
                num_in_map_order = num_markers;
                for (i=0; i<num_markers; i++) map_order[i]=i;
            }

            total_meioses=0; 
            total_meioses_m=0;
            total_meioses_f=0;  
            g_first_pedigree = 0;
            prep_linkage_pedigree(filename); /* LINKAGE has 6 or more fields */
            if (g_association && !g_frequencies) {
                g_families = create_families_from_global_state(num_pedigrees);
                if (!g_jla_datfile) {
                    Vector markers = LinkageMarkerFile_GetMarkers(g_linkage_file);
                    Vector linkage_markers = Vector_CreateEmpty();
                    for (int i = 0; i < Vector_GetSize(markers); i++) {
                        Marker marker = Vector_Get(markers, i);
                        if (UserLinkageMarkerFile_Contains(g_user_linkage_file, Marker_GetName(marker))) {
                            Vector_PushBack(linkage_markers, marker);
                        }
                    }
                    EM_ComputeMarkerFrequencies(linkage_markers, g_families);
                    int number_of_linkage_markers = Vector_GetSize(linkage_markers);
                    int markers_processed = 0;
                    for (int i = 0; i < num_markers; i++) {
                        Marker marker = LinkageMarkerFile_GetMarkerByIndex(g_linkage_file, i);
                        if(!UserLinkageMarkerFile_Contains(g_user_linkage_file, Marker_GetName(marker))) {
                            continue;
                        }
                        markers_processed++;
                    }
                    if (number_of_linkage_markers != markers_processed) {
                        sf(ps, "Wrong number of linkage markers. Wanted %d, got %d.\n", number_of_linkage_markers, markers_processed);
                        pr();
                    }
                    char new_linkage_marker_file[500];
                    snprintf(new_linkage_marker_file, sizeof new_linkage_marker_file, "%s.new", g_linkage_marker_file);
                    LinkageMarkerFile_WriteWithNewLinkageFreqs(g_linkage_file, g_user_linkage_file, linkage_markers, new_linkage_marker_file);
                    restart_ghm_with_new_datfile(new_linkage_marker_file);
                    // Execution should never reach this point.
                    sf(ps, "failed to restart ghm\n");
                    pr();
                    exit(1);
                }
            }
            if (show_expected_recs && num_pedigrees>1) {
                total_total=0.0;
                sf(ps,"TOTAL RECOMBINATION COUNT (%d total meioses):\n",
                        total_meioses); pr();
                print("interval   theta  exp-recs obs-recs\n");
                for (i=0; i<num_in_map_order-1; i++) {
                    theta = dist_to_rec(map_distance[i]);
                    sf(ps,"%4d-%-4d %6.4lf %8.3lf %8.3lf\n",map_order[i]+1,
                            map_order[i+1]+1,theta,theta*total_meioses,total_recs[i]);
                    pr();
                    total_total += total_recs[i];
                }
                sf(ps,"RECTOT = %.2lf\n",total_total); pr(); 
            }
            unarray(total_recs, double);

            if (haplotype_output)
                close_file(haplofp);

#ifdef TDT_ACTIVE
            if (haplotype_output) {
                /* TDT-results */
                int j, k;
                for (k=0; k<1; k++) { /* was 12 */
                    /**********
                      sf(ps,"TDT Summary - GROUP %d - (%d non-original affecteds):\n",
                      k+1, total_affecteds[k]); pr();
                     **********/
                    sf(ps,"TDT Summary - (%d non-original affecteds):\n",
                            total_affecteds[k]); pr();

                    for (i=0; i<num_in_map_order; i++) {
                        sf(ps,"Marker %-3d       trans untrans  Chi2  p-val\n",map_order[i]+1); pr();
                        for (j=0; j<MAX_ALLELE_SIZE; j++) {
                            if (locus[map_order[i]].allele_count[j][k][0] > 0 ||
                                    locus[map_order[i]].allele_count[j][k][1] > 0) {
                                double diff = (double) (locus[map_order[i]].allele_count[j][k][0] -
                                        locus[map_order[i]].allele_count[j][k][1]);
                                double sum = (double)  (locus[map_order[i]].allele_count[j][k][0] +
                                        locus[map_order[i]].allele_count[j][k][1]);
                                double chisq = diff*diff/sum;
                                sf(ps,"M%-3d - Allele %-3d  %4d  %4d   %6.2lf  %.6lf\n",
                                        map_order[i]+1, j,
                                        locus[map_order[i]].allele_count[j][k][TRANSMITTED],
                                        locus[map_order[i]].allele_count[j][k][UNTRANSMITTED],
                                        chisq, chisq_prob(1.0, chisq)); pr();
                            }
                        }
                    }
                }
            }
#else
            (void) total_affecteds;
#endif

            if (single_point_mode) {
                show_expected_recs= temp_ser;
                haplotype_output= temp_hap;
                num_in_map_order= temp_num_in_map_order;
                for (i=0; i<num_in_map_order; i++) {
                    map_order[i] = temp_map[i];
                }
            }

        } else { 
            sf(ps,"error: can't recognize first line of pedigree file (%s)\n",ln);
            pr();
            replicates_error_flag = TRUE;
        }

        scan_done = TRUE; /*** new ***/

    } on_error {
        if (msg == CANTOPEN) {
            sf(ps,"Can't open file '%s'\n",filename); pr();
            replicates_error_flag = TRUE;
        } else {
            relay_messages;
        }
    }
}


void prep_stanford_pedigree(void)
{

  /* eliminated to save space */
  error("old pedigree format - not supported");
}

/*********************************/
/* LINKAGE PEDIGREE READING CODE */
/*********************************/

/***  typedef struct NUC_FAMILY taken out here; it is now in npl.h  ***
      nuc_fam[][] must be matrix and extern for MOD-score analysis  ***/

NUC_FAMILY **nuc_fam;     /*** matrix/extern for MOD-score analysis ***/
static int num_nuc_fam;
int *save_num_nuc_fams_to_peel;  /*** array/extern for MOD-score analysis ***/
int **order_assigned;  /*** use a matrix for MOD-score analysis ***/
                       /*** extern for PMOD ***/

#define MISSING (-1) /* Value to set uncalculated fields to */
#define MALE 1
#define FEMALE 2
static char current_ped[MAX_NAME_SIZE];
static int num_allele_data[MAX_INDIVIDUALS];
static char **pedigree_data;
int allele_offset;
int num_data;
int allele_data[MAX_INDIVIDUALS][MAXLOCI*2];


/* Linkage pedigree auxiliary methods */
void set_mom_index(int,int);
void set_dad_index(int,int);
void check_for_unlinked(int);
void check_for_loop(int,int,int,int,int);
void sort_pedigree(int);
void place_indiv(int);
void fill_nuc_families(int);
int inbred_matings(void);
int mark_pivots(int);
void order_inbred_family(void);
static int list_index;

int num_persons;

/* in pmod.c */
void check_discarded(int);
void check_zero_marker(int);

/* Katz, Aug 2000, Check for Homozygote parents and only calc one prior
 *  prob for each equiv class, look up the rest.
 *  The mask maskHomP has a bit set for each homozygote parent 
 */

void CheckHomozygotes(int num_bits, int real_num_bits, int num_in_ped, 
							 int f_num, int uninformative_bit[], 
							 int f_map_order[], int non_originals) {

  int i, j, jj, k, kk, l, numkids;
  int thebit, theoldbit;


#ifdef PR_19
  printf("---------------doing CheckHomozygotes------------\n\n");
  printf("--NuminPed: %d, numbits: %d, real_numbits: %d, f_num: %d, non_originals: %d--\n", 
			num_in_ped, num_bits, real_num_bits, f_num, non_originals);
#endif


/* initialize */
  for (j=0; j < f_num; j++) 
	 maskHomP[j]=0;

  for (i=0; i < num_in_ped; i++) {
	 if (famtree[famnum][i].nkids == 0) continue;
	 numkids=0;
	 for (jj=0; jj < f_num; jj++) {
		j=f_map_order[jj];

		if ((allele_data[i][allele_offset+(2*j)] ==
			 allele_data[i][allele_offset+(2*j+1)]) &&
			 (allele_data[i][allele_offset+(2*j)] != 0))	{

#ifdef PR_19
		  printf ("Homozygote %d (%s), (has %d,%d) marker: %d(%d), kids=\n",
					 i, famtree[famnum][i].indiv_ID, 
					 allele_data[i][allele_offset+(2*j)],
					 allele_data[i][allele_offset+(2*j+1)],jj,j);
#endif

		  for (kk=0; kk < non_originals; kk++) {
			 k=sorted_ped[famnum][kk];
			 if (k == i) continue;
			 if (famtree[famnum][k].dad_index == i) {
				numkids++;
				thebit = (num_bits - (famtree[famnum][k].inh_vec_pos)*2 -1);
				theoldbit=thebit;
				for (l=0; l < theoldbit; l++)
				  if ((uninformative_bit[l]) && (thebit > 0))
					 thebit--;

#ifdef PR_19
				printf ("     %d(%s) - Dad (bit %d (from %d))\n", 
						  k, famtree[famnum][k].indiv_ID, thebit,famtree[famnum][k].inh_vec_pos);
				if (famtree[famnum][k].dad_bit_uninf)
				  printf ("    *****Dad bit unif %d(%s)*****\n",
							 k, famtree[famnum][k].indiv_ID);

#endif
				if (famtree[famnum][k].discard) {
#ifdef PR_19
				  printf ("   ****Dad, discard this %d(%s)****\n",
							 k, famtree[famnum][k].indiv_ID);
#endif
				  continue;
				}

				if (!(famtree[famnum][k].dad_bit_uninf) && (thebit < real_num_bits))
					 maskHomP[jj] |= (1 << thebit);
				if (numkids == famtree[famnum][i].nkids) break;
			 }

			 if (famtree[famnum][k].mom_index == i) {
				numkids++;
				thebit = (num_bits - (famtree[famnum][k].inh_vec_pos)*2 -2);
				theoldbit=thebit;
				for (l=0; l < theoldbit; l++)
				  if ((uninformative_bit[l]) && (thebit > 0))
					 thebit--;

#ifdef PR_19
				printf ("     %d(%s) = Mom (bit %d (from %d))\n",
						  k, famtree[famnum][k].indiv_ID, thebit,famtree[famnum][k].inh_vec_pos);
				if (famtree[famnum][k].mom_bit_uninf)
				  printf ("    *****Mom bit unif %d(%s)*****\n",
							 k, famtree[famnum][k].indiv_ID);
#endif
				if (famtree[famnum][k].discard) {
#ifdef PR_19
				  printf ("   ****Mom, discard this %d(%s)****\n",
							 k, famtree[famnum][k].indiv_ID);
#endif
				  continue;
				}
					 if (!(famtree[famnum][k].mom_bit_uninf) && (thebit < real_num_bits))
						maskHomP[jj] |= (1 << thebit);

				if (numkids == famtree[famnum][i].nkids) break;
			 }
		  }
		}
#ifdef PR_19
		printf("----Mask for marker %d(from %d): %x\n----", j, jj, maskHomP[jj]);
#endif
	 }
  }
}



void prep_linkage_pedigree(char *filename)
{

  FILE * volatile fpin;
  char ped_name[MAX_NAME_SIZE];
  char indiv_ID[MAX_NAME_SIZE];
  char father_ID[MAX_NAME_SIZE], mother_ID[MAX_NAME_SIZE];
  char *temp_data;
  int sex, i, j, k, datum;
  double rdatum;
  void algebra_power_prep(void);

  /* Open the input file with the LINKAGE format pedigrees, and
     the output file where the inheritance vectors will be written */

  skipped_peds=0;
  num_pedigrees=0;
  num_persons=0;
  max_indiv = 0;
  replicates_error_flag = FALSE;
  save_famtree_allocated = FALSE;

  run {

    fpin = NULL; 
    fpin = open_file(filename,"r");
    strcpy(current_filename, filename);

  } except_when(CANTOPEN) {
    sf(ps,"(linkage_prep_pedigree) Can't open file : %s\n",filename); pr();
    replicates_error_flag = TRUE;
    return;
  }


  /*** MODSCORE: first check how many pedigrees and persons there are ***/

  famnum = 0;


  run {
    list_index=0;
    strcpy(current_ped,"");
    while(1) {
     _fgetln(fpin);
      if (!maxstoken(&ln,sREQUIRED,ped_name,MAX_NAME_SIZE)) {
        /* blank line, read the next line */
      }
      else {
        /* actual line of data */
        if (nullstr(current_ped)) {
          /* Record the new pedigree information for the first pedigree */
          strcpy(current_ped,ped_name);
          num_pedigrees++;
          famnum = num_pedigrees - 1;
        } else if (strcasecmp(current_ped,ped_name) != 0) {
          /* This line is the start of a new pedigree,
             dump out the old one and reset the list index to 0 */
          if (max_indiv < list_index) max_indiv = list_index;
	  strcpy(current_ped,ped_name);
          list_index = 0;
          num_pedigrees++;
	  famnum = num_pedigrees - 1;
        }
	list_index++;
        num_persons++;
      }
    }
    close_file(fpin);
  } m_except {
      when SOFTABORT:
          num_pedigrees = 0;
          close_file(fpin); fpin=NULL;
          relay;
          break;
      when ENDOFILE:
          close_file(fpin); fpin=NULL;
          break;
      default:
          close_file(fpin); fpin=NULL;
          relay;
  }
  if (num_pedigrees == 0) return;
  if (max_indiv < list_index) max_indiv = list_index;
  /*** end of check how many pedigrees and persons there are (MOD-score analysis) ***/
  sf(ps, "\nMaximum number of persons in a pedigree: %i\n", max_indiv); pr();
  max_indiv = max_indiv + 2;
  num_pedigrees_allocated = num_pedigrees;
  mem_limit9 = 2*num_pedigrees;

  sf(ps,"Number of pedigrees: %d\n", num_pedigrees); pr();
  sf(ps,"Total number of persons in pedigrees: %d\n", num_persons); pr();

  run {
    fpin = NULL;
    fpin = open_file(filename,"r");
    strcpy(current_filename, filename);
  } except_when(CANTOPEN) {
    sf(ps,"(PLP) Can't open file: %s\n",filename); pr();
    replicates_error_flag = TRUE;
    return;
  }

  matrix(save_famtree,num_pedigrees,MAX_INDIVIDUALS+5,SAVE_PED_MEMBER);
  save_famtree_allocated = TRUE;
  peds_pmod = num_pedigrees;

  if (algebra)
  {
      saved_models = 0;
      array(pow_prep_vec, mem_limit1, double);
      array(pow_coeff_prep_vec, mem_limit7, double);
      array(vec_class_members, mem_limit4, int);
      matrix(algebra_array, mem_limit1, 12, unsigned int); /* needed for algebra_force, MB - 06/2012 */
      matrix(alg_coeff_combi_check, mem_limit1, mem_limit2, unsigned int);
      for(i=0; i < mem_limit1; i++) {
          for(j=0; j < mem_limit2; j++) {
              alg_coeff_combi_check[i][j] = 0;
          }
      }
      matrix(address_array, mem_limit1, mem_limit2, int);
      matrix(vec_class_pointer, mem_limit4, mem_limit5, int);
      array(list_num_terms, mem_limit4, int);
      array(num_fam_in_vec_class, mem_limit4, int);
      array(unique_terms_line, mem_limit3, int);
      array(unique_coeff, mem_limit3, int);
      array(par_score_address_vec, mem_limit4, double*);
      array(alg_prob_vec, mem_limit7, double);
      unique_alg_terms = 0;
      unique_combis = 0;
      vec_classes = -1; /* meaning that no vec_class is there; vec_class 0 would be the first class then */
      matrix(family_to_vec_class, peds_pmod, mem_limit4, int);
      array(par_score, peds_pmod, double *);
      array(new_vec_to_alg_prob, peds_pmod, int *);
      array(num_classes_in_family, peds_pmod, int);

      for (i = 0; i < peds_pmod; i++) {
          num_classes_in_family[i] = 0;
          for(j=0; j < mem_limit4; j++) {
              family_to_vec_class[i][j] = 0;
          }
      }

      array(datfile_par_score, num_pedigrees, double*);
      array(untotal_datfile_par_score, num_pedigrees, double*);
      array(pen_0, max_indiv, double);
      array(pen_1, max_indiv, double);
      array(pen_2, max_indiv, double);
      array(pen_3, max_indiv, double);
      array(pen_4, max_indiv, double);
      array(pen_5, max_indiv, double);
      array(pen_6, max_indiv, double);
      array(pen_7, max_indiv, double);
      array(pen_8, max_indiv, double);
      array(pen_9, max_indiv, double);
      array(alg_afreq_1, max_indiv, double);
      array(alg_afreq_2, max_indiv, double);
      pen_0[0]=1;
      pen_1[0]=1;
      pen_2[0]=1;
      pen_3[0]=1;
      pen_4[0]=1;
      pen_5[0]=1;
      pen_6[0]=1;
      pen_7[0]=1;
      pen_8[0]=1;
      pen_9[0]=1;
      alg_afreq_1[0]=1;
      alg_afreq_2[0]=1;
      algebra_power_prep();
      if(mobit_simulate) {
          array(permutation_units, peds_pmod, int**);
          array(permutation_units_indexes, peds_pmod, int**);
          array(number_of_permutation_units, peds_pmod, int);
          matrix(number_of_members_in_permutation_unit, peds_pmod, max_indiv, int);
          matrix(number_of_parental_indexes_in_unit, peds_pmod, max_indiv, int);
          matrix(permutation_unit_flipped, peds_pmod, max_indiv, int);
          for (i = 0; i < peds_pmod; i++) {
              matrix(permutation_units[i], max_indiv, max_indiv, int);
              matrix(permutation_units_indexes[i], max_indiv, max_indiv, int);
          }
          for (i = 0; i < peds_pmod; i++) {
              number_of_permutation_units[i] = 0;
              for (j = 0; j < max_indiv; j++) {
                  number_of_members_in_permutation_unit[i][j] = 0;
                  number_of_parental_indexes_in_unit[i][j] = 0;
                  permutation_unit_flipped[i][j] = 0;
                  for (k = 0; k < max_indiv; k++) {
                      permutation_units[i][j][k] = 0;
                  }
              }
          }
          for (i = 0; i < peds_pmod; i++) {
              for (j = 0; j < max_indiv; j++) {
                  for (k = 0; k < max_indiv; k++) {
                      permutation_units_indexes[i][j][k] = 0;
                  }
              }
          }
      }
  }

  g_is_trio = Memory_Calloc(num_pedigrees, sizeof(int));
  g_is_singleton = Memory_Calloc(num_pedigrees, sizeof(int));

  if(compute_sharing) allocate_pair_analysis_storage(); /*** new ***/
  matrix(famtree,num_pedigrees,MAX_INDIVIDUALS+5,PED_MEMBER);
  matrix(sorted_ped, num_pedigrees, MAX_INDIVIDUALS, int);
  matrix(nuc_fam, num_pedigrees, MAX_INDIVIDUALS, NUC_FAMILY);
  matrix(order_assigned, num_pedigrees, MAX_INDIVIDUALS, int);
  matrix(save_bf_original_list, num_pedigrees, MAX_INDIVIDUALS, int);
  matrix(save_bf_non_original_list, num_pedigrees, MAX_INDIVIDUALS, int);
  array(save_bf_originals, num_pedigrees, int);
  array(save_bf_non_originals, num_pedigrees, int);
  array(save_num_nuc_fams_to_peel, num_pedigrees, int);
  array(save_num_in_ped, num_pedigrees, int);
  array(save_real_non_originals, num_pedigrees, int);
  array(save_num_inbred, num_pedigrees, int);
  array(save_old_mom, num_pedigrees, int);
  array(save_num_bits, num_pedigrees, int);
  array(save_num_bits_m, num_pedigrees, int);
  array(save_num_bits_f, num_pedigrees, int);
  array(save_real_num_bits, num_pedigrees, int);
  array(save_real_num_bits_m, num_pedigrees, int);
  array(save_real_num_bits_f, num_pedigrees, int);
  array(save_real_num_vecs, num_pedigrees, int);
  array(save_npadMS, num_pedigrees, int);
  matrix(save_padMask, num_pedigrees, 64, int);
  matrix(save_padShift, num_pedigrees, 64, int);
  matrix(save_nz_bits, num_pedigrees, num_in_map_order, int);
  matrix(save_nz_bits_m, num_pedigrees, num_in_map_order, int);
  matrix(save_nz_bits_f, num_pedigrees, num_in_map_order, int);
  matrix(save_nz_pointer, num_pedigrees, num_in_map_order+1, unsigned long);
  array(save_f_num_in_map_order, num_pedigrees, int);
  matrix(save_f_map_order, num_pedigrees, num_in_map_order, int);
  matrix(save_f_uninformative, num_pedigrees, num_in_map_order, int);
  array(save_uninformative_mask, num_pedigrees, int);
  array(save_flip_mask, num_pedigrees, int*);
  array(save_maskTest, num_pedigrees, int**);
  array(save_pvct, num_pedigrees, double*);
/*  array(save_par_score, num_pedigrees, double*); */
  array(save_unlinked, num_pedigrees, double);
  if (modcalc==1) array(save_p_all_pos, num_pedigrees, double**);
  if (modcalc==2) array(save_p_one_pos, num_pedigrees, double*);

  /* Storing functionality for MOD scores and/or p-value calculation. */
  if (modcalc ==2 || (modcalc ==1 && num_replicates > 0)) {
    array(real_num_vecs_calculated, num_pedigrees, bool);
    for (famnum=0; famnum<num_pedigrees; famnum++) {
      real_num_vecs_calculated[famnum] = FALSE;
    }
  }
  /* here, num_pedigrees still equals num_pedigrees_allocated */

  if (algebra) {  /* pmod together with algebraic calculus */
      if (num_replicates > 0)
      {
          matrix(version_pointer, num_pedigrees, mem_limit8, int); 
      } else {
          matrix(version_pointer, num_pedigrees, 1, int);
      }	
      array(version_counter, num_pedigrees, int);
      array(algebra_analyzed, num_pedigrees, bool);
      array(was_skipped_before, num_pedigrees, bool);
      array(algebra_famtree, num_pedigrees, ALGEBRA_PED_MEMBER**);
      for (i=0; i < num_pedigrees; i++){
          algebra_analyzed[i] = FALSE;
          version_counter[i] = 0;
          was_skipped_before[i] = FALSE;
          if (num_replicates > 0) {
              matrix(algebra_famtree[i], mem_limit8, MAX_INDIVIDUALS+5, ALGEBRA_PED_MEMBER);
          } else {
              matrix(algebra_famtree[i], 1, MAX_INDIVIDUALS+5, ALGEBRA_PED_MEMBER);
          }
      }
  }

  array(algebra_pointer, peds_pmod, int); /* small structures kept independently of calculation mode */
  array(is_currently_in_data, num_pedigrees, int);

  array(famnum_pointer, num_pedigrees, int);
  matrix(pedigree_data, MAXLOCI*2+MAX_PHENOTYPES, 256, char);
  matrix(pedigree_names, num_pedigrees, MAX_NAME_SIZE, char);

  num_pedigrees=0;  /*** count them again while reading the data ***/
  skipped_peds=0;   /*** count number of skipped pedigrees ***/
  strcpy(skipped_peds_char,"");
  famnum_count=0;   /*** will be "unbiased" pedigree-count later ***/
  famnum = 0;       /*** will be num_pedigrees-1 later ***/
  datafile_model_incompatible = FALSE;
  scan_error_flag = FALSE;

  for (famnum=0; famnum<num_pedigrees_allocated; famnum++) {
     allocate_flip_mask[famnum] = 0;
     allocate_maskTest[famnum] = 0;
     allocate_pvct[famnum] = 0;
  }

  run {
    list_index=0;
    strcpy(current_ped,"");
    while(1) {
      _fgetln(fpin);
      if (!maxstoken(&ln,sREQUIRED,ped_name,MAX_NAME_SIZE))
          continue; /* blank line, read the next line */
      /* Start of new pedigree */
      if (strcasecmp(current_ped,ped_name) != 0) {
          /* If there was a previous pedigree, store some
             information about it */
	if (!nullstr(current_ped)) {
	  save_list_index[famnum_count] = list_index;
	  check_zero_marker(list_index);
	  analyze_family(list_index); /* scan the family */
      famnum_count++;
    }
    /* Setup variables for new pedigree */
	  strcpy(current_ped,ped_name);
	  num_pedigrees++;
	  list_index=0;
      famnum = num_pedigrees - 1;
      algebra_pointer[famnum] = famnum_count;
	  famnum_pointer[famnum] = famnum_count;
	}

    // Allocate space for alleles.
    array(famtree[famnum][list_index].a1, num_markers, int);
    array(famtree[famnum][list_index].a2, num_markers, int);
	
	/* Read the individual's index within the pedigree */
	if (!maxstoken(&ln,sREQUIRED,indiv_ID,MAX_NAME_SIZE)) {
	  scan_error_flag = TRUE;
	  sf(ps,"bad input line - no indivdual ID"); error(ps);
	}
	/* Read the parents' indices in the pedigree */
	if (!maxstoken(&ln,sREQUIRED,father_ID,MAX_NAME_SIZE)) {
	  scan_error_flag = TRUE;
	  sf(ps,"bad input line - no paternal ID for indiv %s",
	     indiv_ID); error(ps);
	}
	if (!maxstoken(&ln,sREQUIRED,mother_ID,MAX_NAME_SIZE)) {
	  scan_error_flag = TRUE;
	  sf(ps,"bad input line - no maternal ID for indiv %s",
	     indiv_ID); error(ps);
	}
	/* Read the sex of the individual */
	if (!itoken(&ln,iREQUIRED,&sex)) {
	  scan_error_flag = TRUE;
	  sf(ps,"bad input line - no sex for individual %s",indiv_ID);
	  error(ps);
	}
	
	/* Record the information in the family tree struct */
	strcpy(famtree[famnum][list_index].indiv_ID,indiv_ID);
	strcpy(famtree[famnum][list_index].dad,father_ID);
	strcpy(famtree[famnum][list_index].mom,mother_ID);
	famtree[famnum][list_index].sex=sex;
	
	if (num_replicates > 0) {
	  /* Record all the information in save_famtree struct */
	  strcpy(save_famtree[famnum_count][list_index].current_ped,current_ped);
	  strcpy(save_famtree[famnum_count][list_index].indiv_ID,indiv_ID);
	  strcpy(save_famtree[famnum_count][list_index].dad,father_ID);
	  strcpy(save_famtree[famnum_count][list_index].mom,mother_ID);
	  save_famtree[famnum_count][list_index].sex=sex;
    }
	/* Initialize values to be filled in later */
	famtree[famnum][list_index].nkids=0;
	famtree[famnum][list_index].dad_index=MISSING;
	famtree[famnum][list_index].mom_index=MISSING;
	for (j=0; j<MAX_PHENOTYPES; j++)
	  famtree[famnum][list_index].phenotype[j]=MISSING_PHENO;
	for (j=0; j<MAX_COVARIATES; j++)
	  famtree[famnum][list_index].covariate[j]=0.0;

	/* Check that this isn't a duplicate indiv_ID */
	for (j=0; j<list_index; j++)
	  if (!strcasecmp(famtree[famnum][j].indiv_ID,
			  famtree[famnum][list_index].indiv_ID)) {
	    scan_error_flag = TRUE;
	    sf(ps,"indiv %s appears more than once in pedigree %s",
	       famtree[famnum][list_index].indiv_ID,current_ped); error(ps);
	  }

	/* Read in the genotype and phenotype data */
	j=0;
	while (stoken(&ln,sREQUIRED,pedigree_data[j]))
	  j++;			

	num_data=j;
	if (num_data == 2*num_markers + num_phenotypes + num_covs + 1)
	  allele_offset = 1;
	else if (num_data == 2*num_markers + num_phenotypes + num_covs + 2)
	  allele_offset = 2;
	else {
	  scan_error_flag = TRUE;
	  sf(ps,"number of loci and phenotypes in pedigree file different from number in locus file - ped %s\n",
	     current_ped); 
	  error(ps);
	}

	/* Store the genotype data in static structs */
	temp_data = get_temp_string();
	for (j=0; j<(num_data-num_phenotypes-num_covs); j++) {
          /* '-num_covs' added above (K. Strauch 04/2008) */
	  strcpy(temp_data, pedigree_data[j]);
	  itoken(&temp_data, iREQUIRED, &datum);
	  if (g_association && !g_frequencies) {
		if (j >= allele_offset) {  // allele columns start after j == allele_offset
			Marker marker = LinkageMarkerFile_GetMarkerByIndex(g_linkage_file, (j - allele_offset) / 2);
			// Set markers that are not used for computation of
			// inheritance distribution to missing.
			if (!UserLinkageMarkerFile_Contains(g_user_linkage_file, Marker_GetName(marker))) {
				datum = 0;
			}
		}
	  }
	  allele_data[list_index][j]=datum;
	}
	num_allele_data[list_index]=j;

	if (num_replicates > 0) {
	  save_famtree[famnum_count][list_index].num_allele_data = j;
        /* Store affection status and if available liab class in save_famtree*/
	  for (j=0; j<allele_offset; j++) {
	    strcpy(temp_data, pedigree_data[j]);
	    itoken(&temp_data, iREQUIRED, &datum);
	    allele_data[list_index][j]=datum;
	  }
	  save_famtree[famnum_count][list_index].affection_status=allele_data[list_index][0];
          if (allele_offset == 2)
            save_famtree[famnum_count][list_index].liab_class=allele_data[list_index][1];
        }

	/* Store the phenotype data in static structs */
	for (j=0; j<num_phenotypes; j++) {
	  strcpy(temp_data, pedigree_data[j+2*num_markers+allele_offset]);
	  if (strcasecmp(temp_data, "-")) {
	    rtoken(&temp_data, rREQUIRED, &rdatum);
	    famtree[famnum][list_index].phenotype[j]=rdatum;
	  }
	}

	/* Store the covariate data in static structs */
	for (j=0; j<num_covs; j++) {
	  strcpy(temp_data, pedigree_data[j+2*num_markers+num_phenotypes+allele_offset]);
	  if (strcasecmp(temp_data, "-")) {
	    rtoken(&temp_data, rREQUIRED, &rdatum);
	    famtree[famnum][list_index].covariate[j] = rdatum;
	  }
	}
    // Save marker alleles for JLA analysis
    for (j = 0; j < 2*num_markers; j++) {
      strcpy(temp_data, pedigree_data[j + allele_offset]);
      itoken(&temp_data, iREQUIRED, &datum);
      if (g_association && !g_frequencies) {
          Marker marker = LinkageMarkerFile_GetMarkerByIndex(g_linkage_file, j / 2);
          // Set markers that are not used for computation of
          // inheritance distribution to missing.
          if (!UserLinkageMarkerFile_Contains(g_user_linkage_file, Marker_GetName(marker))) {
              datum = 0;
          }
      }
      if (j % 2 == 0)
          famtree[famnum][list_index].a1[j / 2] = datum;
      else
          famtree[famnum][list_index].a2[j / 2] = datum;
    }
	list_index++;
    }	
    close_file(fpin);
    /* unmatrix(famtree,num_pedigrees_allocated,PED_MEMBER); */
    replicates_error_flag = FALSE;
    fill_save_famtree = TRUE;
    unmatrix(pedigree_data, MAXLOCI*2+MAX_PHENOTYPES, char);
  } m_except {
      when SOFTABORT:
          fill_save_famtree = FALSE;
          replicates_error_flag = TRUE;
          num_pedigrees = 0;
          close_file(fpin); fpin=NULL;
          unmatrix(famtree,num_pedigrees_allocated,PED_MEMBER);
          unmatrix(save_famtree,num_pedigrees_allocated,SAVE_PED_MEMBER);
          unmatrix(pedigree_data, MAXLOCI*2+MAX_PHENOTYPES, char);
          relay;
          break;
      when ENDOFILE:
          if (list_index > 0) {
	    fill_save_famtree = TRUE;
            replicates_error_flag = FALSE;
	    run {
              save_list_index[famnum_count] = list_index;
              check_zero_marker(list_index);
	      analyze_family(list_index);
          if (algebra && (analysis_type != NPL_ANALYSIS))
          {
              if(!g_association) {
              sf(ps, "\nNumber of distinct inheritance vector classes in dataset: %i\n", vec_classes+1);
              pr();
              }
              for (j=0; j < num_pedigrees; j++ )
              {
                  if(g_association) {
                    if (g_is_singleton[algebra_pointer[j]]) {
                      sf(ps, "Pedigree %s is a singleton - no inheritance vector classes defined\n", pedigree_names[algebra_pointer[j]]);
                      pr();
                    }
                  } else {
                  sf(ps, "Number of distinct inheritance vector classes in pedigree %s: %i\n", pedigree_names[algebra_pointer[j]], num_classes_in_family[algebra_pointer[j]]);
                  pr();
                  }
              }
          }
	      save_num_pedigrees = num_pedigrees;
	      if(compute_sharing) extract_warray();  
	      if(compute_sharing) set_weights(-1); /* Sets weights for affectation analysis (default) */
	    }
	    on_error {
              replicates_error_flag = TRUE;
	      fill_save_famtree = FALSE;
	    }
	  }
          close_file(fpin); fpin=NULL;
          unmatrix(pedigree_data, MAXLOCI*2+MAX_PHENOTYPES, char);
          break;
      default:
          close_file(fpin); fpin=NULL;
          fill_save_famtree = FALSE;
          replicates_error_flag = TRUE;
          unmatrix(famtree,num_pedigrees_allocated,PED_MEMBER);
	  unmatrix(save_famtree,num_pedigrees_allocated,SAVE_PED_MEMBER);
          unmatrix(pedigree_data, MAXLOCI*2+MAX_PHENOTYPES, char);
          relay;
    }
}


static int uninformative_bit[64];

void analyze_family(volatile int num_in_ped)
{
  int i,j,k,mom,dad;
  volatile int originals;
  volatile int non_originals;
  volatile int real_non_originals;
  int num_vecs, vec, cur_place, num_bits;
  volatile int num_bits_m;
  volatile int skip_ped;
  volatile int num_bits_f;
  double prob, graph_assign_probs(int,int,int,int**,bool), total;
  double * volatile score=NULL;
  volatile double exp;
  volatile double exp_2;
  double apm_score(int**,int);
  /* double **pvector=NULL; */
  volatile double score_npl_val;
  int num_aff, parents_informative(int);
  int ** volatile affected_allele=NULL;
  int * volatile uninformative_marker=NULL;
  int * volatile num_real_vecs=NULL;
  volatile int num_inbred;
  int * volatile best_pvector=NULL;
  int has_data(int);
  void init_disease_probs(int,PED_MEMBER*);
  void peel(int,int,PED_MEMBER*,NUC_FAMILY*);
  double brute_force_analyze(int,int*,int,int*,PED_MEMBER*);
  void algebra_matrix(int,int*,int,int*,PED_MEMBER*,int, double*);
  void algebra_force(int);
  void algebra_cache_opt(double*, int, int*);
  int algebra_discarded(int, int, bool, PED_MEMBER*, int);
  void mobit_find_permutation_units(int, int*, PED_MEMBER*);
  void mobit_set_permutation_status(int);
  /* int complexity; */
  int new_vec;
  volatile int real_num_bits, real_num_bits_m, real_num_bits_f,
	   new_real_num_bits_m, new_real_num_bits_f;
  /* int num_uninformative_bits; */
  volatile int real_num_vecs;
  volatile int vec_previous;
  int **alleles_to_draw,**haplotype,new_real_num_bits,**inferred;
  int count, vecpos, parent,
      this_indiv;
  volatile int uninformative_mask;
  volatile int this_parent;
  volatile int really_uninformative_bits;
  volatile int really_uninformative_bits_m;
  volatile int really_uninformative_bits_f;
  /*** for MOD ver, *flip_mask removed --> **save_flip_mask now is extern ***/
  /* int found_kid; */
  volatile int inb_fam;
  int dis_geno;
  volatile int old_mom;
  double this_prob;
  int legal_vector, score_mask, score_mask_rdc, score_alloc, iMarker, ibit, k1, k2;
  volatile long amount_to_alloc;
  unsigned long * volatile p_pnt=NULL;
  int * volatile nz_bits=NULL;
  int * volatile nz_bits_m=NULL;
  int * volatile nz_bits_f=NULL;
  unsigned long * volatile nz_pointer=NULL;
  volatile int npl_pnt;
  double * volatile num_par_score=NULL;
  /*** for MOD version, *pvct removed --> **save_pvct now is extern ***/
  volatile int f_num_in_map_order;
  int f_map_order[MAXLOCI], f_uninformative[MAXLOCI];
  volatile int npadMS;
  int padMask[64], padShift[64];
  int     *best_pvector_tmp=NULL;
  int ibit2,lookfor;
  int mom_index_tolook;
  PED_MEMBER *this_famtree;
  NUC_FAMILY *this_nuc_fam;
  double * volatile disease_prob_old_mom;
  double * volatile disease_prob_new_mom;
  int * volatile place_allele_old_mom;
  int * volatile place_allele_new_mom;
  double * volatile this_datfile_par_score;
  double * volatile this_untotal_datfile_par_score;
  volatile int matrix_to_be_done;
  volatile bool analyzed;
  int ped_to_del, num_versions;

  if(algebra)
  analyzed = algebra_analyzed[famnum_pointer[famnum]];

  matrix_to_be_done = 0;
  famnum = num_pedigrees - 1;

  /* For all individuals, look up the person's mom and dad
     and record it in the pedigree structure */
  run {
    if (!pmod_calc_flag) {
      sf(ps,"\nanalyzing pedigree %s...\n",current_ped);
      pr();
    } else {
      strcpy(current_ped,save_current_ped);
    }

    pr_time[10] = (double)clock()/(double)CLOCKS_PER_SEC;

    if (!simulation_alleles_done)
      for (i=0; i<num_in_ped; i++) {
	set_dad_index(i,num_in_ped);
        set_mom_index(i,num_in_ped);
        famtree[famnum][i].analyzed_parent=FALSE; /* for counting inh. vec complexity */
        famtree[famnum][i].dad_bit_uninf=FALSE;
        famtree[famnum][i].mom_bit_uninf=FALSE;
        famtree[famnum][i].kids_found=FALSE; /* used later in bit chopping section */
	famtree[famnum][i].replaced_mom=FALSE;  /* if true, we replaced this mom for loopbreaking */
        famtree[famnum][i].not_peeled=FALSE;
      }
      
    for (i=0; i<num_in_ped; i++) famtree[famnum][i].discard=FALSE;

    if (num_in_ped == 1) {
        // Mark pedigree as singleton
        if (famtree[famnum][0].nkids == 0
            && famtree[famnum][0].dad_index == MISSING
            && famtree[famnum][0].mom_index == MISSING) {
            g_is_singleton[algebra_pointer[famnum]] = 1;
        }
    }

    /* get rid of unlinked individuals - set discard to TRUE */
    if (!g_is_singleton[algebra_pointer[famnum]]) {
        check_for_unlinked(num_in_ped);
    }

    /* at this point we can eliminate from consideration:
        a) untyped individuals with no offspring in the pedigree
        b) those individuals that are unaffected, have no kids, and have 
           informative parents (they add next to nothing to an APM study) */

    for (i=0; i<num_in_ped; i++) {
      if (famtree[famnum][i].discard) continue;
      if (famtree[famnum][i].nkids == 0 && !has_data(i) &&
	  (!include_untyped || (allele_data[i][0] == 0))) {
	  /*** Caution: add last conditions to make sure untyped
	       individuals with no children (except for those with trait
	       phenotype unknwon) will not be discarded in the analysis.
	       - Individuals without marker genotypes but with trait
	       information contribute to the reconstruction of their
	       parents' trait-locus genotypes. Hence, they _do_ contribute
	       to the analysis!
	       A variable include_untyped (together with the command
	       'include untyped') has been introduced which by default
	       is 'yes'; if set to 'no', untyped individuals with no kids
	       are excluded as before.  (This has been added such that
	       users can easily make this version compatible with earlier
	       versions.)  (K. Strauch) ***/

/* ifdef debugging_output removed (K. Strauch 04/2004) */
	  if (!pmod_calc_flag) {
	    sf(ps,"DISCARDING individual %s (pedigree leaf with no genotype data)\n", famtree[famnum][i].indiv_ID); 
	    pr();
	  } else if (replicates_store_result) {
            sf(ps,"DISCARDING individual %s (pedigree leaf with no genotype data)\n", famtree[famnum][i].indiv_ID); 
	    fpr(replifp);
	  }
/* here was endif before */

	  famtree[famnum][i].discard = TRUE;
      }

      if (discard_unaffected) {
	/* discard_unaffected is a flag ("discard") set at command level */
	if ( famtree[famnum][i].nkids == 0 && 
	    allele_data[i][0] != 2 &&
	    parents_informative(i) ) {
#ifdef DEBUGGING_OUTPUT
	  sf(ps,"DISCARDING individual %s (unaffected leaf - informative parents)\n",
	     famtree[famnum][i].indiv_ID); pr();
#endif
	  famtree[famnum][i].discard = TRUE;
	}
      }
    }

    originals = non_originals = 0;
    cur_place=0;

    /* set placeholder alleles for original pedigree members */
    for (i=0; i<num_in_ped; i++) {
      if (famtree[famnum][i].dad_index == MISSING &&
	  famtree[famnum][i].mom_index == MISSING) {
	if (!famtree[famnum][i].discard) {
	  originals++;
	  famtree[famnum][i].place_allele[0]=cur_place; cur_place++;
	  famtree[famnum][i].place_allele[1]=cur_place; cur_place++;
	}
      }

      else if (famtree[famnum][i].dad_index == MISSING ||
	       famtree[famnum][i].mom_index == MISSING) {
	scan_error_flag = TRUE;
        if (g_association && num_in_ped == 2) {
            sf(ps, "Parent-offspring pair detected.  Since you're doing JLA analysis there are 3 options:\n"
                   "\n"
                   "    1) Add dummy parent with missing genotypes to form a trio.\n"
                   "    2) Truncate parent-offspring pair to singleton by removing offspring.\n"
                   "    3) Delete parent-offspring pair.\n"
                   "\n"
                   "Options 1 and 2 preserve association information for JLA analysis.");
            error(ps);
        }
        sf(ps,"person %s has ONE parent",famtree[famnum][i].indiv_ID); error(ps);
      }
      else {
	if (!famtree[famnum][i].discard) non_originals++;
	famtree[famnum][i].place_allele[0]=-1;
	famtree[famnum][i].place_allele[1]=-1;
      }
    }

    /* Now we need to figure out how many loci there are -
       if odd number of alleles, the first is affectation status 
       and there is no liability class column - if even number 
       of alleles, the first is affectation status and the 
       second is liability class */

 
    /* fill in affectation_status and liability_class in PED_MEMBERs */
    for (i=0; i<num_in_ped; i++) {
      famtree[famnum][i].affectation_status = allele_data[i][0];
      if (allele_offset == 2) {
	if (allele_data[i][1] > old_num_liability_classes) {
	  scan_error_flag = TRUE;
	  sf(ps,"liability class %d does not exist (individual %s)",
	     allele_data[i][1], famtree[famnum][i].indiv_ID);
	  error(ps);
	} else famtree[famnum][i].liability_class =
		     recode_liab_class[allele_data[i][1]];
      /* the recode_liab_class[] is new in GH-MODSCORE 04/2004 (K. Strauch)
	 - this allows us to apply the new 'join liability classes' command.
         If this command has not been executed, recode_liab_class[i] == i */
      } else famtree[famnum][i].liability_class = 1;
    }

    /* now let's do a preliminary sort of non-originals so that
       we can accurately determine how much analysis needs to be done */

    /* new_sort_pedigree(); */
    sort_pedigree(num_in_ped);
    resort_pedigree(non_originals);

    /* now read through this sorted list of non_originals until we
       hit the limit of our analytic capabilities */

    really_uninformative_bits = 0;
    num_bits = real_num_bits = 0;
    num_bits_m = num_bits_f = 0;
    real_num_bits_m = real_num_bits_f = 0;
    real_non_originals = non_originals;

                             /*** new (K. Strauch 03/2004) ***/
    if (!pmod_calc_flag) {
        if (g_association && g_is_singleton[algebra_pointer[famnum]]) {
            sf(ps, "Pedigree %s is a singleton\n", current_ped);
            pr();
        } else {
      sf(ps,"Pedigree %s has %d bits (%d nonfounders and %d founders).\n",
    	 current_ped, (2*non_originals)-originals, non_originals, originals); 
      pr();
        }
    } else if (replicates_store_result) {
      sf(ps,"Pedigree %s has %d bits (%d nonfounders and %d founders).\n",
    	 current_ped, (2*non_originals)-originals, non_originals, originals); 
      fpr(replifp);
    }

    for (i=0; i<non_originals; i++) {

        dad = famtree[famnum][sorted_ped[famnum][i]].dad_index;
		  mom = famtree[famnum][sorted_ped[famnum][i]].mom_index;
	
	num_bits += 2;
	num_bits_m += 1;
	num_bits_f += 1;
	new_real_num_bits = real_num_bits+2;
        new_real_num_bits_m = real_num_bits_m+1;
        new_real_num_bits_f = real_num_bits_f+1;
	
	
	if (!famtree[famnum][dad].kids_found && famtree[famnum][dad].dad_index==MISSING && 
	    famtree[famnum][dad].mom_index==MISSING) {
	    /* first kid of this father, hold bit fixed */
	    famtree[famnum][sorted_ped[famnum][i]].dad_bit_uninf = TRUE;
	    new_real_num_bits--;
	    new_real_num_bits_m--;
	    famtree[famnum][dad].kids_found=TRUE;
	    /* debugging message */
#ifdef DEBUGGING_OUTPUT
	    sf(ps,"dropping paternal bit of indiv %s\n",
	       famtree[famnum][sorted_ped[famnum][i]].indiv_ID); pr();
#endif
	    if (famtree[famnum][dad].nkids == 1){
	       really_uninformative_bits++;
	       really_uninformative_bits_m++;
	    }    
	}

	if (!famtree[famnum][mom].kids_found && famtree[famnum][mom].dad_index==MISSING &&
	    famtree[famnum][mom].mom_index==MISSING) {
	    /* first kid of this mother, hold bit fixed */
	    famtree[famnum][sorted_ped[famnum][i]].mom_bit_uninf = TRUE;
	    new_real_num_bits--;
	    new_real_num_bits_f--;
	    famtree[famnum][mom].kids_found=TRUE;
	    /* debugging message */
#ifdef DEBUGGING_OUTPUT
	    sf(ps,"dropping maternal bit of indiv %s\n",
	       famtree[famnum][sorted_ped[famnum][i]].indiv_ID); pr();
#endif
	    if (famtree[famnum][mom].nkids == 1){
	       really_uninformative_bits++;
	       really_uninformative_bits_f++;
	    }   
	}

	if ((new_real_num_bits > max_bits_to_analyze) || (num_bits > AVAILABLE_BITS)) {
	  /* we've hit the limit...can't accept this guy */
	  real_non_originals = i;
	  num_bits -= 2;
	  num_bits_m -= 1;
	  num_bits_f -= 1;
	  break;
	}
	real_num_bits = new_real_num_bits;
	real_num_bits_m = new_real_num_bits_m;
        real_num_bits_f = new_real_num_bits_f;

	if ((new_real_num_bits == max_bits_to_analyze) || (num_bits == AVAILABLE_BITS)) {
	  real_non_originals = i+1;
	  break;
	}
    }

    /* print out a most abject apology regarding our inability
       to analyze beyond a certain point within reasonable
       time and memory constraints */

    skip_ped=FALSE;

    if(!pmod_calc_flag) {
      if (real_non_originals < non_originals) {

        if (skip_large) {
 	    print("WARNING: Pedigree is too large to be computed!\n");
	    print("To analyze you can either turn off the 'skip large' option which will\n");
	    print("allow the program to discard less informative individuals or you may\n");
	    print("raise the 'max bits' option which will allow larger pedigrees to be\n");
	    print("analyzed while raising the time and memory requirements of the program.\n");
	    skip_ped=TRUE; num_pedigrees--; skipped_peds++;
	    strcat(skipped_peds_char," ");
	    strcat(skipped_peds_char,current_ped);
	} else {
	    sf(ps,"WARNING: due to computation time and memory constraints, individual%s\n",
	       (non_originals-real_non_originals > 1)?"s":""); pr();
	    for (i=real_non_originals; i<non_originals; i++) {
	      sf(ps,"%s ",famtree[famnum][sorted_ped[famnum][i]].indiv_ID); pr();
	      famtree[famnum][sorted_ped[famnum][i]].discard = TRUE;
	    }
	    sf(ps,"%s been dropped from the analysis.\n", (non_originals-real_non_originals > 1)?"have":"has"); 
	    pr();
	                       /*** new (K. Strauch 03/2004) ***/
	    sf(ps, "After truncation, the pedigree has %d remaining bits.\n", real_num_bits); pr();
	}
      }
    } else if (replicates_store_result) {
      if (real_non_originals < non_originals) {
        if (skip_large) {
 	    sf(ps, "WARNING: Pedigree is too large to be computed!\n");
	    fpr(replifp);
	    sf(ps,"To analyze you can either turn off the 'skip large' option which will\n");
	    fpr(replifp);
	    sf(ps,"allow the program to discard less informative individuals or you may\n");
	    fpr(replifp);
	    sf(ps,"raise the 'max bits' option which will allow larger pedigrees to be\n");
	    fpr(replifp);
	    sf(ps,"analyzed while raising the time and memory requirements of the program.\n");
	    fpr(replifp);
	    skip_ped=TRUE; num_pedigrees--; skipped_peds++;
	    strcat(skipped_peds_char," ");
	    strcat(skipped_peds_char,current_ped);
	} else {
	    sf(ps,"WARNING: due to computation time and memory constraints, individual%s\n",
	       (non_originals-real_non_originals > 1)?"s":"");
	    fpr(replifp);
	    
	    for (i=real_non_originals; i<non_originals; i++) {
	      sf(ps,"%s ",famtree[famnum][sorted_ped[famnum][i]].indiv_ID); fpr(replifp);
	      famtree[famnum][sorted_ped[famnum][i]].discard = TRUE;
	    }
	    sf(ps,"%s been dropped from the analysis.\n",
	       (non_originals-real_non_originals > 1)?"have":"has");
	    fpr(replifp);
	                       /*** new (K. Strauch 03/2004) ***/
	    sf(ps, "After truncation, the pedigree has %d remaining bits.\n",
	       real_num_bits);
	    fpr(replifp);
	}
      }
    }
  
    /* Now we should go through and discard any originals who
       have had all their kids discarded */

/*****
    for (i=0; i<num_in_ped; i++) {
      if (famtree[famnum][i].dad_index == MISSING &&
	  famtree[famnum][i].mom_index == MISSING && !famtree[famnum][i].discard) {
	  found_kid=FALSE;
	  for (j=0; j<num_in_ped; j++) {
	    if ((famtree[famnum][j].dad_index == i || famtree[famnum][j].mom_index == i) &&
		!famtree[famnum][j].discard) {
	        found_kid=TRUE; break;
	    }
	  }
	  if (!found_kid) { famtree[famnum][i].discard=TRUE; originals--; }
      }
    }
*****/

    int should_skip_pedigree = 0;
    if (g_association && non_originals == 1 && originals == 2) {
        // Mark pedigree as trio
        g_is_trio[algebra_pointer[famnum]] = 1;
    } else if (!g_association && g_is_singleton[algebra_pointer[famnum]]) {
        should_skip_pedigree = 1;
        sf(ps, "Warning: pedigree (%s) is a singleton ...SKIPPING\n", current_ped);
        pr();
    } else if (!g_is_singleton[algebra_pointer[famnum]] && non_originals == 0) {
        should_skip_pedigree = 1;
        sf(ps, "Warning: pedigree (%s) consists entirely of multiple unlinked individuals ...SKIPPING\n", current_ped);
        pr();
    } else if (!g_is_singleton[algebra_pointer[famnum]] && non_originals == 1) {
        should_skip_pedigree = 1;
        sf(ps, "< 2 non-original members in pedigree %s...SKIPPING\n", current_ped);
        if (!pmod_calc_flag) {
            pr();
        } else if (replicates_store_result) {
            fpr(replifp);
        }
    }
    if (should_skip_pedigree) {
        skip_ped = TRUE;
        num_pedigrees--;
        skipped_peds++;
        strcat(skipped_peds_char, " ");
        strcat(skipped_peds_char, current_ped);
    }

    if (!skip_ped && (num_bits > AVAILABLE_BITS)) {
       if (!pmod_calc_flag) {
         sf(ps,"num_bits > %d in pedigree %s...SKIPPING\n", AVAILABLE_BITS, current_ped); 
         pr();
       } else if (replicates_store_result) {
         sf(ps,"num_bits > %d in pedigree %s...SKIPPING\n", AVAILABLE_BITS, current_ped); 
         fpr(replifp);
       } 
      skip_ped = TRUE; num_pedigrees--; skipped_peds++;
      strcat(skipped_peds_char," ");
      strcat(skipped_peds_char,current_ped);
    } 

    if (!skip_ped || (non_originals==1)) {

	    /* Check for impossible loops */
	    for (i=0; i<num_in_ped; i++) {
		    /* We've already know that no one can be their own
		       direct parent so we'll start up a level */
		    if (famtree[famnum][i].dad_index != MISSING) { /* If one parent is missing,both are*/
			    check_for_loop(i,famtree[famnum][i].sex,famtree[famnum][i].dad_index,1,num_in_ped);
			    check_for_loop(i,famtree[famnum][i].sex,famtree[famnum][i].mom_index,1,num_in_ped);
		    }
	    }

	    /* Sort pedigree top down into a list */
	    /* sort_pedigree(num_in_ped); already completed above */

	    /********** bit count completed above 
	      num_bits=0;
	      for (i=0; i<non_originals; i++) {
	      if (!famtree[famnum][sorted_ped[famnum][i]].dad_bit_uninf) num_bits++;
	      if (!famtree[famnum][sorted_ped[famnum][i]].mom_bit_uninf) num_bits++;
	      }
	      if (num_bits > 16) error("too many individuals left...hackprep failed");
	     **********/

	    /* so now we have 2^2*non_originals possible pvectors but only
	       2^real_num_bits of those are interesting...to get them, we will
	       hold the uninformative bits at 0 while allowing all others
	       to be 0 or 1 */

	    /* Fill the array of nuclear families */
	    fill_nuc_families(num_in_ped);

	    /* Count the number of inbred matings in the pedigree and
	       do the right thing if there are any */
	    if (!algebra && !g_association) /* in case of algebraic calculation no loop breaking is needed */
	    {
		    num_inbred = inbred_matings();

		    if (num_inbred >= 1) {
			    if (!pmod_calc_flag) {
				    sf(ps,"family %s is inbred\n",current_ped); pr();
			    } else if (replicates_store_result) {
				    sf(ps,"family %s is inbred\n",current_ped); fpr(replifp);
			    }
			    /*** order_inbred_family(); ***/
			    if (num_inbred == 1) {
				    /* break the first inbred marriage */
				    for (i=0; i<num_nuc_fam; i++) {
					    if (nuc_fam[famnum][i].inbred) {
						    inb_fam = i; 
						    break;
					    }
				    }


				    /* make fake individual (N=num_in_ped) with no parents
				       to be the new wife in this family */
				    old_mom = nuc_fam[famnum][inb_fam].mom;
				    nuc_fam[famnum][inb_fam].mom = num_in_ped;
				    famtree[famnum][num_in_ped].dad_index=MISSING;
				    famtree[famnum][num_in_ped].mom_index=MISSING;
				    famtree[famnum][num_in_ped].nkids = famtree[famnum][old_mom].nkids;


				    famtree[famnum][old_mom].nkids=0; /* temporarily */
				    num_in_ped++;
			    } else {
				    /* more than one inbreeding loop - do by brute force
				       as if there were no inbreeding loops */
				    num_inbred=0;
			    }
		    }

		    if (num_nuc_fam > 1) {
			    if (mark_pivots(num_in_ped) == -1) { 
				    if (!skip_ped) { 
					    skip_ped=TRUE; 
					    num_pedigrees--;  skipped_peds++;
					    strcat(skipped_peds_char," ");
					    strcat(skipped_peds_char,current_ped);
				    }
			    } else {
				    if (num_inbred >= 1) {
					    /* fix variable that were corrupted for mark_pivots */
					    num_in_ped--;

					    /* 	Bugfix for peeling with loops, for brute_force_analyze    */
					    famtree[famnum][old_mom].replaced_mom = TRUE;
					    famtree[famnum][old_mom].nkids=famtree[famnum][num_in_ped].nkids;

				    }
				    if (save_num_nuc_fams_to_peel[famnum] == 0) {
					    /* if we still can't peel after breaking the loop
					       we want to go straight to brute_force and forget
					       the loop breaking stuff */
					    num_inbred=0; /* skip the inbred broken peeling stuff */
				    }
			    }
		    } else {
			    /* Mark the interesting individuals so that they are
			       fully analyzed by the exhaustive method */
			    famtree[famnum][nuc_fam[famnum][0].dad].not_peeled=TRUE;
			    famtree[famnum][nuc_fam[famnum][0].mom].not_peeled=TRUE;
			    for (k=0; k<nuc_fam[famnum][0].num_kids; k++) 
				    famtree[famnum][nuc_fam[famnum][0].kid[k]].not_peeled=TRUE;

			    order_assigned[famnum][0]=0;
			    save_num_nuc_fams_to_peel[famnum]=0;
		    }
	    } else { /* if algebra */
		for (i=0; i < num_in_ped; i++){
		      famtree[famnum][i].not_peeled=TRUE;
		      num_inbred = 0;
		}
	    }		
    }

	if(!skip_ped || (skip_ped && non_originals==1)) {
      /* now connect to graphing code for determining inh. vector probs */
		num_vecs=1;
		for (i=0; i<num_bits; i++) num_vecs *= 2;
		real_num_vecs=1;
		for (i=0; i<real_num_bits; i++) real_num_vecs *= 2;
		
	pr_time[20] = (double)clock()/(double)CLOCKS_PER_SEC;
/*----------------------------------------------------*/
	/* dangerous kludge 
	   meant to take care of repeated appearance of same marker 
	   when merging data sets from several groups */
		f_num_in_map_order =
		Define_family_map_order(originals, real_non_originals,
                            	num_in_map_order, map_order,
								f_map_order,
								f_uninformative,
								0, num_inbred);

	/*	printf("originals, real_non_originals, num_bits, real_num_bits = %d %d %d %d\n", 
		originals, real_non_originals, num_bits, real_num_bits);
		printf("---before FindIllegalBits num_in_map_order = %d\n", num_in_map_order);
	*/	
		matrix (save_maskTest[famnum], f_num_in_map_order, 2, int);
                allocate_maskTest[famnum] = 1;
		/*** why was it '1' instead of '2' before?! ***/
		array (maskHomP, f_num_in_map_order, int);
		/* redundant call to fill_placeholder_alleles(..) to fill uninformative_bit[] array */

		fill_placeholder_alleles_2(0, 0, 0, real_non_originals,0);

		/* determine uninformative_mask for mapping
		   between fs and full representation
		*/  
		uninformative_mask = 0; 
	        for (i=0; i<num_bits; i++) if (uninformative_bit[i]) uninformative_mask+=m_pow2[i];
		
		 /* determine k_full_rep[k_foundersym_rep] */
 
  		k2 = 0;
  		for (k1=0; k1<num_bits; k1++) {
     		   if((uninformative_mask & m_pow2[k1]) != 0) continue;
     		   k_full[k2] = k1;
     		   k2++;
  		}


 		CheckHomozygotes(num_bits, real_num_bits, num_in_ped, f_num_in_map_order,
							  uninformative_bit, f_map_order, non_originals);
		pr_time[30] = (double)clock()/(double)CLOCKS_PER_SEC;
		FindIllegalBits(originals, real_non_originals, num_bits, real_num_bits,
	                	f_num_in_map_order, f_map_order, f_uninformative);
		pr_time[35] = (double)clock()/(double)CLOCKS_PER_SEC;

		/* now obtain NPL score mask */

		Set_npl_score_masks(&score_mask, &score_mask_rdc, &score_alloc,
	                    	real_non_originals, num_bits);
		npl_pnt     = 0;
		vec_previous=-1;
		
		array(p_pnt,      num_in_map_order  , unsigned long);
		array(nz_bits,    num_in_map_order  , int);
		array(nz_bits_m,    num_in_map_order  , int);
		array(nz_bits_f,    num_in_map_order  , int);
		array(nz_pointer, num_in_map_order+1, unsigned long);

		nz_pointer[0] = 0;
		for(iMarker=0; iMarker<f_num_in_map_order; iMarker++) {
			nz_bits[iMarker] = 0;
			nz_bits_m[iMarker] = 0;
			nz_bits_f[iMarker] = 0;
			if(f_uninformative[iMarker]){
				nz_pointer[iMarker+1] = nz_pointer[iMarker] + m_pow2[ nz_bits[iMarker] ];
			} else {
				for(ibit=0; ibit<real_num_bits; ibit++) {
					if ((save_maskTest[famnum][iMarker][0] & m_pow2[ibit]) == 0){
					  nz_bits[iMarker]++;
					  if (k_full[ibit]%2 == 1){
	                                     nz_bits_m[iMarker]++;
	                                   }else{
	                                     nz_bits_f[iMarker]++;
	                                   } 
					  
					}  
									  
				}
				nz_pointer[iMarker+1] = nz_pointer[iMarker] + m_pow2[ nz_bits[iMarker] ];
			}
		}
		amount_to_alloc = nz_pointer[f_num_in_map_order];
/*      sf(ps, "nz_pointer[f_num_in_map_order] = %d\n",
	nz_pointer[f_num_in_map_order]); pr(); */  /* debug (K. Strauch) */


		if (dump_requirements) {
		  print_requirements(num_in_map_order, num_bits, real_num_bits, 
									score_mask_rdc,
									amount_to_alloc, score_alloc, nz_bits,
									f_num_in_map_order,
									f_uninformative);
		}
		
    	if((COMPUTE_SIZE_ONLY || dump_requirements) && !(skip_ped && non_originals==1)) {
		unarray(nz_pointer, unsigned long);
    		unarray(nz_bits,    int);
		unarray(p_pnt,      unsigned long);
		unmatrix (save_maskTest[famnum], f_num_in_map_order,  int);
                allocate_maskTest[famnum] = 0;
		unarray (maskHomP, int);

    		skip_ped = TRUE; num_pedigrees--;  skipped_peds++;
		strcat(skipped_peds_char," ");
		strcat(skipped_peds_char,current_ped);
		}
	}

    if (!skip_ped || (skip_ped && non_originals==1)) {

	this_famtree = famtree[famnum];
	this_nuc_fam = nuc_fam[famnum];

	run {
    	array(save_pvct[famnum], amount_to_alloc, double);
        allocate_pvct[famnum] = 1;

    	if(analysis_type != NPL_ANALYSIS) {

	    /* make list of originals/non-originals for brute_force_analyze;
	       this has been moved here in order to minimize computation time,
	       now it is done only once (not for each vec) (K. Strauch) */

	    if (num_inbred >= 1) num_in_ped++;
	    save_bf_originals[famnum]=0; save_bf_non_originals[famnum]=0;
	    for (i=0; i<num_in_ped; i++) {
		/* Katz, bugfix, Thu Jun  5 17:37:02 2003 */
                if (this_famtree[i].discard)
		    continue;
		if ((this_famtree[i].mom_index != MISSING) &&
		    (this_famtree[this_famtree[i].mom_index].replaced_mom))
		    mom_index_tolook = num_in_ped-1; /* look at the fake guy */
		else mom_index_tolook = this_famtree[i].mom_index;

		if (this_famtree[i].not_peeled) {
		    if (this_famtree[i].dad_index==MISSING && this_famtree[i].mom_index==MISSING) {
			save_bf_original_list[famnum][save_bf_originals[famnum]]=i;
			save_bf_originals[famnum]++;
		    }
		    else if (!this_famtree[this_famtree[i].dad_index].not_peeled ||
			     !this_famtree[mom_index_tolook].not_peeled) {
			save_bf_original_list[famnum][save_bf_originals[famnum]]=i;
			save_bf_originals[famnum]++;
		    } else {
			save_bf_non_original_list[famnum][save_bf_non_originals[famnum]]=i;
			save_bf_non_originals[famnum]++;
		    }
		}
	    }
	    if (save_bf_originals[famnum] > 15) { 
		sf(ps,"too many originals (%d) to run this routine (brute_force)\n",
		   originals);
		pr();
		print("Cannot compute LOD scores for this pedigree\n");
	    }
	    if (num_inbred >= 1) {
	      num_in_ped--;
	      disease_prob_old_mom = this_famtree[old_mom].disease_prob;
	      disease_prob_new_mom = this_famtree[num_in_ped].disease_prob;
	      place_allele_old_mom = this_famtree[old_mom].place_allele;
	      place_allele_new_mom = this_famtree[num_in_ped].place_allele;
	    }

	    if (algebra)
	    {
		    if (!algebra_analyzed[famnum_pointer[famnum]]){		
			    array(par_score[famnum_pointer[famnum]],            real_num_vecs,      double);
			    array(new_vec_to_alg_prob[famnum_pointer[famnum]],            real_num_vecs,      int);
			    array(datfile_par_score[famnum_pointer[famnum]],            real_num_vecs,      double);
			    array(untotal_datfile_par_score[famnum_pointer[famnum]],            real_num_vecs,      double);
		    }
			array(num_par_score,                        1,      double); /* passed on to scan_pedigree */
	    }

	    if (!algebra)
	    {
		    array(num_par_score, real_num_vecs, double);
	    }

	}else {
		/* printf("----allocating nothing\n"); */
		array(num_par_score,                        1,      double); /* need to allocate it because it is passed a sub. argument */
	}
		array(score,                score_alloc,        double);
		matrix(affected_allele,     num_in_ped,         2,             int);
		array(uninformative_marker, num_in_map_order+1, int);
		array(num_real_vecs,        num_in_map_order+1, int);
		array(best_pvector,         num_in_map_order+1, int);

	} except_when (NOMEMORY) {
		print("Can't get enough memory for this scan\n");
    	unarray(best_pvector,         int);
    	unarray(num_real_vecs,        int);
    	unarray(uninformative_marker, int);
    	unmatrix(affected_allele,     num_in_ped,         int);
    	unarray(score,                double);
	
	if (algebra)
	{
    	unarray(par_score[famnum_pointer[famnum]], double);
    	unarray(new_vec_to_alg_prob[famnum_pointer[famnum]], int);
    	unarray(datfile_par_score[famnum_pointer[famnum]], double);
    	unarray(untotal_datfile_par_score[famnum_pointer[famnum]], double);
	}else {
		unarray(num_par_score, double);
	}
	unarray(save_pvct[famnum],    double);
        allocate_pvct[famnum] = 0;
	}
        
	for (i=0; i<f_num_in_map_order; i++) {
		uninformative_marker[i] = (save_maskTest[famnum][i][0] == 0);
		num_real_vecs[i]=0;
	}
	exp = exp_2 = 0.0;
	
	for(i=0; i<f_num_in_map_order; i++) { p_pnt[i] = nz_pointer[i]; }

	pr_time[40] = (double)clock()/(double)CLOCKS_PER_SEC;

	fill_placeholder_alleles_0(0,real_non_originals,0,this_famtree);
	vec = 0;
	uninformative_mask=0;
	for (i=0; i<num_bits; i++) 
		if (uninformative_bit[i]) uninformative_mask+=m_pow2[i];
	npadMS = Setup_local_pad_inhvec_2(uninformative_mask, num_bits,
                                      padMask, padShift);

        if (num_replicates > 0 && saved_models)
          check_discarded(num_in_ped); /* new for PMOD, check if real_num_vecs has changed */
	
	if (algebra)
    {
        if (!analyzed && !skip_ped){
            for (i=0; i < num_in_ped; i++){
                algebra_famtree[famnum_pointer[famnum]][0][i].discard=famtree[famnum][i].discard;
            }
        }	

        if (analyzed && skip_ped){
            if(!was_skipped_before[famnum_pointer[famnum]]){
                ped_to_del = is_currently_in_data[famnum_pointer[famnum]];
                for (k=0; k <= vec_classes; k++){
                    if (family_to_vec_class[ped_to_del][k]){
                        num_fam_in_vec_class[k]--; /* in order to avoid algebra_force for vector classes that are not in replicate but were in data before */	
                    }
                }
            }
        }

        if (skip_ped){
            was_skipped_before[famnum_pointer[famnum]] = TRUE;
        } else{
            was_skipped_before[famnum_pointer[famnum]] = FALSE;
        }

        if (!skip_ped && num_replicates > 0 && modscore_done){
            matrix_to_be_done =  algebra_discarded(num_in_ped, real_num_vecs, matrix_to_be_done, this_famtree, famnum); /* needed to check if algebra_matrix must be run for new version of family */
        }

        this_datfile_par_score = datfile_par_score[algebra_pointer[famnum]];
        this_untotal_datfile_par_score = untotal_datfile_par_score[algebra_pointer[famnum]];
        if (!modscore_done && !skip_ped && mobit_simulate) {
                mobit_find_permutation_units(save_bf_non_originals[famnum], save_bf_non_original_list[famnum], this_famtree);
                mobit_set_permutation_status(algebra_pointer[famnum]);
                }
    } /* end of 'if algebra' */

    int percentage, last_percentage = -1;
	
    for (new_vec=0; new_vec<real_num_vecs; new_vec++) {

		if (new_vec > 0) {
			/*vec = local_pad_inhvec(new_vec, uninformative_mask, num_bits); */
			vec = 0;
			for(i=0; i<npadMS; i++)
				vec |= (new_vec & padMask[i]) << padShift[i];
			fill_placeholder_alleles_1(vec,real_non_originals,
						   this_famtree);
		}

		/* now every member of famtree has two placeholder alleles
		   filled in and we can advance to the prob assignment
		   for each of the markers */

		for (i=0; i<f_num_in_map_order; i++) {
			if(f_uninformative[i] && !TEST_SHORTCUT) {
				prob = 1;
				save_pvct[famnum][ p_pnt[i] ] = prob;
			} else {
				legal_vector = 
				((new_vec & save_maskTest[famnum][i][0]) == save_maskTest[famnum][i][1]);

				if(legal_vector || TEST_SHORTCUT) {

				  if ((new_vec & maskHomP[i]) == 0) {
/* This is the first (representative) of homozygote parent equiv class */
					  prob = graph_assign_probs(originals, real_non_originals, 
														 f_map_order[i], NULL, FALSE); 
				  }
					else {
/* This is a subsequent element of an equiv class, look up the first one */
					  ibit2 = 0;
					  j = 0;
					  lookfor = (new_vec & ~maskHomP[i]);

/*   	  (pvct array ptr is only for variable (nz) bits of inher vec) */
					  for (ibit=0; ibit <= real_num_bits; ibit++)
						 if ((save_maskTest[famnum][i][0] & m_pow2[ibit]) == 0) {
							  j |= ((lookfor & m_pow2[ibit]) >> (ibit - ibit2++));
						 }

#ifdef PR_19
					  if ((lookfor & save_maskTest[famnum][i][0]) != save_maskTest[famnum][i][1])
						 printf (
									"Lookfor is inconsistent, marker %d, lookfor %x, vec %x, maskhom %x, maskt0 %x, maskt1 %x, prob %1.12lf\n",
	i, lookfor, new_vec, maskHomP[i], save_maskTest[famnum][i][0], save_maskTest[famnum][i][1],
									save_pvct[famnum][nz_pointer[i]+j]);
#endif

 				 prob = save_pvct[famnum][nz_pointer[i] +j];

					}

					if (prob > 0.0) num_real_vecs[i]++;

					if(legal_vector) {
						save_pvct[famnum][ p_pnt[i] ] = prob;
						if(!f_uninformative[i]) p_pnt[i]++;
					}

					if(TEST_SHORTCUT) {
						if(!legal_vector && !(prob==0.)) {
						        scan_error_flag = TRUE;
							printf("iM=%3d m=%3d, new_vec=%8x, vec=%8x, is not illegal\n",
							i, f_map_order[i], new_vec, vec);
							exit (1);
						}
					}
				} else {
					prob = 0;
				}

			}
			if(uninformative_marker[i])
				if ((new_vec > 0) && (prob != save_pvct[famnum][nz_pointer[i]])) 
					uninformative_marker[i]=FALSE;
		}

		if (analysis_type != LOD_ANALYSIS) {
		  /* now obtain NPL score for this assignment of placeholder
		     alleles for this pedigree */
/*-----------------------------------------------*/
	if( (vec_previous & score_mask) != (vec & score_mask) ) {
		for (i=0,num_aff=0; i<num_in_ped; i++) {
		  if (this_famtree[i].discard || this_famtree[i].place_allele[0]==-1) continue;

		  if (allele_data[i][0] == 2) {
	    	  /* this individual is affected */
	    	  affected_allele[num_aff][0] = this_famtree[i].place_allele[0];
	    	  affected_allele[num_aff][1] = this_famtree[i].place_allele[1];
	    	  num_aff++;
		  }
		}
		if(num_aff>16) {
		        scan_error_flag = TRUE;
			sf(ps," ******Error num_aff>16 *******\n");
			pr();
			exit (1);
		}
		score_npl_val = apm_score(affected_allele, num_aff);
		score[npl_pnt++] = score_npl_val;
	}
	vec_previous = vec;
/*-----------------------------------------------*/
	exp += score_npl_val;
	exp_2 += (score_npl_val*score_npl_val);
		}
		if (analysis_type != NPL_ANALYSIS) {
			if (num_inbred >= 1) {
				/* break the loop! if algebraic mode is switched on num_inbred is 0 */
				num_par_score[new_vec] = 0.0;

				for (dis_geno=0; dis_geno<4; dis_geno++) {
					init_disease_probs(num_in_ped, this_famtree);

					disease_prob_new_mom[dis_geno]=1.0;
					j = dis_geno^1;
					disease_prob_old_mom[j]=0.0;
					disease_prob_new_mom[j]=0.0;
					j ^= 2;
					disease_prob_old_mom[j]=0.0;
					disease_prob_new_mom[j]=0.0;
					j ^= 1;
					disease_prob_old_mom[j]=0.0;
					disease_prob_new_mom[j]=0.0;

					if (save_num_nuc_fams_to_peel[famnum] > 0)
						peel(vec, num_bits, this_famtree, this_nuc_fam);
					place_allele_new_mom[0] = place_allele_old_mom[0];
					place_allele_new_mom[1] = place_allele_old_mom[1];
					this_prob = brute_force_analyze(
                            save_bf_originals[famnum],
							save_bf_original_list[famnum],
							save_bf_non_originals[famnum],
							save_bf_non_original_list[famnum],
							this_famtree);
					num_par_score[new_vec] += this_prob;
				}

			} else {
				if (!algebra)
					init_disease_probs(num_in_ped, this_famtree);
				if (save_num_nuc_fams_to_peel[famnum] > 0)
					peel(vec, num_bits, this_famtree, this_nuc_fam);
				if (!algebra){
					num_par_score[new_vec] =  brute_force_analyze(save_bf_originals[famnum], save_bf_original_list[famnum], save_bf_non_originals[famnum], save_bf_non_original_list[famnum], this_famtree);
				} else {
					if ((!modscore_done || matrix_to_be_done) && !skip_ped) {
						algebra_matrix(save_bf_originals[famnum], save_bf_original_list[famnum],
								save_bf_non_originals[famnum], save_bf_non_original_list[famnum], this_famtree, new_vec, this_untotal_datfile_par_score);
					}
				}
			}
		} /* end of 'if != NPL_ANALYSIS */
		percentage = (int) (100.0 * (new_vec + 1) / real_num_vecs);
		if ((percentage > last_percentage) && !scan_done) {
			fprintf(stderr, "\r%3d%%", percentage);
			last_percentage = percentage;
		}
      } /* end of loop over real_num_vecs */
    if (!scan_done) {
        fprintf(stderr, "\n");
    }
	if (algebra)
	{
		if ((!modscore_done || matrix_to_be_done) && !skip_ped){
			num_versions = version_counter[famnum_pointer[famnum]]; /* was initialized with 0 */
			is_currently_in_data[famnum_pointer[famnum]] = algebra_pointer[famnum]; 
			version_pointer[famnum_pointer[famnum]][num_versions] = algebra_pointer[famnum];
			version_counter[famnum_pointer[famnum]]++;
			algebra_analyzed[famnum_pointer[famnum]] = TRUE; /* should be initalized with FALSE */
		}
	}

		pr_time[50] = (double)clock()/(double)CLOCKS_PER_SEC;
/*		printf(" finished graph_assign_probs\n");
*/
		for (i=0; i<f_num_in_map_order; i++) {
			if(f_uninformative[i]) p_pnt[i]++;
			if(p_pnt[i] != nz_pointer[i+1]) {
			  scan_error_flag = TRUE;
			  sf(ps," Pointer consistency error: \n"); pr();
			  sf(ps," \t locus, p_pnt, nz_pointer = %4d  %lu %lu \n",
				  i, p_pnt[i], nz_pointer[i+1]); pr();
				exit(1);
			}
		}

      if (new_vec != real_num_vecs) {
	  scan_error_flag = TRUE;
	  error("num_vecs don't match in hackprep");
      }
      
      exp /= real_num_vecs;
      exp_2 /= real_num_vecs;
    
      /* normalize probabilities to sum to 1.0 for each marker */
	if (!skip_ped) {
		for (i=0; i<f_num_in_map_order; i++) {
		  total=0.0;
		  for(unsigned long h=nz_pointer[i]; h<nz_pointer[i+1]; h++) total += save_pvct[famnum][h];
		  if (total == 0.0) {
		    if (!pmod_calc_flag) {
		      sf(ps,"ERROR(AF): bad inheritance in pedigree %s, marker %s\n", current_ped, locus[f_map_order[i]].name); 
		      pr();
		      sf(ps,"SKIPPING THIS PEDIGREE\n"); 
                      pr(); nl();
		    } else if (replicates_store_result) { 
		      sf(ps,"ERROR(AF): bad inheritance in pedigree %s, marker %s\n", current_ped, locus[f_map_order[i]].name); 
	              fpr(replifp);
		      sf(ps,"SKIPPING THIS PEDIGREE\n"); 
	              fpr(replifp); fnl(replifp);
		    }
                    if (replicates_count == 0 && !pmod_calc_flag) nl();
		    skip_ped = TRUE; num_pedigrees--; skipped_peds++; 
		    strcat(skipped_peds_char," ");
		    strcat(skipped_peds_char,current_ped);
		    break;
	    	  } else {
		    for(unsigned long h=nz_pointer[i]; h<nz_pointer[i+1]; h++)
		      save_pvct[famnum][h] /= total;
		    if(f_uninformative[i]) 
		      save_pvct[famnum][nz_pointer[i]] = 1/real_num_vecs;
		    /* for(j=nz_pointer[i]; j<nz_pointer[i+1]; j++) 
		      save_pvct[famnum][j] *= real_num_vecs; K. Strauch, to calculate large maps */
		  }
		}
	}

	/* calculate single marker entropy */
#ifdef DUMP_SINGLE_P_INFOC 
		if (!skip_ped) {
			for (i=0; i<f_num_in_map_order; i++) {
				k = 0;
				total=0.0;
				for(j=nz_pointer[i]; j<nz_pointer[i+1]; j++) {
					if(save_pvct[famnum][j] != 0) { total += (-1.0) * save_pvct[famnum][j] * log(save_pvct[famnum][j]); }
					else             { k++; }
				}
				printf("                        ");
				if(f_uninformative[i]) {
					printf("iM = %3d  *\n", i);
				} else {
					printf("iM = %3d  more_zeroes= (%10d/%10d)  (%2d/%2d) s-info = %6.4f\n",
			        		 i, k, m_pow2[nz_bits[i]],
							 real_num_bits - nz_bits[i], real_num_bits, 
							 1 - total/(real_num_bits*log(2)));
				}
			}
		}
#endif

	/* Check for uninformative markers */
	j = 0;  for(i=0; i<f_num_in_map_order; i++) if(uninformative_marker[i]) j++;
	if(j) {

#if 0
 Debugging only:
		sf(ps,"-------------------------------------------------- \n"); pr();
		sf(ps,"WARNING: the following marker(s) are uninformative \n"); pr();
		sf(ps,"-------------------------------------------------- \n"); pr();
		sf(ps,"  index(ordered)  index(unordered)  position\n"); pr();
		total = 0.;
		for (i=0; i<f_num_in_map_order; i++) {
			if(uninformative_marker[i]) {
				sf(ps,"          %3d             %3d      %12.4e\n",
				i, map_order[i], total); pr();
			}
			total += map_distance[i];
		}
#endif
                if (!pmod_calc_flag) {
                  sf(ps,"\nin pedigree %s:", current_ped); 
		  pr(); 
		  sf(ps,"\nf_map_order[0:%2d] = ", f_num_in_map_order - 1); 
		  pr(); 
		  for(iMarker=0; iMarker<f_num_in_map_order; iMarker++) {
		    sf(ps,"%3d", f_map_order[iMarker]);  pr();
		  }

		  sf(ps,"\n"); pr(); 
		  sf(ps,"                    "); pr(); 
		  for(iMarker=0; iMarker<f_num_in_map_order; iMarker++) {
			if(uninformative_marker[iMarker]) { 
			  if(f_uninformative[iMarker]) {
				 sf(ps,"  *"); pr(); 
			  } 
			  else {
				 sf(ps,"  X"); pr();
			  } /* X--> it was not caught before */
			} else {
			  sf(ps,"  ."); pr();
			}
		  }
		  sf(ps,"\n"); pr();
		  sf(ps,"  (The markers marked above with a * or X are uninformative.)\n"); pr();
          if (g_association) {
	          sf(ps,"  (* --> no contribution to inheritance vector distribution )\n"); pr();
	          sf(ps,"  (X --> still used for prior/left/right probabilities      )\n"); pr();
          } else {
	          sf(ps,"  (* --> automatically removed                              )\n"); pr();
	          sf(ps,"  (X --> still used for prior/left/right probabilities      )\n"); pr();
          }
	   	  if (!(!skip_ped  &&  j == f_num_in_map_order)) nl();
		  if (!g_association && !skip_ped  &&  j == f_num_in_map_order) {
		    sf(ps,"ERROR: no informative markers in pedigree %s\n", current_ped); pr(); 
		    sf(ps, "SKIPPING PEDIGREE %s\n", current_ped); pr(); 
		  }
                  nl();
		} else if (replicates_store_result) {
		  sf(ps,"\nin pedigree %s:", current_ped); 
                  fpr(replifp);
		  sf(ps,"\nf_map_order[0:%2d] = ", f_num_in_map_order - 1); 
		  fpr(replifp);
		  for(iMarker=0; iMarker<f_num_in_map_order; iMarker++) {
		    sf(ps,"%3d", f_map_order[iMarker]); 
		    fpr(replifp);
		  }
		  sf(ps,"\n"); fpr(replifp);
		  sf(ps,"                    "); fpr(replifp);
		  for(iMarker=0; iMarker<f_num_in_map_order; iMarker++) {
			if(uninformative_marker[iMarker]) { 
			  if(f_uninformative[iMarker]) {
				 sf(ps,"  *"); fpr(replifp);
			  } 
			  else {
				 sf(ps,"  X"); fpr(replifp);
			  } /* X--> it was not caught before */
			} else {
			  sf(ps,"  . "); fpr(replifp);
			}
		  }
                  sf(ps,"\n"); fpr(replifp);
	          sf(ps,"  (The markers marked above with a * or X are uninformative.)\n");
	          fpr(replifp);
              if (g_association) {
                  sf(ps,"  (* --> no contribution to inheritance vector distribution )\n");
                  fpr(replifp);
                  sf(ps,"  (X --> still used for prior/left/right probabilities      )\n");
                  fpr(replifp);
              } else {
                  sf(ps,"  (* --> automatically removed                              )\n");
                  fpr(replifp);
                  sf(ps,"  (X --> still used for prior/left/right probabilities      )\n");
                  fpr(replifp);
              }
		  if (!(!skip_ped  &&  j == f_num_in_map_order)) fnl(replifp);
		  if (!skip_ped  &&  j == f_num_in_map_order) {
		    sf(ps,"ERROR: no informative markers in pedigree %s\n", current_ped);
    fpr(replifp);
		    sf(ps, "SKIPPING PEDIGREE %s\n", current_ped);
		    fpr(replifp);
		    fnl(replifp);
		  }
	       }
	       if (!g_association && !skip_ped  &&  j == f_num_in_map_order) {
	       	 skip_ped = TRUE; num_pedigrees--; skipped_peds++;
		 strcat(skipped_peds_char," ");
		 strcat(skipped_peds_char,current_ped);
	       }
	}
	
	if ((!skip_ped) && (analysis_type != NPL_ANALYSIS)){
		total=0.0;
		if (!algebra)
		{
			for(j=0; j<real_num_vecs; j++)
				total += num_par_score[j];
		}else {
			for(j=0; j<real_num_vecs; j++)
				total += this_untotal_datfile_par_score[j];
			for(j=0; j<real_num_vecs; j++){ /* normalization needed here for datfile model */ 
				this_datfile_par_score[j] = this_untotal_datfile_par_score[j]/total;
			}
		}

		if (total == 0.0) {
			if (!pmod_calc_flag) {
			sf(ps,"disease pvectors sum to 0.0 - SKIPPING PEDIGREE %s\n", current_ped); 
			pr(); 
			datafile_model_incompatible = TRUE;
                        /*** if MOD-score analysis is done later,
                             then exit from modscore routine with error
                             since at least one pedigree has been discarded
                             due to incompatibility with the trait phenotypes
			***/
                         nl();
			 } else if (replicates_store_result) {
			  sf(ps,"disease pvectors sum to 0.0 - SKIPPING PEDIGREE %s\n", current_ped); 
			  fpr(replifp);
			  datafile_model_incompatible = TRUE;
			  fnl(replifp);
			}
			skip_ped = TRUE; num_pedigrees--; skipped_peds++;
		 	strcat(skipped_peds_char," ");
			strcat(skipped_peds_char,current_ped);
		} else {
			if (!algebra){
			for(j=0; j<real_num_vecs; j++) num_par_score[j] /= total;
			}
		}
	}
	
		if(EXIT_EARLY) { 
		  scan_error_flag = TRUE;
		  if (replicates_count == 0 && !pmod_calc_flag) {
		    sf(ps," ------ EXIT_EARLY ------\n"); 
                    pr();
                  }
		  skip_ped = TRUE; 
		}

    } /* if (!skip_ped) */

    if (!skip_ped && !g_is_singleton[algebra_pointer[famnum]]) {
      /* call scan stuff now that pvector is filled in */
      /* count original parents of analyzed members */
      /* complexity=0; */
        if (g_first_pedigree == 0) {
            g_first_pedigree = 1;
        } else if (g_first_pedigree == 1) {
            g_first_pedigree = -1;
        }
      total_meioses += num_bits - really_uninformative_bits;
      total_meioses_m += num_bits_m - really_uninformative_bits_m;
      total_meioses_f += num_bits_f - really_uninformative_bits_f;	
      /***** now that we correctly cut out first kids bits for each
	     original parent, we should actually be passing no
	     redundant vectors (except in the double missing parent
	     case) so we'll leave complexity=0 *****/
      /**********
      for (i=0; i<num_in_ped; i++) {
	if (famtree[famnum][i].dad_index==MISSING && famtree[famnum][i].mom_index==MISSING &&
	    famtree[famnum][i].analyzed_parent) complexity++;
      }
      if (num_bits < non_originals*2) complexity -= (non_originals*2-num_bits);
      if (complexity < 0) complexity=0;
      **********/

      /* make bitmasks */
      uninformative_mask=0; count=1;
      array(save_flip_mask[famnum], num_bits, int);
      allocate_flip_mask[famnum] = 1;
      for (i=0; i<num_bits; i++) {
	if (uninformative_bit[i]) {
	    uninformative_mask+=count;
	    save_flip_mask[famnum][i]=0;
	
	    if (i%2 == 1) { vecpos = (num_bits-i-1)/2; parent=MALE; }
	    else          { vecpos = (num_bits-i-2)/2; parent=FEMALE; }
	    for (j=0; j<real_non_originals; j++) {
	      if (famtree[famnum][sorted_ped[famnum][j]].inh_vec_pos == vecpos) {
		/* mark this person and sibs pat or mat bit */
		this_indiv = sorted_ped[famnum][j];
		if (parent == MALE) this_parent=famtree[famnum][this_indiv].dad_index;
		else                this_parent=famtree[famnum][this_indiv].mom_index;
		break;
	      }
	    }
	    for (j=0; j<real_non_originals; j++) {
	        if (parent == MALE) {
		  if (famtree[famnum][sorted_ped[famnum][j]].dad_index == this_parent) {
		    vecpos=famtree[famnum][sorted_ped[famnum][j]].inh_vec_pos;
		    save_flip_mask[famnum][i]+=m_pow2[num_bits-(2*vecpos)-1];
		  }
		} else {
		  if (famtree[famnum][sorted_ped[famnum][j]].mom_index == this_parent) {
		    vecpos=famtree[famnum][sorted_ped[famnum][j]].inh_vec_pos;
		    save_flip_mask[famnum][i]+=m_pow2[num_bits-(2*vecpos)-2];
		  }
		}
	    }
	}
	count *= 2;
      }

      if(compute_sharing) make_member_list(current_ped, num_in_ped); /*** new - alt 10/18 ***/

/*      printf("call scan_pedigree \n");
*/	  pr_time[60] = (double)clock()/(double)CLOCKS_PER_SEC;
      scan_pedigree(nz_bits, nz_bits_m, nz_bits_f, nz_pointer, save_pvct[famnum], num_par_score, f_num_in_map_order, f_uninformative, f_map_order,
		      real_num_bits, real_num_bits_m, real_num_bits_f, score, score_mask_rdc, exp, exp_2,
		      num_pedigrees-1,best_pvector,
		      num_bits, num_bits_m, num_bits_f, uninformative_mask,save_flip_mask[famnum],
		      save_maskTest[famnum], -1);

      pr_time[80] = (double)clock()/(double)CLOCKS_PER_SEC;

	pr_time_all +=  pr_time[80] - pr_time[10];
	for(i=11; i<200; i++) { 
		pr_time[i] = pr_time[i] - pr_time[10];
	}

#ifdef DUMP_TIME 
	  printf("time 40,60,80,all=%12.4e %12.4e %12.4e %12.4e\n",
				pr_time[40], pr_time[60], pr_time[80], pr_time_all);

/*	printf("cum% 40,60,80,all=%12.3f %12.3f %12.3f\n",
	100*pr_time[40]/pr_time[80], 
	100*pr_time[60]/pr_time[80], 
	100*pr_time[80]/pr_time[80]);
*/	
	  printf("     40,60,80,all=%12.3f %12.3f %12.3f\n",
				100*pr_time[40]/pr_time[80], 
				100*(pr_time[60]-pr_time[40])/pr_time[80], 
				100*(pr_time[80]-pr_time[60])/pr_time[80]);
#endif

	if (haplotype_output) {
	/* need to replace eliminated bits in best_pvectors so that
	   the haplotyping display can do everything easily */
	/*****
	for (i=0; i<num_bits; i++) {
	  if (uninformative_bit[i]) {
	    add_bit(num_bits-num_uninformative_bits,i,
		    best_pvector,num_in_map_order);
	    num_uninformative_bits--;
	  }
	}
	*****/

	npl_max = npl_pos = apm_stat[num_pedigrees-1][2];
	for (i=0; i<num_w_pos; i++) {
	  if (apm_stat[num_pedigrees-1][i] > npl_max) 
	    npl_max = apm_stat[num_pedigrees-1][i];
	}

	matrix(alleles_to_draw,num_in_ped,num_in_map_order*2,int);
	matrix(inferred,num_in_ped,num_in_map_order*2,int);
	matrix(haplotype,originals*2,num_markers,int);
	array(best_pvector_tmp, num_in_map_order, int)

	/* Map best_vector from f_map_order to map_order */	
	for(i=0; i<f_num_in_map_order; i++)
		best_pvector_tmp[i] = best_pvector[i];
	j=0;
	for(i=0; i<f_num_in_map_order; i++) {
		best_pvector[j] = best_pvector_tmp[i];
		while((map_distance[j] == 0.) && (j<num_in_map_order)) {
			best_pvector[++j] = best_pvector_tmp[i]; 
		}
		j++;
	}

	sf(ps,"*****  %s  %.3lf\n",current_ped,npl_max);
	fpr(haplofp);
    ps[0] = '\0'; pr();
	orig_haplotypes(num_in_ped,best_pvector,alleles_to_draw,haplotype,inferred);

	if (postscript_output)
	  draw_pedigree(num_in_ped,num_in_map_order,map_order,current_ped,
			alleles_to_draw,best_pvector,sorted_ped[famnum],
			real_non_originals,haplotype,originals,inferred);
	unarray(best_pvector_tmp, int)
	unmatrix(haplotype,originals*2,int);
	unmatrix(inferred,num_in_ped,int);
	unmatrix(alleles_to_draw,num_in_ped,int);

      }

    } else if (non_originals==1) {
      /* dump haplotypes of single affected member pedigrees */
      if (haplotype_output) {
	/* need to replace eliminated bits in best_pvectors so that
	   the haplotyping display can do everything easily */

	matrix(alleles_to_draw,num_in_ped,num_in_map_order*2,int);
	matrix(inferred,num_in_ped,num_in_map_order*2,int);
	matrix(haplotype,originals*2,num_markers,int);

	sf(ps,"*****  %s  0.000\n",current_ped);
	fpr(haplofp);


	for (i=0; i<num_in_map_order; i++) best_pvector[i]=0;
	orig_haplotypes(num_in_ped,best_pvector,alleles_to_draw,haplotype,inferred);

	if (postscript_output)
	  draw_pedigree(num_in_ped,num_in_map_order,map_order,current_ped,
			alleles_to_draw,best_pvector,sorted_ped[famnum],
			real_non_originals,haplotype,originals,inferred);
	unmatrix(haplotype,originals*2,int);
	unmatrix(inferred,num_in_ped,int);
	unmatrix(alleles_to_draw,num_in_ped,int);
      }
    }
    if (!skip_ped) {
      save_num_in_ped[famnum] = num_in_ped;
      save_real_non_originals[famnum] = real_non_originals;
      save_num_inbred[famnum] = num_inbred;
      save_old_mom[famnum] = old_mom;
      save_num_bits[famnum] = num_bits;
      save_real_num_bits[famnum] = real_num_bits;
      save_real_num_vecs[famnum_count] = real_num_vecs;

      if (modcalc == 2 || (modcalc == 1 && num_replicates > 0)) {
        /* if (!real_num_vecs_calculated) {
          array(save_par_score[famnum_count], real_num_vecs, double); 
          for (j=0; j<save_real_num_vecs[famnum_pointer[famnum]]; j++)
            save_par_score[famnum_count][j] = par_score[j];
          * allocate memory for saved model vectors for those families 
             that have not been analyzed since start *
         } */
        real_num_vecs_calculated[famnum_count]= TRUE;
      }
      save_num_bits_m[famnum] = num_bits_m;
      save_real_num_bits_m[famnum] = real_num_bits_m;
      save_num_bits_f[famnum] = num_bits_f;
      save_real_num_bits_f[famnum] = real_num_bits_f;
      save_npadMS[famnum] = npadMS;

      if ((modcalc || num_replicates > 0) && (modscore_done && saved_models)) {
        if (!saved_model_allocated[famnum_count]) {
          matrix(saved_model_vectors[famnum_count], saved_models,
          save_real_num_vecs[famnum_count], double);
          saved_model_allocated[famnum_count] = TRUE;
        }
        if (diff_discard) {
          unmatrix(saved_model_vectors[famnum_count], saved_models, double);
          matrix(saved_model_vectors[famnum_count], saved_models, real_num_vecs, double);
          /* Calculate families new for each QUICK calc, since saved_model_vectors changed !!! */
          for (i=0; i< saved_models; i++)
             quick_model_famnum[famnum_count][i] = 0;
          /* unarray(save_par_score[famnum_count], double);
          array(save_par_score[famnum_count], real_num_vecs, double);
          for (j=0; j<save_real_num_vecs[famnum_pointer[famnum]]; j++)
            save_par_score[famnum_count][j] = par_score[j]; */
        }
      }

      for (i=0; i<64; i++) {
	save_padMask[famnum][i] = padMask[i];
	save_padShift[famnum][i] = padShift[i];
      }
      for (i=0; i<num_in_map_order; i++){
	save_nz_bits[famnum][i] = nz_bits[i];
	save_nz_bits_m[famnum][i] = nz_bits_m[i];
	save_nz_bits_f[famnum][i] = nz_bits_f[i];
      }	
      for (i=0; i<num_in_map_order+1; i++)
	save_nz_pointer[famnum][i] = nz_pointer[i];
      save_f_num_in_map_order[famnum] = f_num_in_map_order;
      for (i=0; i<num_in_map_order; i++) {
	save_f_map_order[famnum][i] = f_map_order[i];
	save_f_uninformative[famnum][i] = f_uninformative[i];
      }
      save_uninformative_mask[famnum] = uninformative_mask;
      strcpy (pedigree_names[algebra_pointer[famnum]], current_ped);
      if (g_is_singleton[algebra_pointer[famnum]]) {
          pedigrees_calculated++;
      }
    }

    if (skip_ped) {

      if (allocate_flip_mask[famnum] == 1) {
        unarray (save_flip_mask[famnum], int);
        allocate_flip_mask[famnum] = 0;
      }
     if (allocate_maskTest[famnum] == 1) {
        unmatrix (save_maskTest[famnum], f_num_in_map_order, int);
        allocate_maskTest[famnum] = 0;
      }
     if (allocate_pvct[famnum] == 1) {
        unarray (save_pvct[famnum], double);
        allocate_pvct[famnum] = 0;
      }
   }
  } on_exit {
    /* free memory */
      /** unarray(flip_mask, int); 
        * deleted for MOD version **/
    unarray(best_pvector,         int);
    unarray(num_real_vecs,        int);
    unarray(uninformative_marker, int);
    unmatrix(affected_allele,     num_in_ped,         int);
    unarray(score,                double);
      /** unmatrix(pvector, num_in_map_order+1, double); **/
    if(!algebra)
	    unarray(num_par_score,            double); 
    /** unarray(pvct,                 double); 
        * deleted for MOD version **/
    unarray(nz_pointer, unsigned long);
    unarray(nz_bits,    int);
    unarray(nz_bits_m,    int);
    unarray(nz_bits_f,    int);
    unarray(p_pnt,      unsigned long);
      /** unmatrix (maskTest, f_num_in_map_order, int); 
        * deleted for MOD version **/
    unarray (maskHomP, int);
    relay_messages;
  }
}

int Define_family_map_order( int originals, int non_originals,
                             int num_in_map_order, int *map_order,
							 int *f_map_order,
							 int *f_uninformative,
							 int print_flag, int num_inbred)
{
	/* dangerous kludge 
	   meant to take care of repeated appearance of same marker 
	   when merging data sets from several groups */

	int i, j, iMarker, n_genotyped, last_marker;
	int i1, i2, i3, i4, im;
    int f_num = 0;
    /* int process_marker; */
	int *uninformative, *save_ngen;
	double total;


	if(single_point_mode) {
		for(iMarker=0; iMarker<num_in_map_order; iMarker++) {
			f_map_order[iMarker]   = map_order[iMarker];
			f_uninformative[iMarker] = FALSE;
			f_num=num_in_map_order;
		}
		return (f_num);	
	}
	
	array(uninformative, num_in_map_order+1, int);
	array(save_ngen,     num_in_map_order+1, int);

	for(iMarker=0; iMarker<num_in_map_order; iMarker++) {
		uninformative[iMarker] = TRUE;
		save_ngen[iMarker]     = 0;
	}
	for(iMarker=0; iMarker<num_in_map_order; iMarker++) {
		n_genotyped = 0;
		for (i=0,j=0; i<originals+non_originals; i++,j++) {
			if (famtree[famnum][j].discard) { i--; continue; }
			if (allele_data[j][allele_offset+(2*map_order[iMarker])] != 0 ||
			    allele_data[j][allele_offset+(2*map_order[iMarker])+1] != 0) 
				n_genotyped++;
		}

/* 		Bugfix, 1/02 v3R3*/
		if (num_inbred >=1)
		  uninformative[iMarker] = (n_genotyped < 1);
		else
		  uninformative[iMarker] = (n_genotyped < 2);

		save_ngen[iMarker]     = n_genotyped;
	}
/*	printf("Found uninformative marker(s): \n");
	i = 0;
	for(iMarker=0; iMarker<num_in_map_order; iMarker++) {
		if(uninformative[iMarker]) {
			i++;
			if(i > 10) { i=1;  printf("\n"); }
			printf("%4d", map_order[iMarker]);
		}
	}
	printf("\n");
*/	
	i1 = i2 = -1;
	total = 0;
	if(print_flag & 1) printf("  iM map_order d1     d2    n_genotyped \n");
	for(iMarker=0; iMarker<num_in_map_order; iMarker++) {
	
		last_marker = (iMarker == num_in_map_order - 1);
		if(i1 == -1) i1 = iMarker;
		i2 = iMarker;
		
		if(print_flag & 1) {
			printf(" %3d  %3d  %7.2lf %7.2lf %2d",
			iMarker, map_order[iMarker], map_distance[iMarker], 
			total, save_ngen[iMarker]);
			if(uninformative[iMarker]) {
				printf(" *\n");
			} else {
				printf("  \n");
			}
		}

		total += map_distance[iMarker];
		
		if(!last_marker) { if(map_distance[iMarker] == 0.) continue; }
		
		
/*		if(last_marker) {
			process_marker = TRUE;
		} else {
			process_marker = (map_distance[iMarker] != 0.);
		}
*/		
		/* process completed group */
		i3 = 0;
		for(i=i1; i<=i2; i++) { 
			if(!uninformative[i]) {
				i3++;     /* # of informative in group */
				i4 = i;   /* last informative index */
			}
		}
		if(i3 == 0) { /* no marker in the group is informative.  Pick first*/
			i4 = i1;
			if(print_flag & 2) printf("                                 ");
			if(i1 == i2) {
				if(print_flag & 2) { printf("WARNING uninformative marker      : (%4d,%4d)\n",
				iMarker, map_order[iMarker]); }
			} else {
				if(print_flag & 2) { sf(ps,"WARNING uninformative marker group: (%4d,%4d)-(%4d,%4d)\n",
				i1, map_order[i1], i2, map_order[i2]); pr(); }
			}
		} else if(i3 > 1) { /* can't have more than one informative at the same spot */
		        scan_error_flag = TRUE;
			sf(ps,"**** Error two informative markers at the same position\n"); pr();
			sf(ps,"     check markers (ordered, unordered) index:\n"); pr();
			for(i=i1; i<=i2; i++) { sf(ps,"(%4d,%4d) ", i, map_order[i]); pr(); }
			sf(ps,"\n"); pr();
			
			
			
			
			for(im=i1; im<=i2; im++) {
				printf(" im, map_order = %4d %4d \n", im, map_order[im]);
				for (i=0,j=0; i<originals+non_originals; i++,j++) {
					if (famtree[famnum][j].discard) { i--; continue; }
					printf(" %s\t  %2d  %2d \n", famtree[famnum][j].indiv_ID,
					allele_data[j][allele_offset+(2*map_order[im])],
					allele_data[j][allele_offset+(2*map_order[im])+1]);
				}
			}
			
			
			
			exit(1);
		}
		f_map_order[f_num]   = map_order[i4];       /* previous group */
		f_uninformative[f_num] = uninformative[i4];
		f_num++;
		i1 = i2 = -1;
		if(print_flag & 1) printf("                                 select (%4d,%4d)\n", 
			i4, map_order[i4]);
	}

	if(print_flag & 4) { 
		printf("f_map_order[0:%2d] = ", f_num - 1);
		for(iMarker=0; iMarker<f_num; iMarker++) {
			printf("%3d", f_map_order[iMarker]);
		}
		printf("\n");
		printf("                    ");
		for(iMarker=0; iMarker<f_num; iMarker++) {
			if(f_uninformative[iMarker]) { printf("  *"); } else { printf("   "); }
		}
		printf("\n");
	}
    
	unarray(uninformative, int);
	unarray(save_ngen,     int);
	
	return (f_num);
}

void Set_npl_score_masks(int *score_mask, int *score_mask_rdc, int*score_alloc,
                         int real_non_originals, int num_bits)
{
	int i, j, s_mask, s_mask_rdc, bits_used;
	s_mask = 0;
	for (i=0; i<real_non_originals; i++) {
		/* if (not affected) && (does not have kids) */
		if ((allele_data[sorted_ped[famnum][i]][0] != 2) && 
		    (famtree[famnum][sorted_ped[famnum][i]].nkids==0)) {
			j = famtree[famnum][sorted_ped[famnum][i]].inh_vec_pos;
			s_mask |= 3 << (num_bits - j*2 - 2);
		}
	}

	s_mask_rdc = j = bits_used = 0;
	for (i=0; i<num_bits; i++) {
		if(!uninformative_bit[i]) {
			if(s_mask & m_pow2[i]) {
				s_mask_rdc |= m_pow2[j];
			} else {
				bits_used++;
			}
			j++;
		}
	}
	
	s_mask     ^= 0xffffffff;
	s_mask_rdc ^= 0xffffffff;
	*score_mask     = s_mask;
	*score_mask_rdc = s_mask_rdc;
	*score_alloc    = m_pow2[bits_used];
/*	printf("-score_mask = %8x \n", s_mask);	
	printf("-s_mask_rdc = %8x \n", s_mask_rdc);
	printf("score_alloc = %8d \n", *score_alloc); 
*/

}

int parents_informative(int ind)
{
    int dad,mom,dad_sum,mom_sum,i;

    dad = famtree[famnum][ind].dad_index;
    mom = famtree[famnum][ind].mom_index;

    dad_sum=mom_sum=0;
    for (i=allele_offset; i<num_allele_data[dad]; i++) {
        if (allele_data[dad][i] > 0) dad_sum++;
	if (allele_data[mom][i] > 0) mom_sum++;
    }
    if (dad_sum == num_allele_data[dad]-allele_offset &&
	mom_sum == num_allele_data[mom]-allele_offset) return(TRUE);
    else return(FALSE);
}

int has_data(int ind) 
{
  int i, sum, pos;

  sum=0;
  /**********
  for (i=allele_offset; i<num_allele_data[ind]; i++) {
      if (allele_data[ind][i] > 0) sum++;
  }
  **********/
  for (i=0; i<num_in_map_order; i++) {
      pos = (map_order[i]*2) + allele_offset;
    if (allele_data[ind][pos] > 0) sum++;
    if (allele_data[ind][pos+1] > 0) sum++;
  }

  if (sum > 0) return(TRUE);
  else return(FALSE);
}

void set_dad_index (int indiv,int num_in_ped)
{
  /* Find the index of dad in the the pedigree list and
     do basic pedigree checks */

  bool found;
  int i;

  /* If the LINKAGE pedigree has a 0 for dad or mom
     that means the parent is missing */
  if (!strcmp(famtree[famnum][indiv].dad,"0")) {
    famtree[famnum][indiv].dad_index = MISSING;
  } else {
    /* Go through the pedigree to find dad */
    found=FALSE;
    for (i=0; i<num_in_ped; i++) {
      if (!strcasecmp(famtree[famnum][indiv].dad, famtree[famnum][i].indiv_ID)) {
	/* Check that the father isn't impossible */
	if (i == indiv) {
	  scan_error_flag = TRUE;
	  sf(ps, "individual %s is own father", famtree[famnum][indiv].indiv_ID);
	  error(ps);
	}
	if (famtree[famnum][i].sex != MALE) {
	  scan_error_flag = TRUE;
	  sf(ps, "father of indiv %s is female", famtree[famnum][indiv].indiv_ID); error(ps);
	}
	/* No errors, record the info */
	found=TRUE;
	famtree[famnum][indiv].dad_index=i;
	famtree[famnum][i].nkids++; /* Increment dad's # of kids */
	break;
      }
    }
    if (!found) {
      scan_error_flag = TRUE;
      sf(ps,"can't find father of indiv %s (%s) in pedigree",
	 famtree[famnum][indiv].indiv_ID,famtree[famnum][indiv].dad);
      error(ps);
    }
  }
}

void set_mom_index (int indiv,int num_in_ped)
{
  /* Find the index of mom in the the pedigree list and
     do some basic pedigree checks */

  bool found;
  int i;

  /* If the LINKAGE pedigree has a 0 for mom or dad
     that means the parent is missing */
  if (!strcmp(famtree[famnum][indiv].mom,"0")) {
    famtree[famnum][indiv].mom_index = MISSING;
  } else {
    /* Go through the pedigree to find mom */
    found=FALSE;
    for (i=0; i<num_in_ped; i++) {
      if (!strcasecmp(famtree[famnum][indiv].mom, famtree[famnum][i].indiv_ID)) {
	/* Check that the mother isn't impossible */
	if (i == indiv) {
	  scan_error_flag = TRUE;
	  sf(ps, "individual %s is own mother",
	     famtree[famnum][indiv].indiv_ID);
	  error(ps);
	}
	if (famtree[famnum][i].sex != FEMALE) {
	  scan_error_flag = TRUE;
	  sf(ps, "mother of indiv %s is male", 
	     famtree[famnum][indiv].indiv_ID); error(ps);
	}
	/* No errors, record the info */
	found=TRUE;
	famtree[famnum][indiv].mom_index=i;
	famtree[famnum][i].nkids++; /* Increment mom's # of kids */
	break;
      }
    }
    if (!found) {
      scan_error_flag = TRUE;
      if (g_association && num_in_ped == 2) {
          sf(ps, "Parent-offspring pair detected.  Since you're doing JLA analysis there are 3 options:\n"
                 "\n"
                 "    1) Add dummy parent with missing genotypes to form a trio.\n"
                 "    2) Truncate parent-offspring pair to singleton by removing offspring.\n"
                 "    3) Delete parent-offspring pair.\n"
                 "\n"
                 "Options 1 and 2 preserve association information for JLA analysis.");
          error(ps);
      }
      sf(ps,"can't find mother of indiv %s (%s) in pedigree",
	 famtree[famnum][indiv].indiv_ID,famtree[famnum][indiv].mom);
      error(ps);
    }
  }
}


void check_for_unlinked (int num_in_ped)
{
  /* Everyone in the pedigree should either have
     parents or kids, otherwise they have no links
     to the rest of the pedigree and can be excluded
     from analysis.  Finds any such individuals, prints
     a warning and removes them from the famtree list */

  int i,num_unlinked=0,*unlinked,new_num_in_ped;

  array(unlinked,MAX_INDIVIDUALS,int);
  new_num_in_ped = num_in_ped;

  for (i=0; i<new_num_in_ped; i++)
    if (famtree[famnum][i].nkids==0 &&
	famtree[famnum][i].dad_index==MISSING &&
	famtree[famnum][i].mom_index==MISSING) {
      unlinked[num_unlinked]=i;
      num_unlinked++;
    }

  if (num_unlinked > 0) {

    /* ifdef debugging_output removed (K. Strauch 04/2004) */
    if (!pmod_calc_flag) {
      if (num_unlinked==1) {
        sf(ps, "pedigree %s contains an unlinked individual (%s) that will not be analyzed\n",
	 current_ped, famtree[famnum][unlinked[0]].indiv_ID); pr();

      } else {
        sf(ps, "pedigree %s contains unlinked individuals that will not be analyzed\n",
	 current_ped); pr();
      }
    } else if (replicates_store_result) {
      if (num_unlinked==1) {
        sf(ps, "pedigree %s contains an unlinked individual (%s) that will not be analyzed\n",
	 current_ped, famtree[famnum][unlinked[0]].indiv_ID); fpr(replifp);
      } else {
        sf(ps, "pedigree %s contains unlinked individuals that will not be analyzed\n",
	 current_ped); fpr(replifp);
      }
    }
    
    /* here was endif before */

    for (i=0; i<num_unlinked; i++) {
      /* Remove unlinked indivs from the famtree list */
      famtree[famnum][unlinked[i]].discard=TRUE;
    }
  }
  unarray(unlinked,int);
}

void check_for_loop (int start_indiv, int sex, int new_index, 
		     int level, int max_levels)
{

  /* Make sure that start_indiv does not create a loop in
     the pedigree.  Recursively check all the ancestors */

  int dad, mom; /* dad and mom of the individual where we
		   are up in the pedigree */
  
  if (level >= max_levels) {
    sf(ps,"loop detected in pedigree %s\n",current_ped); pr();
    return;
  }

  dad = famtree[famnum][new_index].dad_index;
  mom = famtree[famnum][new_index].mom_index;

  if (dad != MISSING) {
    if (sex == MALE)
      if (!strcasecmp(famtree[famnum][dad].indiv_ID,
		      famtree[famnum][start_indiv].indiv_ID)) {
	scan_error_flag = TRUE;
	sf(ps,"individual %s is his own ancestor",
	   famtree[famnum][start_indiv].indiv_ID); error(ps);
      }
    check_for_loop(start_indiv,sex,dad,level+1,max_levels);
  }
  if (mom != MISSING) {
    if (sex == FEMALE)
      if (!strcasecmp(famtree[famnum][mom].indiv_ID,
		      famtree[famnum][start_indiv].indiv_ID)) {
	scan_error_flag = TRUE;
	sf(ps,"individual %s is her own ancestor",
	   famtree[famnum][start_indiv].indiv_ID); error(ps);
      }
    check_for_loop(start_indiv,sex,mom,level+1,max_levels);
  }
}  


/*******************************************************
 Sort the pedigree into a list with all ancestors first
 *******************************************************/
void sort_pedigree (int num_in_ped)
{

  /* Take the famtree list of PED_MEMBERs and
     make a list of its indices such that each
     member is present after his or her parents */

  int i;

  if (num_in_ped > MAX_INDIVIDUALS) {
    scan_error_flag = TRUE;
    sf(ps,"too many individuals in pedigree %s - max %d",
       current_ped, MAX_INDIVIDUALS); error(ps);
  }

  for (i=0; i<num_in_ped; i++) {
    if (famtree[famnum][i].discard) placed[i]=TRUE;
    else placed[i]=FALSE;
  }
  num_placed=0;

  for (i=0; i<num_in_ped; i++) {
    /* add each member of the pedigree to the sorted list */
    if (!placed[i]) place_indiv(i);    
  }

}

void place_indiv (int ped_index)
{
  /* Place each individual after recursively
     placing any parents */
  
  /* Only place the individual if they are a non-original */
  if (famtree[famnum][ped_index].dad_index == MISSING &&
      famtree[famnum][ped_index].mom_index == MISSING) {
    /* Individual is an original */
    placed[ped_index]=TRUE;
    return;
  }
      
  /* Check that dad has been added to the list */
  if (!placed[famtree[famnum][ped_index].dad_index])
    place_indiv(famtree[famnum][ped_index].dad_index);

  /* Check that mom has been added to the list */
  if (!placed[famtree[famnum][ped_index].mom_index])
    place_indiv(famtree[famnum][ped_index].mom_index);

  /* All ancestors in the list, add the current
     individual */
  sorted_ped[famnum][num_placed]=ped_index;
  placed[ped_index]=TRUE;
  num_placed++;

}



void fill_placeholder_alleles_0(int vec, int non_originals, int for_hap,
			 PED_MEMBER *this_famtree)
{
    int i, indiv, mom, dad, num_bits, vec_pos;

    num_bits = non_originals*2;

    if (!pmod_calc_flag) {
        if(g_association && g_is_singleton[algebra_pointer[famnum]]) {
            // do nothing
        } else {
      if (vec==0 && !for_hap) print("using non-originals: ");
        }
    }

    for (i=0; i<non_originals; i++) {
        /* sorted_ped[famnum][] = index of non_originals in correct order */
	indiv = sorted_ped[famnum][i];
	dad = this_famtree[indiv].dad_index;
	mom = this_famtree[indiv].mom_index;

	vec_pos = num_bits - 2*i - 1;

        this_famtree[indiv].place_allele[0] = 
	  this_famtree[dad].place_allele[(vec>>vec_pos)&1];

        this_famtree[indiv].place_allele[1] = 
	  this_famtree[mom].place_allele[(vec>>(vec_pos-1))&1];

        if (!pmod_calc_flag) {
            if(g_association && g_is_singleton[algebra_pointer[famnum]]) {
                // do nothing
            } else {
                if (vec==0 && !for_hap) {
                    sf(ps," %s",this_famtree[indiv].indiv_ID);
                    pr();
                }
            }
        }
	    
	if (vec==0 && !for_hap) { 
	  this_famtree[indiv].inh_vec_pos = i;
	  this_famtree[dad].analyzed_parent=TRUE;
	  this_famtree[mom].analyzed_parent=TRUE;
	  if (this_famtree[indiv].dad_bit_uninf)
	    uninformative_bit[vec_pos] = TRUE;
	  else uninformative_bit[vec_pos] = FALSE;
	  if (this_famtree[indiv].mom_bit_uninf)
	    uninformative_bit[vec_pos-1] = TRUE;
	  else uninformative_bit[vec_pos-1] = FALSE;	  
	}
    }

    if (!pmod_calc_flag) {
        if(g_association && g_is_singleton[algebra_pointer[famnum]]) {
            // do nothing
        } else {
            if (vec==0 && !for_hap) {
                nl();
            }
        }
    }
}


void fill_placeholder_alleles_1(int vec, int non_originals,
			   PED_MEMBER *this_famtree)
{
    int i, indiv, num_bits, vec_pos;

    num_bits = non_originals*2;

    for (i=0; i<non_originals; i++) {
        /* sorted_ped[famnum][] = index of non_originals in correct order */
	indiv = sorted_ped[famnum][i];

	vec_pos = num_bits - 2*i - 1;

	this_famtree[indiv].place_allele[0] =
	  this_famtree[this_famtree[indiv].dad_index].place_allele[(vec>>vec_pos)&1];
	this_famtree[indiv].place_allele[1] = 
	  this_famtree[this_famtree[indiv].mom_index].place_allele[(vec>>(vec_pos-1))&1];
    }
}


double graph_assign_probs(int originals,int non_originals,int marker,
			  int **haplotype, bool record_haplotype)
{
    /* graph_assign_probs formerly had 'int vec' as first argument */
    /* allele_data[][]  holds the allele data for each member of
       famtree as follows:

       allele_data[indiv][0] = affectation value

       allele_data[indiv][allele_offset+(2*i)] and
       allele_data[indiv][allele_offset+(2*i)+1]  = alleles for ith marker
    
       famtree[famnum][indiv].place_allele [0] and [1] are the two place holder
       alleles for the individual...for originals, these represent the 
       nodes of the graph, for non-originals, these represent the obligate
       assignments given the assignments of the originals and the 
       particular inheritance vector (vec) we are working on
    */
    int i, j, num_edges;
    void create_nodes(int), AddEdge(int,int,int,int),
         process_graph(ASSIGN_STRUCT*), free_assign(ASSIGN_STRUCT*);
    ASSIGN_STRUCT *assign_list,*alist;
    double prob, total_prob;

    create_nodes(originals*2);
    
    num_edges=0;
    for (i=0,j=0; i<originals+non_originals; i++,j++) {
      
        if (famtree[famnum][j].place_allele[0]==-1) { i--; continue; }
	if (famtree[famnum][j].discard) { i--; continue; }

	/* Changed the following from && to || to allow
	   processing of (A,0) genotypes - MJD 3/98 */

        if (allele_data[j][allele_offset+(2*marker)] != 0 ||
	    allele_data[j][allele_offset+(2*marker)+1] != 0) {

	    /******
	   sf(ps,"IN: (%d,%d,%d,%d) \n", 
	       famtree[famnum][j].place_allele[0],
	       famtree[famnum][j].place_allele[1],
	       allele_data[j][allele_offset+(2*marker)],
	       allele_data[j][allele_offset+(2*marker)+1] ); pr();
	    ******/
	    /******
	    if((vec < 1) && (marker<3) ){
	    	printf("--v=%d\t m=%d\t j=%d\t %s\t-- ",
		       vec, marker, j, famtree[famnum][j].indiv_ID);
		   sf(ps,"IN: (%d,%d,%d,%d) \n", 
		       famtree[famnum][j].place_allele[0],
		       famtree[famnum][j].place_allele[1],
		       allele_data[j][allele_offset+(2*marker)],
		       allele_data[j][allele_offset+(2*marker)+1] ); pr();
	    }
	    ******/
	    
	    AddEdge( famtree[famnum][j].place_allele[0],
		     famtree[famnum][j].place_allele[1],
		     allele_data[j][allele_offset+(2*marker)],
		     allele_data[j][allele_offset+(2*marker)+1] );
	    num_edges++;
	}
    }

    if (num_edges > 0) { 
        single (assign_list, ASSIGN_STRUCT);
	assign_list->next=NULL; assign_list->num_alleles=0;
        process_graph(assign_list);
	/*** free_assign(assign_list); WHY WAS THIS HERE? ***/
	/******
	if (marker == 0) {
		alist=assign_list;
		if (alist->num_alleles != 0){
			printf("--v=%x\t m=%d\t--", vec, marker);
			while(alist != NULL) {
				printf(":%d:", alist->num_alleles);
				for(i=0; i<alist->num_alleles; i++) {
					printf("%d ",alist->allele_list[i]);
				}
				alist = alist->next;
			}
			printf("--\n");
		} else {
			printf("..v=%x\t m=%d\t \n", vec, marker);
		}
	}
	******/
	alist=assign_list; total_prob=0.0;

	if (record_haplotype == TRUE) {
	  for (i=0; i<(originals*2); i++)
	    haplotype[i][marker] = alist->force_list[i];
	  /***** debugging output
	  sf(ps,"marker %d: ",marker+1); pr(); 
	  for (i=0; i<alist->num_alleles; i++) {
	    sf(ps,"%d ",alist->allele_list[i]); pr();
	  }
	  nl();
	  *****/
	}
	
	if (alist->num_alleles == 0) {
	  /* no consistent assignments, prob=0.0 */
	} else {
	  while (alist!=NULL) {
	    if (alist->num_alleles == 0) { prob=0.0; }
	    else {
	      prob = 1.0;
	      for (i=0; i<alist->num_alleles; i++) {
		if (alist->allele_list[i] == 0) {
		  /* add nothing to prob - this node was held at zero */
		} else if (locus[marker].allele_freq[alist->allele_list[i]] == 0.0) {
		  scan_error_flag = TRUE;
		  sf(ps,"non-existant allele %d at marker %d in graph_probs",
		     alist->allele_list[i],marker+1); error(ps);
		} else {
		  prob *= locus[marker].allele_freq[alist->allele_list[i]];
		  /* to normalize as Linkage does, prob would also have to
		     be divided by locus[marker].allele_sum here each time */
		}
	      }
	    }
	    total_prob += prob;
	    alist = alist->next;
	  }
	}
	free_assign(assign_list);
	return(total_prob);
    } else {
      /* all individuals are missing, set all vec probs = 1 */
      return(1.0);
    }
  }


void free_assign(ASSIGN_STRUCT *alist)
{
    if (alist->next != NULL) free_assign(alist->next);
    unsingle(alist, ASSIGN_STRUCT);
}


void fill_nuc_families(int num_in_ped)
{
  
  int i,j,kid[MAX_KIDS];
  int num_kids, found, dad, mom;
  bool checked[MAX_INDIVIDUALS];


  for (i=0; i<num_in_ped; i++)
    checked[i]=FALSE;
  
  num_nuc_fam=0;
  
  for (i=0; i<num_in_ped; i++) {
    if (famtree[famnum][i].discard) continue;
    if (checked[i]) continue;

    if (famtree[famnum][i].nkids > 0) {

      /* Make a list of all the kids of the individual */
      num_kids=0;
      for (j=0; j<num_in_ped; j++){
	if (num_kids == famtree[famnum][i].nkids) break;
	if (famtree[famnum][j].dad_index == i ||
	    famtree[famnum][j].mom_index == i) {
	  if (famtree[famnum][j].discard) continue;
	  kid[num_kids]=j;
	  num_kids++;
	}
      }
      
      if (num_kids > 0) {
	nuc_fam[famnum][num_nuc_fam].pivot = -1;
	nuc_fam[famnum][num_nuc_fam].inbred=FALSE;
	nuc_fam[famnum][num_nuc_fam].in_loop=FALSE;
	nuc_fam[famnum][num_nuc_fam].loop_indiv=FALSE;
	nuc_fam[famnum][num_nuc_fam].num_common_anc=0;
	nuc_fam[famnum][num_nuc_fam].dad = famtree[famnum][kid[0]].dad_index;
	nuc_fam[famnum][num_nuc_fam].mom = famtree[famnum][kid[0]].mom_index;
	nuc_fam[famnum][num_nuc_fam].num_kids=0;
	for (j=0; j<num_kids; j++) {
	  if (famtree[famnum][kid[j]].mom_index == nuc_fam[famnum][num_nuc_fam].mom &&
	      famtree[famnum][kid[j]].dad_index == nuc_fam[famnum][num_nuc_fam].dad) {
	    nuc_fam[famnum][num_nuc_fam].kid[nuc_fam[famnum][num_nuc_fam].num_kids]=kid[j];
	    nuc_fam[famnum][num_nuc_fam].num_kids++;
	  }
	}
	num_nuc_fam++;
      }

      checked[nuc_fam[famnum][num_nuc_fam-1].dad]=TRUE;
      checked[nuc_fam[famnum][num_nuc_fam-1].mom]=TRUE;
    }
  }

  /* Final check to ensure that all kids have their nuclear
     family represented - just to cover the unusual case where
     both parents are also parents of other nuclear pedigrees */

  for (i=0; i<num_in_ped; i++) {
    if (famtree[famnum][i].discard) continue;
    dad = famtree[famnum][i].dad_index;
    mom = famtree[famnum][i].mom_index;
    if (dad == MISSING || mom == MISSING) continue;

    found=FALSE;
    for (j=0; j<num_nuc_fam; j++) {
      if (nuc_fam[famnum][j].dad == dad && nuc_fam[famnum][j].mom == mom) {
	found=TRUE; break;
      }
    }

    if (!found) {
      /* Create a new nuc_fam */
      num_kids=0;
      for (j=0; j<num_in_ped; j++){
	if (famtree[famnum][j].dad_index == dad && famtree[famnum][j].mom_index == mom) {
	  if (famtree[famnum][j].discard) continue;
	  kid[num_kids]=j;
	  num_kids++;
	}
      }
      if (num_kids > 0) {
	print("Special double half-sib case detected\n");
	nuc_fam[famnum][num_nuc_fam].pivot = -1;
	nuc_fam[famnum][num_nuc_fam].inbred=FALSE;
	nuc_fam[famnum][num_nuc_fam].in_loop=FALSE;
	nuc_fam[famnum][num_nuc_fam].loop_indiv=FALSE;
	nuc_fam[famnum][num_nuc_fam].num_common_anc=0;
	nuc_fam[famnum][num_nuc_fam].dad = dad;
	nuc_fam[famnum][num_nuc_fam].mom = mom;
	nuc_fam[famnum][num_nuc_fam].num_kids=0;
	for (j=0; j<num_kids; j++) {
	  if (famtree[famnum][kid[j]].mom_index == nuc_fam[famnum][num_nuc_fam].mom &&
	      famtree[famnum][kid[j]].dad_index == nuc_fam[famnum][num_nuc_fam].dad) {
	    nuc_fam[famnum][num_nuc_fam].kid[nuc_fam[famnum][num_nuc_fam].num_kids]=kid[j];
	    nuc_fam[famnum][num_nuc_fam].num_kids++;
	  }
	}
	num_nuc_fam++;
      }
    }
  }

}

int mark_pivots (int num_in_ped)
{
  /* Make a list of all the pivots in the extended family.
     Then each family who has one member in the pivot
     list is peripheral and we assign them unless we have
     already assigned num_nuc_fam-1 (num_nuc_fam is the
     number of nuclear families in the extended pedigree). */
  
  int i,j,k;
  int num_pivots=0;
  int num_assigned,fam_pivot,num_kids;
  int assigned[MAX_INDIVIDUALS];
  int possible_pivot[MAX_INDIVIDUALS], poss, assigned_this_round;
  bool done;
  

  /* Make the list of pivots -- pivots are individuals
     who have not been discarded, who are non-originals,
     and who have at least one child */
  for (i=0; i<num_in_ped; i++) {
    possible_pivot[i]=0;

    if (famtree[famnum][i].discard) continue;
    if (famtree[famnum][i].mom_index != MISSING &&
	famtree[famnum][i].dad_index != MISSING) {

      /* Count up non-discarded kids */
      num_kids=0;
      if (famtree[famnum][i].nkids > 0)  {
	for (j=0; j<num_in_ped; j++) {
	  if (famtree[famnum][j].discard) continue;
	  if (famtree[famnum][j].mom_index == i ||
	      famtree[famnum][j].dad_index == i)
	    num_kids++;
	}
      }
      if (num_kids > 0) {
	possible_pivot[i]=1; /* has kids and parents */
	num_pivots++;
      }
    }
  }

  if (num_pivots < (num_nuc_fam-1)) {
    /* In this case we have a string of marriages with
       half sibs on the next level, so the pivots will
       be individuals with more than one partner (and may
       actually be original) */
    for (i=0; i<num_nuc_fam; i++) 
      for (j=i+1; j<num_nuc_fam; j++) {
	/* Make a list of individuals shared between two nuc fams */
	if (nuc_fam[famnum][i].dad == nuc_fam[famnum][j].dad) {
	  num_pivots++;
	  possible_pivot[nuc_fam[famnum][i].dad]++;
	}
	if (nuc_fam[famnum][i].mom == nuc_fam[famnum][j].mom) {
	  num_pivots++;
	  possible_pivot[nuc_fam[famnum][i].mom]++;
	}
      }
  }


  for (i=0; i<num_nuc_fam; i++)
    assigned[i]=FALSE;

  /* Assign pivots within nuclear families */
  num_assigned = 0;
  done=FALSE;
  while(!done) {
    assigned_this_round=FALSE;

    for (i=0; i<num_nuc_fam; i++) {
      if (assigned[i]) continue;

      poss=0;
      if (possible_pivot[nuc_fam[famnum][i].dad]) { poss++; fam_pivot=nuc_fam[famnum][i].dad; }
      if (possible_pivot[nuc_fam[famnum][i].mom]) { poss++; fam_pivot=nuc_fam[famnum][i].mom; }
      for (k=0; k<nuc_fam[famnum][i].num_kids; k++) {
	if (possible_pivot[nuc_fam[famnum][i].kid[k]]) {
	  poss++; fam_pivot=nuc_fam[famnum][i].kid[k];
	}
      }

      if (poss == 1) {
	  /* use it, mark the members as done, subtract 1
	     from possible_pivot[member] */
	  assigned_this_round=TRUE;
	  nuc_fam[famnum][i].pivot = fam_pivot;
	  order_assigned[famnum][num_assigned]=i;
	  num_assigned++;
	  assigned[i]=TRUE;
	  possible_pivot[fam_pivot]--;
	  if (num_assigned == (num_nuc_fam-1)) {
	    done=TRUE; break;
	  }
      } else if (poss == 0) {
	  /* we have a nuclear family that is not connected to 
	     the rest of the pedigree - ERROR */
	  sf(ps,"pedigree contains disconnected arms...SKIPPING\n"); 
	  pr(); 
	  sf(ps,"This means that either two pedigrees that are not related\n");
	  pr(); 
	  sf(ps,"or are related only by a marriage with uninformative\n");
	  pr(); 
	  sf(ps,"children are present with the same name and should be\n");
	  pr(); 
	  sf(ps,"divided into two separate pedigrees.\n");
	  pr(); 
	  return(-1);
	  /* returns -1 to indicate to skip the pedigree above */
      }
    }
   
    if (!assigned_this_round) {
      /* can't assign any more, must have hit some sort of loop */
      done=TRUE;
    }
  }

  /* Mark everyone in the unassigned nuclear families as requiring
     further analysis */

  for (i=0; i<num_in_ped; i++) famtree[famnum][i].not_peeled=FALSE;
  
  for (i=0; i<num_nuc_fam; i++) {
    if (assigned[i]) continue;
    
    famtree[famnum][nuc_fam[famnum][i].dad].not_peeled = TRUE;
    famtree[famnum][nuc_fam[famnum][i].mom].not_peeled = TRUE;
    for (k=0; k<nuc_fam[famnum][i].num_kids; k++) 
      famtree[famnum][nuc_fam[famnum][i].kid[k]].not_peeled = TRUE;
  }

  if(algebra || g_association)
  save_num_nuc_fams_to_peel[famnum] = 0; /* MB - 04/2012 - no families to peel */
  else
  save_num_nuc_fams_to_peel[famnum] = num_assigned; /* MB - 10/2013 - to allow for numeric mode and peeling */

  /* Now peeling can be done on all the nuclear families in the
     assigned list, then a brute_force analysis can be applied
     to the remaining individuals not peeled off - yielding one
     analysis method for all types of pedigrees */

  /* All that remains is to recognize long inbreeding loops and
     to subsequently break them (a la Linkage - but automatically)
     or to develop an inbred peeling method  - this has now been
     done in the analyze_family code! */

  return(1);
}


void order_inbred_family (void)
{
  int i,j,k,mom_anc,dad_anc;
  int max_to_common,*loop_order,num_in_loop=0,num_assigned;
  bool *in_loop,found,*checked;
  int inbred_family = -1;

  array(loop_order,num_nuc_fam,int);
  array(checked,num_nuc_fam,bool);
  array(in_loop,num_nuc_fam,bool);
  for (i=0; i<num_nuc_fam; i++) {
    in_loop[i]=FALSE;
    checked[i]=FALSE;
  }

  /* Find the inbred family */
  for (i=0; i<num_nuc_fam; i++) {
    if (nuc_fam[famnum][i].inbred) {
      inbred_family=i;
      break;
    }
  }
  checked[inbred_family]=TRUE;
  in_loop[inbred_family]=TRUE;
  famtree[famnum][nuc_fam[famnum][inbred_family].mom].in_loop=TRUE;
  famtree[famnum][nuc_fam[famnum][inbred_family].dad].in_loop=TRUE;
  loop_order[num_in_loop]=inbred_family;
  num_in_loop++;
  if (nuc_fam[famnum][inbred_family].mom_to_common >
      nuc_fam[famnum][inbred_family].dad_to_common)
    max_to_common = nuc_fam[famnum][inbred_family].mom_to_common;
  else
    max_to_common = nuc_fam[famnum][inbred_family].dad_to_common;

  
  /* Starting with the inbred nuclear family, work
     upward through the families that are in the inbreeding
     loop */
  mom_anc = nuc_fam[famnum][inbred_family].mom;
  dad_anc = nuc_fam[famnum][inbred_family].dad;
  for (i=0; i<=max_to_common; i++) {
    if (i<=nuc_fam[famnum][inbred_family].mom_to_common) {
      /* Find the nuclear family that mom is a part 
	 of.  We'll always be peeling upward, so the
	 indiv will always be among the children */
      for (j=0; j<num_nuc_fam; j++) {
	if (checked[j]) continue;
	found = FALSE;
	for (k=0; k<nuc_fam[famnum][j].num_kids; k++)
	  if (nuc_fam[famnum][j].kid[k] == mom_anc) {
	    found=TRUE;
	    break;
	  }
	if (found) {
	  /* found the family mom of her ancestor is a part of, one
	     of the parents of this nuclear family will be the
	     next 1/2 of a superpivot */
	  if (famtree[famnum][nuc_fam[famnum][j].mom].mom_index != MISSING &&
	      famtree[famnum][nuc_fam[famnum][j].mom].dad_index != MISSING &&
	      famtree[famnum][nuc_fam[famnum][j].mom].nkids > 0)
	    mom_anc = nuc_fam[famnum][j].mom;
	  else
	    mom_anc = nuc_fam[famnum][j].dad;
	  checked[j] = TRUE;
	  nuc_fam[famnum][j].in_loop = TRUE;
	  nuc_fam[famnum][j].loop_indiv = mom_anc;
	  famtree[famnum][mom_anc].in_loop=TRUE;
	  loop_order[num_in_loop]=j;
	  num_in_loop++;
	}
      }
    }
    /* Now do the same routine for dad */
    if (i<=nuc_fam[famnum][inbred_family].dad_to_common) {
      for (j=0; j<num_nuc_fam; j++) {
	if (checked[j]) continue;
	found = FALSE;
	for (k=0; k<nuc_fam[famnum][j].num_kids; k++)
	  if (nuc_fam[famnum][j].kid[k] == dad_anc){
	    found=TRUE;
	    break;
	  }
	if (found) {
	  /* found the family mom of her ancestor is a part of, one
	     of the parents of this nuclear family will be the
	     next 1/2 of a superpivot */
	  if (famtree[famnum][nuc_fam[famnum][j].mom].mom_index != MISSING &&
	      famtree[famnum][nuc_fam[famnum][j].mom].dad_index != MISSING &&
	      famtree[famnum][nuc_fam[famnum][j].mom].nkids > 0)
	    dad_anc = nuc_fam[famnum][j].mom;
	  else
	    dad_anc = nuc_fam[famnum][j].dad;
	  checked[j] = TRUE;
	  nuc_fam[famnum][j].in_loop = TRUE;
	  nuc_fam[famnum][j].loop_indiv = dad_anc;
	  famtree[famnum][dad_anc].in_loop=TRUE;
	  loop_order[num_in_loop]=j;
	  num_in_loop++;
	}
      }
    }
  }

  num_assigned=0;
  if (num_in_loop < num_nuc_fam) {
    /* In this case there are peripheral families outside
       the inbreeding loop where we can assign pivots in
       the usual fashion */
    num_assigned++;
  }
  for (i=0; i<num_in_loop; i++) {
    order_assigned[famnum][num_assigned] = loop_order[i];
    num_assigned++;
  }
  
  unarray(in_loop,bool);
  unarray(loop_order,int);
  unarray(checked,bool);
}
  
static int num_mom, num_dad;

int inbred_matings (void)
{
  /* Return the number of inbred matings */
  
  int i,j,k;
  int mom,dad;
  int *mom_anc, *dad_anc;
  int num_inbred_matings=0;
  int num_in_common;


  array(mom_anc,MAX_INDIVIDUALS,int);
  array(dad_anc,MAX_INDIVIDUALS,int);

  /* If a person has > 0 kids make a list of the
     person's ancestors and the partner(s)'s ancestors
     and make sure they have none in common */
  for (i=0; i<num_nuc_fam; i++) {
    mom = nuc_fam[famnum][i].mom;
    dad = nuc_fam[famnum][i].dad;

    /* Make a list of each parents ancestors */
    num_mom=0;
    num_dad=0;
    get_mat_ancestors(mom,mom_anc);
    get_pat_ancestors(dad,dad_anc);

    /* Compare the lists */
    num_in_common=0;
    for (k=0; k<num_mom; k++) {
      for (j=0; j<num_dad; j++) {

/* Note:  to_common stuff only used in order_inbred_family, which is
 *  not called nomore
 *  --Katz, 1/22/02
 */

	if (mom_anc[k] == dad_anc[j]) {
	  if (num_in_common == 0) {
	    nuc_fam[famnum][i].mom_to_common = (k+2)/2-1;
	    nuc_fam[famnum][i].dad_to_common = (j+2)/2-1;
	  }
	  num_in_common++;
	}
      }
    }
    if (num_in_common > 0) {
      num_inbred_matings++;
      nuc_fam[famnum][i].inbred=TRUE;
      nuc_fam[famnum][i].num_common_anc=num_in_common;
    }

	 if (((dad == famtree[famnum][mom].dad_index) && dad != MISSING) || 
		  ((mom == famtree[famnum][dad].mom_index) && mom != MISSING)) {
      num_inbred_matings++;
      nuc_fam[famnum][i].inbred=TRUE;
      nuc_fam[famnum][i].num_common_anc=num_in_common;
	 }

  }

  unarray(mom_anc,int);
  unarray(dad_anc,int);
  
  return(num_inbred_matings);
}
     

void get_mat_ancestors(int indiv,int *list)
{

  if (famtree[famnum][indiv].mom_index != MISSING) {
    list[num_mom]=famtree[famnum][indiv].mom_index;
    num_mom++;
    get_mat_ancestors(famtree[famnum][indiv].mom_index,list);
  }
  if (famtree[famnum][indiv].dad_index != MISSING) {
    list[num_mom]=famtree[famnum][indiv].dad_index;
    num_mom++;
    get_mat_ancestors(famtree[famnum][indiv].dad_index,list);
  }

}
void get_pat_ancestors(int indiv,int *list)
{

  if (famtree[famnum][indiv].mom_index != MISSING) {
    list[num_dad]=famtree[famnum][indiv].mom_index;
    num_dad++;
    get_pat_ancestors(famtree[famnum][indiv].mom_index,list);
  }
  if (famtree[famnum][indiv].dad_index != MISSING) {
    list[num_dad]=famtree[famnum][indiv].dad_index;
    num_dad++;
    get_pat_ancestors(famtree[famnum][indiv].dad_index,list);
  }

}

  
    
void init_disease_probs(int num_in_ped, PED_MEMBER *this_famtree)
{
  /* sets each pedigree member's a-priori disease allele frequencies */
  /* called before peel() for each inheritance vector */

  int i,aff_stat,liab_class;
  double *current_disease_prob, *current_apriori, *current_penetrance;

  for (i=0; i<num_in_ped; i++) {
    aff_stat = this_famtree[i].affectation_status;
    liab_class = this_famtree[i].liability_class;
    current_disease_prob = this_famtree[i].disease_prob;
    if (this_famtree[i].dad_index==MISSING && this_famtree[i].mom_index==MISSING) {
      current_apriori = apriori_disease[liab_class][aff_stat];
      current_disease_prob[HOM_AFF]   = current_apriori[HOM_AFF];
      current_disease_prob[HET_UA]    = current_apriori[HET_UA];
      current_disease_prob[HET_AU]    = current_apriori[HET_AU];
      current_disease_prob[HOM_UNAFF] = current_apriori[HOM_UNAFF];
    } else {  /* non-founder */
      current_penetrance = penetrance[liab_class];
      if (aff_stat == 1) {
	current_disease_prob[HOM_AFF]   = 1.0 - current_penetrance[HOM_AFF];
	current_disease_prob[HET_UA]    = 1.0 - current_penetrance[HET_UA];
	current_disease_prob[HET_AU]    = 1.0 - current_penetrance[HET_AU];
	current_disease_prob[HOM_UNAFF] = 1.0 - current_penetrance[HOM_UNAFF];
      } else if (aff_stat == 2) {
	current_disease_prob[HOM_AFF]   = current_penetrance[HOM_AFF];
	current_disease_prob[HET_UA]    = current_penetrance[HET_UA];
	current_disease_prob[HET_AU]    = current_penetrance[HET_AU];
	current_disease_prob[HOM_UNAFF] = current_penetrance[HOM_UNAFF];
      } else {  /* UNKNOWN */
	current_disease_prob[HOM_AFF]   = 1.0; /* changed from 0.25 to 1 for correct numeric results in algebraic formulation before normalization takes place, MB 06/2012 */
	current_disease_prob[HET_UA]    = 1.0;
	current_disease_prob[HET_AU]    = 1.0;
	current_disease_prob[HOM_UNAFF] = 1.0;
      }
    }
  }
}

/* expects global nuc_fam[], num_nuc_fams to be filled in */


void peel(int inh_vec, int num_bits, PED_MEMBER *this_famtree,
	  NUC_FAMILY *this_nuc_fam)
{
    int nf, f, i, i_shift, j, k, this_pivot, this_spouse, kid_allele;
    int kid_bit, vec_pos;
    int kid_array_1[4], kid_array_2[4];
    int num_kids;
    double this_prob, new_prob[4];
    double *current_disease_prob, *current_disease_prob2;
    double this_prob_array[16];
    unsigned int kid_allele_bitmask;

    for (f=0; f<save_num_nuc_fams_to_peel[famnum]; f++) {
      nf = order_assigned[famnum][f];
      num_kids = this_nuc_fam[nf].num_kids;

      this_pivot = this_nuc_fam[nf].pivot;
      if (this_pivot == MISSING) {
	scan_error_flag = TRUE;
	error("no pivot listed in non-final nuclear family");
      }

      if (this_pivot == this_nuc_fam[nf].dad || this_pivot == this_nuc_fam[nf].mom) {
	/* PEEL ONTO PARENT */
	if (this_pivot == this_nuc_fam[nf].dad) {
	  this_spouse = this_nuc_fam[nf].mom;
	  current_disease_prob = this_famtree[this_spouse].disease_prob;
	  for (j=0; j<4; j++)
	    this_prob_array[j] = (this_prob_array[4|j] =
	      (this_prob_array[8|j] = (this_prob_array[12|j] =
				       current_disease_prob[j])));
	  for (k=0; k<num_kids; k++) {
	    kid_bit = this_famtree[this_nuc_fam[nf].kid[k]].inh_vec_pos;
	    vec_pos = num_bits - 2*kid_bit - 1;
	    kid_allele_bitmask = ((inh_vec >> vec_pos) & 1)
		               + ((inh_vec >> (vec_pos-1)) & 1) * 2;
	    current_disease_prob =
	      this_famtree[this_nuc_fam[nf].kid[k]].disease_prob;
	    for (i=0; i<4; i++) {
	      kid_array_1[i]
		= ( (i & 1)       & ~kid_allele_bitmask)
		| (((i & 2) >> 1) &  kid_allele_bitmask);
	      kid_array_2[i]
		= (((i & 1) << 1) & ~kid_allele_bitmask)
		| ( (i & 2)       &  kid_allele_bitmask);
	    }
	    for (i=0; i<4; i++) {
	      i_shift = i<<2;
	      for (j=0; j<4; j++) {
		kid_allele = kid_array_1[i] | kid_array_2[j];
		this_prob_array[i_shift|j] *= current_disease_prob[kid_allele];
	      }
	    }
	  }
	} else {  /* pivot is mom */
	  this_spouse = this_nuc_fam[nf].dad;
	  current_disease_prob = this_famtree[this_spouse].disease_prob;
	  for (j=0; j<4; j++)
	    this_prob_array[j] = (this_prob_array[4|j] =
	      (this_prob_array[8|j] = (this_prob_array[12|j] =
				       current_disease_prob[j])));
	  for (k=0; k<num_kids; k++) {
	    kid_bit = this_famtree[this_nuc_fam[nf].kid[k]].inh_vec_pos;
	    vec_pos = num_bits - 2*kid_bit - 1;
	    kid_allele_bitmask = ((inh_vec >> vec_pos) & 1)
		               + ((inh_vec >> (vec_pos-1)) & 1) * 2;
	    current_disease_prob =
	      this_famtree[this_nuc_fam[nf].kid[k]].disease_prob;
	    for (i=0; i<4; i++) {
	      kid_array_1[i]
		= (((i & 1) << 1) & ~kid_allele_bitmask)
		| ( (i & 2)       &  kid_allele_bitmask);
	      kid_array_2[i]
		= ( (i & 1)       & ~kid_allele_bitmask)
		| (((i & 2) >> 1) &  kid_allele_bitmask);
	    }
	    for (i=0; i<4; i++) {
	      i_shift = i<<2;
	      for (j=0; j<4; j++) {
		kid_allele = kid_array_1[i] | kid_array_2[j];
		this_prob_array[i_shift|j] *= current_disease_prob[kid_allele];
	      }
	    }
	  }
        }
	current_disease_prob = this_famtree[this_pivot].disease_prob;
	for (i=0; i<4; i++) {
	  i_shift = i<<2;
	  this_prob = this_prob_array[i_shift];
	  for (j=1; j<4; j++) this_prob += this_prob_array[i_shift|j];
	  current_disease_prob[i] *= this_prob;
	}

      } else {
	/* PEEL ONTO KID */

	current_disease_prob = this_famtree[this_nuc_fam[nf].dad].disease_prob;
	current_disease_prob2 =
	  this_famtree[this_nuc_fam[nf].mom].disease_prob;
	for (i=0; i<4; i++) {
	  new_prob[i] = 0.0;
	  i_shift = i<<2;
	  for (j=0; j<4; j++) this_prob_array[i_shift|j] =
		      current_disease_prob[i] * current_disease_prob2[j];
	}
	for (k=0; k<num_kids; k++) if (this_nuc_fam[nf].kid[k] != this_pivot) {
	  kid_bit = this_famtree[this_nuc_fam[nf].kid[k]].inh_vec_pos;
	  vec_pos = num_bits - 2*kid_bit - 1;
	  kid_allele_bitmask = ((inh_vec >> vec_pos) & 1)
	                     + ((inh_vec >> (vec_pos-1)) & 1) * 2;
	  current_disease_prob =
	    this_famtree[this_nuc_fam[nf].kid[k]].disease_prob;
	  for (i=0; i<4; i++) {
	    kid_array_1[i]
	      = ( (i & 1)       & ~kid_allele_bitmask)
	      | (((i & 2) >> 1) &  kid_allele_bitmask);
            kid_array_2[i]
	      = (((i & 1) << 1) & ~kid_allele_bitmask)
	      | ( (i & 2)       &  kid_allele_bitmask);
	  }
	  for (i=0; i<4; i++) {
	    i_shift = i<<2;
	    for (j=0; j<4; j++) {
	      kid_allele = kid_array_1[i] | kid_array_2[j];
	      this_prob_array[i_shift|j] *= current_disease_prob[kid_allele];
	    }
          }
	}
	kid_bit = this_famtree[this_pivot].inh_vec_pos;
	vec_pos = num_bits - 2*kid_bit - 1;
	kid_allele_bitmask = ((inh_vec >> vec_pos) & 1)
	                   + ((inh_vec >> (vec_pos-1)) & 1) * 2;
	current_disease_prob =
	  this_famtree[this_pivot].disease_prob;
	for (i=0; i<4; i++) {
          kid_array_1[i]
	    = ( (i & 1)       & ~kid_allele_bitmask)
	    | (((i & 2) >> 1) &  kid_allele_bitmask);
	  kid_array_2[i]
	    = (((i & 1) << 1) & ~kid_allele_bitmask)
	    | ( (i & 2)       &  kid_allele_bitmask);
	}
	for (i=0; i<4; i++) {
	  i_shift = i<<2;
	  for (j=0; j<4; j++) {
	    kid_allele = kid_array_1[i] | kid_array_2[j];
	    new_prob[kid_allele] += this_prob_array[i_shift|j];
	  }
	}
	/* save the new probs for the pivot child */
	for (i=0; i<4; i++) current_disease_prob[i] *= new_prob[i];
      }
    }
}


void orig_haplotypes (int num_in_ped, int *iv_idx, int **draw_alleles,
		      int **haplotype, int **inferred)
{

  /* Fill in the haplotypes for the originals in ped */

  int i, j, num_orig=0, num_nonorig=0;
  int *original, indiv;
  int lookup_force(int);
  int all;

  array(original,num_in_ped,int);
  for (i=0; i<num_in_ped; i++) {
    if ( famtree[famnum][i].mom_index == MISSING &&
	 famtree[famnum][i].dad_index == MISSING ) {
      if (!famtree[famnum][i].discard) {
	original[num_orig] = i;
	num_orig++;
      }
    } else { if (!famtree[famnum][i].discard) num_nonorig++; }
  }

  for (i=0; i<num_orig*2; i++)
    for (j=0; j<num_markers; j++)
      haplotype[i][j]=0;

  /* Fill in the haplotype matrix */
  for (i=0; i<num_in_map_order; i++) {
    /* Call the allele dropping code so that the node_array
       and the EdgeList are filled in */
    fill_placeholder_alleles_0(iv_idx[i],num_nonorig,1,famtree[famnum]);
    for (j=0; j<num_in_ped; j++) {
      draw_alleles[j][i*2]=famtree[famnum][j].place_allele[0];
      draw_alleles[j][i*2+1]=famtree[famnum][j].place_allele[1];
      if (allele_data[j][allele_offset+(map_order[i]*2)] == 0 ||
	  allele_data[j][allele_offset+(map_order[i]*2)+1] == 0) {
	  inferred[j][i*2]=TRUE; inferred[j][i*2+1]=TRUE;
      } else {
	  inferred[j][i*2]=FALSE; inferred[j][i*2+1]=FALSE;
      }
    }
    
    graph_assign_probs(num_orig,num_nonorig,map_order[i],haplotype,TRUE);

#ifdef TDT_ACTIVE

    /* keep TDT statistics */

    for (int k=0; k<1; k++) { /* was 12 */
        int dad0, dad1, mom0, mom1, dkid, mkid;

      int count_it=TRUE;

      /**********
      count_it=FALSE;
      if (k==0 && npl_pos > 2.0) count_it=TRUE;
      else if (k==1 && npl_pos > 1.5) count_it=TRUE;
      else if (k==2 && npl_pos > 1.0) count_it=TRUE;
      else if (k==3 && npl_pos > 0.5) count_it=TRUE;
      else if (k==4 && npl_pos > 0.0) count_it=TRUE;
      else if (k==5 && npl_pos < 0.0) count_it=TRUE;
      else if (k==6 && npl_max > 2.0) count_it=TRUE;
      else if (k==7 && npl_max > 1.5) count_it=TRUE;
      else if (k==8 && npl_max > 1.0) count_it=TRUE;
      else if (k==9 && npl_max > 0.5) count_it=TRUE;
      else if (k==10 && npl_max > 0.0) count_it=TRUE;
      else if (k==11 && npl_max < 0.0) count_it=TRUE;
      **********/

      if (!count_it) continue;

     for (j=0; j<num_in_ped; j++) {
       
       if (famtree[famnum][j].affectation_status != AFFECTED) continue;
       if (famtree[famnum][j].discard) continue;
       if (famtree[famnum][j].dad_index==MISSING || famtree[famnum][j].mom_index==MISSING) 
	 continue;

       if (i == 0) total_affecteds[k]++;

       dad0 = haplotype[famtree[famnum][famtree[famnum][j].dad_index].place_allele[0]][map_order[i]];
       dad1 = haplotype[famtree[famnum][famtree[famnum][j].dad_index].place_allele[1]][map_order[i]];
       mom0 = haplotype[famtree[famnum][famtree[famnum][j].mom_index].place_allele[0]][map_order[i]];
       mom1 = haplotype[famtree[famnum][famtree[famnum][j].mom_index].place_allele[1]][map_order[i]];
       dkid = haplotype[famtree[famnum][j].place_allele[0]][map_order[i]];
       mkid = haplotype[famtree[famnum][j].place_allele[1]][map_order[i]];

       if (dad0 <= 0 || dad1 <= 0) {
	 if (mom0 <= 0 || mom1 <= 0) {
	    continue; /* both parents are missing in part */
	 } else {
	    if (mkid==mom0 && dkid!=mom1 && mom0!=mom1 && mkid!=dkid) {
	      locus[map_order[i]].allele_count[mom0][k][TRANSMITTED] += 1;
	      locus[map_order[i]].allele_count[mom1][k][UNTRANSMITTED] += 1;
	    } else if (mkid==mom1 && dkid!=mom0 && mom0!=mom1 && mkid!=dkid) {
	      locus[map_order[i]].allele_count[mom1][k][TRANSMITTED] += 1;
	      locus[map_order[i]].allele_count[mom0][k][UNTRANSMITTED] += 1;
	    } else {
	      continue;
	    }
	  }
	} else {
	  if (mom0 <= 0 || mom1 <= 0) {
	    if (dkid==dad0 && mkid!=dad1 && dad0!=dad1 && mkid!=dkid) {
	      locus[map_order[i]].allele_count[dad0][k][TRANSMITTED] += 1;
	      locus[map_order[i]].allele_count[dad1][k][UNTRANSMITTED] += 1;
	    } else if (dkid==dad1 && mkid!=dad0 && dad0!=dad1 && mkid!=dkid) {
	      locus[map_order[i]].allele_count[dad1][k][TRANSMITTED] += 1;
	      locus[map_order[i]].allele_count[dad0][k][UNTRANSMITTED] += 1;
	    } else {
	      continue;
	    }
	  } else {
	    /* both parents present */
	    if (dad0 != dad1) {
	      if (dkid==dad0) {
		locus[map_order[i]].allele_count[dad0][k][TRANSMITTED] += 1;
		locus[map_order[i]].allele_count[dad1][k][UNTRANSMITTED] += 1;
	      } else {
		locus[map_order[i]].allele_count[dad1][k][TRANSMITTED] += 1;
		locus[map_order[i]].allele_count[dad0][k][UNTRANSMITTED] += 1;
	      }
	    }
	    if (mom0 != mom1) {
	      if (mkid==mom0) {
		locus[map_order[i]].allele_count[mom0][k][TRANSMITTED] += 1;
		locus[map_order[i]].allele_count[mom1][k][UNTRANSMITTED] += 1;
	      } else {
		locus[map_order[i]].allele_count[mom1][k][TRANSMITTED] += 1;
		locus[map_order[i]].allele_count[mom0][k][UNTRANSMITTED] += 1;
	      }
	    }
	  }
	}

      } /* for each ped member (j) */
     } /* for k = 0 to 12 */

#else
    (void) total_affecteds;
#endif

    } /* for each marker (i) */


  /* Print out the haplotype matrix */
  if (display_scores && haplotype_output) {
    print("\nHAPLOTYPES OF ORIGINAL INDIVIDUALS: \n");
    for (i=0; i<(num_orig*2); i++) {
      indiv = i/2;
      if (i%2 == 0) {
	sf(ps,"\nindiv %s:\t",famtree[famnum][original[indiv]].indiv_ID); pr();
      } else { print("        \t"); }
      
      if (i%2 == 0) {
	sf(ps,"%s\t0\t0\t%d\t",famtree[famnum][original[indiv]].indiv_ID,
	   famtree[famnum][original[indiv]].affectation_status); fpr(haplofp);
      } else {
	sf(ps,"\t\t\t\t"); fpr(haplofp);
      }
      
    for (j=0; j<num_in_map_order; j++) {
      sf(ps,"%-3d ",haplotype[i][map_order[j]]); pr();
      fpr(haplofp);
    }
      nl(); fnl(haplofp);
    }
  }

  /* dump haplos of non-originals to haplofp */
  for (i=0; i<num_in_ped; i++) {
    if (famtree[famnum][i].discard) continue;
    if (famtree[famnum][i].dad_index==MISSING || famtree[famnum][i].mom_index==MISSING) 
      continue;
   
    sf(ps,"%s\t%s\t%s\t%d\t",famtree[famnum][i].indiv_ID,
       famtree[famnum][i].dad, famtree[famnum][i].mom, famtree[famnum][i].affectation_status);
    fpr(haplofp);
    for (j=0; j<num_in_map_order; j++) {
      all = draw_alleles[i][j*2];
      sf(ps,"%-3d ",haplotype[all][map_order[j]]);
      fpr(haplofp);
    }
    sf(ps,"\n\t\t\t\t"); fpr(haplofp);
    for (j=0; j<num_in_map_order; j++) {
      all = draw_alleles[i][(j*2)+1];
      sf(ps,"%-3d ",haplotype[all][map_order[j]]);
      fpr(haplofp);
    }
    fnl(haplofp);
  }
  
  unarray(original,int);
}
    
#define real_eq(x,y) (fabs((x)-(y))<.0000000001)

void remove_bit(int num_bits, int bit_to_remove, double **pvector, double *score,
	   int num_markers)
{
    int i, j, num_vecs, bitval, newcount;
    double *newlist;

    num_vecs=1; bitval=1;
    for (i=0; i<num_bits; i++) num_vecs*=2;
    for (i=0; i<bit_to_remove; i++) bitval*=2;
    array(newlist, num_vecs, double);

    for (j=0; j<num_markers; j++) {
      newcount=0;
      for (i=0; i<num_vecs; i++) {
	if ((i>>bit_to_remove) % 2 == 0) {
	  newlist[newcount] = pvector[j][i];
	  newcount++;
	  if (!real_eq(pvector[j][i],pvector[j][i+bitval])) {
	    sf(ps,"vectors %d and %d differ (marker %d), bit %d is relevant\n",
	       i,i+bitval,j+1,bit_to_remove); pr();
	    sf(ps,"%.12lf vs. %.12lf\n",pvector[j][i],pvector[j][i+bitval]);
	    pr();
	  }
	}
      }
      if (newcount != num_vecs/2) {
	scan_error_flag = TRUE;
	error("in remove_bit");
      }
      for (i=0; i<newcount; i++) pvector[j][i]=newlist[i]*2.0;
    }

    newcount=0;
    for (i=0; i<num_vecs; i++) {
      if ((i>>bit_to_remove) % 2 == 0) {
	newlist[newcount] = score[i];
	newcount++;
	if (score[i] != score[i+bitval]) {
	  sf(ps,"scores for vectors %d and %d differ, bit %d is relevant\n",
	     i,i+bitval,bit_to_remove); pr();
	  break;
	}
      }
    }
    if (newcount != num_vecs/2) {
      scan_error_flag = TRUE;
      error("in remove_bit");
    }
    for (i=0; i<newcount; i++) score[i]=newlist[i];
}


void add_bit(int num_bits, int bit_to_insert, int *pvec, int num_loci)
{
    int i, j, newpvec, pow2;

    for (i=0; i<num_loci; i++) {
      j=0; newpvec=0; pow2=1;
      while (j < bit_to_insert) {
	newpvec += ((pvec[i] >> j)%2)*pow2;
	pow2 *= 2;
	j++;
      }
      j++; pow2 *= 2;
      while (j-1 < num_bits) {
	newpvec += ((pvec[i] >> (j-1))%2)*pow2;
	pow2 *= 2;
	j++;
      }
      pvec[i] = newpvec;
    }
}

/**********
int new_acceptable_vector(int vec, int mask)
{
    if ((vec&mask) > 0) return(FALSE);
    else return(TRUE);
}
**********/

int acceptable_vector(int vec, int num_bits)
{
    int i;
    /* for a vector to be acceptable, all uninformative bits must be 0 */

    for (i=0; i<num_bits; i++) {
      if (uninformative_bit[i]) {
	if ((vec>>i)%2 == 1) return(FALSE);
      }
    }
    return(TRUE);
}

void resort_pedigree(int non_originals)
{
    int i, temp, any_changes_made=FALSE;

    for (i=0; i<non_originals-1; i++) {
        if (famtree[famnum][sorted_ped[famnum][i]].affectation_status != AFFECTED &&

	    (famtree[famnum][sorted_ped[famnum][i+1]].affectation_status == AFFECTED ||
	     (famtree[famnum][sorted_ped[famnum][i]].nkids == 0 && 
	      famtree[famnum][sorted_ped[famnum][i+1]].nkids > 0)) &&

	    (famtree[famnum][sorted_ped[famnum][i+1]].dad_index != sorted_ped[famnum][i] &&
	     famtree[famnum][sorted_ped[famnum][i+1]].mom_index != sorted_ped[famnum][i]) ) {
	    /* we can flip them */
	    temp = sorted_ped[famnum][i];
	    sorted_ped[famnum][i] = sorted_ped[famnum][i+1];
	    sorted_ped[famnum][i+1] = temp;
	    any_changes_made=TRUE;
	}
    }
    if (any_changes_made) resort_pedigree(non_originals);
}

double brute_force_analyze(int originals, int *original_list,
			   int non_originals, int *non_original_list,
			   PED_MEMBER *this_famtree)
{
    unsigned long or_vec, pow4_originals;
    int node_assign[MAX_INDIVIDUALS*2]; 
    int i, indiv, a0, a1, this_dis_loc, n1, n2;
    int or_vec_bit_0[MAX_INDIVIDUALS], or_vec_bit_1[MAX_INDIVIDUALS];
    double this_prob = 1.0;
    double sum_prob[15];

    for (i=0; i<originals; i++) {
	n1 = this_famtree[original_list[i]].place_allele[0];
	n2 = this_famtree[original_list[i]].place_allele[1];
	node_assign[n1] = 2*i;
	node_assign[n2] = 2*i + 1;
	sum_prob[i] = 0.0;
    }
    for (i=0; i<non_originals; i++) {
	indiv = non_original_list[i];
	or_vec_bit_0[indiv]
	    = node_assign[this_famtree[indiv].place_allele[0]];
	or_vec_bit_1[indiv]
	    = node_assign[this_famtree[indiv].place_allele[1]];
    }

    pow4_originals = 1 << 2*originals;

    for (or_vec=0; or_vec<pow4_originals; or_vec++) {
        this_prob = 1.0;
      for (i=0; i<non_originals; i++) {
	indiv = non_original_list[i];
	a0 = (or_vec >> or_vec_bit_0[indiv]) & 1;
	a1 = (or_vec >> or_vec_bit_1[indiv]) & 1;
	this_prob *= this_famtree[indiv].disease_prob[(a1<<1)|a0];
      }

      for (i=0; i<originals; i++) {
	this_dis_loc = (or_vec >> (2*i)) & 3;
	sum_prob[i] += this_prob *
	  this_famtree[original_list[i]].disease_prob[this_dis_loc];
	if (this_dis_loc < 3) break;
	this_prob = sum_prob[i];
	sum_prob[i] = 0.0;
      }

    }
    return(this_prob);
}

int local_pad_inhvec(int vec, int uninf_mask, int num_bits)
{
    int i, newvec, realcount;

    newvec=0; realcount=0;
    for (i=0; i<num_bits; i++) {
      if ((uninf_mask&m_pow2[i]) > 0) {
        /* bit inserted here */
      } else {
        if ((vec&m_pow2[realcount]) > 0) {
          newvec += m_pow2[i];
        }
        realcount++;
      }
    }
    return(newvec);
}

/*----------------------------------------------------*/
/*----------------------------------------------------*/
int Setup_local_pad_inhvec_2(int uninf_mask, int num_bits,
                         int padMask[], int padShift[])
{
	int i, fs_bit, npadMS;
	
	for(i=0; i<64; i++) { padMask[ i] = padShift[i] = -1; }
	
	fs_bit = 0;
	for(i=0; i<num_bits; i++) {
		if(uninf_mask & m_pow2[i]) continue;
		padMask[ fs_bit] = m_pow2[fs_bit];
		padShift[fs_bit] = i - fs_bit;
		fs_bit++;
	}
	
	npadMS = 0;
	for(i=1; i<fs_bit; i++) {
		if(padShift[npadMS] == padShift[i]) {
			padMask[npadMS] |= padMask[i];
		} else {
			npadMS++;
			padMask[ npadMS] = padMask[ i];
			padShift[npadMS] = padShift[i];
		}
	}
	npadMS++;
	
/*	printf(" ------------------------ start i, padMask, padShift\n");
	for(i=0; i<npadMS; i++) 
		printf(" %6d  %8x %8x\n", i, padMask[ i], padShift[i]);
	printf(" ------------------------ end   i, padMask, padShift\n");
*/	
	return(npadMS);
}

/*----------------------------------------------------*/
/*----------------------------------------------------*/
void fill_placeholder_alleles_2(int print_flag, int iFirstMember,
                           BITS vec, int non_originals, int for_hap)
{
	int i,mom,dad,num_bits;

	num_bits = non_originals*2;
	
	for (i=iFirstMember; i<non_originals; i++) {
		famtree[famnum][sorted_ped[famnum][i]].place_allele[0] = -1;
		famtree[famnum][sorted_ped[famnum][i]].place_allele[1] = -1;
	}

	if (vec==0 && !for_hap) {
		if(print_flag) {
            if(g_association && g_is_singleton[algebra_pointer[famnum]]) {
                // do nothing
                } else {
                    print("using non-originals: ");
                }
        }
		for (i=0; i<non_originals; i++) {
            if(print_flag) {
                if(g_association && g_is_singleton[algebra_pointer[famnum]]) {
                    // do nothing
                } else {
                    sf(ps," %s",famtree[famnum][sorted_ped[famnum][i]].indiv_ID);
                    pr();
                }
            }
			dad = famtree[famnum][sorted_ped[famnum][i]].dad_index;
			mom = famtree[famnum][sorted_ped[famnum][i]].mom_index;
			famtree[famnum][sorted_ped[famnum][i]].inh_vec_pos = i;
			famtree[famnum][dad].analyzed_parent=TRUE;
			famtree[famnum][mom].analyzed_parent=TRUE;
			uninformative_bit[num_bits-(2*i)-1]=FALSE;
			uninformative_bit[num_bits-(2*i)-2]=FALSE;	  
			if (famtree[famnum][sorted_ped[famnum][i]].dad_bit_uninf)
				uninformative_bit[num_bits-(2*i)-1] = TRUE;
			if (famtree[famnum][sorted_ped[famnum][i]].mom_bit_uninf)
				uninformative_bit[num_bits-(2*i)-2] = TRUE;
		}
        if (!pmod_calc_flag) {
            if(print_flag) {
                if(g_association && g_is_singleton[algebra_pointer[famnum]]) {
                    //do nothing
                } else {
                    nl();
                }
            }
        }
	}

	for (i=iFirstMember; i<non_originals; i++) {
		/* sorted_ped[famnum][] = index of non_originals in correct order */
		dad = famtree[famnum][sorted_ped[famnum][i]].dad_index;
		mom = famtree[famnum][sorted_ped[famnum][i]].mom_index;

		famtree[famnum][sorted_ped[famnum][i]].place_allele[0] = 
		famtree[famnum][dad].place_allele[(vec>>(num_bits-(2*i)-1))%2];

		famtree[famnum][sorted_ped[famnum][i]].place_allele[1] = 
		famtree[famnum][mom].place_allele[(vec>>(num_bits-(2*i)-2))%2];
	}
}
/*----------------------------------------------------*/










/*----------------------------------------------------*/

/* Katz, Thu May  1 13:17:10 2003 Bugfix on Mac, change emi to dynamic
 * allocation and a few other changes so stack frame not too big.
 */

int  FindIllegalBits(int originals, int real_non_originals,
                     int num_bits, int real_num_bits,
					 int f_num_in_map_order, int *f_map_order, int *f_uninformative)
{
	int i, j, dad, mom, nSibships, nSibs, ifamily;
	int mask, nPhaseFixed, iPhaseFixed, maskPhaseFixed;
	int inh_vec_pos;
	int arrSibships[MAX_INDIVIDUALS];
	int arrSibs[MAX_INDIVIDUALS];
	int uninformative_mask;
	int sib_mask, sib_inf_mask;
	int iMarker, marker, nSibs_inf, arrSibs_inf[9], 
	    arrSibs_inf_rep[MAX_INDIVIDUALS];
	int allele1, allele2, allele3, allele4;
	int i1, inList;
	int nSibs_inf_bits, nSibMS, sibMask[22], sibShift[22];
	BITS vec, vec0;
	int iFmem, iSib, iSib_inf, num_edges;
	int ibit, ibit_s, sum_s0[MAX_INDIVIDUALS*2], sum_s1[MAX_INDIVIDUALS*2];
	int ibit_r, ibit_o;
	int ** volatile emi0=NULL;
    int ** volatile emi1=NULL;
	int iSpMem, nSpMem, spMemID[MAX_INDIVIDUALS], cur_place, spFirstMember;
/*-------------------------------------------- include 	grandparents algorithm */
#define B_POS_OFFS  5
	int nbit_pos, bit_pos[MAX_INDIVIDUALS*B_POS_OFFS*2];
	int indID, nPar_bits, nF_bits;
	int nSibs_inf_bits_all, nSibs_inf_bits_wp;
	int nSibs_inf_vecs_all, nSibs_inf_vecs_wp;
/*-------------------------------------------- matrix num_in_map_order here */
/*-------------------------------------------- position */
	int *legalVec=NULL;
	ASSIGN_STRUCT *assign_list,*alist;
	void create_nodes(int), AddEdge(int,int,int,int),
	     process_graph(ASSIGN_STRUCT*), free_assign(ASSIGN_STRUCT*);

#ifdef PR_00
	  printf("----------- FindIllegalBits ------------\n");
#endif

	  run {
		 matrix(emi0, MAXLOCI, MAX_INDIVIDUALS*2, int);
		 matrix(emi1, MAXLOCI, MAX_INDIVIDUALS*2, int);
	  } except_when (NOMEMORY) {
		 printf("Unable to allocate enough memory in FindIllegalBits\n");
		 unmatrix(emi0, MAXLOCI, int);
		 unmatrix(emi1, MAXLOCI, int);
		 emi0=NULL;
		 emi1=NULL;
	  }
	for(i=0; i<f_num_in_map_order; i++) inconsistent_genotypes[i] = FALSE;
	
	for(i=0; i<MAXLOCI; i++) {
		for(j=0; j<MAX_INDIVIDUALS*2; j++) {
			emi0[i][j] = emi1[i][j] = 9;
		}
	}
	
	vec0 = 0;

	uninformative_mask=0;
	for (i=0; i<num_bits; i++) {
		if (uninformative_bit[i]) uninformative_mask |= m_pow2[i];
	}
#ifdef PR_01
	printf("uninformative_mask=%x \n", uninformative_mask);
#endif
	
	/* Define nuclear families ---------------------------------------*/
	nSibships=0;
	for (i=0; i<real_non_originals; i++) {
		/* sorted_ped[famnum][] = index of real_non_originals in correct order */
		dad = famtree[famnum][sorted_ped[famnum][i]].dad_index;
		mom = famtree[famnum][sorted_ped[famnum][i]].mom_index;
		for(j=0; j<nSibships; j++) {
			if( (dad==arrSibships[j*2]) && (mom==arrSibships[j*2+1]) ) break;
		}
		if( j==nSibships ) {
			arrSibships[nSibships*2]  =dad;
			arrSibships[nSibships*2+1]=mom;
			nSibships++;
		}
	}
	if( (nSibships*2-1) > MAX_INDIVIDUALS ) { 
	        scan_error_flag = TRUE;
		printf(" ******Error arrSibships overflow *******\n");
		exit (1);
	}
	
	/* Execute per nuclear family ---------------------------------------*/
	for(ifamily=0; ifamily<nSibships; ifamily++) {
		dad=arrSibships[ifamily*2];
		mom=arrSibships[ifamily*2+1];

#ifdef PR_03
		printf("sibship of %s  %s\n", famtree[famnum][dad].indiv_ID, 
		                                        famtree[famnum][mom].indiv_ID);
#endif
		
		/* Define sibship (array arrSibs[] & sib_mask -----*/
		nSibs=0;
		for(i=0; i<real_non_originals; i++) {
			if( (dad==famtree[famnum][sorted_ped[famnum][i]].dad_index) && 
			    (mom==famtree[famnum][sorted_ped[famnum][i]].mom_index) ) {
				arrSibs[nSibs]=sorted_ped[famnum][i];
				nSibs++;
			}
		}
/* Here define nSibs_all for discarded indiv. */

#ifdef PR_04
		printf("                  (nSibs=%d)  ", nSibs);
		for(i=0; i<nSibs; i++) printf("%s  ", famtree[famnum][ arrSibs[i] ].indiv_ID);
#endif
		sib_mask = j = nPhaseFixed = iPhaseFixed = 0;
		for(i=0; i<nSibs; i++) {
			j         = famtree[famnum][ arrSibs[i] ].inh_vec_pos;
			mask      = m_pow2[ num_bits - j*2 -1 ] +
			            m_pow2[ num_bits - j*2 -2 ];
			sib_mask |= mask;
			if((mask & uninformative_mask) != 0) { 
				nPhaseFixed++; 
				iPhaseFixed=i;
				maskPhaseFixed=mask;
			}
		}
		if(nPhaseFixed > 1) { 
		        scan_error_flag = TRUE;
			printf(" ******Error nPhaseFixed > 1 *****\n"); exit (1);
		}

#ifdef PR_04
			printf("  sib_mask=%x", sib_mask);
			if(nPhaseFixed==0) printf("  -- no PhaseFixed sib -- ");
			else printf("  PhaseFixed=%s, maskPhaseFixed=%x ", 
			famtree[famnum][ arrSibs[iPhaseFixed] ].indiv_ID, maskPhaseFixed);
			printf("\n");
#else
			(void) maskPhaseFixed; /* silence "set but not used" warning */
#endif

		/* itterate over the marker list.  Informative sibship set is 
		marker dependent */
		/*------- execute per marker ---------------------------*/
		for(iMarker=0; iMarker<f_num_in_map_order; iMarker++) {
			marker = f_map_order[iMarker];
			
			/* find a set of informative sibs (unique genotype) */
			/* if no "phase fixer" just 
			set first sib (e.g. consanguineous pedigree) */
 			if(nPhaseFixed==0) iPhaseFixed=0;
			
			/* now define set of informative sibs */
			arrSibs_inf[0]=arrSibs[iPhaseFixed];
			nSibs_inf=1;
			for(i=0; i<nSibs; i++) {
				if(i==iPhaseFixed) { continue; }
				i1=arrSibs[i];
				allele1=allele_data[i1][allele_offset+(2*marker)];
				allele2=allele_data[i1][allele_offset+(2*marker)+1];
				if(allele1==0 && allele2==0) { 
					arrSibs_inf_rep[arrSibs[i]]=-1;
					continue; 
				}
				inList=0;
				for(j=1; j<nSibs_inf; j++) {
					i1=arrSibs_inf[j];
					allele3=allele_data[i1][allele_offset+(2*marker)];
					allele4=allele_data[i1][allele_offset+(2*marker)+1];
					if(COMP_GENOTYPES(allele1,allele2,allele3,allele4)) {
						inList=1;
						arrSibs_inf_rep[arrSibs[i]]=j;
						break;
					}
				}
				if(!inList) { 
					arrSibs_inf[nSibs_inf]=arrSibs[i];
					arrSibs_inf_rep[arrSibs[i]]=nSibs_inf;
					nSibs_inf++;
				}
			}
			sib_inf_mask=0;
			for(i=0; i<nSibs_inf; i++) {
				j         = famtree[famnum][ arrSibs_inf[i] ].inh_vec_pos;
				mask      = m_pow2[ num_bits - j*2 -1 ] +
				            m_pow2[ num_bits - j*2 -2 ];
				sib_inf_mask |= mask;
			}

#ifdef PR_05
			  printf("                  marker=%d\t nSibs_inf=%d\t arrSibs_inf=", 
						marker, nSibs_inf);
			for(i=0; i<nSibs_inf; i++) printf("%s  ", 
														 famtree[famnum][ arrSibs_inf[i] ].indiv_ID);
			printf("  sib_inf_mask=%x\n", sib_inf_mask);
#endif

/*---- max number of informative sibs = 9 = 1*phaseFixer + 4*(A,B) + 4*(A,0) */
			if(nSibs_inf>9) {
			        scan_error_flag = TRUE;
				printf(" ******Error nSibs_inf overflow *******\n");
				exit (1);
			}

			nSibs_inf_bits = 2*(nSibs_inf-1);
			nbit_pos       = 0;

			/*  for array (struct)  bit_pos[]: 
			    indID    index for famtree[famnum][indID].x access
			    ibit_s   bit position for (inf.sibs+parent) inheritance vector 
			    ibit_r   bit position for founder symmetry reduced inheritance vector
			    i1       flag that:
				         -1  the bit belongs to an individual ungenotyped 
				             in this marker => cannot be constrained.
				          1  bit participates in the inf.sibs+parent 
						     inheritance vector.
				          0  bit represented in the inf.sibs+parent inh. vector.
			    ibit_o   bit position for original inheritance vector*/
			
			/* all sibs but the phase fixer */
			for(i=0; i<nSibs; i++) {
				if(i==iPhaseFixed) { continue; }
				iSib_inf    = arrSibs_inf_rep[arrSibs[i]];
				indID       = arrSibs[i];
				inh_vec_pos = famtree[famnum][indID].inh_vec_pos;

				/* ibit paternal=1, maternal=0 */
				for(ibit=1; ibit >= 0; ibit--) {
					ibit_r = ibit_o = (num_bits - inh_vec_pos*2 - 2) + ibit;
					for(j=0; j<ibit_o; j++) if(uninformative_bit[j]) ibit_r--;
					if(iSib_inf==-1) {
						/* this marker is ungenotyped, meiosis cannot be constrained*/
						ibit_s =  0;
						i1     = -1;
					} else {
						ibit_s = 2*(nSibs_inf - 1- iSib_inf)    + ibit;
						i1 = 0;
					}
					for(j=1; j<nSibs_inf; j++) {
						if(indID==arrSibs_inf[j]) {i1=1; continue;}
					}
					bit_pos[B_POS_OFFS*nbit_pos]     = indID;
					bit_pos[B_POS_OFFS*nbit_pos + 1] = ibit_s;
					bit_pos[B_POS_OFFS*nbit_pos + 2] = ibit_r;
					bit_pos[B_POS_OFFS*nbit_pos + 3] = i1;
					bit_pos[B_POS_OFFS*nbit_pos + 4] = ibit_o;
					nbit_pos++;
				}
			}

			/* parents if grandparents exist */
			nPar_bits = 0;
			for(i=0; i<2; i++) {
				if(i==0) indID = mom;
				else     indID = dad;

				/* require both grandparents in pedigree */
				if( (famtree[famnum][indID].dad_index==MISSING) || 
	    			(famtree[famnum][indID].mom_index==MISSING)   ) continue;

				inh_vec_pos = famtree[famnum][indID].inh_vec_pos;

				for(ibit=0; ibit<2; ibit++) {
					ibit_r = ibit_o = (num_bits - inh_vec_pos*2 - 2) + ibit;
					for(j=0; j<ibit_o; j++) if(uninformative_bit[j]) ibit_r--;

					/* bit fixed by founder symmetry?*/
					if((uninformative_mask & m_pow2[ibit_o]) != 0) continue;

					if(emi0[marker][ibit_r] == 1) {
						vec0 |= 1 << ibit_o;
					} else if( (emi0[marker][ibit_r] != 0) &&
		    			       (emi0[marker][ibit_r] != 1)   ) {

						nPar_bits++;
						ibit_s = nSibs_inf_bits - 1 + nPar_bits;
						bit_pos[B_POS_OFFS*nbit_pos]     = indID;
						bit_pos[B_POS_OFFS*nbit_pos + 1] = ibit_s;
						bit_pos[B_POS_OFFS*nbit_pos + 2] = ibit_r;
						bit_pos[B_POS_OFFS*nbit_pos + 3] = 1;
						bit_pos[B_POS_OFFS*nbit_pos + 4] = ibit_o;
						nbit_pos++;
					}
				}
			}

			/* reference phase (phase fixer) sib */
			nF_bits     = 0;
			indID       = arrSibs[iPhaseFixed];
			inh_vec_pos = famtree[famnum][indID].inh_vec_pos;

			/* ibit paternal=1, maternal=0 */
			for(ibit=1; ibit >= 0; ibit--) {
				ibit_r = ibit_o = (num_bits - inh_vec_pos*2 - 2) + ibit;
				for(j=0; j<ibit_o; j++) if(uninformative_bit[j]) ibit_r--;

				/* bit fixed by founder symmetry?*/
				if((uninformative_mask & m_pow2[ibit_o]) != 0) continue;

				nF_bits++;
				ibit_s = nSibs_inf_bits - 1 + nPar_bits + nF_bits;
				bit_pos[B_POS_OFFS*nbit_pos]     = indID;
				bit_pos[B_POS_OFFS*nbit_pos + 1] = ibit_s;
				bit_pos[B_POS_OFFS*nbit_pos + 2] = ibit_r;
				bit_pos[B_POS_OFFS*nbit_pos + 3] = 1;
				bit_pos[B_POS_OFFS*nbit_pos + 4] = ibit_o;
				nbit_pos++;
			}

			nSibs_inf_bits_wp  = nSibs_inf_bits    + nPar_bits;
			nSibs_inf_bits_all = nSibs_inf_bits_wp + nF_bits;
			nSibs_inf_vecs_wp  = m_pow2[ nSibs_inf_bits_wp ];
			nSibs_inf_vecs_all = m_pow2[ nSibs_inf_bits_all ];
			
#ifdef PR_15
			for(i=0; i<nbit_pos; i++) {
			  indID  = bit_pos[B_POS_OFFS*i];
			  ibit_s = bit_pos[B_POS_OFFS*i + 1];
			  ibit_r = bit_pos[B_POS_OFFS*i + 2];
			  i1     = bit_pos[B_POS_OFFS*i + 3];
			  ibit_o = bit_pos[B_POS_OFFS*i + 4];
			  printf("                  ");
			  printf("%s ibit_r=%3d  ibit_s=%3d  %2d  ibit_o=%3d \n", 
						famtree[famnum][ indID ].indiv_ID, ibit_r, ibit_s, i1, ibit_o);
			}
#endif


			array (legalVec, nSibs_inf_vecs_all, int);
			for(i=0; i<nSibs_inf_vecs_all; i++) legalVec[i]=0;
			for(ibit=0; ibit<nSibs_inf_bits_all; ibit++) {
				sum_s0[ibit] = sum_s1[ibit] = 0;
			}
			
			/* build a mask-shift array */
			for(i=0; i<22; i++) { sibShift[i] = -1; }
			for(i=0; i<nbit_pos; i++) {
				ibit_s = bit_pos[B_POS_OFFS*i + 1];
				i1     = bit_pos[B_POS_OFFS*i + 3];
				ibit_o = bit_pos[B_POS_OFFS*i + 4];
				if(i1==1) {
					sibMask[ibit_s]  = 1 << ibit_s;
					sibShift[ibit_s] = ibit_o - ibit_s;
				}
			}
			nSibMS=0;
			for(i=1; i<nSibs_inf_bits_all; i++) {
				if(sibShift[i]==sibShift[nSibMS]) {
					sibMask[nSibMS] |= sibMask[i];
				} else {
					nSibMS++;
					sibMask[nSibMS]  = sibMask[i];
					sibShift[nSibMS] = sibShift[i];
				}
			}
			nSibMS++;
			/* END build a mask-shift array */
			
#ifdef PR_16
			for(i=0; i<nSibMS; i++){
			  printf("                  MaskShift %3d  %8x  %d\n", 
						i, sibMask[i], sibShift[i]);
			}
#endif

#ifdef PR_17
			for(i=0; i<nSibs_inf_vecs_all; i++) {
			  vec = 0;
			  for(j=0; j<nSibMS; j++){ 
				 if(sibShift[j]>0) vec |= (i & sibMask[j]) <<  sibShift[j];
				 else              vec |= (i & sibMask[j]) >> -sibShift[j];
			  }
			  printf("  i=%6x  vec=%6x \n", i, vec);
			}
#endif


			/* Define founders and members of subPedigree */
			nSpMem=0;
			cur_place=originals*2;
			spFirstMember=real_non_originals;
			for(iFmem=0; iFmem < nSibs_inf+6; iFmem++) {
				if(iFmem==0) {
					j=famtree[famnum][dad].dad_index;
				} else if(iFmem==1) {
					j=famtree[famnum][dad].mom_index;
				} else if(iFmem==2) {
					j=famtree[famnum][mom].dad_index;
				} else if(iFmem==3) {
					j=famtree[famnum][mom].mom_index;
				} else if(iFmem==4) {
					j=dad;
				} else if(iFmem==5) {
					j=mom;
				} else {
					iSib = iFmem - 6;
					j=arrSibs_inf[iSib];
				}
				
				if(j==MISSING) continue;
				spMemID[nSpMem++] = j;
				
				/* check if grandparents whole pedigree founders.
				   Error due to single parent individual already checked */
				if(iFmem<4){
					if(famtree[famnum][j].dad_index != MISSING) {
						famtree[famnum][j].place_allele[0]=cur_place++;
						famtree[famnum][j].place_allele[1]=cur_place++;
					}
				}
				
				/* find the firsd individual that needs placeholder alleles
				   set by inheritance pattern.  dad, mom, or sibs. */
				if(iFmem>3) {
					for(i=0; i<real_non_originals; i++) {
						if((j==sorted_ped[famnum][i]) && (i<spFirstMember)) 
							spFirstMember=i;
					}
				}
			}

#ifdef PR_18
			printf("spFirstMember=%2d nSpMem=%d : ", spFirstMember, nSpMem);
			for(iSpMem=0; iSpMem<nSpMem; iSpMem++) {
			  j=spMemID[iSpMem];
			  printf("%s ", famtree[famnum][ j ].indiv_ID);
			}
			printf("  vec0=%8x", vec0);
			printf("\n");
#endif

			
			/* Check all possible inheritance vectors in nuclear family */
			for(i=0; i<nSibs_inf_vecs_all; i++) {
				vec = vec0;
				for(j=0; j<nSibMS; j++){ 
					if(sibShift[j]>0) vec |= (i & sibMask[j]) <<  sibShift[j];
					else              vec |= (i & sibMask[j]) >> -sibShift[j];
				}
				fill_placeholder_alleles_2(0, spFirstMember, 
				                           vec, real_non_originals,0);
				create_nodes(cur_place);
				num_edges=0;
				for(iSpMem=0; iSpMem<nSpMem; iSpMem++) {
					j=spMemID[iSpMem];
					
					if (allele_data[j][allele_offset+(2*marker)] == 0 &&
					allele_data[j][allele_offset+(2*marker)+1] == 0) continue;
					
					AddEdge( famtree[famnum][j].place_allele[0],
					         famtree[famnum][j].place_allele[1],
					         allele_data[j][allele_offset+(2*marker)],
					         allele_data[j][allele_offset+(2*marker)+1] );
					num_edges++;
				}

				if (num_edges > 0) { 
					single (assign_list, ASSIGN_STRUCT);
					assign_list->next=NULL; assign_list->num_alleles=0;
					process_graph(assign_list);
					alist=assign_list;
					
					if (alist->num_alleles != 0) {
						legalVec[i]=1;
						for(ibit=0; ibit<nSibs_inf_bits_all; ibit++) {
							if(m_pow2[ibit] & i) sum_s1[ibit]++;
							else               sum_s0[ibit]++;
						}
					} 
					free_assign(assign_list);
				} else {
					/* all individuals are missing, all patterns legal */
					legalVec[i]=1;
					for(ibit=0; ibit<nSibs_inf_bits_all; ibit++) {
						if(m_pow2[ibit] & i) sum_s1[ibit]++;
						else               sum_s0[ibit]++;
					}
				}
				
				/*-------------- save in emi array */
				/* Save result in emi[marker][ibit_r][1/0] */  

				if(i==(nSibs_inf_vecs_wp-1)) {
					if((sum_s1[0] + sum_s0[0]) == 0) {
						/* no solution for this phase fix */
						for(j=0; j<nbit_pos; j++) {
							ibit_s = bit_pos[B_POS_OFFS*j + 1];
							ibit_r = bit_pos[B_POS_OFFS*j + 2];

							/* set only if not set already (for parents)*/
							if( (emi1[marker][ibit_r]>1) && 
							    (ibit_s<nSibs_inf_bits)   ) {
								emi1[marker][ibit_r] = 8;
							}
						}
					} else {
						for(j=0; j<nbit_pos; j++) {
							ibit_s = bit_pos[B_POS_OFFS*j + 1];
							ibit_r = bit_pos[B_POS_OFFS*j + 2];

							if(bit_pos[B_POS_OFFS*j + 3] == -1) {
						       /* this marker is ungenotyped, meiosis cannot be constrained*/
								i1=2;
							} else {
								if(     sum_s1[ibit_s] == 0) { i1=0; }
								else if(sum_s0[ibit_s] == 0) { i1=1; }
								else                           { i1=2; }
							}
							/* set only if not set already (for parents)*/
							if( (emi1[marker][ibit_r]>1) && 
							    (ibit_s<nSibs_inf_bits_wp)   ) {
								emi1[marker][ibit_r] = i1;
							}
						}
					}
				} 
				if(i==(nSibs_inf_vecs_all-1)) {
					/* In case no informative bits exist (two founders with a single
					child) the single legal pattern has 
					(sum_s1[0] + sum_s0[0]) == 0) but  (legalVec[1] == 1) */
					if( ((sum_s1[0] + sum_s0[0]) == 0) && (legalVec[0] == 0) ) {
/*						printf(" ******Error Inconsistent genotypes\n");
						printf(" family of %s  %s\n", famtree[famnum][dad].indiv_ID, 
	                                        	famtree[famnum][mom].indiv_ID);
						printf(" marker=%d\t nSibs_inf=%d\t arrSibs_inf=", 
						marker, nSibs_inf);
						for(iSib=0; iSib<nSibs_inf; iSib++) printf("%s  ", 
						famtree[famnum][ arrSibs_inf[iSib] ].indiv_ID);
						printf("  sib_inf_mask=%x\n", sib_inf_mask);
						printf(" *************************************\n");
*/
						if (!pmod_calc_flag) {
						  sf(ps,"ERROR(FIB): bad inheritance in pedigree %s, marker %s\n", 
						    current_ped, locus[marker].name); 
						  pr();
						  sf(ps,"       family of %s  %s\n", famtree[famnum][dad].indiv_ID,
						    famtree[famnum][mom].indiv_ID); 
						  pr();
						  sf(ps,"SKIPPING THIS PEDIGREE\n"); 
						  pr();
						} else if (replicates_store_result) {
						  sf(ps,"ERROR(FIB): bad inheritance in pedigree %s, marker %s\n", 
						    current_ped, locus[marker].name); 
						  fpr(replifp);
						  sf(ps,"       family of %s  %s\n", famtree[famnum][dad].indiv_ID, 
						    famtree[famnum][mom].indiv_ID); 
						  fpr(replifp);
						  sf(ps,"SKIPPING THIS PEDIGREE\n"); 
						  fpr(replifp);
						}	

						inconsistent_genotypes[iMarker] = TRUE; 
						/* since this is the last itteration (last in list of 
						   possible inheritance vectors) we do not need 
						   a break statement here*/
						/* exit (1); */
					} else {
						for(j=0; j<nbit_pos; j++) {
							ibit_s = bit_pos[B_POS_OFFS*j + 1];
							ibit_r = bit_pos[B_POS_OFFS*j + 2];

							if(bit_pos[B_POS_OFFS*j + 3] == -1) {
						       /* this marker is ungenotyped, meiosis cannot be constrained*/
								i1=2;
							} else {
								if(     sum_s1[ibit_s] == 0) { i1=0; }
								else if(sum_s0[ibit_s] == 0) { i1=1; }
								else                         { i1=2; }
							}
							
							/* set only if not set already (for parents)*/
							if(emi0[marker][ibit_r]>1) {
								emi0[marker][ibit_r] = i1;
							}
						}
					}
				}
				/*-------------- END save in emi array */
				
			}
			/* END Check all possible inheritance vectors in nuclear family */
			
#ifdef PR_06
			printf("                  sum 0-1(ibit) ");
			for(ibit=nSibs_inf_bits-1; ibit>=0; ibit--) {
			  printf("%d-%d(%d)  ", 
						sum_s0[ibit], sum_s1[ibit], ibit);
			}
			printf("\n");
#endif

			unarray (legalVec, int);
		} /*-------END execute per marker-------------------------*/
	} /* ---------------------------------END Execute per nuclear family */
	
#if defined(PR_12) || defined(PR_13)
		printf("\n");
		for(int ipass=1; ipass>=0; ipass--) {
            int bitcnt_fix, bitcnt_use;
            int emi_to_check;
            int ibmask;
			if((ipass == 0) && !PR_12) continue;
			if((ipass == 1) && !PR_13) continue;
			printf("                          non-originals:     ");
			for(i=0; i<real_non_originals; i++) {
				printf("  %2s", famtree[famnum][sorted_ped[famnum][i]].indiv_ID);
			}
			printf("\n");
			double total   = 0.0;
			int jMarker = 0;
			for(iMarker=0; iMarker<f_num_in_map_order; iMarker++) {
				marker = f_map_order[iMarker];
				double map_position = total;
				while(map_distance[jMarker] == 0.) jMarker++;
				total += map_distance[jMarker];
				jMarker++;
				printf("                  e[%1d]  iM=%3d  m=%3d %7.2lf", 
				ipass, iMarker, marker, map_position);
				ibit_r  = real_num_bits-1;
				bitcnt_use = bitcnt_fix = 0;

				if (ipass==0)
				  emi_to_check=emi0[marker][ibit_r];
				else
				  emi_to_check=emi1[marker][ibit_r];

				for(ibit=num_bits-1; ibit>=0; ibit--) {
					ibmask=m_pow2[ibit];
					if(ibit & 1) printf("  ");
					if((ibmask & uninformative_mask) != 0) {
						printf("_");
					}
					else if(emi_to_check==2) {
						printf("x");
						bitcnt_use++;
					}
					else {
					  printf("%1d", emi_to_check);
						bitcnt_use++;
						bitcnt_fix++;
					}
					if((ibmask & uninformative_mask) == 0) ibit_r--;
				}
				printf("  (%2d/%2d)",bitcnt_fix,bitcnt_use);
				if(f_uninformative[iMarker]) printf("*");
				printf("\n");
			}
			printf("\n");
		}
#else
		(void) f_uninformative;	/* disable unused parameter warning */
#endif


	/* make a mask/test per marker */
	for(iMarker=0; iMarker<f_num_in_map_order; iMarker++) {
		marker = f_map_order[iMarker];
/*		printf("marker = f_map_order[%2d] =%d\n", iMarker, marker); */
		
		/* if no possible solution just set save_maskTest[][][0:1] = 1,0 */
		if(inconsistent_genotypes[iMarker]) {
			save_maskTest[famnum][iMarker][0] = 1;
			save_maskTest[famnum][iMarker][1] = 0;
		} else {
			save_maskTest[famnum][iMarker][0] = save_maskTest[famnum][iMarker][1] = 0;
			for(ibit=0; ibit<real_num_bits; ibit++) {
				j = emi0[marker][ibit];
				if((j==0) || (j==1)) save_maskTest[famnum][iMarker][0] |= m_pow2[ibit];
				if(j==1)             save_maskTest[famnum][iMarker][1] |= m_pow2[ibit];
			}

/* Debugging only:
			printf("                  m=%3d,  maskTest %8x\n", 
			iMarker, save_maskTest[famnum][iMarker][0]);
			printf("                          maskTest %8x\n", 
			save_maskTest[famnum][iMarker][1]);
*/

		}

	}

#ifdef PR_00
	printf("----------- FindIllegalBits END---------\n");
#endif

	if (emi0)
	  unmatrix(emi0, MAXLOCI, int);
	if (emi1)
	  unmatrix(emi1, MAXLOCI, int);

	return 0;
}

void print_requirements(int num_in_map_order, int num_bits, int real_num_bits, int score_mask_rdc,
                        int amount_to_alloc, int score_alloc, int *nz_bits,
						int f_num_in_map_order,
		                int *f_uninformative)
{
	int           i, iMarker, nuninformative, max_bitc, max_bitc_n, reduced_num_markers;
	int           print_mode;
	double        num_elements, memsize;
	double        d_num_elements[200], d_memsize[200];
	double        d_num_in_map_order, d_vectors, d_amount_to_alloc, 
	              d_score_alloc, d_one_meg,
				  disease_priors_0, transition_tmp_0,
				  disease_priors_1, transition_tmp_1;

	/*  print_mode = 0  Print details
	                 1  Print summary
	*/
	print_mode = 1;
	
	d_num_in_map_order = (double)num_in_map_order;
	d_vectors          = (double)m_pow2[real_num_bits];
	if(amount_to_alloc < 0) {  /* overflow */
		num_elements = 0;
		for (i=0; i<f_num_in_map_order; i++) {
			num_elements += (double)m_pow2[nz_bits[i]];
		}
		d_amount_to_alloc  = num_elements;
	} else {
		d_amount_to_alloc  = (double)amount_to_alloc;
	}
	d_score_alloc      = (double)score_alloc;
	d_one_meg          = (double)m_pow2[20];
	nuninformative     = 0;
	max_bitc = 0;
    max_bitc_n = 0;
	for(iMarker=0; iMarker<f_num_in_map_order; iMarker++) {
		if(f_uninformative[iMarker])     nuninformative++;
		if(max_bitc < nz_bits[iMarker])  {
			max_bitc = nz_bits[iMarker];
			max_bitc_n = 0;  
			if(iMarker>0) max_bitc_n = nz_bits[iMarker-1];
			if(iMarker<(f_num_in_map_order-1)) {
				if(max_bitc_n<nz_bits[iMarker+1]) max_bitc_n = nz_bits[iMarker+1];
			}
		}
	}
/*	printf(" max_bitc, max_bitc_n = %3d, %3d \n", max_bitc,max_bitc_n);
*/	
	disease_priors_0 = 2 * d_vectors;
	disease_priors_1 = d_score_alloc;
	if(analysis_type != NPL_ANALYSIS) 
		disease_priors_1 += d_vectors;
	
	transition_tmp_0 = 2.5 * d_vectors;
	if((scan_steps != 1) || (nuninformative > 0)) {
		transition_tmp_1  = 1.5 * (double)m_pow2[max_bitc];
		transition_tmp_1 += 1.5 * (double)m_pow2[max_bitc_n];
	} else {
		transition_tmp_1 = 1.5 * (double)m_pow2[max_bitc];
	}
	
	reduced_num_markers = f_num_in_map_order - nuninformative;

	/*  0- 99 --> previous allocation
	  100-199 --> current allocation */
	
	/* 0, 100 p+l+r storage*/	
	d_num_elements[  0] = 3 * d_num_in_map_order * d_vectors;
	d_num_elements[100] = 3 * d_amount_to_alloc;
	
	/* 90, 190 all exponential*/
	d_num_elements[ 90] = 3 * d_num_in_map_order * d_vectors
	                    + disease_priors_0
	                    + transition_tmp_0;
	d_num_elements[190] = 3 * d_amount_to_alloc
	                    + disease_priors_1
						+ transition_tmp_1;

	/* 91, 191 p+l+r using reduced map.  */
	d_num_elements[ 91] = 3 * (double)reduced_num_markers * d_vectors;
	d_num_elements[191] = d_num_elements[100];

	/* 92, 192 all using reduced map.  assume that positions with uninf. ignored*/
	d_num_elements[ 92] = 3 * (double)reduced_num_markers * d_vectors
	                    + disease_priors_0
	                    + transition_tmp_0;
	d_num_elements[192] = 3 * d_amount_to_alloc
	                    + disease_priors_1
						+ 1.5 * (double)m_pow2[max_bitc];
	
	/* 93, 193 all using reduced map.  assume that positions with uninf. USED*/
	d_num_elements[ 93] = 3 * (double)f_num_in_map_order * d_vectors
	                    + disease_priors_0
	                    + transition_tmp_0;
	d_num_elements[193] = d_num_elements[190];
	
	for(i=0; i<200; i++) d_memsize[i] = (d_num_elements[i]*8)/d_one_meg ;

	/* COMPUTE_SIZE_ONLY, flag to interrupt calculation */
/*	if((real_num_bits < 15) || (d_memsize[192] > 4096))return;
*/

	
	if(print_mode==0) {
	  sf(ps,
		"num_bits, real_num_bits, score_mask_rdc, reduced_num_markers = %2d %2d %8x %d\n", 
		 num_bits, real_num_bits, score_mask_rdc, reduced_num_markers);
		pr();
		sf(ps, "N-->%9.3f          p+l+r       all    p+l+r-rmap  all-rmap  all-rmap2\n", (d_vectors*8)/d_one_meg);
		pr();
		sf(ps,"mem.(MBytes) v2.0 = %9.3f  %9.3f  %9.3f  %9.3f  %9.3f\n", 
		d_memsize[  0], d_memsize[ 90], d_memsize[ 91], d_memsize[ 92], d_memsize[ 93]);
		pr();
		sf(ps,"mem.(MBytes) v2.1 = %9.3f  %9.3f  %9.3f  %9.3f  %9.3f\n", 
		d_memsize[100], d_memsize[190], d_memsize[191], d_memsize[192], d_memsize[193]);
		pr();
	}
	if(print_mode==1) {
		sf(ps,"Ped. size =(2n-f)=(%3d -%3d) =%3d bits\n", num_bits,
		num_bits - real_num_bits, real_num_bits); pr();
		sf(ps," Memory requirements for version 2.1 (2.0) = %10.3f (%10.3f) Mbytes \n",
		d_memsize[190],d_memsize[90]);
		pr();
		sf(ps,"Memory requirements per Marker (Mbytes):\n"); pr();
		for (i=0; i<f_num_in_map_order; i++) {
			num_elements = m_pow2[nz_bits[i]];
			memsize = (num_elements*3*8)/d_one_meg;
			sf(ps," %3d %10.3f", i, memsize); pr();
			if(f_uninformative[i]){
				sf(ps,"  *\n"); pr();
			} else {
				sf(ps,"\n"); pr();
			}
		}
        if(nuninformative != 0) {
            if (g_association) {
                sf(ps,"Markers followed by a * do not contribute to the inheritance vector distribution. \n");
                pr();
            } else {
                sf(ps,"Markers followed by a * are removed because they are uninformative. \n");
                pr();
            }
        }
	}

}


void algebra_matrix(int originals, int *original_list, int non_originals, int *non_original_list,
PED_MEMBER *this_famtree, int new_vec, double *this_untotal_datfile_par_score)
{
    unsigned long or_vec, pow4_originals, use_or_vec;
    pow4_originals = 1 << 2*originals;
    int node_assign[MAX_INDIVIDUALS*2];
    int i, j, k, indiv, a0, a1, this_dis_loc, n1, n2, affect_stat, found_in_unique, alg_line, coeff, found_in_pointer;
    int or_vec_bit_0[MAX_INDIVIDUALS], or_vec_bit_1[MAX_INDIVIDUALS];
    unsigned int alg_pen[10], alg_afreq[2];
    int num_terms, match_class;
    int match_pointer = 0;
    int *private_unique_terms;
    int *new_vec_pointer, *pointer_to_compare;
    double par_prob;
    unsigned int *coeff_storage, temp_power;
    unsigned int *temp_algebra;
    unsigned int old_mem_limit2;
    int *this_new_vec_to_alg_prob;
    array(coeff_storage, pow4_originals, unsigned int); /* private, intermediate coefficient storage */
    array(private_unique_terms, pow4_originals, int); /* private, intermediate power storage */
    array(new_vec_pointer, pow4_originals, int); /* private, intermediate pointer storage; needed to compare with global address_array to check if powers-coefficient combination has already been found */  
    par_prob = 0.0;
    this_new_vec_to_alg_prob = new_vec_to_alg_prob[algebra_pointer[famnum]];

    for (i=0; i<originals; i++) {
        indiv = original_list[i];
	n1 = this_famtree[original_list[i]].place_allele[0];
	n2 = this_famtree[original_list[i]].place_allele[1];
	node_assign[n1] = 2*i;
	node_assign[n2] = 2*i + 1;
    }
    for (i=0; i<non_originals; i++) {
	indiv = non_original_list[i];
	or_vec_bit_0[indiv]
	    = node_assign[this_famtree[indiv].place_allele[0]];
	or_vec_bit_1[indiv]
	    = node_assign[this_famtree[indiv].place_allele[1]];
    }
    num_terms=0;
    for (or_vec=0; or_vec < pow4_originals; or_vec++) { /* loops over or_vecs, i.e., founder disease genotypes*/
        for (i=0; i < 10; i++){
            alg_pen[i] = 0;
        }/* set to 0 for next or_vec */ 
        for(i=0; i < 2; i++){
            alg_afreq[i] = 0; /* set to 0 for next or_vec */
        }

        if (mobit_simulate) {
            for (int i = 0; i < number_of_permutation_units[algebra_pointer[famnum]]; i++) {
                for (int j = 0; j < number_of_members_in_permutation_unit[algebra_pointer[famnum]][i]; j++) {
                    indiv = permutation_units[algebra_pointer[famnum]][i][j];
                    affect_stat = this_famtree[indiv].affectation_status;
                    a0 = (or_vec >> or_vec_bit_0[indiv]) & 1;
                    a1 = (or_vec >> or_vec_bit_1[indiv]) & 1;
                    if (permutation_unit_flipped[algebra_pointer[famnum]][i]) {
                        this_dis_loc = ((a0<<1)|a1);
                    } else {
                        this_dis_loc = ((a1<<1)|a0);
                    }
                    if (affect_stat) alg_pen[((affect_stat & 2) << 1)|this_dis_loc]++;
                }
            }
        } else {
            for (i=0; i < non_originals; i++) { /* loop over non-originals first, extract disease genotype and save corresponding penetrance */
                indiv = non_original_list[i];
                affect_stat = this_famtree[indiv].affectation_status;
                a0 = (or_vec >> or_vec_bit_0[indiv]) & 1;
                a1 = (or_vec >> or_vec_bit_1[indiv]) & 1;
                this_dis_loc = ((a1<<1)|a0);
                if (affect_stat) alg_pen[((affect_stat & 2) << 1)|this_dis_loc]++;
            }
        }

        for (i=0; i < originals; i++) { /* loop over originals, look up disease genotype in or_vec and save corresponding penetrance */
            indiv = original_list[i];
            affect_stat = this_famtree[indiv].affectation_status;	
            this_dis_loc = (or_vec >> (2*i)) & 3;
            if (this_dis_loc == 1 || this_dis_loc == 2){ /* needed since origin of founder chromosomes unknown, therefore the heterozygote penetrances must be averaged */
                if (affect_stat == 1) alg_pen[8]++;
                if (affect_stat == 2) alg_pen[9]++;
            }
            else if (affect_stat){
                alg_pen[((affect_stat & 2) << 1)|this_dis_loc]++;
            }
        }

        for(i=0; i < 2*originals; i++){ /* get allele_freq from originals */
            use_or_vec = (or_vec >> i) & 1;
            alg_afreq[use_or_vec]++;
        }
        found_in_unique = 0;
        for (i=0; i <= unique_alg_terms; i++){ /* loop over lines in algebra_array already filled to check for same power combination */
            if (found_in_unique == 1){
                break;
            }
            if (i == mem_limit1){ /* increase memory allocation if necessary */
                mem_limit1 += mem_increment1;
                array_realloc(pow_prep_vec, mem_limit1, double);
                matrix_realloc(algebra_array, i, mem_limit1, 12, unsigned int);
                matrix_realloc(alg_coeff_combi_check, i, mem_limit1, mem_limit2, unsigned int);
                for(j=i; j < mem_limit1; j++){
                    for(k=0; k < mem_limit2; k++){
                        alg_coeff_combi_check[j][k]=0;
                    }
                }
                matrix_realloc(address_array, i, mem_limit1, mem_limit2, int);
            }
            if (i < unique_alg_terms){
                temp_algebra = algebra_array[i];
                for(j=0; j < 10; j++){
                    if (temp_algebra[j] != alg_pen[j]) break;
                }
                if (j == 10){
                    for(j=10; j < 12; j++){ 
                        if(temp_algebra[j] != alg_afreq[j-10]) break;
                    }
                }
                if(j == 12){
                    for (k=0; k <= num_terms; k++){ /* is combination hit for the 1st time for this new_vec? */
                        if (k < num_terms){
                            if (private_unique_terms[k] == i){ /* no, it has already occured, so that only coefficient must be updated */
                                coeff_storage[k]++;
                                found_in_unique = 1;
                                break;
                            }
                        }
                        if (k == num_terms){ /* it is a new combination for this new_vec, so coefficient must be initialized */
                            private_unique_terms[k] = i; /* line of algebra_array corresponding to power combination found for this new_vec is stored */
                            coeff_storage[k] = 1;
                            /* current_pointer[k] = &(pow_prep_vec[i]); */
                            num_terms++;
                            found_in_unique = 1;
                            break;
                        }
                    }
                }
            }
            if (i == unique_alg_terms){ /* combination is unique over all new_vecs investigated so far */

                j=0;
                while(j < 10){
                    algebra_array[i][j]=alg_pen[j]; /* fill in penetrance powers */
                    j++;
                }
                while(j < 12){
                    algebra_array[i][j]=alg_afreq[j-10]; /* same for disease allele frequency */
                    j++;
                }

                pow_prep_vec[i] = 1.0;
                temp_algebra = algebra_array[i];

                temp_power = temp_algebra[0];
                if (temp_power){
                    pow_prep_vec[i]*=pen_0[temp_power];
                }
                temp_power = temp_algebra[1];
                if (temp_power){
                    pow_prep_vec[i]*=pen_1[temp_power];
                }
                temp_power = temp_algebra[2];
                if (temp_power){
                    pow_prep_vec[i]*=pen_2[temp_power];
                }
                temp_power = temp_algebra[3];
                if (temp_power){
                    pow_prep_vec[i]*=pen_3[temp_power];
                }
                temp_power = temp_algebra[4];
                if (temp_power){
                    pow_prep_vec[i]*=pen_4[temp_power];
                }
                temp_power = temp_algebra[5];
                if (temp_power){
                    pow_prep_vec[i]*=pen_5[temp_power];
                }
                temp_power = temp_algebra[6];
                if (temp_power){
                    pow_prep_vec[i]*=pen_6[temp_power];
                }
                temp_power = temp_algebra[7];
                if (temp_power){
                    pow_prep_vec[i]*=pen_7[temp_power];
                }
                temp_power = temp_algebra[8];
                if (temp_power){
                    pow_prep_vec[i]*=pen_8[temp_power];
                }
                temp_power = temp_algebra[9];
                if (temp_power){
                    pow_prep_vec[i]*=pen_9[temp_power];
                }
                temp_power = temp_algebra[10];
                if (temp_power){
                    pow_prep_vec[i]*=alg_afreq_1[temp_power];
                }
                temp_power = temp_algebra[11];
                if (temp_power){
                    pow_prep_vec[i]*=alg_afreq_2[temp_power];
                }
                coeff_storage[num_terms] = 1; /* coefficient must be initialized */
                private_unique_terms[num_terms] = i; /* new entry with line number in algebra_array is added */
                num_terms++;
                unique_alg_terms++; 
                break;
            }
        }
    }
   
    if(num_terms > mem_limit5)
    {
	    while(num_terms > mem_limit5){
	    mem_limit5 += mem_increment5;
	    }
	    total_realloc(vec_class_pointer, mem_limit4, mem_limit4, mem_limit5, int);
    } /* number of unique combis for each vec_class is max. 2^2f; for 8 founders: 65536 combinations */
    
    match_class=0;
    for(i=0; i < num_terms; i++){ /* check whether the combination of powers and coefficients is new; if yes, then store combination in two arrays: one for the powers, one for the coefficients; this gives the link between the position in the pow_coeff_prep_vec and the underlying combination for the algebra_calculated function; if no, then point to corresponding address in pow_coeff_prep_vec */  
      alg_line = private_unique_terms[i];
      coeff = coeff_storage[i];

      if (coeff > mem_limit2)
      { /* increase memory allocation if necessary */
       	old_mem_limit2 = mem_limit2;
	while(coeff > mem_limit2){
	mem_limit2 += mem_increment2;
	}
	total_realloc(alg_coeff_combi_check, mem_limit1, mem_limit1, mem_limit2, unsigned int);
		     for(j=0; j < mem_limit1; j++){
			     for(k=old_mem_limit2; k < mem_limit2; k++){
				     alg_coeff_combi_check[j][k]=0;
			     }
		     }
	total_realloc(address_array, mem_limit1, mem_limit1, mem_limit2, int);
      }

      if (alg_coeff_combi_check[alg_line][coeff])
      {
	match_class++;
	new_vec_pointer[i] = address_array[alg_line][coeff];
      }
      
      if (!alg_coeff_combi_check[alg_line][coeff])
      {
	alg_coeff_combi_check[alg_line][coeff] = 1;
	if(unique_combis == mem_limit3)
	{
		mem_limit3 += mem_increment3;
		array_realloc(unique_terms_line, mem_limit3, int);
		array_realloc(unique_coeff, mem_limit3, int);
	}
	if(unique_combis == mem_limit7)
	{
        mem_limit7 += mem_increment7;
        array_realloc(pow_coeff_prep_vec, mem_limit7, double);
        array_realloc(alg_prob_vec, mem_limit7, double);
	}
	unique_terms_line[unique_combis] = alg_line;
	unique_coeff[unique_combis] = coeff;
	new_vec_pointer[i] = unique_combis;
	address_array[alg_line][coeff] = unique_combis;
	pow_coeff_prep_vec[unique_combis] = coeff*pow_prep_vec[alg_line];
	unique_combis++;
      }
    }
    
    if (match_class < num_terms)
    {
      vec_classes++;
	if(vec_classes == mem_limit4)
	{
		mem_limit4 += mem_increment4;
		array_realloc(vec_class_members, mem_limit4, int);
		matrix_realloc(vec_class_pointer, mem_limit4 - mem_increment4, mem_limit4, mem_limit5, int);
		array_realloc(num_fam_in_vec_class, mem_limit4, int);
		array_realloc(par_score_address_vec, mem_limit4, double*);
		array_realloc(list_num_terms, mem_limit4, int);
		total_realloc(family_to_vec_class, peds_pmod, mem_limit4, mem_limit4, int);
		for(i=0; i < peds_pmod; i++) {
			for(j = mem_limit4 - mem_increment4; j < mem_limit4; j++) {
				family_to_vec_class[i][j] = 0;
			}
		}
	}
      vec_class_members[vec_classes] = 1;
      if(!family_to_vec_class[algebra_pointer[famnum]][vec_classes]){ /* not needed since new vec_class is discovered and is therefore also new in family */ 
	      family_to_vec_class[algebra_pointer[famnum]][vec_classes] = 1;
	      num_classes_in_family[algebra_pointer[famnum]]++;
      }
      num_fam_in_vec_class[vec_classes] = 1;
      
      for(i=0; i < num_terms; i++){
      vec_class_pointer[vec_classes][i] = new_vec_pointer[i];
      par_prob += *(pow_coeff_prep_vec + new_vec_pointer[i]);
      }
      list_num_terms[vec_classes] = num_terms;
      this_untotal_datfile_par_score[new_vec] = par_prob;
      par_score_address_vec[vec_classes] = &(this_untotal_datfile_par_score[new_vec]);
      this_new_vec_to_alg_prob[new_vec] = vec_classes;
    }
    
    if (match_class == num_terms)
    {
      for(i=0; i <= vec_classes; i++){  /* current number of vector classes in dataset starting with 0 */
	match_pointer=0;
	pointer_to_compare = vec_class_pointer[i];
	if(num_terms == list_num_terms[i]) { /* fast pre-check if num_terms is the same */
	  for(j=0; j < num_terms; j++){
	    for(k=0; k < num_terms; k++){
	      found_in_pointer = 0;
	      if(new_vec_pointer[j] == pointer_to_compare[k]){
		match_pointer++;
		found_in_pointer = 1;
		break; /* go to next term of new_vec and compare with vector class terms */
		}
	    }
	      if (k == (num_terms - 1) && !found_in_pointer){
		  break;
		}
	      }
	  }
	if (match_pointer == num_terms) { 
	  vec_class_members[i]++; /* count up after storing because storing starts with 0, counting with 1 */
	  this_untotal_datfile_par_score[new_vec] = *(par_score_address_vec[i]);
	  this_new_vec_to_alg_prob[new_vec] = i;
	  if(!family_to_vec_class[algebra_pointer[famnum]][i]){ /* is vec_class already in family? */ 
		  family_to_vec_class[algebra_pointer[famnum]][i] = 1;
		  num_classes_in_family[algebra_pointer[famnum]]++;
		  num_fam_in_vec_class[i]++;
	  }

	  break;
	  }
	}
	if (match_pointer != num_terms) {
	  vec_classes++;
	if(vec_classes == mem_limit4)
	{
		mem_limit4 += mem_increment4;
		array_realloc(vec_class_members, mem_limit4, int);
		matrix_realloc(vec_class_pointer, mem_limit4 - mem_increment4, mem_limit4, mem_limit5, int);
		array_realloc(num_fam_in_vec_class, mem_limit4, int);
		array_realloc(par_score_address_vec, mem_limit4, double*);
		array_realloc(list_num_terms, mem_limit4, int);
		total_realloc(family_to_vec_class, peds_pmod, mem_limit4, mem_limit4, int);
		for(i=0; i < peds_pmod; i++) {
			for(j = mem_limit4 - mem_increment4; j < mem_limit4; j++) {
				family_to_vec_class[i][j] = 0;
			}
		}
	}
	  vec_class_members[vec_classes] = 1;
	  num_fam_in_vec_class[vec_classes] = 1;
	  for(i=0; i < num_terms; i++){
	    vec_class_pointer[vec_classes][i] = new_vec_pointer[i];
        par_prob += *(pow_coeff_prep_vec + new_vec_pointer[i]);
	  }
	  this_untotal_datfile_par_score[new_vec] = par_prob;
          par_score_address_vec[vec_classes] = &(this_untotal_datfile_par_score[new_vec]);
          this_new_vec_to_alg_prob[new_vec] = vec_classes;
	  list_num_terms[vec_classes] = num_terms;
	  if(!family_to_vec_class[algebra_pointer[famnum]][vec_classes]){ /* is vec_class already in family? */ 
		  family_to_vec_class[algebra_pointer[famnum]][vec_classes] = 1;
		  num_classes_in_family[algebra_pointer[famnum]]++;
	  }
	}
    }
    
    unarray(private_unique_terms, int);
    unarray(coeff_storage, int);
    unarray(new_vec_pointer, int);
}

void algebra_power_prep(void)
{
  int i;
    pen_0[1] = 1 - penetrance[modliab][HOM_UNAFF];
    pen_1[1] = 1 - penetrance[modliab][HET_AU];
    pen_2[1] = 1 - penetrance[modliab][HET_UA];
    pen_3[1] = 1 - penetrance[modliab][HOM_AFF];
    pen_4[1] = penetrance[modliab][HOM_UNAFF];
    pen_5[1] = penetrance[modliab][HET_AU];
    pen_6[1] = penetrance[modliab][HET_UA];
    pen_7[1] = penetrance[modliab][HOM_AFF];
    pen_8[1] = 1 - (penetrance[modliab][HET_AU] + penetrance[modliab][HET_UA])/2;
    pen_9[1] = (penetrance[modliab][HET_AU] + penetrance[modliab][HET_UA])/2;
    alg_afreq_1[1] = 1 - affected_freq;
    alg_afreq_2[1] = affected_freq;
    
  for (i=2; i < max_indiv; i++){
    
    pen_0[i] = pen_0[1]*pen_0[i-1]; 
    pen_1[i] = pen_1[1]*pen_1[i-1]; 
    pen_2[i] = pen_2[1]*pen_2[i-1]; 
    pen_3[i] = pen_3[1]*pen_3[i-1]; 
    pen_4[i] = pen_4[1]*pen_4[i-1]; 
    pen_5[i] = pen_5[1]*pen_5[i-1]; 
    pen_6[i] = pen_6[1]*pen_6[i-1]; 
    pen_7[i] = pen_7[1]*pen_7[i-1]; 
    pen_8[i] = pen_8[1]*pen_8[i-1]; 
    pen_9[i] = pen_9[1]*pen_9[i-1]; 
    alg_afreq_1[i] = alg_afreq_1[1]*alg_afreq_1[i-1]; 
    alg_afreq_2[i] = alg_afreq_2[1]*alg_afreq_2[i-1];
    }
}


void algebra_calculated(void)
{
  unsigned int current_power;
  int i, alg_power, coeff;
  unsigned int *current_algebra;
  
  for (i=0; i < unique_alg_terms; i++){
       
       pow_prep_vec[i] = 1.0;
       current_algebra = algebra_array[i];
     
       current_power = current_algebra[0];
       if (current_power){
	 pow_prep_vec[i]*=pen_0[current_power];
       }
       current_power = current_algebra[1];
       if (current_power){
	 pow_prep_vec[i]*=pen_1[current_power];
       }
       current_power = current_algebra[2];
       if (current_power){
	 pow_prep_vec[i]*=pen_2[current_power];
       }
       current_power = current_algebra[3];
       if (current_power){
	 pow_prep_vec[i]*=pen_3[current_power];
       }
       current_power = current_algebra[4];
       if (current_power){
	 pow_prep_vec[i]*=pen_4[current_power];
       }
       current_power = current_algebra[5];
       if (current_power){
	 pow_prep_vec[i]*=pen_5[current_power];
       }
       current_power = current_algebra[6];
       if (current_power){
	 pow_prep_vec[i]*=pen_6[current_power];
       }
       current_power = current_algebra[7];
       if (current_power){
	 pow_prep_vec[i]*=pen_7[current_power];
       }
       current_power = current_algebra[8];
       if (current_power){
	 pow_prep_vec[i]*=pen_8[current_power];
       }
       current_power = current_algebra[9];
       if (current_power){
	 pow_prep_vec[i]*=pen_9[current_power];
       }
       current_power = current_algebra[10];
       if (current_power){
	 pow_prep_vec[i]*=alg_afreq_1[current_power];
       }
       current_power = current_algebra[11];
       if (current_power){
	 pow_prep_vec[i]*=alg_afreq_2[current_power];
       }
    }
    for(i=0; i < unique_combis; i++){
      alg_power = unique_terms_line[i];
      coeff = unique_coeff[i];
      pow_coeff_prep_vec[i] = pow_prep_vec[alg_power]*coeff;
    }
}

void algebra_force(int vec_class)
{
  int i;
  int *power_address;
  int num_terms;
  double this_alg_prob_vec;
  
  power_address = vec_class_pointer[vec_class];
  num_terms = list_num_terms[vec_class];
  this_alg_prob_vec = alg_prob_vec[vec_class];
  
  this_alg_prob_vec = 0.0;
  for (i=0; i < num_terms; i++){
    this_alg_prob_vec += *(pow_coeff_prep_vec + *power_address++);
  }
  alg_prob_vec[vec_class] = this_alg_prob_vec;
}

void algebra_cache_opt(double *this_par_score, int this_srnvecs, int *this_new_vec_to_alg_prob)
{
	int i;

		for (i=0; i < this_srnvecs; i++){
			this_par_score[i] = *(alg_prob_vec + *this_new_vec_to_alg_prob++);
		}
}

int algebra_discarded (int ind, int real_num_vecs, int matrix_to_be_done, PED_MEMBER *this_famtree, int famnum)
{
	int i, j, k, l;
	unsigned int old_mem_limit8;
	int found_version, analyzed, num_versions, ped_to_del;
	int *this_version_pointer;
	num_versions = version_counter[famnum_pointer[famnum]]; /* was initialized with 0 */
	this_version_pointer = version_pointer[famnum_pointer[famnum]];
	analyzed = algebra_analyzed[famnum_pointer[famnum]]; /* was initialized with FALSE */
	if (!analyzed) {
		matrix_to_be_done = 1;
		/* version_counter[famnum_pointer[famnum]] = 1; this happens in algebra_matrix anyway */ 
	        /* for (i=0; i < ind; i++) {
			algebra_famtree[famnum_pointer[famnum]][0][i].discard = this_famtree[i].discard;
		} this must happen when family is considered for the first time in analyze_family */
	} else{    
		found_version = 0;
		for (j=0; j <= num_versions; j++){
			if(found_version) break;
			if (j == num_versions){
				peds_pmod++; /* should be initialized with num_pedigrees-1 before any discarding */
				/* Allocate space for another family version. */
				par_score_realloc(par_score, peds_pmod - 1, peds_pmod, double);
				array_realloc(algebra_pointer, peds_pmod, int);
				par_score_realloc(new_vec_to_alg_prob, peds_pmod - 1, peds_pmod, int);
				array_realloc(num_classes_in_family, peds_pmod, int);
				matrix_realloc(family_to_vec_class, peds_pmod - 1, peds_pmod, mem_limit4, int);
				num_classes_in_family[peds_pmod - 1] = 0;
				for (l = 0; l < mem_limit4; l++) {
				    family_to_vec_class[peds_pmod - 1][l] = 0;
				}

				if(peds_pmod == mem_limit7)
				{
                    mem_limit7 += mem_increment7;
                    array_realloc(pow_coeff_prep_vec, mem_limit7, double);
                    array_realloc(alg_prob_vec, mem_limit7, double);
				}

				if (num_versions == mem_limit8){ /* increase memory allocation if necessary */
					old_mem_limit8 = mem_limit8;
					mem_limit8 += mem_increment8;
				        matrix_realloc(algebra_famtree[famnum_pointer[famnum]], old_mem_limit8, mem_limit8, MAX_INDIVIDUALS+5, ALGEBRA_PED_MEMBER);
                                        array_realloc(version_pointer[famnum_pointer[famnum]], mem_limit8, int);
				}
				algebra_pointer[famnum] = peds_pmod - 1;
				/* version_counter[famnum_pointer[famnum]]++; this will be done in algebra_matrix */
				array(par_score[algebra_pointer[famnum]], real_num_vecs, double);
			        array(new_vec_to_alg_prob[algebra_pointer[famnum]], real_num_vecs, int);
				array(datfile_par_score[algebra_pointer[famnum]], real_num_vecs, double);
				array(untotal_datfile_par_score[algebra_pointer[famnum]], real_num_vecs, double);
				matrix_to_be_done = 1;
				for (i=0; i < ind; i++){
					algebra_famtree[famnum_pointer[famnum]][j][i].discard = this_famtree[i].discard;
				}
				ped_to_del = is_currently_in_data[famnum_pointer[famnum]];
				for (k=0; k <= vec_classes; k++){
					if (family_to_vec_class[ped_to_del][k]){
						num_fam_in_vec_class[k]--; /* in order to avoid algebra_force for vector classes that are not in replicate but were in data before */	
					}
				}
			} else{
				for (i=0; i <= ind; i++){
					if (i == ind){
						found_version = 1;
						algebra_pointer[famnum] = this_version_pointer[j];
						if(is_currently_in_data[famnum_pointer[famnum]] != this_version_pointer[j]){
							ped_to_del = is_currently_in_data[famnum_pointer[famnum]];
							for (k=0; k <= vec_classes; k++){
								if (family_to_vec_class[ped_to_del][k]){
									num_fam_in_vec_class[k]--; /* in order to avoid algebra_force for vector classes that are not in replicate but were in data before */	
								}
							}
							is_currently_in_data[famnum_pointer[famnum]] = this_version_pointer[j];	
							ped_to_del = is_currently_in_data[famnum_pointer[famnum]]; /* same as before but now with pedigree that is in dataset */
							for (k=0; k <= vec_classes; k++){
								if (family_to_vec_class[ped_to_del][k]){
									num_fam_in_vec_class[k]++; /* in order to restore class membership */	
								}
							}
						}
					} else{
						if (algebra_famtree[famnum_pointer[famnum]][j][i].discard != this_famtree[i].discard) {
							break;
						}
					}
				}
			}
		}
	}
	return (matrix_to_be_done);
}

void mobit_find_permutation_units(int non_originals, int *non_original_list, PED_MEMBER *this_famtree)
{
    int *parents_index;
    array(parents_index, 2, int);
    int indiv;
    int num_members;
    int found_in_unit;
    for (int i = 0; i < non_originals; i++) {
        parents_index[0] = 0;
        parents_index[1] = 0;
        indiv = non_original_list[i];
        parents_index[0] = this_famtree[indiv].dad_index;
        parents_index[1] = this_famtree[indiv].mom_index;
        found_in_unit = 0;
        for (int j = 0; j < number_of_permutation_units[algebra_pointer[famnum]]; j++) {
            for (int k = 0; k < number_of_parental_indexes_in_unit[algebra_pointer[famnum]][j]; k++) {
                if (parents_index[0] == permutation_units_indexes[algebra_pointer[famnum]][j][k] || parents_index[1] == permutation_units_indexes[algebra_pointer[famnum]][j][k]) {
                    found_in_unit = 1;
                    break;
                }
            }
            if (found_in_unit) {
                num_members = number_of_members_in_permutation_unit[algebra_pointer[famnum]][j];
                permutation_units[algebra_pointer[famnum]][j][num_members] = indiv;
                number_of_members_in_permutation_unit[algebra_pointer[famnum]][j]++;
                int found_index_0 = 0;
                int found_index_1 = 0;
                for (int k = 0; k < number_of_parental_indexes_in_unit[algebra_pointer[famnum]][j]; k++) {
                    if (parents_index[0] == permutation_units_indexes[algebra_pointer[famnum]][j][k]) {
                        found_index_0 = 1;
                    }
                    if (parents_index[1] == permutation_units_indexes[algebra_pointer[famnum]][j][k]) {
                        found_index_1 = 1;
                    }
                }
                if (!found_index_0) {
                    permutation_units_indexes[algebra_pointer[famnum]][j][number_of_parental_indexes_in_unit[algebra_pointer[famnum]][j]] = parents_index[0];
                    number_of_parental_indexes_in_unit[algebra_pointer[famnum]][j]++;
                }
                if (!found_index_1) {
                    permutation_units_indexes[algebra_pointer[famnum]][j][number_of_parental_indexes_in_unit[algebra_pointer[famnum]][j]] = parents_index[1];
                    number_of_parental_indexes_in_unit[algebra_pointer[famnum]][j]++;
                }
                break;
            }
        }
        if (!found_in_unit) {
            permutation_units[algebra_pointer[famnum]][number_of_permutation_units[algebra_pointer[famnum]]][0] = indiv;
            permutation_units_indexes[algebra_pointer[famnum]][number_of_permutation_units[algebra_pointer[famnum]]][0] = parents_index[0];
            permutation_units_indexes[algebra_pointer[famnum]][number_of_permutation_units[algebra_pointer[famnum]]][1] = parents_index[1];
            number_of_members_in_permutation_unit[algebra_pointer[famnum]][number_of_permutation_units[algebra_pointer[famnum]]] = 1;
            number_of_parental_indexes_in_unit[algebra_pointer[famnum]][number_of_permutation_units[algebra_pointer[famnum]]] = 2;
            number_of_permutation_units[algebra_pointer[famnum]]++;
        }
    }
    sf(ps, "\nNumber of permutation units in pedigree %d: %d\n", algebra_pointer[famnum] + 1, number_of_permutation_units[algebra_pointer[famnum]]);
    pr();
    for (int j = 0; j < number_of_permutation_units[algebra_pointer[famnum]]; j++) {
        sf(ps, "\nNumber of members in permutation unit %d: %d", j + 1, number_of_members_in_permutation_unit[algebra_pointer[famnum]][j]);
                pr();
        sf(ps, "\nParental indexes in permutation unit %d:", j + 1);
        pr();
        for (int k = 0; k < number_of_parental_indexes_in_unit[algebra_pointer[famnum]][j]; k++) {
            sf(ps, " %d", permutation_units_indexes[algebra_pointer[famnum]][j][k] + 1);
            pr();
        }
        sf(ps, "\nUnit members in unit %d:", j + 1);
        pr();
        for (int k = 0; k < number_of_members_in_permutation_unit[algebra_pointer[famnum]][j]; k++) {
            sf(ps, " %d", permutation_units[algebra_pointer[famnum]][j][k] + 1);
            pr();
        }
        print("\n");
    }
    print("\n");
    unarray(parents_index, int);
}

void mobit_set_permutation_status(int famnum)
{
    double random_number;
    if (g_random_seed >= 0) {
        init_gen_rand(g_random_seed);
    } else {
        g_random_seed = time(NULL);
        init_gen_rand(g_random_seed);
    }
    for (int i = 0; i < number_of_permutation_units[algebra_pointer[famnum]]; i++) {
        random_number = genrand_open_open();
        sf(ps,"Random number for MOBIT simulations: %.4f for permutation unit %d\n", random_number, i + 1);
        pr();
        if (random_number < 0.5) {
            permutation_unit_flipped[algebra_pointer[famnum]][i] = 1;
        }
    }
    print("\n\n");
}
