/* GENEHUNTER - Kruglyak, Daly, Reeve-Daly, and Lander
   Whitehead Institute for Biomedical Research, Copyright 1996
   modified by Konstantin Strauch, Johannes Dietter, Manuel Mattheisen,
   and Markus Brugger.    Copyright (c) 1999-2019 */

/* glibc feature test macro requirements for erf() */
#define _XOPEN_SOURCE
#include <math.h>

#include <search.h>
#define INC_LIB
#define INC_SHELL  /* for error() */
#include "npl.h"


void find_sharing_probs(int position, int number_bits, int vec, double vector_probs);
void norm_sharing_probs(int position, double like);


/* Katz, Aug, 2000, this was in newcombo.h in ghk version */
#define real_eq(x,y) (fabs((x)-(y))<1e-10)
#define real_eq12(x,y) (fabs((x)-(y))<1e-12)
#define real_eq03(x,y) (fabs((x)-(y))<1e-03)
#define real_eq02(x,y) (fabs((x)-(y))<1e-02)

/* externs */
double **apm_stat,**raw_stat,**p_value,**apm_info,**apm_entropy,**pvec_entropy;
double *wpos, *wpos_m, *wpos_f, *null_variance;
int num_w_pos;
extern int ind_sp;
int num_mod_pos, first_mod_pos, last_mod_pos;  /*** for MOD version ***/
double *total_lod;
double **sub_lods;

double *total_recs;

/* exact p-value computation */
double **prior_pscore, **pscore_list;
int *num_pscores;

/* local */
double *map_position, *map_position_m, *map_position_f; /* extern for pmod */
static double **lvector, **rvector, **aux, *tempvector, *newp; 
static int n, n_m, n_f, N, rn, rn_f, rn_m, bigN, *padded_vec;
static double theta[MAXLOCI], theta_m[MAXLOCI], theta_f[MAXLOCI];
static double **pvector;
static double *apm_score;
static double expscore, expscore_2;
static int *best_pvector, **saved_partner;
static int uninf_mask, *flip_mask;
int *row_partner;
int *bit_count, *real_bit_count;
static int k_full[64];
int k_org[64];
int *m_wgt;
int *f_wgt;
int *m_wgt_s;
int *f_wgt_s;
static int lbtc8mask=0xff, mod8[AVAILABLE_BITS], lbtc8[256];
static int    **maskTest;
static int    *nz_bits, *nz_bits_m,*nz_bits_f;
static unsigned long *nz_pointer;
static double *num_par_score, *pvct, *lvct, *rvct;
static int f_num_in_map_order, *f_uninformative, *f_map_order;
static int score_mask_rdc;

static double *this_p_pos=NULL;
static int really_modcalc=FALSE;
static int this_mod_pos;

int *seq_sim_single;             /*** for PMOD ***/

int num_mapped_loci;

typedef struct {
  int p_bitc;
  int p_bitc_m;
  int p_bitc_f;
  int n_size;
  unsigned long n_pntf;
  unsigned long n_pntl;
  int nffs;
  int nffs_m;
  int nffs_f;
  int nfbit_sex[33];
  int nMaskShift;
  int p_address0;
  int p_mask_v_bits;
  int p_legal;
  double l_theta[33];
  double l_ctheta[33]; 
  double l_theta_m[33];
  double l_ctheta_m[33];
  double l_theta_f[33];
  double l_ctheta_f[33];     
  int p_v_flip_mask[32];
  int p_f_flip_mask[32];
  double *tempvct;
  int    *bit_cnt;
  int    *bit_cnt_m;
  int    *bit_cnt_f;
  int **maskShift;
} TRANSSTRUCT;


typedef struct {
  double maxlike;
  double minlike;
  double lodlike;
  double like;
  double statval; 
  double varsum;
  double un_normalized_entropy;
} TRDOTSTRUCT;

  struct pos_at_marker{
    int ind_marker_pos;
    double pos_m;
    double pos_f;
  };
  typedef struct pos_at_marker POS_AT_MARKER; 



/* internal procedures */
double likelihood(int,int,int,double,double,
		  double*,
		  double*,double*,double*,int);
double likelihood_sp(int,int,int,double,double,double,double,
		  double*,
		  double*,double*,double*,int);		  
		  
double left(double*,int), right(double*,int);
void reduce(int,int,int), makerec(double, double*), 
     new_reduce(int,int,int), makerec_sp(double, double, double *);
void exciting_reduce(int,int), haplo_reduce(int,int);
void speedy_reduce(double,double*);
void convolve_fft(double,double*), convolve_count_fft(double,double*);
void fft(double*);
void init_sex_weights(int, int*, int*);
void s_wgt_free(int*, int*);
void Setup_flip_masks(int, int *, int *, int *, int *);
void Setup_flip_masks_sp(int, int *, int *, int *, int *, int *, int *, int*);
void convolve_fft_2_sp(int, double, double, int *, int*, double *);
void convolve_fft_2(int, double, int *, double *);
void Setup_MaskShift(int ,int , int, int, TRANSSTRUCT *);
void Setup_legal_vectors(int, int *, int, int *, int, int *);
void Allocate_MCh_vars(void);
void Free_MCh_vars(void);
void Load_pvct(void);
double Compute_lr_vectors(void);
double Compute_lr_vectors_sp(void);
double Get_apm_score(int);
void TransSetupConvolve(double, int, int, double *, TRANSSTRUCT *);
void TransSetupConvolve_sp(double, double, int, int, double *, TRANSSTRUCT *,
                        int*, int*);
void TransAndMap(TRANSSTRUCT *, double *);
void TransAndMap_sp(TRANSSTRUCT *ts, double*, int*, int*);
void TransBetweenAndDot(TRANSSTRUCT *tsleft, TRANSSTRUCT *tsright,
								int pos_in_scan, TRDOTSTRUCT *tdot, double *tmpdot,
								double *tmp3, double *tmp4);
void TransBetweenAndDot_sp(TRANSSTRUCT *tsleft, TRANSSTRUCT *tsright,
								int pos_in_scan, TRDOTSTRUCT *tdot, double *tmpdot,
								double *tmp3, double *tmp4);
void TransFree(TRANSSTRUCT *);
void Check_results_01(int, unsigned long, int , int, int, int, double *, double *);
void modify_setup_lodscores(void);
void init_sharing_probs(int number_vecs, int number_bits);
void count_recs_2(int num_loci, double like, int *bit_cnt, double *observed_recs);
void count_recs_2_sp(int num_loci, double like, int* bit_cnt_m, int* bit_cnt_f, double *observed_recs);
void compute_haplotype(int num_loci);
int pad_inhvec(int vec);
int lookup_mask(int vec);
int diff_bits(int a, int b);
int min_diff_bits(int new, int old, int *new_vec);


void scan_pedigree(int *nz_bits_s, int *nz_bits_s_m, int *nz_bits_s_f, unsigned long *nz_pointer_s, double *pvct_s,
	      double *num_par_score_s,
	      int f_num_in_map_order_s, int *f_uninformative_s, int *f_map_order_s,
              int real_num_bits, int real_num_bits_m_s, int real_num_bits_f_s,
              double *score, int score_mask_rdc_s, double exp, double exp2,
              int pednum, int *best_pv, int num_bits, int num_bits_m_s, int num_bits_f_s, 
	      int u_mask, int *f_mask, int **maskT, int this_mod_pos_s)
{    
    void setup_lodscores(void);
    void setup_lodscores_map_sup(void);
    void calc_apm_scores_sp(int,double,double*);
    void calc_apm_scores(int,double,double*);
    int i,j=0,num_scores,*score_count;
    double score_npl_val;
    double *score_list,tempmax,tempmin,default_entropy, this_prob,*lod;
    void allocate_apm_storage(void), free_apm_storage(void);
    static int previously_allocated_apm=FALSE;

    nz_bits    = nz_bits_s;
    nz_bits_m  = nz_bits_s_m;
    nz_bits_f  = nz_bits_s_f;
    nz_pointer = nz_pointer_s;
    pvct       = pvct_s;
    num_par_score = num_par_score_s;
    f_num_in_map_order = f_num_in_map_order_s;
    f_uninformative    = f_uninformative_s;
    f_map_order        = f_map_order_s;
	
    apm_score = score;
    score_mask_rdc = score_mask_rdc_s;
    expscore = exp;
    expscore_2 = exp2;
    best_pvector = best_pv;
    uninf_mask = u_mask;
    flip_mask = f_mask;

    this_mod_pos = this_mod_pos_s;
    /*** -1: not for mod, i.e., called from scan pedigrees command
	 0...num_w_pos-1: called from modscore.c, i.e., modcalc single ***/

    n = num_bits;
    n_m = num_bits_m_s;
    n_f = num_bits_f_s;
    rn = real_num_bits;
    rn_m= real_num_bits_m_s;
    rn_f= real_num_bits_f_s;
    N = m_pow2[rn];
    bigN = m_pow2[n];

	/* fill bitcount lookup table */
	for(i=0; i<AVAILABLE_BITS; i++) mod8[i] = i/8;
	for(unsigned long h=0; h<m_pow2[8]; h++) {
		lbtc8[h] = 0;
		for(j=0; j<8; j++) lbtc8[h] += ((h & m_pow2[j]) ? 1:0);
	}
	
    /*r-full-alloc  allocate_lod_vars(real_num_bits); */
	
	maskTest = maskT;
	Allocate_MCh_vars();
	/* Load_pvct(); */
	
    /*r-old-alloc  array(row_partner, N, int); for *reduce */
    /* r-full-alloc array(bit_count, N, int); */
    /* r-full-alloc array(real_bit_count, N, int); */
    /* r-old-alloc matrix(saved_partner, n, N, int); for speedy_reduce  */
    /* for (j=0; j<n; j++) for (i=0; i<N; i++) saved_partner[j][i]=-1; */

	/* define bit count for each vector */
    /*r-full-alloc  removed computation for real_bit_count, bit_count */

    if (this_mod_pos==-1 && g_first_pedigree==1) {
	
	/*setup_lodscores();*/ /*  only call setup the first time */
	setup_lodscores_map_sup();

	modify_setup_lodscores();
	/* Allocate the apm_stat array  - free first if necessary */
	if (previously_allocated_apm) free_apm_storage();
	allocate_apm_storage();
	pedigrees_calculated=0;
	previously_allocated_apm=TRUE;
	array(total_recs, num_in_map_order, double);
	for (i=0; i<num_in_map_order; i++) total_recs[i]=0.0;
    }
    if (this_mod_pos==-1) {
	if (modcalc == 1)
	    matrix(save_p_all_pos[famnum], num_mod_pos, 1<<rn, double);
	if (modcalc == 2)
	    array(save_p_one_pos[famnum], 1<<rn, double);
    }
    
    /**********
       compute the list of possible scores for 2 purposes:
       1) so we can compute Shannon's entropy for score function 
          in our likelihood function
       2) so we can keep a running probability distribution of
          the sum of raw scores for a p-value over all pedigrees 
    **********/
    array(score_list, 100000, double);
    array(score_count, 100000, int); 

    num_scores=0; tempmax=0.0; tempmin=1.0;
    for (i=0; i<N; i++) {
	  score_npl_val = Get_apm_score(i);
      if (score_npl_val > tempmax) tempmax = score_npl_val;
      if (score_npl_val < tempmin) tempmin = score_npl_val;
      for (j=0; j<num_scores; j++)
	if (score_npl_val == score_list[j]) { score_count[j]++; break; }
      
      if (j == num_scores) {
	  /* this one not in list */
	  score_list[num_scores] = score_npl_val;
	  score_count[num_scores] = 1;
	  num_scores++;
      }
    }
    if (j>=100000) 
	error("Overwrote score_list[], score_count[]");
	
/*	printf("num_scores = %d \n", num_scores);
*/
    default_entropy=0.0;
    for (i=0; i<num_scores; i++) {
      this_prob = (double) score_count[i] / (double) N;
      default_entropy += (-1.0)*this_prob*log(this_prob);
    }
    array(lod, num_w_pos, double);

    if (this_mod_pos==-1) {
      array(prior_pscore[pednum], num_scores, double);
      array(pscore_list[pednum], num_scores, double);
    }

    
    if (ind_sp==1){
    
       calc_apm_scores_sp(pednum,default_entropy,
		    lod);
    
    } else if (ind_sp==0) {
    
       calc_apm_scores(pednum,default_entropy,
		    lod);
		    
    }	    
     

    if (this_mod_pos==-1)
	null_variance[pednum] = expscore_2 - (expscore*expscore);

    if (this_mod_pos==-1) for (i=0; i<num_w_pos; i++) {
        total_lod[i] += lod[i]; sub_lods[pednum][i] = lod[i];
    }
    unarray(lod, double);

    /***** debugging output
    sf(ps,"this max = %.4lf, this min = %.4lf, num different scores = %d\n",
       tempmax,tempmin,num_scores); pr();
    *****/

    if (this_mod_pos==-1) for (i=0; i<num_scores; i++) {
      pscore_list[pednum][i]= score_list[i];
      prior_pscore[pednum][i]= (double) score_count[i] / (double) N;
      /***** debugging output OFF *****/
      /*****
      sf(ps,"score=%.4lf, prior=%.6lf, post=%.6lf\n",
	 score_list[i], prior_pscore[pednum][i], post_pscore[pednum][0][i]);
      pr();
      *****/
    }
    if (this_mod_pos==-1) num_pscores[pednum]= num_scores;
        
    if (this_mod_pos==-1) pedigrees_calculated++;

    unarray(score_count, int);
    unarray(score_list, double);

    /*r-old-alloc unmatrix(saved_partner, n, int); */
    /*r-old-alloc narray(row_partner,int); */
    /*r-full-alloc  unarray(bit_count,int); */
    /*r-full-alloc  unarray(real_bit_count,int); */
	Free_MCh_vars();
    /*r-full-alloc  free_lod_vars(real_num_bits); */
}

void allocate_apm_storage(void)
{
    int i;

    matrix (apm_stat, num_pedigrees_allocated, num_w_pos, double);
    matrix (raw_stat, num_pedigrees_allocated, num_w_pos, double);
    matrix (p_value, num_pedigrees_allocated, num_w_pos, double);
    matrix (apm_info, num_pedigrees_allocated, num_w_pos, double);
    matrix (apm_entropy, num_pedigrees_allocated, num_w_pos, double);
    matrix (pvec_entropy, num_pedigrees_allocated, num_w_pos, double);

    array(prior_pscore, num_pedigrees_allocated, double*);
    array(pscore_list, num_pedigrees_allocated, double*);
    array(num_pscores, num_pedigrees_allocated, int);

    array(total_lod, num_w_pos, double);
    for (i=0; i<num_w_pos; i++) total_lod[i]=0.0;
    matrix (sub_lods, num_pedigrees_allocated, num_w_pos, double);

    array(null_variance, num_pedigrees_allocated, double);
}

void free_apm_storage(void)
{
    int i;
    /***** print("freeing memory\n"); *****/
    unmatrix (apm_stat, num_pedigrees_allocated, double);
    unmatrix (raw_stat, num_pedigrees_allocated, double);
    unmatrix (p_value, num_pedigrees_allocated, double);
    unmatrix (apm_info, num_pedigrees_allocated, double);
    unmatrix (apm_entropy, num_pedigrees_allocated, double);
    unmatrix (pvec_entropy, num_pedigrees_allocated, double);

    for (i=0; i<pedigrees_calculated; i++) {
      unarray(pscore_list[i], double);
      unarray(prior_pscore[i], double);
    }
    
    unarray(prior_pscore, double*);
    unarray(pscore_list,  double*);
    unarray(num_pscores, int);

    unarray(total_lod, double);
    unmatrix(sub_lods, num_pedigrees_allocated, double);

    unarray(null_variance, double);
}

void Allocate_MCh_vars(void)
{
    long amount_to_alloc;

	amount_to_alloc = nz_pointer[f_num_in_map_order];

	run {
		array(lvct, amount_to_alloc, double);
		array(rvct, amount_to_alloc, double);
	} except_when (NOMEMORY) {
		Free_MCh_vars();
		error("Can't get enough memory for this scan (l/rvct)\n");
	}
}

void Free_MCh_vars(void)
{
	unarray(lvct, double);
	unarray(rvct, double);
}

void allocate_lod_vars(int num_bits)
{
    long amount_to_alloc;

    amount_to_alloc=m_pow2[num_bits];

    run {

      matrix (lvector, num_in_map_order+2,  amount_to_alloc, double);
      matrix (rvector, num_in_map_order+2,  amount_to_alloc, double);
      /*r-old-alloc matrix (aux, amount_to_alloc, 30, double); for *reduce */
      array (tempvector, amount_to_alloc, double);
      /*r-old-alloc array (newp, amount_to_alloc, double); for speedy_reduce */
      /*r-old-alloc array (padded_vec, amount_to_alloc, int); for speedy_reduce */
      /*r-old-alloc for (int i=0; i<amount_to_alloc; i++) padded_vec[i] = pad_inhvec(i); for speedy_reduce */

    } except_when (NOMEMORY) {

      print("Can't get enough memory for this scan\n");
      free_lod_vars();
      
    }
}

void free_lod_vars(void)
{
    /* free_lod_vars formerly had 'int num_bits' as argument */
    /*long amount_to_free;*/

    /*amount_to_free=m_pow2[num_bits];*/

    /*r-old-alloc unarray (padded_vec, int); for speedy_reduce */
    /*r-old-alloc unarray (newp, double); for speedy_reduce */
    unarray (tempvector, double);
    /* unmatrix(aux, amount_to_free, double);  for *reduce */
    unmatrix(rvector,num_in_map_order+2,double);
    unmatrix(lvector,num_in_map_order+2,double);
}

void setup_lodscores(void)
{
  int i, j, num_pos;
  double total;
  double pos, start_position, end_position, this_dist;
  int interval;
  double increment;
  array(wpos, MAX_POSITIONS, double);
  array(wpos_m, MAX_POSITIONS, double);
  array(wpos_f, MAX_POSITIONS, double);
  array (map_position, num_markers+2, double);
  array (map_position_m, num_markers+2, double);
  array (map_position_f, num_markers+2, double);
  
  
  /* Set up the map information needed */
  
  /* for the male map ... */
    
  num_mapped_loci = f_num_in_map_order;
  
  if (single_point_mode) {
      num_w_pos = num_markers;
      return;
  }
  
  total = 0.0;
  j=0;
  for (i=0; i<num_mapped_loci-1; i++) {
    map_position_m[i]=total;
    while(map_distance_m[j] == 0.) j++;
    this_dist = map_distance_m[j];
    j++;
    total += this_dist;
    theta_m[i] = dist_to_rec(this_dist);
/*	printf(" ilocus, theta, map_position, total %2d  %12.4e  %12.4e  %12.4e \n",
	i, theta[i], map_position[i], total);
*/}
  map_position_m[i] = total;
  theta_m[i] = .5; /* for mapping disease locus at infinity */
  
  /* Currently only acceptable setting scan_steps = 1 
	if(scan_steps != 1) {
		scan_steps = 1;
		printf("******* Warning **********************************\n");
		printf("Currently only acceptable setting scan_steps = 1\n");
		printf("scan_steps set to default = 1 \n");
		printf("**************************************************\n");
	}
  */
  /* Tally the number of positions and save their position, so
     that we can correctly allocate the apm_stat array */
  num_pos=0;
  if (scan_steps > 0) {
    /* Calculate a set number of positions between each pair of markers */
    for (interval=0; interval<num_mapped_loci+1; interval++) {
      /* Deal with first and last intervals */
      if (interval==0) 
	start_position=(-1.0)*off_end_dist_m;
      else
	start_position=map_position_m[interval-1];
      if (interval==num_mapped_loci)
	end_position=map_position_m[num_mapped_loci-1]+off_end_dist_m;
      else
	end_position=map_position_m[interval];
      /* Calculate the increment size for the interval */
      increment = (end_position-start_position)/(double)scan_steps;
      /* so that the final step is taken and to prevent 
	 rounding problems in summations */
      if (interval==num_mapped_loci) 
	end_position+=increment/4.0;
      else
	end_position-=increment/4.0;
      /* if off_end==0 have to get the last point */
      if (interval==num_mapped_loci && increment==0.00) {
	end_position += 0.5;
	increment = 1.0;
      }
      /* Now we're ready to actually count and store the positions
	 for the interval */
      for (pos=start_position; pos<end_position; pos+=increment) {
	wpos_m[num_pos]=pos;
	num_pos++;
      }
    }
  } else {
    /* Move down the map at a set distance interval */
    start_position = (-1.0)*off_end_dist_m;
    end_position = map_position_m[num_mapped_loci-1]+off_end_dist_m;
    for (pos=start_position; pos<end_position; pos+=scan_increment_m) {
      wpos_m[num_pos] = pos;
      num_pos++;
    }
  }

  
  
  /* for the female map ... */
  
  
  num_mapped_loci = f_num_in_map_order;
  total = 0.0;
  j=0;
  for (i=0; i<num_mapped_loci-1; i++) {
    map_position_f[i]=total;
    while(map_distance_f[j] == 0.) j++;
    this_dist = map_distance_f[j];
    j++;
    total += this_dist;
    theta_f[i] = dist_to_rec(this_dist);
/*	printf(" ilocus, theta, map_position, total %2d  %12.4e  %12.4e  %12.4e \n",
	i, theta[i], map_position[i], total);
*/}
  map_position_f[i] = total;
  theta_f[i] = .5; /* for mapping disease locus at infinity */
  
  /* Currently only acceptable setting scan_steps = 1 
	if(scan_steps != 1) {
		scan_steps = 1;
		printf("******* Warning **********************************\n");
		printf("Currently only acceptable setting scan_steps = 1\n");
		printf("scan_steps set to default = 1 \n");
		printf("**************************************************\n");
	}
  */
  /* Tally the number of positions and save their position, so
     that we can correctly allocate the apm_stat array */
  num_pos=0;
  if (scan_steps > 0) {
    /* Calculate a set number of positions between each pair of markers */
    for (interval=0; interval<num_mapped_loci+1; interval++) {
      /* Deal with first and last intervals */
      if (interval==0) 
	start_position=(-1.0)*off_end_dist_f;
      else
	start_position=map_position_f[interval-1];
      if (interval==num_mapped_loci)
	end_position=map_position_f[num_mapped_loci-1]+off_end_dist_f;
      else
	end_position=map_position_f[interval];
      /* Calculate the increment size for the interval */
      increment = (end_position-start_position)/(double)scan_steps;
      /* so that the final step is taken and to prevent 
	 rounding problems in summations */
      if (interval==num_mapped_loci) 
	end_position+=increment/4.0;
      else
	end_position-=increment/4.0;
      /* if off_end==0 have to get the last point */
      if (interval==num_mapped_loci && increment==0.00) {
	end_position += 0.5;
	increment = 1.0;
      }
      /* Now we're ready to actually count and store the positions
	 for the interval */
      for (pos=start_position; pos<end_position; pos+=increment) {
	wpos_f[num_pos]=pos;
	num_pos++;
      }
    }
  } else {
    /* Move down the map at a set distance interval */
    start_position = (-1.0)*off_end_dist_f;
    end_position = map_position_f[num_mapped_loci-1]+off_end_dist_f;
    for (pos=start_position; pos<end_position; pos+=scan_increment_f) {
      wpos_f[num_pos] = pos;
      num_pos++;
    }
  }
  

  if (num_pos > MAX_POSITIONS) {
    sf(ps,"Number of positions to be calculated exceeds the max of %d\n",
       MAX_POSITIONS);
    error(ps);
  }
    
  num_w_pos = num_pos;

  for (j=0; j<num_w_pos; j++) wpos[j] = (wpos_m[j] + wpos_f[j]) / 2.0;
  for (i=0; i<num_mapped_loci; i++){
     theta[i] = 0.5*(theta_m[i] + theta_f[i]);
     map_position[i] = 0.5*(map_position_m[i]+map_position_f[i]);
  }     

  if (modcalc) {    /*** i.e., for both 'global' and 'single' ***/
    num_mod_pos = 0;
    for (j=0; j<num_w_pos; j++)
      if (((wpos[j]>=lowpos) && (wpos[j]<=highpos)) ||
	  real_eq12(wpos[j], lowpos) || real_eq12(wpos[j], highpos)) {
	num_mod_pos++;
	last_mod_pos = j;
      }      
    if (num_mod_pos == 0) {
	sf(ps, "no valid genetic positions in interval between %.2lf and %.2lf cM.\n", lowpos, highpos);
	error(ps);
    }
    if (pmod_calc_flag && best_position)
      first_mod_pos = last_mod_pos = save_max_pos;
    else
      first_mod_pos = last_mod_pos + 1 - num_mod_pos;
  }
} 


void setup_lodscores_map_sup(void)
{
  int i, j, num_pos, interval;
  double total, increment,pos, start_position,
         end_position, this_dist;
  POS_AT_MARKER* pos_marker;	 
 
  void build_sxp_pos(POS_AT_MARKER* pos_marker);

  array(wpos, MAX_POSITIONS, double);
  array(wpos_m, MAX_POSITIONS, double);
  array(wpos_f, MAX_POSITIONS, double);
  array(map_position, num_markers+2, double);
  array(map_position_m, num_markers+2, double);
  array(map_position_f, num_markers+2, double);
  array(pos_marker, MAX_POSITIONS, POS_AT_MARKER);
  /* Set up the map information needed */
  num_mapped_loci = f_num_in_map_order;

  if (single_point_mode) {
      num_w_pos = num_markers;
      return;
  }

  total = 0.0;
  j=0;
  for (i=0; i<num_mapped_loci-1; i++) {
    map_position[i]=total;
    while(map_distance[j] == 0.) j++;
    this_dist = map_distance[j];
    j++;
    total += this_dist;
    theta[i] = dist_to_rec(this_dist);
/*	printf(" ilocus, theta, map_position, total %2d  %12.4e  %12.4e  %12.4e \n",
	i, theta[i], map_position[i], total);
*/}
  map_position[i] = total;
  theta[i] = .5; /* for mapping disease locus at infinity */
  
  total = 0.0;
  j=0;
  for (i=0; i<num_mapped_loci-1; i++) {
    map_position_m[i]=total;
    while(map_distance_f[j] == 0.) j++;
    this_dist = map_distance_m[j];
    j++;
    total += this_dist;
    theta_m[i] = dist_to_rec(this_dist);
	/*printf(" ilocus, theta, map_position, total %2d  %12.4e  %12.4e  %12.4e \n",
	i, theta_m[i], map_position_m[i], total);
       */
       /* printf("%.4lf ", theta_m[i]); */
  }
  /* printf("\n"); */
  map_position_m[i] = total;
  theta_m[i] = .5; /* for mapping disease locus at infinity */
  
  total = 0.0;
  j=0;
  for (i=0; i<num_mapped_loci-1; i++) {
    map_position_f[i]=total;
    while(map_distance[j] == 0.) j++;
    this_dist = map_distance_f[j];
    j++;
    total += this_dist;
    theta_f[i] = dist_to_rec(this_dist);
	/*printf(" ilocus, theta, map_position, total %2d  %12.4e  %12.4e  %12.4e \n",
	i, theta_f[i], map_position_f[i], total);
       */
        /* printf("%.4lf ", theta_f[i]); */
  }
  /* printf("\n"); */
  map_position_f[i] = total;
  theta_f[i] = .5; /* for mapping disease locus at infinity */

 /*  for (i=0; i<num_mapped_loci; i++){
     theta[i] = 0.5*(theta_m[i] + theta_f[i]);
     //printf("%.4lf ", theta[i]);
     map_position[i] = 0.5*(map_position_m[i]+map_position_f[i]);
  }
  //printf("\n");  */

  /* Currently only acceptable setting scan_steps = 1 
	if(scan_steps != 1) {
		scan_steps = 1;
		printf("******* Warning **********************************\n");
		printf("Currently only acceptable setting scan_steps = 1\n");
		printf("scan_steps set to default = 1 \n");
		printf("**************************************************\n");
	}
  */
  /* Tally the number of positions and save their position, so
     that we can correctly allocate the apm_stat array */
  num_pos=0;
  if (scan_steps > 0) {
    /* Calculate a set number of positions between each pair of markers */
    for (interval=0; interval<num_mapped_loci+1; interval++) {
      /* Deal with first and last intervals */
      if (interval==0) 
	start_position=(-1.0)*off_end_dist;
      else
	start_position=map_position[interval-1];
      if (interval==num_mapped_loci)
	end_position=map_position[num_mapped_loci-1]+off_end_dist;
      else
	end_position=map_position[interval];
      /* Calculate the increment size for the interval */
      increment = (end_position-start_position)/(double)scan_steps;
      /* so that the final step is taken and to prevent 
	 rounding problems in summations */
      if (interval==num_mapped_loci) 
	end_position+=increment/4.0;
      else
	end_position-=increment/4.0;
      /* if off_end==0 have to get the last point */
      if (interval==num_mapped_loci && increment==0.00) {
	end_position += 0.5;
	increment = 1.0;
      }
      /* Now we're ready to actually count and store the positions
	 for the interval */
      for (pos=start_position; pos<end_position; pos+=increment) {
        if (pos == start_position && interval != 0) {
	  pos_marker[num_pos].ind_marker_pos = 1;
	  pos_marker[num_pos].pos_m = map_position_m[interval-1];
	  pos_marker[num_pos].pos_f = map_position_f[interval-1];
	}else{
	  pos_marker[num_pos].ind_marker_pos = 0;
	}  
	wpos[num_pos]=pos;
	num_pos++;
      }
    }
  } else {
    /* Move down the map at a set distance interval */
    start_position = (-1.0)*off_end_dist;
    end_position = map_position[num_mapped_loci-1]+off_end_dist;
    for (pos=start_position; pos<end_position; pos+=scan_increment) {
      pos_marker[num_pos].ind_marker_pos = 0;
      wpos[num_pos] = pos;
      num_pos++;
    }
  }

  if (num_pos > MAX_POSITIONS) {
    sf(ps,"Number of positions to be calculated exceeds the max of %d\n",
       MAX_POSITIONS);
    error(ps);
  }

  num_w_pos = num_pos;

  build_sxp_pos(pos_marker);

  /*** the following for modscore (for both, 'global' and 'single') 
       or p-Value-calculation: ***/
  if (modcalc  || (num_replicates > 0)) {
    num_mod_pos = 0;
    for (j=0; j<num_w_pos; j++)
      if (((wpos[j]>=lowpos) && (wpos[j]<=highpos)) ||
            real_eq12(wpos[j], lowpos) || real_eq12(wpos[j], highpos)) {
        num_mod_pos++;
        last_mod_pos = j;
      }
    if (num_mod_pos == 0) {
	sf(ps, "no valid genetic positions in interval between %.2lf and %.2lf cM.\n", lowpos, highpos);
	error(ps);
    }
    if (pmod_calc_flag && best_position)
      first_mod_pos = last_mod_pos = save_max_pos;
    else
      first_mod_pos = last_mod_pos + 1 - num_mod_pos;
  }
  unarray(pos_marker, POS_AT_MARKER);
}


void modify_setup_lodscores(void)
{
	int i;
	int pos_a, pos_b;

	pos_a = 0; 
	pos_b = num_w_pos;
	for(i=0; i<=(pos_b-pos_a); i++) {
		wpos_m[i] = wpos_m[i+pos_a];
	}
	for(i=0; i<=(pos_b-pos_a); i++) {
		wpos_f[i] = wpos_f[i+pos_a];
	}	
	num_w_pos = pos_b - pos_a;
}


void build_sxp_pos(POS_AT_MARKER* pos_marker){
  int i, low_bound, intrv, nr_bound, use_sp_map=1;
  double r, df, dm;
  double *bound_m, *bound_f, *bound_a, bm, bf, ba;
  double dmap = MIN_MAP_MARKER_DISTANCE;
  char marker_name[MAX_NAME_SIZE];
  MARKER *act_marker;
  MARKER *first_user_marker;
  MARKER *last_user_marker;
  MARKER *first_chrom_marker;
  MARKER *last_chrom_marker;
  ENTRY itmt_htable;
    
  int interval_search(double* bound, int count, int low_lim, int ipos);
  
  array(bound_m, MAX_NUMBER_MAP_MARKER, double); 
  array(bound_f, MAX_NUMBER_MAP_MARKER, double); 
  array(bound_a, MAX_NUMBER_MAP_MARKER, double);
  
  /*determine first and last marker of user*/
  
  first_user_marker = NULL;
  last_user_marker  = NULL; 
  
  strcpy(marker_name, locus[f_map_order[0]].name);
  itmt_htable.key = uppercase(crunch(marker_name));
  if (hsearch(itmt_htable, FIND)){                 
    first_user_marker = (struct marker*)(hsearch(itmt_htable, FIND))->data;
  }else{
    sf(ps,"first marker of user not found in map, should not happen in this program state");
    error(ps);
  }    
  strcpy(marker_name, locus[f_map_order[num_mapped_loci-1]].name);
  itmt_htable.key = uppercase(crunch(marker_name));
  if (hsearch(itmt_htable, FIND)){                 
    last_user_marker = (struct marker*)(hsearch(itmt_htable, FIND))->data;
  }else{
    sf(ps,"last marker of user not found in map, should not happen in this program state");
    error(ps);
  }
  
  /* determine first and last marker of chromosome */
  
  first_chrom_marker = first_user_marker;
  last_chrom_marker  = last_user_marker;   
  
  act_marker = first_user_marker;
  while (act_marker->prior && !strcmp(act_marker->chromosome, 
                                      act_marker->prior->chromosome)){
        act_marker = act_marker->prior;
  }
  if(act_marker!=NULL) first_chrom_marker = act_marker;
  
  act_marker = last_user_marker;
  while (act_marker->next && !strcmp(act_marker->chromosome,
                                     act_marker->next->chromosome)){ 
         act_marker = act_marker->next;
  } 
  if(act_marker!=NULL) last_chrom_marker = act_marker;
  
  /* determine interval bounds on chromosome for lod-score calculation */
    
  if (wpos[0]+first_user_marker->gen_pos_a < first_chrom_marker->gen_pos_a){
  
     /* if map has only one marker */
    if (last_chrom_marker->order_number != first_chrom_marker->order_number){    
      r = (last_chrom_marker->gen_pos_f-first_chrom_marker->gen_pos_f)/
          (last_chrom_marker->gen_pos_m-first_chrom_marker->gen_pos_m);
    }else{          
      r = 1.;      
    }  
    
    bound_f[0] = 1./(1.+r)*(2.*r*(wpos[0]+first_user_marker->gen_pos_a) +
                            first_chrom_marker->gen_pos_f - 
                            r*first_chrom_marker->gen_pos_m) - 
                 first_user_marker->gen_pos_f;
    
    bound_m[0] = 2.*(wpos[0]+first_user_marker->gen_pos_a) - bound_f[0] -
                 first_user_marker->gen_pos_f - first_user_marker->gen_pos_m;
    
    bound_a[0] = wpos[0]-0.0001; /* to include the first position */
    
    act_marker = first_chrom_marker;
    i=1;
    
  }else if (wpos[0]+first_user_marker->gen_pos_a >= first_chrom_marker->gen_pos_a){
    
    act_marker = first_chrom_marker;
    while (act_marker->gen_pos_a < wpos[0]+first_user_marker->gen_pos_a){
	 act_marker = act_marker->next;
    }
    if (act_marker != first_chrom_marker) act_marker = act_marker->prior;
    bound_m[0] = act_marker->gen_pos_m - first_user_marker->gen_pos_m;
    bound_f[0] = act_marker->gen_pos_f - first_user_marker->gen_pos_f;
    bound_a[0] = act_marker->gen_pos_a - first_user_marker->gen_pos_a;
    act_marker = act_marker->next;
    i=1;
  }  
        
  while (act_marker && act_marker->prior != last_chrom_marker &&
         act_marker->gen_pos_a < wpos[num_w_pos-1] + 
	                         first_user_marker->gen_pos_a){	
     
       bm = act_marker->gen_pos_m - first_user_marker->gen_pos_m;
       bf = act_marker->gen_pos_f - first_user_marker->gen_pos_f;
       ba = act_marker->gen_pos_a - first_user_marker->gen_pos_a;
     
       if (use_sp_map){
         
	 if (ba > bound_a[i-1] + dmap && !real_eq(ba,bound_a[i-1])) {	
           bound_m[i] = bm;
           bound_f[i] = bf;
           bound_a[i] = ba;
	   i++;
         }           
       
       }else if(!use_sp_map){
       
         if (act_marker->ind_user_marker==1) {	
           bound_m[i] = bm;
           bound_f[i] = bf;
           bound_a[i] = ba;
	   i++;
         }
       }
       
       act_marker = act_marker->next; 	  
  }
  
  if (act_marker && act_marker->prior != last_chrom_marker){  
     bound_m[i] = act_marker->gen_pos_m - first_user_marker->gen_pos_m;
     bound_f[i] = act_marker->gen_pos_f - first_user_marker->gen_pos_f;
     bound_a[i] = act_marker->gen_pos_a - first_user_marker->gen_pos_a;
  }else{
     /*  if offend is beyond the map extrapolate by drawing a straight */
     /*  line from 0 to the last marker of the chromosome. */
     if (last_chrom_marker->order_number != first_chrom_marker->order_number){   
       r = (last_chrom_marker->gen_pos_f-first_chrom_marker->gen_pos_f)/
           (last_chrom_marker->gen_pos_m-first_chrom_marker->gen_pos_m);
     }else{
       r = 1.;
     } 
 
    bound_f[i] = 1./(1.+r)*(2.*r*(wpos[num_w_pos-1]+first_user_marker->gen_pos_a) +
                            first_chrom_marker->gen_pos_f - 
                            r*first_chrom_marker->gen_pos_m) - 
                 first_user_marker->gen_pos_f;
    
    bound_m[i] = 2.*(wpos[num_w_pos-1]+first_user_marker->gen_pos_a) - bound_f[i] -
                 first_user_marker->gen_pos_f - first_user_marker->gen_pos_m;
    
    bound_a[i] = wpos[num_w_pos-1]; 
  
  }
  
  nr_bound = i; 
   
  low_bound = 0;
  intrv = 0;
  for (i=0; i<num_w_pos; i++){
  
     if (pos_marker[i].ind_marker_pos==0){
        
       /*get index of map interval which contains wpos[i], actually lower bound*/
       intrv = interval_search(bound_a, nr_bound, low_bound, i);
       df = bound_f[intrv+1]-bound_f[intrv];
       dm = bound_m[intrv+1]-bound_m[intrv];
     
       if (!real_eq(df,0.0) && !real_eq(dm,0.0)){
        
         r = df/dm;
         wpos_m[i] = (r*bound_m[intrv]+2*wpos[i]-bound_f[intrv]) / (1.+r);
         wpos_f[i] = 2*wpos[i]-wpos_m[i];
       
       }else if (!real_eq(df,0.0) && real_eq(dm,0.0)){
       
         wpos_m[i] = bound_m[intrv];
         wpos_f[i] = 2*wpos[i]-wpos_m[i];
       
       }else if (real_eq(df,0.0) && !real_eq(dm,0.0)){
      
         wpos_f[i] = bound_f[intrv];
         wpos_m[i] = 2*wpos[i]-wpos_f[i];
      
       }else if (real_eq(df,0.0) && real_eq(dm,0.0)){  
         
         wpos_f[i] = bound_f[intrv];
         wpos_m[i] = bound_m[intrv];
       
       }  
      
     }else if(pos_marker[i].ind_marker_pos==1){
     
         wpos_f[i] = pos_marker[i].pos_f;
         wpos_m[i] = pos_marker[i].pos_m;
     
     }
              
     if (i>0){
       if (wpos_f[i]<wpos_f[i-1] || wpos_m[i]<wpos_m[i-1]){
         sf(ps,"positions inconsistent in build_sxp_pos");
	 error(ps);
       } 
     }
  }

  unarray(bound_m, double); 
  unarray(bound_f, double); 
  unarray(bound_a, double); 
 
}


int interval_search(double* bound, int count, int low_lim, int ipos){
  int low, high, mid;
  
  low = low_lim;
  high = count-1;
  
  while (low<=high){
    mid = (low+high)/2;
    if (wpos[ipos] <= bound[mid]) {
       high = mid-1;
    }else if (wpos[ipos] > bound[mid+1]) {
       low = mid+1;
    }else{
       return mid;
    }     
  }
  return -1;           
}


void calc_apm_scores(int ped_num,
		     double default_entropy, double *saved_lod)
{

  /* calc_apm_scores formerly had 'int originals' as argument */
  int i, j, pos, count;
  /* int left_i; */
  int ilocus, locus_l, locus_r, locus_haplo;
  double apmstat,varnull,variance,entropy;
  volatile double like;
  double thetal, thetar, pventropy, lodlike, this_lod, our_total;
  /* double dum; */
  double observed_recs[MAXLOCI], real_map_sum, our_map_sum;
  double *this_par_score = NULL;

  /*** for MOD ver, unlinked removed --> *save_unlinked now is extern ***/
  int    haplo_viterbi;

    int k, bits_on;
    unsigned long pnt;
    volatile int alloc_bit_cnt;
	int     nvfs, nffs, p_v_flip_mask[32], p_f_flip_mask[32], andmask, im;
	int     * volatile bit_cnt;
	double  p, l_theta, obs_recs;
	double  score_npl_val;
	void Setup_flip_masks(int, int *, int *, int *, int *);

/*  printf("----compute_sharing = %1d \n", compute_sharing);
*/
  if(this_mod_pos==-1 && compute_sharing) init_sharing_probs(N, n);

/****** start of inserted part for modcalc single (K. Strauch 04/2004) ******/

  if (this_mod_pos != -1) {   /* i.e., call from modscore (modcalc single) */

    num_mapped_loci = f_num_in_map_order;
    Compute_lr_vectors();
    
    array(bit_cnt, 1, int);
    alloc_bit_cnt = TRUE;

    /* find nearest informative marker left/right.  
       Here locus_l/locus_r = -1 means that left/right vector is not used.*/
    /* left */
    ilocus = 0;  locus_l = -1;  thetal = -1;
    while (ilocus < num_mapped_loci && (map_position[ilocus] <= wpos[this_mod_pos] || real_eq12(map_position[ilocus], wpos[this_mod_pos]))) {
	/* real_eq12 inserted (K. Strauch 04/2004) */
	if(!f_uninformative[ilocus]) locus_l = ilocus;
	ilocus++;
    }
    if(locus_l != -1) {
	thetal = dist_to_rec(wpos[this_mod_pos]-map_position[locus_l]);
	if (real_eq(thetal, 0.0)) thetal = 0.0;  /* K. Strauch 05/2004 */
    }
    if(locus_l == 0  && real_eq(thetal, 0.0)) locus_l = -1;

    /* right */
    ilocus = num_mapped_loci-1;  locus_r = -1;  thetar = -1;
    while (ilocus >=0 && (map_position[ilocus] >= wpos[this_mod_pos] || real_eq12(map_position[ilocus], wpos[this_mod_pos]))) {
	/* real_eq12 inserted (K. Strauch 04/2004) */
	if(!f_uninformative[ilocus]) locus_r = ilocus;
	ilocus--;
    }
    if(locus_r != -1) {
	thetar = dist_to_rec(map_position[locus_r]-wpos[this_mod_pos]);
	if (real_eq(thetar, 0.0)) thetar = 0.0;  /* K. Strauch 05/2004 */
    }
    if(locus_r == num_mapped_loci-1  && real_eq(thetar, 0.0)  &&  locus_l !=-1) locus_r = -1;

    /* find if at locus stated in marker data file.
       If so we might need to find haplotype */
    /* not needed for call from modscore (modcalc single),
       but still left here for completeness (K. Strauch 04/2004) */
    locus_haplo = -1;
    for(ilocus=0; ilocus<num_mapped_loci; ilocus++)
	if(real_eq12(wpos[this_mod_pos], map_position[ilocus]))
	    locus_haplo = ilocus;  /* now real_eq12 (K. Strauch 05/2004) */

    if(locus_l==-1  &&  locus_r==-1) {
	printf(" ******Error in calc_apm_scores() did not find any informative markers *******\n");
	exit (1);
    }

    really_modcalc = TRUE;
    this_p_pos = save_p_one_pos[famnum];
    if (algebra)
    this_par_score = datfile_par_score[algebra_pointer[famnum]];

    like = likelihood(locus_haplo,locus_l,locus_r,
		      thetal,thetar,
		      &apmstat,&variance,
		      &pventropy,&lodlike,this_mod_pos);

    if(alloc_bit_cnt) unarray(bit_cnt, int);
    alloc_bit_cnt = FALSE;
  
/******  end of inserted part for modcalc single (K. Strauch 04/2004) ******/

  } else if (!single_point_mode) {
    num_mapped_loci = f_num_in_map_order;

/*	printf("compute l/r vectors \n");
*/
    /*compute left vectors*/
    /*r-full-alloc  unlinked = left(theta, num_mapped_loci); */
    /*compute right vectors*/
    /*r-full-alloc  dum = right(theta, num_mapped_loci-1); */

    save_unlinked[famnum_pointer[famnum]] = Compute_lr_vectors();
/*	printf("unlinked=%12.4e \n", unlinked); */

	/* printf("search_vector \n"); */
	/* search_vector(); */    /* search p/l/rvector for unique values */
  
    if (display_scores && !scan_done) {
      if (analysis_type == NPL_ANALYSIS) 
	print("position  NPL score  p-value    information\n");
      else if (analysis_type == LOD_ANALYSIS) 
	print("position  LOD score    information\n");
      else /* BOTH */
	print("position  LOD score    NPL score  p-value    information\n");
    }

	/* prepare bitcount for fft */
	j = 0;
	for(i=0; i<num_mapped_loci; i++)
		if(!f_uninformative[i]) j++;
	/* if any one of the loci is uninformative we need bit_cnt[] */
	if(!show_expected_recs){
		array(bit_cnt, 1, int);
		alloc_bit_cnt = TRUE;
	} else {
		run {
			array(bit_cnt, N, int);
			alloc_bit_cnt = TRUE;
		} except_when (NOMEMORY) {
			unarray(bit_cnt, int);
			alloc_bit_cnt = FALSE;
			error("Can't get enough memory to store bit_cnt\n");
		}
		/* make all flip masks "fully enclosed" */ 
		Setup_flip_masks(0, &nvfs, p_v_flip_mask,
		                    &nffs, p_f_flip_mask);
		for(i=0; i<N; i++) {
			bit_cnt[i] = 0;
			for(j=0; j<=mod8[rn]; j++) {
				bit_cnt[i] += lbtc8[ (i >> (8*j)) & lbtc8mask ];
			}
			for(im=0; im<nvfs; im++) {
				andmask = i & p_v_flip_mask[im];
				bits_on = 0;
				for(k=0; k<=mod8[rn]; k++) 
					bits_on += lbtc8[ (andmask >> (8*k)) & lbtc8mask ];
				bit_cnt[i] += bits_on & 1;
			}
		}
	}

	
	for (pos=0; pos<num_w_pos; pos++) {
/*		left_i = 0;
		while (left_i < num_mapped_loci && map_position[left_i] <= wpos[pos]) {
			++left_i;
			
		}
		if (left_i) {
			thetal = dist_to_rec(wpos[pos]-map_position[left_i-1]);
		} else {
			thetal = 0.0;
		}
		if (left_i < num_mapped_loci) { 
			thetar = dist_to_rec(map_position[left_i]-wpos[pos]);
		} else {
			thetar = 0.0;
		}
*/		
		
		/* find nearest informative marker left/right.  
		   Here locus_l/locus_r = -1 means that left/right vector is not used.*/
		/* left */
		ilocus = 0;  locus_l = -1;  thetal = -1;
		while (ilocus < num_mapped_loci && (map_position[ilocus] <= wpos[pos] || real_eq12(map_position[ilocus], wpos[pos]))) {
		    /* real_eq12 inserted (K. Strauch 04/2004) */
		    if(!f_uninformative[ilocus]) locus_l = ilocus;
		    ilocus++;
		}
		if(locus_l != -1) {
			thetal = dist_to_rec(wpos[pos]-map_position[locus_l]);
			if (real_eq(thetal, 0.0)) thetal = 0.0;
                        /* K. Strauch 05/2004 */
		}
		if(locus_l == 0  && real_eq(thetal, 0.0)) locus_l = -1;

		/* right */
		ilocus = num_mapped_loci-1;  locus_r = -1;  thetar = -1;
		while (ilocus >=0 && (map_position[ilocus] >= wpos[pos] || real_eq12(map_position[ilocus], wpos[pos]))) {
		    /* real_eq12 inserted (K. Strauch 04/2004) */
		    if(!f_uninformative[ilocus]) locus_r = ilocus;
		    ilocus--;
		}
		if(locus_r != -1) {
			thetar = dist_to_rec(map_position[locus_r]-wpos[pos]);
			if (real_eq(thetar, 0.0)) thetar = 0.0;
			/* K. Strauch 05/2004 */
		}
		if(locus_r == num_mapped_loci-1  && real_eq(thetar, 0.0)  &&  locus_l !=-1) locus_r = -1;
		
		/* find if at locus stated in marker data file.
		   If so we might need to find haplotype */
		locus_haplo = -1;
		for(ilocus=0; ilocus<num_mapped_loci; ilocus++)
		    if(real_eq12(wpos[pos], map_position[ilocus])) locus_haplo = ilocus;  /* now real_eq12 (K. Strauch 05/2004) */

/*      print for debugging
		if(pos==0) printf("pos, wpos[pos], locus_l, thetal, locus_r, thetar, locus_haplo=\n");
		printf("%3d  %7.2lf,  %2d, %5.2lf,  %2d, %5.2lf, %3d",
		 pos, wpos[pos], locus_l, thetal, locus_r, thetar, locus_haplo);
 */

		if(locus_l==-1  &&  locus_r==-1) {
			printf(" ******Error in calc_apm_scores() did not find any informative markers *******\n");
			exit (1);
		}
		
      if (modcalc==1 && pos >= first_mod_pos && pos <= last_mod_pos) {
	really_modcalc = TRUE;
	this_p_pos = save_p_all_pos[famnum][pos-first_mod_pos];
      } else really_modcalc = FALSE;

      /***  p_score / post_pscore left out here - this saves memory!  ***/

	  like = likelihood(locus_haplo,locus_l,locus_r,
	        thetal,thetar,
			&apmstat,&variance,
			&pventropy,&lodlike,pos);

      if (lodlike <= 0.0) this_lod=-10000.0;
      else this_lod = log10(lodlike/save_unlinked[famnum_pointer[famnum]]);
      saved_lod[pos] = this_lod;

      entropy=0.0;
    
      count=0; 
      for (i=0; i<N; i++) {
	score_npl_val = Get_apm_score(i);
	if (apmstat < score_npl_val) count++;
	else if (real_eq(apmstat,score_npl_val)) count++;
      }

      p_value[ped_num][pos] = (double)count / (double)N;
      raw_stat[ped_num][pos] = apmstat;

      varnull = expscore_2 - (expscore*expscore);
      if (real_eq(varnull,0.0)) { 
        apm_stat[ped_num][pos]=0.0;
	apm_info[ped_num][pos]=0.0;
	apm_entropy[ped_num][pos] = 0.0;
      } else {
        apm_stat[ped_num][pos] = (apmstat-expscore)/sqrt(varnull);
	apm_info[ped_num][pos] = 1.0 - variance/varnull;
	apm_entropy[ped_num][pos] = 1.0 - entropy/default_entropy;
      }
      /* pventropy -= (originals*log(2)); */
      /*pvec_entropy[ped_num][pos] = 1.0 - pventropy/((n-originals)*log(2));*/

      pvec_entropy[ped_num][pos] = 1.0 - pventropy/(rn*log(2));

      if (display_scores && !scan_done) {
          if (analysis_type == NPL_ANALYSIS) {
              sf(ps,"%7.2lf  %9.6lf %9.6lf    %9.6lf\n",
                  wpos[pos], apm_stat[ped_num][pos],
                  p_value[ped_num][pos], pvec_entropy[ped_num][pos]);
              pr();
          } else if (analysis_type == LOD_ANALYSIS) {
              sf(ps,"%7.2lf  %10.6lf    %9.6lf\n",
                  wpos[pos], saved_lod[pos], pvec_entropy[ped_num][pos]);
              pr();
          } else { /* BOTH */
              sf(ps,"%7.2lf  %10.6lf    %9.6lf %9.6lf   %9.6lf\n",
                  wpos[pos], saved_lod[pos], apm_stat[ped_num][pos],
                  p_value[ped_num][pos], pvec_entropy[ped_num][pos]);
              pr();
          }
      }
    }

    if (display_scores && !scan_done) nl();
	
    if (show_expected_recs && num_mapped_loci > 1) {
      print("OBSERVED RECOMBINATION:\n");
      real_map_sum = our_map_sum = our_total = 0.0;
/*    count_recs(num_mapped_loci,like,observed_recs); */
      count_recs_2(num_mapped_loci, like, bit_cnt, observed_recs);
	  print("interval   theta   observed \n");
      j=0;
	  for (i=0; i<num_in_map_order-1; i++) {
	if(map_distance[i] != 0.) {
		obs_recs = observed_recs[j];
		l_theta  = theta[j];
		j++;
	} else {
		obs_recs = l_theta = 0.;
	}
	sf(ps,"%4d-%-4d %7.4lf  %7.4lf (%.4lf/%d)\n",map_order[i]+1,
	   map_order[i+1]+1,l_theta,obs_recs/n,obs_recs, n);
	pr();
	
	real_map_sum += rec_to_dist(l_theta);
	our_map_sum += rec_to_dist(obs_recs/(double)n);
	our_total += obs_recs;
	total_recs[i] += obs_recs;
      }
      sf(ps,"real map = %.2lf, observed map = %.2lf, total recs = %.2lf\n",real_map_sum,our_map_sum, our_total);
      pr();
    }
	haplo_viterbi = (haplo_method == 0);
    if (haplotype_output && haplo_viterbi) {
		printf("haplotype using viterbi\n");
      /* haplo_prob(num_mapped_loci); */
        compute_haplotype(num_mapped_loci);
    }
	
	if(alloc_bit_cnt) unarray(bit_cnt, int);
	alloc_bit_cnt = FALSE;
  
  } else { /*** if SINGLE_POINT_MODE ***/

    if (analysis_type == NPL_ANALYSIS) 
      print("position   NPL score  p-value   information\n");    
    else if (analysis_type == LOD_ANALYSIS)
      print("position   LOD score    information\n");    
    else
      print("position   LOD score    NPL score   p-value   information\n");    

    if(analysis_type != NPL_ANALYSIS) {
	    save_unlinked[famnum_pointer[famnum]]=0.0;
	    if(algebra)
	    {
		    for (i=0; i<N; i++) 
			    save_unlinked[famnum_pointer[famnum]] += this_par_score[i] / (double)N;
	    }else
		    for (i=0; i<N; i++) 
			    save_unlinked[famnum_pointer[famnum]] += num_par_score[i] / (double)N;
    } else {
	    save_unlinked[famnum_pointer[famnum]] = 1.;
    }

    for (pos=0; pos<num_markers; pos++) {
      apmstat=0.0; variance=0.0; pventropy=0.0; lodlike=0.0;

      pnt = nz_pointer[pos];
	  for (i=0; i<N; i++) {
	if(!f_uninformative[pos]) {
		if((i & maskTest[pos][0]) != maskTest[pos][1]) {
			p = 0;
		} else {
			p = pvct[pnt++];
		}
	} else {
		p = 1 / (double)N;
	}
	if((compute_sharing) && (p != 0.)) 
		find_sharing_probs(pos, n, i, p); /*** new - FIX 6/99 ***/	
	
	if(analysis_type != NPL_ANALYSIS)
	{	
		if(algebra)
			lodlike += p*this_par_score[i];
		else
			lodlike += p*num_par_score[i];
	}
	score_npl_val = Get_apm_score(i);
	apmstat += p*score_npl_val;
	variance += p*score_npl_val*score_npl_val;
	if (p > 0.0) 
	  pventropy += (-1.0)*p*log(p);
	  }
      variance -= (apmstat*apmstat);
      
      if (lodlike <= 0.0) this_lod=-10000.0;
      else this_lod = log10(lodlike/save_unlinked[famnum_pointer[famnum]]);
      saved_lod[pos] = this_lod;

      entropy=0.0;
    
      count=0; 
      for (i=0; i<N; i++) {
	score_npl_val = Get_apm_score(i);
	if (apmstat < score_npl_val) count++;
	else if (real_eq(apmstat,score_npl_val)) count++;
      }

      p_value[ped_num][pos] = (double)count / (double)N;
      
      raw_stat[ped_num][pos] = apmstat;

      varnull = expscore_2 - (expscore*expscore);
      if (real_eq(varnull,0.0)) { 
        apm_stat[ped_num][pos]=0.0;
	apm_info[ped_num][pos]=0.0;
	apm_entropy[ped_num][pos] = 0.0;
      } else {
        apm_stat[ped_num][pos] = (apmstat-expscore)/sqrt(varnull);
	apm_info[ped_num][pos] = 1.0 - variance/varnull;
	apm_entropy[ped_num][pos] = 1.0 - entropy/default_entropy;
      }
      /* pventropy -= (originals*log(2)); */
      pvec_entropy[ped_num][pos] = 1.0 - pventropy/(rn*log(2));

      if (analysis_type == NPL_ANALYSIS) {
	sf(ps,"%-9s %9.6lf %9.6lf   %9.6lf\n",
	   locus[pos].name, apm_stat[ped_num][pos],
	   p_value[ped_num][pos], pvec_entropy[ped_num][pos]);
	pr();
      } else if (analysis_type == LOD_ANALYSIS) {
	sf(ps,"%-9s %10.6lf    %9.6lf\n",
	   locus[pos].name, saved_lod[pos], pvec_entropy[ped_num][pos]);
	pr();
      } else {
	sf(ps,"%-9s %10.6lf    %9.6lf %9.6lf   %9.6lf\n",
	   locus[pos].name, saved_lod[pos], apm_stat[ped_num][pos],
	   p_value[ped_num][pos], pvec_entropy[ped_num][pos]);
	pr();
      }
    }
  }
}


void calc_apm_scores_sp(int ped_num,
		     double default_entropy, double *saved_lod)
{
  /* calc_apm_scores_sp formerly had 'int originals' as argument */
  int i, j, pos, count, k1, k2;
  /* int left_i; */
  int ilocus, locus_l, locus_r, locus_haplo;
  double apmstat,varnull,variance,entropy;
  volatile double like;
  double thetal_m, thetar_m, thetal_f, thetar_f, pventropy, lodlike, this_lod, our_total;
  double observed_recs[MAXLOCI], real_map_sum, our_map_sum;
  /* double dum; */
  /*** for MOD ver, unlinked removed --> *save_unlinked now is extern ***/
  int    haplo_viterbi;
  
  int     k, bits_on;
  unsigned long pnt;
  volatile int alloc_bit_cnt;
  int     nvfs, nffs, p_v_flip_mask[32], p_f_flip_mask[32], andmask, im;
  int     nffs_m, nffs_f, nfbit_sex[33];
  int * volatile bit_cnt;
  int * volatile bit_cnt_m;
  int * volatile bit_cnt_f;
  int * volatile m_wgt_loc;
  int * volatile f_wgt_loc;
  double  p, l_theta, obs_recs;
  double  score_npl_val;
  void Setup_flip_masks_sp(int, int *, int *, int *, int *, int*, int*, int*);
  double *this_par_score = NULL;

/*  printf("----compute_sharing = %1d \n", compute_sharing);
*/
  if(this_mod_pos==-1 && compute_sharing) init_sharing_probs(N, n);
  
   /* determine k_full_rep[k_foundersym_rep] */
  k2 = 0;
  for (k1=0; k1<n; k1++) {
     if((uninf_mask & m_pow2[k1]) != 0) continue;
     k_full[k2] = k1;
     k2++;
  }

/****** start of inserted part for modcalc single (K. Strauch 04/2004) ******/

  if (this_mod_pos != -1) {  /** i.e., call from modscore (modcalc single **/

    num_mapped_loci = f_num_in_map_order;
    Compute_lr_vectors_sp();

    array(bit_cnt, 1, int);
    alloc_bit_cnt = TRUE;
    
    run {
	array (m_wgt, N, int);
	array (f_wgt, N , int); 
	array (m_wgt_s, N, int);
	array (f_wgt_s, N , int); 
    } except_when (NOMEMORY) {
	unarray(m_wgt, int);
	unarray(f_wgt, int); 
	unarray(m_wgt_s, int);
	unarray(f_wgt_s, int); 
	error("Can't get enough memory for this scan (local vars)\n");
    }
    
    /* find nearest informative marker left/right.  
       Here locus_l/locus_r = -1 means that left/right vector is not used.*/
    /* left */
    
    ilocus = 0;  locus_l = -1;  thetal_m = -1; thetal_f = -1;
    while (ilocus < num_mapped_loci && (map_position_m[ilocus] <= wpos_m[this_mod_pos] || real_eq12(map_position_m[ilocus], wpos_m[this_mod_pos]))) {
	/* real_eq12 inserted (K. Strauch 04/2004) */
	if(!f_uninformative[ilocus]) locus_l = ilocus;
	ilocus++;
    }
    if(locus_l != -1) {
	thetal_m = dist_to_rec(wpos_m[this_mod_pos]-map_position_m[locus_l]);
	thetal_f = dist_to_rec(wpos_f[this_mod_pos]-map_position_f[locus_l]);
	if (real_eq(thetal_m, 0.0)) thetal_m = 0.0;  /* K. Strauch 05/2004 */
        if (real_eq(thetal_f, 0.0)) thetal_f = 0.0;  /* K. Strauch 05/2004 */
    }
    if(locus_l == 0  && real_eq(min(thetal_m,thetal_f), 0.0)) locus_l = -1;
    
    

    /* right */
    ilocus = num_mapped_loci-1;  locus_r = -1;  thetar_m = -1; thetar_f = -1;
    while (ilocus >=0 && (map_position_m[ilocus] >= wpos_m[this_mod_pos] || real_eq12(map_position_m[ilocus], wpos_m[this_mod_pos]))) {
	/* real_eq12 inserted (K. Strauch 04/2004) */
	if(!f_uninformative[ilocus]) locus_r = ilocus;
	ilocus--;
    }
    if(locus_r != -1) {
	thetar_m = dist_to_rec(map_position_m[locus_r]-wpos_m[this_mod_pos]);
  	thetar_f = dist_to_rec(map_position_f[locus_r]-wpos_f[this_mod_pos]);
	if (real_eq(thetar_m, 0.0)) thetar_m = 0.0;  /* K. Strauch 05/2004 */
        if (real_eq(thetar_f, 0.0)) thetar_f = 0.0;  /* K. Strauch 05/2004 */
    }
    if(locus_r == num_mapped_loci-1  && real_eq(min(thetar_m, thetar_f),0.)  &&  locus_l !=-1) locus_r = -1;

    /* find if at locus stated in marker data file.
       If so we might need to find haplotype */
    /* not needed for call from modscore (modcalc single),
       but still left here for completeness (K. Strauch 04/2004) */
    locus_haplo = -1;
    for(ilocus=0; ilocus<num_mapped_loci; ilocus++)
	if(real_eq12(wpos_m[this_mod_pos], map_position_m[ilocus]))
	    locus_haplo = ilocus;  /* now real_eq12 (K. Strauch 05/2004) */

    if(locus_l==-1  &&  locus_r==-1) {
	printf(" ******Error in calc_apm_scores() did not find any informative markers *******\n");
	exit (1);
    }

    really_modcalc = TRUE;
    this_p_pos = save_p_one_pos[famnum];
    if(algebra)
    this_par_score = datfile_par_score[algebra_pointer[famnum]];

    like = likelihood_sp(locus_haplo,locus_l,locus_r,
                      thetal_m,thetar_m,thetal_f,thetar_f,
		      &apmstat,&variance,
		      &pventropy,&lodlike,this_mod_pos);
    /*** bit_cnt not used at all in funct likelihood! (K. Strauch) ***/
    
    
    if(alloc_bit_cnt) unarray(bit_cnt, int);
    alloc_bit_cnt = FALSE;
    
    s_wgt_free(m_wgt, f_wgt);
    s_wgt_free(m_wgt_s, f_wgt_s);
  
/******  end of inserted part for modcalc single (K. Strauch 04/2004) ******/

  } else if (!single_point_mode) {
    num_mapped_loci = f_num_in_map_order;

/*	printf("compute l/r vectors \n");
*/
    /*compute left vectors*/
    /*r-full-alloc  unlinked = left(theta, num_mapped_loci); */
    /*compute right vectors*/
    /*r-full-alloc  dum = right(theta, num_mapped_loci-1); */

    save_unlinked[famnum_pointer[famnum]] = Compute_lr_vectors_sp();
/*	printf("unlinked=%12.4e \n", unlinked); */

	/* printf("search_vector \n"); */
	/* search_vector(); */    /* search p/l/rvector for unique values */
  
    if (display_scores && !scan_done) {
      if (analysis_type == NPL_ANALYSIS) 
        print("pos.(avg. female  male)   NPL score  p-value    information\n");
      else if (analysis_type == LOD_ANALYSIS) 
        print("pos.(avg. female  male)   LOD score    information\n");
      else /* BOTH */
        print("pos.(avg. female  male)   LOD score    NPL score p-value     information\n");
    }
    
    
         run {
	         array (m_wgt, N, int);
	         array (f_wgt, N , int); 
		 array (m_wgt_s, N, int);
	         array (f_wgt_s, N , int); 
        } except_when (NOMEMORY) {
	         unarray(m_wgt, int);
	         unarray(f_wgt, int); 
		 unarray(m_wgt_s, int);
		 unarray(f_wgt_s, int); 
	         error("Can't get enough memory for this scan (local vars)\n");
        }
		
	/* prepare bitcount for fft */
	j = 0;
	for(i=0; i<num_mapped_loci; i++)
		if(!f_uninformative[i]) j++;
	/* if any one of the loci is uninformative we need bit_cnt[] */
	if(!show_expected_recs){
		array(bit_cnt, 1, int);
		array(bit_cnt_m, 1, int);
		array(bit_cnt_f, 1, int);
		alloc_bit_cnt = TRUE;
	} else {
		run {
			array(bit_cnt, N, int);
			array(bit_cnt_m, N, int);
			array(bit_cnt_f, N, int);
			alloc_bit_cnt = TRUE;
		} except_when (NOMEMORY) {
			unarray(bit_cnt, int);
			unarray(bit_cnt_m, int);
			unarray(bit_cnt_f, int);
			alloc_bit_cnt = FALSE;
			error("Can't get enough memory to store bit_cnt\n");
		}
		
		run {
	         array (m_wgt_loc, N, int);
	         array (f_wgt_loc, N , int); 
		} except_when (NOMEMORY) {
	         unarray(m_wgt_loc, int);
	         unarray(f_wgt_loc, int); 
		 error("Can't get enough memory for this scan (local vars)\n");
                }
		/* make all flip masks "fully enclosed"   */
		Setup_flip_masks_sp(0, &nvfs, p_v_flip_mask,
		                    &nffs, &nffs_m, &nffs_f, nfbit_sex, p_f_flip_mask);
				    

		init_sex_weights(0,m_wgt_loc, f_wgt_loc);

		for (i=0; i<N; i++) {
		   bit_cnt_m[i] = m_wgt_loc[i];
		   bit_cnt_f[i] = f_wgt_loc[i];
		   for (im=0; im<nvfs; im++) {
		       andmask = i & p_v_flip_mask[im];
		       bit_cnt_m[i] += m_wgt_loc[andmask] & 1;
		       bit_cnt_f[i] += f_wgt_loc[andmask] & 1;
		    }
		}				    
				    
			    
		for(i=0; i<N; i++) {
			bit_cnt[i] = 0;
			for(j=0; j<=mod8[rn]; j++) {
				bit_cnt[i] += lbtc8[ (i >> (8*j)) & lbtc8mask ];
			}
			for(im=0; im<nvfs; im++) {
				andmask = i & p_v_flip_mask[im];
				bits_on = 0;
				for(k=0; k<=mod8[rn]; k++) 
					bits_on += lbtc8[ (andmask >> (8*k)) & lbtc8mask ];
				bit_cnt[i] += bits_on & 1;
			}
		}
	
	}
	
	
	
	
	for (pos=0; pos<num_w_pos; pos++) {
/*		left_i = 0;
		while (left_i < num_mapped_loci && map_position[left_i] <= wpos[pos]) {
			++left_i;
			
		}
		if (left_i) {
			thetal = dist_to_rec(wpos[pos]-map_position[left_i-1]);
		} else {
			thetal = 0.0;
		}
		if (left_i < num_mapped_loci) { 
			thetar = dist_to_rec(map_position[left_i]-wpos[pos]);
		} else {
			thetar = 0.0;
		}
*/		
		
		/* find nearest informative marker left/right.  
		   Here locus_l/locus_r = -1 means that left/right vector is not used.*/
		/* left */
		ilocus = 0;  locus_l = -1;  thetal_m = -1; thetal_f = -1;
		while (ilocus < num_mapped_loci && (map_position_m[ilocus] <= wpos_m[pos] || real_eq12(map_position_m[ilocus], wpos_m[pos]))) {
		    /* real_eq12 inserted (K. Strauch 04/2004) */
		    if(!f_uninformative[ilocus]) locus_l = ilocus;
		    ilocus++;
		}
		if(locus_l != -1) {
			thetal_m = dist_to_rec(wpos_m[pos]-map_position_m[locus_l]);
			thetal_f = dist_to_rec(wpos_f[pos]-map_position_f[locus_l]);
			if (real_eq(thetal_m, 0.0)) thetal_m = 0.0;
			if (real_eq(thetal_f, 0.0)) thetal_f = 0.0;
                        /* K. Strauch 05/2004 */
		}
		if(locus_l == 0  && real_eq(min(thetal_m,thetal_f),0.)) locus_l = -1;
		
		
		/* right */
		ilocus = num_mapped_loci-1;  locus_r = -1;  thetar_m = -1; thetar_f = -1;
		while (ilocus >=0 && (map_position_m[ilocus] >= wpos_m[pos] || real_eq12(map_position_m[ilocus], wpos_m[pos]))) {
		    /* real_eq12 inserted (K. Strauch 04/2004) */
		    if(!f_uninformative[ilocus]) locus_r = ilocus;
		    ilocus--;
		}
		if(locus_r != -1){ 
			thetar_m = dist_to_rec(map_position_m[locus_r]-wpos_m[pos]);
			thetar_f = dist_to_rec(map_position_f[locus_r]-wpos_f[pos]);
			if (real_eq(thetar_m, 0.0)) thetal_m = 0.0;
			if (real_eq(thetar_f, 0.0)) thetal_f = 0.0;
                        /* K. Strauch 05/2004 */
		}	
		if(locus_r == num_mapped_loci-1  && real_eq(min(thetar_m, thetar_f),0.0)  &&  locus_l !=-1) locus_r = -1;
		
				
		
		/* find if at locus stated in marker data file.
		   If so we might need to find haplotype */
		locus_haplo = -1;
		for(ilocus=0; ilocus<num_mapped_loci; ilocus++)
		    if(real_eq12(wpos_m[pos], map_position_m[ilocus])) locus_haplo = ilocus;  /* now real_eq12 (K. Strauch 05/2004) */

/*      print for debugging
		if(pos==0) printf("pos, wpos[pos], locus_l, thetal, locus_r, thetar, locus_haplo=\n");
		printf("%3d  %7.2lf,  %2d, %5.2lf,  %2d, %5.2lf, %3d",
		 pos, wpos[pos], locus_l, thetal, locus_r, thetar, locus_haplo);
 */

		if(locus_l==-1  &&  locus_r==-1) {
			printf(" ******Error in calc_apm_scores() did not find any informative markers *******\n");
			exit (1);
		}
		
      if (modcalc==1 && pos >= first_mod_pos && pos <= last_mod_pos) {
	really_modcalc = TRUE;
	this_p_pos = save_p_all_pos[famnum][pos-first_mod_pos];
      } else really_modcalc = FALSE;

      /***  p_score / post_pscore left out here - this saves memory!  ***/
      
         	like = likelihood_sp(locus_haplo,locus_l,locus_r,
	        thetal_m,thetar_m,thetal_f,thetar_f,
			&apmstat,&variance,
			&pventropy,&lodlike,pos);		
	  /*** bit_cnt not used at all in funct likelihood! (K. Strauch) ***/
	  
      if (lodlike <= 0.0) this_lod=-10000.0;
      else this_lod = log10(lodlike/save_unlinked[famnum_pointer[famnum]]);
      saved_lod[pos] = this_lod;

      entropy=0.0;
    
      count=0; 
      for (i=0; i<N; i++) {
	score_npl_val = Get_apm_score(i);
	if (apmstat < score_npl_val) count++;
	else if (real_eq(apmstat,score_npl_val)) count++;
      }

      p_value[ped_num][pos] = (double)count / (double)N;
      raw_stat[ped_num][pos] = apmstat;

      varnull = expscore_2 - (expscore*expscore);
      if (real_eq(varnull,0.0)) { 
        apm_stat[ped_num][pos]=0.0;
	apm_info[ped_num][pos]=0.0;
	apm_entropy[ped_num][pos] = 0.0;
      } else {
        apm_stat[ped_num][pos] = (apmstat-expscore)/sqrt(varnull);
	apm_info[ped_num][pos] = 1.0 - variance/varnull;
	apm_entropy[ped_num][pos] = 1.0 - entropy/default_entropy;
      }
      /* pventropy -= (originals*log(2)); */
      /*pvec_entropy[ped_num][pos] = 1.0 - pventropy/((n-originals)*log(2));*/

      pvec_entropy[ped_num][pos] = 1.0 - pventropy/(rn*log(2));

      if (display_scores && !scan_done) {
          if (analysis_type == NPL_ANALYSIS) {
              sf(ps,"%7.2lf %7.2lf %7.2lf  %9.6lf  %9.6lf  %9.6lf\n",
                  wpos[pos], wpos_f[pos], wpos_m[pos], apm_stat[ped_num][pos],
                  p_value[ped_num][pos], pvec_entropy[ped_num][pos]);
              pr();
          } else if (analysis_type == LOD_ANALYSIS) {
              sf(ps,"%7.2lf %7.2lf %7.2lf %10.6lf    %9.6lf\n",
                  wpos[pos], wpos_f[pos],wpos_m[pos], saved_lod[pos], pvec_entropy[ped_num][pos]);
              pr();
          } else { /* BOTH */
              sf(ps,"%7.2lf %7.2lf %7.2lf %10.6lf    %9.6lf %9.6lf   %9.6lf\n",
                  wpos[pos], wpos_f[pos], wpos_m[pos], saved_lod[pos], apm_stat[ped_num][pos],
                  p_value[ped_num][pos], pvec_entropy[ped_num][pos]);
              pr();
          }
      }
    }
    
    		
    s_wgt_free(m_wgt, f_wgt);
    s_wgt_free(m_wgt_s, f_wgt_s);

    if (display_scores && !scan_done) nl();
	
    if (show_expected_recs && num_mapped_loci > 1) {
      print("OBSERVED RECOMBINATION:\n");
      real_map_sum = our_map_sum = our_total = 0.0;
/*    count_recs(num_mapped_loci,like,observed_recs); */
      count_recs_2_sp(num_mapped_loci, like, bit_cnt_m, 
                         bit_cnt_f, observed_recs);
     
      print("interval   theta   observed \n");
      j=0;
      for (i=0; i<num_in_map_order-1; i++) {
	if(map_distance[i] != 0.) {
		obs_recs = observed_recs[j];
        	l_theta  = theta[j];
		j++;
	} else {
		obs_recs = l_theta = 0.;
	}
	sf(ps,"%4d-%-4d %7.4lf  %7.4lf (%.4lf/%d)\n",map_order[i]+1,
	   map_order[i+1]+1,l_theta,obs_recs/n,obs_recs, n);
	pr();
	
	real_map_sum += rec_to_dist(l_theta);
	our_map_sum += rec_to_dist(obs_recs/(double)n);
	our_total += obs_recs;
	total_recs[i] += obs_recs;
      }
      sf(ps,"real map = %.2lf, observed map = %.2lf, total recs = %.2lf\n",real_map_sum,our_map_sum, our_total);
      pr();
      s_wgt_free(m_wgt_loc, f_wgt_loc);
    }
	haplo_viterbi = (haplo_method == 0);
    if (haplotype_output && haplo_viterbi) {
		printf("haplotype using viterbi\n");
      /* haplo_prob(num_mapped_loci); */
        compute_haplotype(num_mapped_loci);
    }
	
	if(alloc_bit_cnt){
	  unarray(bit_cnt, int);
	  unarray(bit_cnt_m, int);
	  unarray(bit_cnt_f, int);
	}  
	alloc_bit_cnt = FALSE;
  
  } else { /*** if SINGLE_POINT_MODE ***/

    if (analysis_type == NPL_ANALYSIS) 
      print("position   NPL score  p-value   information\n");    
    else if (analysis_type == LOD_ANALYSIS)
      print("position   LOD score    information\n");    
    else
      print("position   LOD score    NPL score   p-value   information\n");    

    if(analysis_type != NPL_ANALYSIS) {
	    save_unlinked[famnum_pointer[famnum]]=0.0;
        for (i=0; i<N; i++) {
            if(algebra)
            { 
                save_unlinked[famnum_pointer[famnum]] += this_par_score[i] / (double)N;
            } else
                save_unlinked[famnum_pointer[famnum]] += num_par_score[i] / (double)N;
        }
    } else {
	    save_unlinked[famnum_pointer[famnum]] = 1.0;
    }

    for (pos=0; pos<num_markers; pos++) {
      apmstat=0.0; variance=0.0; pventropy=0.0; lodlike=0.0;

      pnt = nz_pointer[pos];
	  for (i=0; i<N; i++) {
	if(!f_uninformative[pos]) {
		if((i & maskTest[pos][0]) != maskTest[pos][1]) {
			p = 0;
		} else {
			p = pvct[pnt++];
		}
	} else {
		p = 1 / (double)N;
	}
	if((compute_sharing) && (p != 0.)) 
		find_sharing_probs(pos, n, i, p); /*** new - FIX 6/99 ***/	
	
	if(analysis_type != NPL_ANALYSIS) 
	{
		if(algebra)
			lodlike += p*this_par_score[i];
		else
			lodlike += p*num_par_score[i];
	}
	score_npl_val = Get_apm_score(i);
	apmstat += p*score_npl_val;
	variance += p*score_npl_val*score_npl_val;
	if (p > 0.0) 
	  pventropy += (-1.0)*p*log(p);
	  }
      variance -= (apmstat*apmstat);
      
      if (lodlike <= 0.0) this_lod=-10000.0;
      else this_lod = log10(lodlike/save_unlinked[famnum_pointer[famnum]]);
      saved_lod[pos] = this_lod;

      entropy=0.0;
    
      count=0; 
      for (i=0; i<N; i++) {
	score_npl_val = Get_apm_score(i);
	if (apmstat < score_npl_val) count++;
	else if (real_eq(apmstat,score_npl_val)) count++;
      }

      p_value[ped_num][pos] = (double)count / (double)N;
      
      raw_stat[ped_num][pos] = apmstat;

      varnull = expscore_2 - (expscore*expscore);
      if (real_eq(varnull,0.0)) { 
        apm_stat[ped_num][pos]=0.0;
	apm_info[ped_num][pos]=0.0;
	apm_entropy[ped_num][pos] = 0.0;
      } else {
        apm_stat[ped_num][pos] = (apmstat-expscore)/sqrt(varnull);
	apm_info[ped_num][pos] = 1.0 - variance/varnull;
	apm_entropy[ped_num][pos] = 1.0 - entropy/default_entropy;
      }
      /* pventropy -= (originals*log(2)); */
      pvec_entropy[ped_num][pos] = 1.0 - pventropy/(rn*log(2));

      if (analysis_type == NPL_ANALYSIS) {
	sf(ps,"%-9s %9.6lf %9.6lf   %9.6lf\n",
	   locus[pos].name, apm_stat[ped_num][pos],
	   p_value[ped_num][pos], pvec_entropy[ped_num][pos]);
	pr();
      } else if (analysis_type == LOD_ANALYSIS) {
	sf(ps,"%-9s %10.6lf    %9.6lf\n",
	   locus[pos].name, saved_lod[pos], pvec_entropy[ped_num][pos]);
	pr();
      } else {
	sf(ps,"%-9s %10.6lf    %9.6lf %9.6lf   %9.6lf\n",
	   locus[pos].name, saved_lod[pos], apm_stat[ped_num][pos],
	   p_value[ped_num][pos], pvec_entropy[ped_num][pos]);
	pr();
      }
    }
  }
}

/*===============================================================*/

void count_recs_2(int num_loci, double like, int *bit_cnt, double *observed_recs)
{
	int i, loc, lloc, rloc, no_info;
	double rectotal, l_theta;
	double * volatile p_l;
	double * volatile p_r;
	void copy_vector(double *, int, double *);
	void convolve_fft_2(int, double, int *, double *);
	void convolve_count_fft_2(int, double, int *, double *);

	run { 
		array(p_l, N, double); 
		array(p_r, N, double); 
	} except_when (NOMEMORY) {
		unarray(p_r, double);
		unarray(p_l, double);
		error("Can't get enough memory to store p_l[], p_r[] in count_recs\n");
	}
	
	for (loc=0; loc < num_loci-1; loc++) {
	
		no_info = FALSE;
		
		/* Compute left vectoror in interval */
		if(!f_uninformative[loc]) {
			copy_vector(lvct, loc, p_l);
		} else {
			for(lloc=loc-1; lloc>=0; lloc--)
				if(!f_uninformative[lloc]) break;
			if(lloc>=0) {
				l_theta = dist_to_rec(map_position[loc] - map_position[lloc]);
				copy_vector(lvct, lloc, p_l);
				convolve_fft_2(rn, l_theta, bit_cnt, p_l);
			} else {
				no_info = TRUE;
			}
		}
		
		/* convolve_count left vectoror in interval */
		convolve_count_fft_2(rn, theta[loc], bit_cnt, p_l);
		
		/* Compute right vectoror in interval */
		if(!f_uninformative[loc+1]) {
			copy_vector(rvct, loc+1, p_r);
		} else {
			for(rloc=loc+1; rloc<num_loci; rloc++)
				if(!f_uninformative[rloc]) break;
			if(rloc<num_loci) {
				copy_vector(rvct, rloc, p_r);
				l_theta = dist_to_rec(map_position[rloc] - map_position[loc+1]);
				convolve_fft_2(rn, l_theta, bit_cnt, p_r);
			} else {
				no_info = TRUE;
			}
		}
		
		rectotal=0.0;
		if(no_info) {
			observed_recs[loc] = theta[loc] * n;
		} else {
			for(i=0; i<N; i++)
				rectotal += p_l[i] * p_r[i];
			observed_recs[loc] = rectotal/like;
		}
	}
	unarray(p_r, double);
	unarray(p_l, double);
}

void count_recs_2_sp(int num_loci, double like, int* bit_cnt_m,
                int* bit_cnt_f, double *observed_recs)
{
	int i, loc, lloc, rloc, no_info;
	double rectotal, l_theta_m, l_theta_f;
	double * volatile p_l;
	double * volatile p_r;
	void copy_vector(double *, int, double *);
	void convolve_fft_2_sp(int nbits, double theta_m, double theta_f, 
	                       int *bit_cnt_m, int *bit_cnt_f, double *pv);
	void convolve_count_fft_2_sp(int nbits, double theta_m, double theta_f, 
	                             int *bit_cnt_m, int *bit_cnt_f, double *pv);

	run { 
		array(p_l, N, double); 
		array(p_r, N, double); 
	} except_when (NOMEMORY) {
		unarray(p_r, double);
		unarray(p_l, double);
		error("Can't get enough memory to store p_l[], p_r[] in count_recs\n");
	}
	
	for (loc=0; loc < num_loci-1; loc++) {
	
		no_info = FALSE;
		
		/* Compute left vectoror in interval */
		if(!f_uninformative[loc]) {
			copy_vector(lvct, loc, p_l);
		} else {
			for(lloc=loc-1; lloc>=0; lloc--)
				if(!f_uninformative[lloc]) break;
			if(lloc>=0) {
				l_theta_m = dist_to_rec(map_position_m[loc] - map_position_m[lloc]);
				l_theta_f = dist_to_rec(map_position_f[loc] - map_position_f[lloc]);				
				copy_vector(lvct, lloc, p_l);
				convolve_fft_2_sp(rn, l_theta_m, l_theta_f, bit_cnt_m, bit_cnt_f, p_l);
				
			} else {
				no_info = TRUE;
			}
		}
		
		/* convolve_count left vectoror in interval */
/*		convolve_count_fft_2(rn, theta[loc], bit_cnt, p_l);*/
                convolve_count_fft_2_sp(rn, theta_m[loc], theta_f[loc],
		                        bit_cnt_m, bit_cnt_f, p_l);
		
		/* Compute right vectoror in interval */
		if(!f_uninformative[loc+1]) {
			copy_vector(rvct, loc+1, p_r);
		} else {
			for(rloc=loc+1; rloc<num_loci; rloc++)
				if(!f_uninformative[rloc]) break;
			if(rloc<num_loci) {
				copy_vector(rvct, rloc, p_r);
				l_theta_m = dist_to_rec(map_position_m[rloc] - map_position_m[loc+1]);
				l_theta_f = dist_to_rec(map_position_f[rloc] - map_position_f[loc+1]);
				convolve_fft_2_sp(rn, l_theta_m, l_theta_f, bit_cnt_m, bit_cnt_f, p_r);
			} else {
				no_info = TRUE;
			}
		}
		
		rectotal=0.0;
		if(no_info) {
			observed_recs[loc] = theta[loc] * n;
		} else {
			for(i=0; i<N; i++)
				rectotal += p_l[i] * p_r[i];
			observed_recs[loc] = rectotal/like;
		}
	}
	unarray(p_r, double);
	unarray(p_l, double);
}

void copy_vector(double *vct, int loc, double *pv)
{
	int i;
	unsigned long pnt;

	pnt = nz_pointer[loc];
	for(i=0; i<N; i++) {
		if((i & maskTest[loc][0]) == maskTest[loc][1]) 
			{ pv[i] = vct[pnt++]; } else { pv[i] = 0; }
	}
}

void convolve_count_fft_2(int nbits, double theta, int *bit_cnt, double *pv)
{
	int i, Nstates;
	double trans[100], factor;
	void fft_2(int, double *);

	Nstates = m_pow2[nbits];
	trans[0] = 1.0;
	factor = 1.0 - 2.0*theta;
	for (i=1; i<100; ++i)
		trans[i] = trans[i-1]*factor;

	fft_2(nbits, pv);
	for(i=0; i<Nstates; i++)
    		pv[i] *= (theta*(n-bit_cnt[i])*trans[bit_cnt[i]] -
		theta*bit_cnt[i]*
		trans[((bit_cnt[i]-1)>=0) ? (bit_cnt[i]-1) : 0]);
	fft_2(nbits, pv);
	for(i=0; i<Nstates; i++) {
		pv[i] /= Nstates;
		if (pv[i] < 0.0) pv[i] = -pv[i];
	}
}

void convolve_count_fft_2_sp(int nbits, double theta_m, double theta_f, int *bit_cnt_m, int *bit_cnt_f, double *pv)
{	
	int i, Nstates;
	double trans_m[100], trans_f[100], factor_m, factor_f, T_s_m, T_s_f, T_m, T_f;
	void fft_2(int, double *);

	Nstates = m_pow2[nbits];

	trans_m[0] = 1.0;
	trans_f[0] = 1.0;
	factor_m = 1.0 - 2.0*theta_m;
	factor_f = 1.0 - 2.0*theta_f;
	for (i=1; i<100; ++i){
		trans_m[i] = trans_m[i-1]*factor_m;
	        trans_f[i] = trans_f[i-1]*factor_f;
        }
	
	fft_2(nbits, pv);
	
	for (i=0; i<Nstates; i++){
	        
		T_s_m = theta_m*(n_m-bit_cnt_m[i])*trans_m[bit_cnt_m[i]] -
			theta_m*bit_cnt_m[i]*
			trans_m[((bit_cnt_m[i]-1)>=0) ? (bit_cnt_m[i]-1) : 0];
		T_s_f = theta_f*(n_f-bit_cnt_f[i])*trans_f[bit_cnt_f[i]] -
			theta_f*bit_cnt_f[i]*
			trans_f[((bit_cnt_f[i]-1)>=0) ? (bit_cnt_f[i]-1) : 0];
		T_m = trans_m[bit_cnt_m[i]];
		T_f = trans_f[bit_cnt_f[i]];
		
		pv[i] *= (T_s_m*T_f+T_s_f*T_m);
		 
	}		 
		
	fft_2(nbits, pv);
	for(i=0; i<Nstates; i++) {
	    pv[i] /= Nstates;
	    if (pv[i] < 0.0) pv[i] = -pv[i];
	}

}


/*===============================================================*/

void count_recs(int num_loci, double like, double *observed_recs)
{
    int i,loc;
    double rectotal;

    for (loc=0; loc < num_loci-1; loc++) {

        rectotal=0.0;

        for (i=0; i<N; ++i)
	  tempvector[i] = lvector[loc][i];

	convolve_count_fft(theta[loc], tempvector);

	for (i=0; i<N; ++i) rectotal += rvector[loc+1][i]*tempvector[i];

	observed_recs[loc] = rectotal/like;
    }
}

int **aux_index;

void compute_haplotype(int num_loci)
{
    int i, j, loc, **path_matrix, max_i, vec, height, temp, bit;
    double *p, max_p, *t;
    /* FILE *fp; */
    char *binary_output(int);
    int tmp_max_i, last_max_i, reverse_mask(int,int);
    /* int best_mask; */
    /* int num_steps; */
    int ngnr;
    unsigned long pnt;
    
    /*if (ind_sp==1){
      ngnr = n_m*n_f;
      array(t, n_m*n_f+2, double);
      array(t, MAXLOCI, double);
      matrix (aux_index, N, n_m*n_f+2, int);
      matrix(aux, N, n_m*n_f+2, double);
    }else{*/
      ngnr = n; 
      array(t, MAXLOCI, double);
      matrix (aux_index, N, 30, int);
      matrix(aux, N, 30, double);
    /* } */

    array(p, N, double);
    matrix (path_matrix, num_loci, N, int);    
    array(row_partner, N, int);

/*    for (i=0; i<N; i++) p[i]=pvector[0][i];
*/
/*	printf(" Computing Viterbi, num_loci = %d\n", num_loci); */
	pnt = nz_pointer[0];
	for(i=0; i<N; i++) {
		if(f_uninformative[0]) {
			p[i] = 1 / (double)N;
		} else {
			if((i & maskTest[0][0]) != maskTest[0][1]) {
				p[i] = 0;
			} else {
				p[i] = pvct[pnt++];
			}
		}
	}

    for (loc=1; loc<num_loci; loc++) {
        if (ind_sp==1){
	 makerec(theta[loc-1], t);
	  /* makerec_sp(theta[loc-1], theta_m[loc-1], theta_f[loc-1], t); */
	}else{
	  makerec(theta[loc-1], t);
	}    
	
	for (i=0; i<N; i++) 
	  for (j=0; j<ngnr+1; j++) {
	    aux[i][j] = p[i]*t[j];
	    aux_index[i][j] = -1;
	  }
	
	haplo_reduce(bigN,ngnr+1);

	pnt = nz_pointer[loc];
	for (i=0; i<N; i++) {
/*	    p[i] = aux[i][0]*pvector[loc][i];
*/		
		if(f_uninformative[loc]) {
			p[i] = aux[i][0] * (1/(double)N);
		} else {
			if((i & maskTest[loc][0]) != maskTest[loc][1]) {
				p[i] = 0;
			} else {
				p[i] = aux[i][0] * pvct[pnt++];
			}
		}
	    /* path_matrix[loc][i] = aux_index[i][0]; */

	    height=1;  temp=aux_index[i][0];
	    vec=pad_inhvec(i);
	    for (j=0; j<ngnr; j++) {
	      bit = temp%2;
	      temp >>= 1;
	      height *= 2;
	      if (bit) {
		  vec = height*(vec/height) + (height-1) - (vec%height);
		  /* vec = lookup_mask(vec); */
	      }
	    }
	    path_matrix[loc][i]=vec;
	}
    }

    /* get largest at end */
    /* fp = open_file("newvectors.out","w"); */
    /* sf(ps,"MAXIMUM LIKELIHOOD HAPLOTYPE:\n"); fpr(fp); */
    max_p=0.0; max_i=-1;
    for (i=0; i<N; i++) {
      if (p[i] > max_p) { max_p=p[i]; max_i=i; }
    }

    max_i = pad_inhvec(max_i); /* the critical missing link */

    /* sf(ps,"max p[%d] = %.10g\n",max_i,max_p); fpr(fp); */

    /* walk back from largest through maximal path */

    for (loc=num_loci-1; loc>=0; loc--) {
      best_pvector[loc] = max_i;
      last_max_i = max_i;
      tmp_max_i = lookup_mask(max_i);
      max_i = path_matrix[loc][tmp_max_i];
      max_i = reverse_mask(max_i, last_max_i);
    }

/*    sf(ps,"marker %d: %s\n",map_order[0]+1,binary_output(best_pvector[0])); 
      fpr(fp);  */

/**********
    for (loc=1; loc<num_loci; loc++) {
      num_steps = min_diff_bits(best_pvector[loc],best_pvector[loc-1],
				&best_mask);
      sf(ps,"marker %d: %s  (%d recs)\n", map_order[loc]+1,
	 binary_output(best_pvector[loc]), num_steps);
      fpr(fp);
      diff_bits(best_pvector[loc],best_pvector[loc-1]));
    }
**********/

/*     close_file(fp); */

    unarray(row_partner, int);
    unarray(t, double);
    unmatrix(aux, N, double);
    unmatrix(aux_index, N, int);
    unmatrix(path_matrix, num_loci, int);
    unarray(p, double);

}    

void haplo_prob(int num_loci)
{
    int num_recs, loc;
    double t[MAXLOCI], total_prob;
    char *binary_output(int);
    FILE *fp;

    fp = open_file("newvectors.out","w");
    
    sf(ps,"BEST p[i], MINIMUM RECS HAPLOTYPE:\n"); fpr(fp);
    sf(ps,"marker %d: %s\n",f_map_order[0]+1,binary_output(best_pvector[0])); 
    fpr(fp);

    total_prob=pvector[0][best_pvector[0]];

    for (loc=1; loc<num_loci; loc++) {
        makerec(theta[loc-1], t);
	num_recs = diff_bits(best_pvector[loc-1], best_pvector[loc]);
	total_prob *= t[num_recs];
	total_prob *= pvector[loc][best_pvector[loc]];
	sf(ps,"marker %d: %s  (%d recs)\n", f_map_order[loc]+1,
	 binary_output(best_pvector[loc]), num_recs);
	fpr(fp);
    }
    sf(ps,"total prob = %.10g\n",total_prob); fpr(fp);

    close_file(fp);
}



int running_bestrecs[MAXLOCI], running_secondrecs[MAXLOCI];
int secondbest_pvector[MAXLOCI];

double likelihood(int locus_haplo, int locus_l, int locus_r, 
		  double thetal, double thetar,
		  double *apmstat, double *variance, 
		  double *pvec_entropy, double *ret_lodlike,
		  int pos_in_scan)
{

	int i, num_maxlike, best_new_choice, best_num_steps, num_steps;
	double p, plod, maxlike, minlike, totlike;
	double lodlike, like, statval, varsum, entropy;
	static FILE *fp;
	static int like_fp_open=TRUE;
	static int choice;
	static double first_prob, second_prob;
	char *binary_output(int);
/*	int skip_this_haplo_code = TRUE, best_mask, best_i;
    km for checking alt. haplo code */
	int skip_this_haplo_code = FALSE, best_mask;
	volatile int ilocus;
	volatile unsigned long pnt;
	double un_normalized_entropy;
	volatile double pvct_max;
	volatile double pvct_value;
	double score_npl_val;
	int    at_informative_marker;
	double *tmpdot;
	double *this_par_score = NULL;
	if(algebra)
	this_par_score = datfile_par_score[algebra_pointer[famnum]];
	TRANSSTRUCT tsleft, tsright;
	TRDOTSTRUCT tdot;

	if (this_mod_pos==-1 && like_fp_open && !skip_this_haplo_code) {
	  fp=open_file("vectors.out","w"); like_fp_open=FALSE;
	}
	
	like = lodlike = 0.0; statval = 0.0; varsum = 0.0;
	maxlike=0.0; 
	minlike=1.0;
	un_normalized_entropy = 0;
	at_informative_marker = (thetal == 0.0) || (thetar == 0.0);
	tmpdot = (double *) 0;


	if ( !at_informative_marker ) {
		/* need transition away from informative marker */
		/* compute prob. per inheritance vector and store in tmp1[] */

	  /* for Haplotyping between markers, but slows things down */
	  if (this_mod_pos==-1 && haplotype_output && (locus_haplo != -1) && (haplo_method == 1)) {
		 run {
			array(tmpdot, N, double);
		 } except_when (NOMEMORY) {
			unarray(tmpdot, double);
			tmpdot = (double *) 0;
			error("Can't get enough memory to compute haplotypes between markers\n");
		 }
	  }

	  /* off end, only one transition necessary */
		if((locus_l == -1) || (locus_r == -1)){

		  if (locus_r == -1) {

			TransSetupConvolve(thetal, locus_l, -1, &lvct[0], &tsleft);	
			TransBetweenAndDot(&tsleft, (TRANSSTRUCT *)NULL,
									 pos_in_scan, &tdot, 
									 tmpdot, 0, 0);
         TransFree(&tsleft);
		  }

		  if (locus_l == -1) {

			TransSetupConvolve(thetar, locus_r, -1, &rvct[0], &tsright);	
			TransBetweenAndDot((TRANSSTRUCT *)NULL, &tsright,
									 pos_in_scan, &tdot, 
									 tmpdot, 0, 0);
			TransFree(&tsright);

		  }

		}  /* if not at informative marker and off one end */

		else {

			TransSetupConvolve(thetal, locus_l, -1, &lvct[0], &tsleft);
			TransSetupConvolve(thetar, locus_r, -1, &rvct[0], &tsright);	
			TransBetweenAndDot(&tsleft, &tsright,
									 pos_in_scan, &tdot, tmpdot, 0, 0);
         TransFree(&tsleft);
			TransFree(&tsright);

		} /* if NOT off one end and not informative marker */

		maxlike=tdot.maxlike;
		minlike=tdot.minlike;
		lodlike=tdot.lodlike;
		like=tdot.like;
		statval=tdot.statval;
		varsum=tdot.varsum;
		un_normalized_entropy=tdot.un_normalized_entropy;

  } /* if !informative marker */  

	else {
		if(locus_l >= 0) { ilocus = locus_l; } else { ilocus = locus_r; }
		pnt = nz_pointer[ilocus];
		for(i=0; i<N; i++) {
	  		score_npl_val = Get_apm_score(i);
			if((i & maskTest[ilocus][0]) != maskTest[ilocus][1]) {
			  if (really_modcalc) this_p_pos[i] = 0.0;
			  continue;
			}
			if(locus_l == -1) {
				p    = rvct[pnt];
			} else if(locus_r == -1) {
				p    = lvct[pnt];
			} else {
			        /*** the following was if(rvct...) before
				     (K. Strauch 03/2004) ***/
				if(pvct[pnt] != 0.) {
					p    = lvct[pnt] * (rvct[pnt]/pvct[pnt]);
				} else {
					p    = 0.;
				}
			}
			pnt++;

			if(analysis_type != NPL_ANALYSIS) 
			{ 
				if(algebra)
				{
					plod = p * this_par_score[i]; 
				} else {
					plod = p * num_par_score[i];
				}	       
			} else { 
				plod = p;
			}

			if (really_modcalc) this_p_pos[i] = p;
			if (p > maxlike) maxlike=p;
			if (p < minlike && p > 0.0) minlike = p;
			lodlike += plod;
			like += p;
			statval += (p*score_npl_val);
			varsum  += (p*score_npl_val*score_npl_val);
			/* SUM(-(p[i]/like)*log(p[i]/like) = 
		            	   (1/like)*SUM(-p[i]*log(p[i])) + log(like)  */
			if(p > 0) un_normalized_entropy += (-1.0) * p * log(p);
			
			if(this_mod_pos==-1 && compute_sharing) find_sharing_probs(pos_in_scan, n, i, p); /*** new - FIX 6/99 ***/
		}
	}
	
	/* output for haplotyping code */
	skip_this_haplo_code = (haplo_method != 1);
	if (this_mod_pos==-1 && haplotype_output && (locus_haplo != -1) && !skip_this_haplo_code) {
		if (locus_haplo == 0) {
/*			printf("haplotype using maximum probability path\n");
*/			sf(ps,"haplo_ratio = %.4lf\n",haplo_ratio); fpr(fp);
			first_prob = second_prob = 1.0;
		}
		best_new_choice=-1; best_num_steps=n+1;
		/* we're at a marker, do the haplotyping stuff */
		sf(ps,"at marker %d\n",f_map_order[locus_haplo]+1); fpr(fp);
		maxlike /= like;
		num_maxlike=0;
		
		if(at_informative_marker) { pnt = nz_pointer[locus_haplo];
		} else                    { pvct_value = 1/m_pow2[rn]; }
		
		for(i=0; i<N; i++) {
			if(at_informative_marker) {
				if((i & maskTest[ilocus][0]) != maskTest[ilocus][1]) continue;
				if(locus_l == -1) {
					p    = rvct[pnt];
				} else if(locus_r == -1) {
					p    = lvct[pnt];
				} else {
				    /*** the following was if(rvct...) before
					 (K. Strauch 10/2007) ***/
					if(pvct[pnt] != 0.) {

						p    = lvct[pnt] * (rvct[pnt]/pvct[pnt]);
					} else {
						p = 0.;
					}
				}
				pvct_value = pvct[pnt];
				pnt++;
			} else {
			  if (tmpdot != (double *) 0)   /* for safety, should always be */
				 p=tmpdot[i];
			  else {
				 error("Internal error: Null tmp pointer in compute haplotype between markers\n");
				 p=0;
			  }
			}
			if (real_eq(p/like,maxlike)) { 
				sf(ps,"p[%s] = %.15lf\n",binary_output(i),p/like); fpr(fp);
				num_maxlike++;
				if (locus_haplo > 0) {
					/* see if this is a good step */
					num_steps = min_diff_bits(pad_inhvec(i),choice,&best_mask);
					if (num_steps < best_num_steps) {
						best_new_choice = best_mask;
						best_num_steps  = num_steps;
						pvct_max        = pvct_value;
					}
				} else {
					/* first locus, doesn't matter which one we take */
					if (best_new_choice == -1) {
						best_new_choice = pad_inhvec(i);
						pvct_max        = pvct_value;
					}
				}
			}
		}
		choice = best_new_choice;
		totlike = (double)num_maxlike * maxlike;
		sf(ps,"num_max_pvectors = %d, amounting to %.6lf of total prob\n",
		num_maxlike, totlike); fpr(fp);
		if (locus_haplo==0) { 
			sf(ps,"CHOICE = %s\n\n",binary_output(choice)); 
			running_bestrecs[locus_haplo]=0;
			first_prob = pvct_max;
		} else { 
			sf(ps,"CHOICE = %s  (implying %d recs)\n\n",
			binary_output(choice), best_num_steps); 
			running_bestrecs[locus_haplo] = running_bestrecs[locus_haplo-1]+best_num_steps;
			first_prob *= pvct_max;
			first_prob *= pow(theta[locus_haplo-1], (double) best_num_steps);
			first_prob *= pow(1.0-theta[locus_haplo-1], (double)(n-best_num_steps));
		}
	    fpr(fp);
	    best_pvector[locus_haplo] = choice; /* to be returned to pedprep */

	    fflush(fp);
        /* 'int loci' = num_mapped_loci was argument of likelihood function */
	    /* if (index == loci) fclose(fp); */
	}
	
	if (tmpdot != (double *) 0){
	  unarray(tmpdot, double);
	  tmpdot = (double *) 0;  /* just in case */
	}


	if (like < 1e-280) {
	    if (pos_in_scan == 0) {
print("WARNING: map likelihood=0.0 - this usually indicates that there is\n");
print("an obligate recombinant between markers separated by 0.0 recombination\n");
print("In cases where more than 50 markers are in use it is possible that\n");
print("this error condition indicates that the map probability is smaller than\n");
print("the smallest number supported by this architecture - try a subset of\n");
print("the markers and see if the problem persists.\n");
	    }
        *apmstat=expscore; /* NPL = 0.0 */
	    *variance=0.0;
	    entropy=0.0;
	} else {
	    *apmstat = statval/like;
	    *variance = varsum/like - ((*apmstat)*(*apmstat));
		entropy = un_normalized_entropy/like + log(like);
		if(this_mod_pos==-1 && compute_sharing) norm_sharing_probs(pos_in_scan, like); /*** new - FIX 6/99 ***/
/*r-full-alloc  
	    entropy=0.0;
	    for (i=0; i<N; i++) p[i] /= like;
	    for (i=0; i<N; i++) {
	      if (p[i] > 0.0) entropy += (-1.0)*p[i]*log(p[i]);
	    }
*/
	}

	/*r-full-alloc  find_sharing_probs(pos_in_scan, N, n, p); */ /*** new ***/

	*pvec_entropy = entropy;
	/*r-full-alloc  unarray(p, double); */
	/*r-full-alloc  unarray(plod, double); */

	*ret_lodlike = lodlike;
	return(like);
}


double likelihood_sp(int locus_haplo, int locus_l, int locus_r, 
		  double thetal_m, double thetar_m, double thetal_f, double thetar_f,
		  double *apmstat, double *variance, 
		  double *pvec_entropy, double *ret_lodlike,
		  int pos_in_scan)
{

	int i, num_maxlike, best_new_choice, best_num_steps, num_steps,
	    num_steps_m, num_steps_f, min_diff_bits_m, min_diff_bits_f;
	volatile int best_num_steps_m, best_num_steps_f;
	double p, plod, maxlike, minlike, totlike;
	double lodlike, like, statval, varsum, entropy;
	static FILE *fp;
	static int like_fp_open=TRUE;
	static int choice;
	static double first_prob, second_prob;
	char *binary_output(int);
/*	int skip_this_haplo_code = TRUE, best_mask, best_i;
    km for checking alt. haplo code */
	int skip_this_haplo_code = FALSE, best_mask;
	volatile int ilocus;
	volatile unsigned long pnt;
	double un_normalized_entropy;
	volatile double pvct_max;
	volatile double pvct_value;
	double score_npl_val;
	int    at_informative_marker;
	double *tmpdot;
   	TRANSSTRUCT tsleft, tsright;
   	TRDOTSTRUCT tdot;
	double *this_par_score = NULL;
	if(algebra)
	this_par_score = datfile_par_score[algebra_pointer[famnum]];
	
	int min_diff_bits_sp(int new, int old, int *new_vec, int *min_diff_bits_m, 
                     int *min_diff_bits_f);
	
	if (this_mod_pos==-1 && like_fp_open && !skip_this_haplo_code) {
	  fp=open_file("vectors.out","w"); like_fp_open=FALSE;
	}
	
	like = lodlike = 0.0; statval = 0.0; varsum = 0.0;
	maxlike=0.0; 
	minlike=1.0;
	un_normalized_entropy = 0;
	at_informative_marker = (min(thetal_m,thetal_f) == 0.0) || (min(thetar_m,thetar_f) == 0.0);
	tmpdot = (double *) 0;

	if ( !at_informative_marker ) {
		/* need transition away from informative marker */
		/* compute prob. per inheritance vector and store in tmp1[] */

	  /* for Haplotyping between markers, but slows things down */
	  if (haplotype_output && (locus_haplo != -1) && (haplo_method == 1)) {
		 run {
			array(tmpdot, N, double);
		 } except_when (NOMEMORY) {
			unarray(tmpdot, double);
			tmpdot = (double *) 0;
			error("Can't get enough memory to compute haplotypes between markers\n");
		 }
	  }

	  /* off end, only one transition necessary */
		if((locus_l == -1) || (locus_r == -1)){

		  if (locus_r == -1) {
		  
			TransSetupConvolve_sp(thetal_m,thetal_f, locus_l, -1, &lvct[0], &tsleft, m_wgt, f_wgt);	
			TransBetweenAndDot_sp(&tsleft, (TRANSSTRUCT *)NULL,
					   pos_in_scan, &tdot, tmpdot, 0, 0);
         		TransFree(&tsleft);
		  }

		  if (locus_l == -1) {
		  
		  	TransSetupConvolve_sp(thetar_m, thetar_f, locus_r, -1, &rvct[0], &tsright, m_wgt_s, f_wgt_s);	
			TransBetweenAndDot_sp((TRANSSTRUCT *)NULL, &tsright,
					   pos_in_scan, &tdot, tmpdot, 0, 0);
			TransFree(&tsright);

		  }

		}  /* if not at informative marker and off one end */

		else {
		
 
			TransSetupConvolve_sp(thetal_m, thetal_f, locus_l, -1, &lvct[0], &tsleft, m_wgt, f_wgt);
			TransSetupConvolve_sp(thetar_m, thetar_f, locus_r, -1, &rvct[0], &tsright, m_wgt_s, f_wgt_s);	
			TransBetweenAndDot_sp(&tsleft, &tsright,
					   pos_in_scan, &tdot, tmpdot, 0, 0);
         		TransFree(&tsleft);
			TransFree(&tsright);
		} /* if NOT off one end and not informative marker */

		maxlike=tdot.maxlike;
		minlike=tdot.minlike;
		lodlike=tdot.lodlike;
		like=tdot.like;
		statval=tdot.statval;
		varsum=tdot.varsum;
		un_normalized_entropy=tdot.un_normalized_entropy;

  }
	 /* if !informative marker */  

	else {
		if(locus_l >= 0) { ilocus = locus_l; } else { ilocus = locus_r; }
		pnt = nz_pointer[ilocus];
		for(i=0; i<N; i++) {
	  		score_npl_val = Get_apm_score(i);
			if((i & maskTest[ilocus][0]) != maskTest[ilocus][1]) {
			  if (really_modcalc) this_p_pos[i] = 0.0;
			  continue;
			}
			if(locus_l == -1) {
				p    = rvct[pnt];
			} else if(locus_r == -1) {
				p    = lvct[pnt];
			} else {
			        /*** the following was if(rvct...) before
				     (K. Strauch 03/2004) ***/
				if(pvct[pnt] != 0.) {
					p    = lvct[pnt] * (rvct[pnt]/pvct[pnt]);
				} else {
					p    = 0.;
				}
			}
			pnt++;

			if(analysis_type != NPL_ANALYSIS) { 
				if(algebra)
					plod = p * this_par_score[i];
				else
					plod = p * num_par_score[i];
			} else { 
				plod = p;
			}

			if (really_modcalc) this_p_pos[i] = p;
			if (p > maxlike) maxlike=p;
			if (p < minlike && p > 0.0) minlike = p;
			lodlike += plod;
			like += p;
			statval += (p*score_npl_val);
			varsum  += (p*score_npl_val*score_npl_val);
			/* SUM(-(p[i]/like)*log(p[i]/like) = 
		            	   (1/like)*SUM(-p[i]*log(p[i])) + log(like)  */
			if(p > 0) un_normalized_entropy += (-1.0) * p * log(p);
			
			if(this_mod_pos==-1 && compute_sharing) find_sharing_probs(pos_in_scan, n, i, p); /*** new - FIX 6/99 ***/
		}
	}
	
	/* output for haplotyping code */
	skip_this_haplo_code = (haplo_method != 1);
	if (this_mod_pos==-1 && haplotype_output && (locus_haplo != -1) && !skip_this_haplo_code) {
		if (locus_haplo == 0) {
/*			printf("haplotype using maximum probability path\n");
*/			sf(ps,"haplo_ratio = %.4lf\n",haplo_ratio); fpr(fp);
			first_prob = second_prob = 1.0;
		}
		best_new_choice=-1; best_num_steps=n+1;
		/* we're at a marker, do the haplotyping stuff */
		sf(ps,"at marker %d\n",f_map_order[locus_haplo]+1); fpr(fp);
		maxlike /= like;
		num_maxlike=0;
		
		if(at_informative_marker) { pnt = nz_pointer[locus_haplo];
		} else                    { pvct_value = 1/m_pow2[rn]; }
		
		for(i=0; i<N; i++) {
			if(at_informative_marker) {
				if((i & maskTest[ilocus][0]) != maskTest[ilocus][1]) continue;
				if(locus_l == -1) {
					p    = rvct[pnt];
				} else if(locus_r == -1) {
					p    = lvct[pnt];
				} else {
				    /*** the following was if(rvct...) before
					 (K. Strauch 10/2007) ***/
					if(pvct[pnt] != 0.) {
						p    = lvct[pnt] * (rvct[pnt]/pvct[pnt]);
					} else {
						p = 0.;
					}
				}
				pvct_value = pvct[pnt];
				pnt++;
			} else {
			  if (tmpdot != (double *) 0)   /* for safety, should always be */
				 p=tmpdot[i];
			  else {
				 error("Internal error: Null tmp pointer in compute haplotype between markers\n");
				 p=0;
			  }
			}
			if (real_eq(p/like,maxlike)) { 
				sf(ps,"p[%s] = %.15lf\n",binary_output(i),p/like); fpr(fp);
				num_maxlike++;
				if (locus_haplo > 0) {
					/* see if this is a good step */
					/* num_steps = min_diff_bits(pad_inhvec(i),choice,&best_mask); */
					num_steps = min_diff_bits_sp(pad_inhvec(i), choice, &best_mask, &min_diff_bits_m, &min_diff_bits_f);
					num_steps_m = min_diff_bits_m;
					num_steps_f = min_diff_bits_f;
					min_diff_bits_sp(pad_inhvec(i),choice,&best_mask,&min_diff_bits_m, &min_diff_bits_f);
					if (num_steps < best_num_steps) {
						best_new_choice = best_mask;
						best_num_steps  = num_steps;
						best_num_steps_m  = num_steps_m;
						best_num_steps_f  = num_steps_f;
						pvct_max        = pvct_value;
					}
				} else {
					/* first locus, doesn't matter which one we take */
					if (best_new_choice == -1) {
						best_new_choice = pad_inhvec(i);
						pvct_max        = pvct_value;
					}
				}
			}
		}
		choice = best_new_choice;
		totlike = (double)num_maxlike * maxlike;
		sf(ps,"num_max_pvectors = %d, amounting to %.6lf of total prob\n",
		num_maxlike, totlike); fpr(fp);
		if (locus_haplo==0) { 
			sf(ps,"CHOICE = %s\n\n",binary_output(choice)); 
			running_bestrecs[locus_haplo]=0;
			first_prob = pvct_max;
		} else { 
			sf(ps,"CHOICE = %s  (implying %d recs)\n\n",
			binary_output(choice), best_num_steps); 
			running_bestrecs[locus_haplo] = running_bestrecs[locus_haplo-1]+best_num_steps;
			first_prob *= pvct_max;
			first_prob *= (pow(theta_m[locus_haplo-1], (double) best_num_steps_m)*pow(theta_f[locus_haplo-1], (double) best_num_steps_f));
			first_prob *= (pow(1.0-theta_m[locus_haplo-1], (double)(n_m-best_num_steps_m))*pow(1.0-theta_f[locus_haplo-1], (double)(n_f-best_num_steps_f)));
		}
	    fpr(fp);
	    best_pvector[locus_haplo] = choice; /* to be returned to pedprep */

	    fflush(fp);
	    /* if (index == loci) fclose(fp); */
	}
	
	if (tmpdot != (double *) 0){
	  unarray(tmpdot, double);
	  tmpdot = (double *) 0;  /* just in case */
	}


	if (like < 1e-280) {
	    if (pos_in_scan == 0) {
print("WARNING: map likelihood=0.0 - this usually indicates that there is\n");
print("an obligate recombinant between markers separated by 0.0 recombination\n");
print("In cases where more than 50 markers are in use it is possible that\n");
print("this error condition indicates that the map probability is smaller than\n");
print("the smallest number supported by this architecture - try a subset of\n");
print("the markers and see if the problem persists.\n");
	    }
        *apmstat=expscore; /* NPL = 0.0 */
	    *variance=0.0;
	    entropy=0.0;
	} else {
	    *apmstat = statval/like;
	    *variance = varsum/like - ((*apmstat)*(*apmstat));
		entropy = un_normalized_entropy/like + log(like);
		if(this_mod_pos==-1 && compute_sharing) norm_sharing_probs(pos_in_scan, like); /*** new - FIX 6/99 ***/
/*r-full-alloc  
	    entropy=0.0;
	    for (i=0; i<N; i++) p[i] /= like;
	    for (i=0; i<N; i++) {
	      if (p[i] > 0.0) entropy += (-1.0)*p[i]*log(p[i]);
	    }
*/
	}

	/*r-full-alloc  find_sharing_probs(pos_in_scan, N, n, p); */ /*** new ***/

	*pvec_entropy = entropy;
	/*r-full-alloc  unarray(p, double); */
	/*r-full-alloc  unarray(plod, double); */

	*ret_lodlike = lodlike;
	return(like);
}



char *binary_output(int num)
{
    char *tempret;
    int i;

    tempret = get_temp_string();
    strcpy(tempret,"");
    for (i=n-1; i>=0; i--) {
      sf(ps,"%d",(num>>i)%2);
      strcat(tempret,ps);
    }
    return(tempret);
}

int diff_bits(int a, int b)
{
    int xor, i, ones;

    xor = a^b;
    /* now count the 1's */
    ones=0;
    for (i=n-1; i>=0; i--) {
      if (((xor>>i)%2) == 1) ones++;
    }
    return(ones);
}

int min_diff_bits(int new, int old, int *new_vec)
{
    int xor, i, j, ones, newxor, newones;
    int best_vec;

    xor = new^old;
    /* now count the 1's */
    ones=0;
    for (i=n-1; i>=0; i--) {
      if (((xor>>i)%2) == 1) ones++;
    }
    if (ones < 2) { *new_vec=new; return(ones); }
    
    /* else try each mask and see if it improves the recombination count */
    best_vec=new;

    for (i=0; i<n; i++) {
      if (flip_mask[i] > 0) {
	newxor = xor^flip_mask[i];
	newones=0;
	for (j=n-1; j>=0; j--) { if (((newxor>>j)%2) == 1) newones++; }
	if (newones < ones) {
	  /* accept this flip */
	  best_vec = (best_vec^flip_mask[i]);
	}
      }
    }

    xor = old^best_vec;
    ones=0;
    for (i=n-1; i>=0; i--) {
      if (((xor>>i)%2) == 1) ones++;
    }

    *new_vec=best_vec;
    return(ones);
}


int min_diff_bits_sp(int new, int old, int *new_vec, int *min_diff_bits_m, 
                     int *min_diff_bits_f)
{
    int xor, i, j, ones, newxor, newones, newones_m, newones_f;
    int best_vec, ones_m, ones_f;

    xor = new^old;
    /* now count the 1's */
    ones=0;
    ones_m=0;
    ones_f=0;
    for (i=n-1; i>=0; i--) {
      if (((xor>>i)%2) == 1){
         ones++;
	 ones_m += (i%2);
	 ones_f += (1-i%2);
      }
    }
    if ((ones) < 2) { *new_vec=new; return(ones); }
    
    /* else try each mask and see if it improves the recombination count */
    best_vec=new;

    for (i=0; i<n; i++) {
      if (flip_mask[i] > 0) {
	newxor = xor^flip_mask[i];
	newones = 0;
	newones_m = 0;
	newones_f = 0;
	for (j=n-1; j>=0; j--) { 
	   if (((newxor>>j)%2) == 1){
	      newones_m++;
	      newones_m += (i%2);
	      newones_f += (1-i%2);; 
	   }   
	}
	if (newones < ones) {
	  /* accept this flip */
	  best_vec = (best_vec^flip_mask[i]);
	}
      }
    }

    xor = old^best_vec;
    ones=0;
    ones_m=0;
    ones_f=0;
    for (i=n-1; i>=0; i--) {
      if (((xor>>i)%2) == 1){
        ones++;
        ones_m += (i%2);
        ones_f += (1-i%2);
      } 
    }

    *new_vec=best_vec;
    *min_diff_bits_m = ones_m;
    *min_diff_bits_f = ones_f;
    return(ones);
}


double left(double *theta, int loci)
{

	int i, l;
	double like;
	for (i=0; i<N; ++i) lvector[0][i] = pvector[0][i];

	for (l=0; l<loci; ++l)
	{
		for (i=0; i<N; ++i)
		  tempvector[i] = lvector[l][i];

		convolve_fft(theta[l], tempvector);

		for (i=0; i<N; ++i) 
		  lvector[l+1][i]=pvector[l+1][i]*tempvector[i];
	}

	like = 0.;
	for (i=0; i<N; ++i) like += lvector[loci][i];
	return(like);
}

double right(double *theta, int loci)
{

	int i, l;
	double like;
	for (i=0; i<N; ++i) rvector[loci][i] = pvector[loci][i];

	for (l=loci; l>0; --l)
	{
	    for (i=0; i<N; i++) 
	      tempvector[i]=rvector[l][i];
	    convolve_fft(theta[l-1], tempvector);
	    for (i=0; i<N; ++i) 
	      rvector[l-1][i]=tempvector[i]*pvector[l-1][i];
	}

	like = 0.;
	for (i=0; i<N; ++i) like += rvector[0][i];
	return(like);
}

/* compute the transition probabilities for 0,1,...,n recombinations */
void makerec(double theta, double *t)
{
	int i;
	double ctheta, rat;
    /* double rat_m, rat_f, ctheta_m, ctheta_f; */

	ctheta = 1.-theta;
	/* ctheta_m = 1.-theta_m; */
	/* ctheta_f = 1.-theta_f; */
	rat = theta/ctheta;
	/* rat_m = theta_m/ctheta_m; */
	/* rat_f = theta_f/ctheta_f; */
	t[0] = 1.;
	for (i=0; i<n; ++i) t[0]*=ctheta;
	for (i=1; i<n+1; ++i) t[i] = t[i-1]*rat;
	
	/*for (i=0; i<n; ++i){
	    if (i%2 == 1) { 
	      t[0] *= ctheta_m; 
	    }else{ 
	      t[0] *= ctheta_f;
	    }//t[0]*=ctheta; 
	} 
	for (i=1; i<n+1; ++i){
	    if (i%2 == 1) { 
	      t[i] = t[i-1]*rat_m; 
	    }else{ 
	      t[i] = t[i-1]*rat_f; 
	    }//t[i] = t[i-1]*rat;
	} 
	*/   
}


void makerec_sp(double theta_m, double theta_f, double *t)
{
	int i, i_m, i_f, i_g;
	double ctheta_m, ctheta_f, rat_m, rat_f,
	       t_m[64], t_f[64];

	ctheta_m = 1.-theta_m;
	ctheta_f = 1.-theta_f;
	rat_m = theta_m/ctheta_m;
	rat_f = theta_f/ctheta_f;
	
	t[0] = 1.;
	t_m[0] = 1.;
	t_f[0] = 1.;
	
        for (i=0; i<n_m; ++i) t_m[0]*=ctheta_m;
	for (i=1; i<n_m+1; ++i) t_m[i] = t_m[i-1]*rat_m;	
	for (i=0; i<n_f; ++i) t_f[0]*=ctheta_f;
	for (i=1; i<n_f+1; ++i) t_f[i] = t_f[i-1]*rat_f;
	
	i_g=0;
	for (i_f=0; i_f<n_f; ++i_f){
	   for (i_m=0; i_m<n_m; ++i_m){
	      t[i_g] = t_m[i_m]*t_f[i_f];
	      i_g++;
	   }
	}      
}

/* recursively compute the vector-matrix product */
void reduce(int tc, int h, int w)
{
	int i, j, half;

	if (w>1) {
	half = h/2;
	for (i=0; i<half; ++i)
	for (j=0; j<w; ++j)
		aux[tc+i][j] += aux[tc+h-i-1][w-j-1];

	for (i=tc+half; i<tc+h; ++i)
	for (j=0; j<w-1; ++j)
		aux[i][j] = aux[i-half][j+1];

	reduce(tc,half,w-1);
	reduce(tc+half,half,w-1);
	} /* if */
} /* end reduce */

void convolve_fft(double theta, double *pv)
{
    int i,j;
    double trans[100], factor;

    trans[0] = 1.0;
    factor = 1.0 - 2.0*theta;
    for (j=1; j<=n; ++j)
      trans[j] = trans[j-1]*factor;

    fft(pv);

    for (i=0; i<N; i++) {

      /*****
      sf(ps,"%d\t%d\t%d\t%f\n",i,real_bit_count[i],bit_count[i],pv[i]);
      pr();
      *****/

      pv[i] *= trans[bit_count[i]];

      /* zero out all higher order components */
      /*
      if (real_bit_count[i] > 1)
        pv[i] = 0;
      */
    }

    fft(pv);  /* should I divide by N = 2^rn here to normalize? */

    for (i=0; i<N; i++)
      pv[i] /= N;
}


void convolve_count_fft(double theta, double *pv)
{
    int i,j;
    double trans[100], factor;

    trans[0] = 1.0;
    factor = 1.0 - 2.0*theta;
    for (j=1; j<=n; ++j)
      trans[j] = trans[j-1]*factor;

    fft(pv);

    for (i=0; i<N; i++) {

      /*****
      sf(ps,"%d\t%d\t%d\t%f\n",i,real_bit_count[i],bit_count[i],pv[i]);
      pr();
      *****/

      pv[i] *= (theta*(n-bit_count[i])*trans[bit_count[i]] -
		theta*bit_count[i]*
		trans[((bit_count[i]-1)>=0) ? (bit_count[i]-1) : 0]);

      /* zero out all higher order components */
      /*
      if (real_bit_count[i] > 1)
        pv[i] = 0;
      */
    }

    fft(pv);  /* should I divide by N = 2^rn here to normalize? */

    for (i=0; i<N; i++)
      pv[i] /= N;
}


void fft(double *pv)
{
    int i,j,ej,flipi;
    double temp;

    for (j=0; j<rn; j++) {

      ej = m_pow2[j];

      for (i=0; i<N; i++) {

        if ((i & ej) == 0) {

	  flipi = i ^ ej; /* should this by "+" instead of "^" ? */
	  temp = pv[i];
	  pv[i] = temp + pv[flipi];
	  pv[flipi] = temp - pv[flipi];

	}
      }
    }
}

void speedy_reduce(double theta, double *pv)
{
    int i,j,ej;
    int virtual_i, virtual_partner, partner;
    double ctheta;

    ctheta=1.0-theta;

    for (j=0; j<n; j++) {

      for (i=0; i<N; i++) {

	if (saved_partner[j][i] >= 0) {
	  partner = saved_partner[j][i];
	  newp[i] = pv[i]*ctheta + pv[partner]*theta;
	} else {
	 
	  ej = m_pow2[j];
	  /* expand me to full size */
	  /* virtual_i = pad_inhvec(i); */
	  virtual_i = padded_vec[i];

	  /* find partner with bit j flipped */
	  virtual_partner = virtual_i ^ ej;
	  partner = lookup_mask(virtual_partner);

	  newp[i] = pv[i]*ctheta + pv[partner]*theta;
	  saved_partner[j][partner]=i;
	  saved_partner[j][i]=partner;
	}
      }

      for (i=0; i<N; i++) pv[i] = newp[i];
    }
}



void exciting_reduce(int height, int width)
{
    int half,row,col,partner,virtual_partner,virtual_row;
    int lookup_mask(int), pad_inhvec(int);

    if (width > 1) {
      /* row_partner is now globally allocated */
      for (row=0; row<N; row++) row_partner[row]=-1;

      for (row=0; row<N; row++) {
      
        if ((partner=row_partner[row]) >= 0 ) {
	  for (col=0; col<width; col++) 
	    aux[row][col] = aux[partner][width-col-1];
	  
	} else {
	  virtual_row = pad_inhvec(row);
	  virtual_partner=  height*(virtual_row/height) + 
	    (height-1) - virtual_row%height;
	  partner = lookup_mask(virtual_partner);
	  if (partner == row) {
	    /* partner with itself */
	    half = (width+1)/2;
	    for (col=0; col<half; col++) 
	      aux[row][col] += aux[row][width-col-1];
	    for (col=half; col<width; col++) 
	      aux[row][col] = aux[row][width-col-1];
	  } else {
	    /* partner with another row */
	    row_partner[partner]=row;
	    for (col=0; col<width; col++)
	      aux[row][col] += aux[partner][width-col-1];
	  }
	}
      }
      exciting_reduce(height/2, width-1);
    }
}

int lookup_mask(int vec)
{
    int andmask,newvec,i,fmask,realcount;

    andmask = vec&uninf_mask;
    
    for (i=0; i<n; i++) {
      if ((andmask&(m_pow2[i])) > 0) { 
	fmask = flip_mask[i];
	newvec = vec^fmask; 
	vec = newvec; 
      }
    }

    /* now we have the original vector flipped such that all uninformative
       positions now have 0's in them - now we must return the value of 
       this vector with those bits taken out entirely */
    newvec=0; realcount=0;
    for (i=0; i<n; i++) {
      if ((uninf_mask&m_pow2[i]) > 0) {
	/* bit to be ignored */
      } else {
	if ((vec&m_pow2[i]) > 0) {
	  newvec += m_pow2[realcount];
	}
	realcount++;
      }
    }
    return(newvec);
}



int reverse_mask(int vec1, int vec2)
{
    int andmask,newvec,i,fmask,newvec1;

    andmask = vec2&uninf_mask;
    
    for (i=0; i<n; i++) {
      if ((andmask&(m_pow2[i])) > 0) { 
	fmask = flip_mask[i];

	newvec = vec2^fmask; 
	vec2 = newvec;

	newvec1 = vec1^fmask;
	vec1 = newvec1;
      }
    }
    return(vec1);
}


int pad_inhvec(int vec)
{
    int i, newvec, realcount;

    newvec=0; realcount=0;
    for (i=0; i<n; i++) {
      if ((uninf_mask&m_pow2[i]) > 0) {
	/* bit inserted here */
      } else {
	if ((vec&m_pow2[realcount]) > 0) {
	  newvec += m_pow2[i];
	}
	realcount++;
      }
    }
    return(newvec);
}

void haplo_reduce(int height, int width)
{
    int half,row,col,partner,virtual_partner,virtual_row;
    int lookup_mask(int), pad_inhvec(int), this_bit, tmpvec;
    /* int pp, vpp; */

    if (width > 1) {
      for (row=0; row<N; row++) row_partner[row]=-1;

      for (row=0; row<N; row++) {
      
        if ((partner=row_partner[row]) >= 0 ) {

	  for (col=0; col<width; col++) {
	    aux[row][col] = aux[partner][width-col-1];
	    this_bit = aux_index[partner][width-col-1] % 2; 

	    if (this_bit) {
	      if (aux_index[row][col] >= 0) {
		tmpvec = aux_index[row][col];
		aux_index[row][col] = tmpvec << 1;
	      } else {
		aux_index[row][col] = 0;
	      }
	    } else {
	      if (aux_index[row][col] >= 0) {
		aux_index[row][col] = aux_index[partner][width-col-1];
		aux_index[row][col] += 1;
		/* tmpvec = aux_index[partner][width-col-1];
		   aux_index[row][col] = tmpvec << 1; already shifted */
	      } else {
		aux_index[row][col] = 1;
	      }
	    }
	    /* aux_index[row][col] = aux_index[partner][width-col-1];  */
	  }

	} else {
	  virtual_row = pad_inhvec(row);
	  virtual_partner=  height*(virtual_row/height) + 
	    (height-1) - virtual_row%height;
	  partner = lookup_mask(virtual_partner);
	  if (partner == row) {

	    /* partner with itself */
	    half = (width+1)/2;
	    for (col=0; col<half; col++) {
	      if (aux[row][col] < aux[row][width-col-1]) {
		aux[row][col] = aux[row][width-col-1];
		/* aux_index[row][col] = virtual_partner; */
		if (aux_index[row][width-col-1] >= 0) {
		  tmpvec = aux_index[row][width-col-1];
		  aux_index[row][col] = tmpvec << 1;
		  aux_index[row][col] += 1;
		} else {
		  aux_index[row][col]=1;
		}
	      } else {
		if (aux_index[row][col] >= 0) {
		  tmpvec = aux_index[row][col];
		  aux_index[row][col] = tmpvec << 1;
		} else {
		  aux_index[row][col] = 0;
		}
	      }
	    }

	    for (col=half; col<width; col++) {
	      aux[row][col] = aux[row][width-col-1];
	      this_bit = aux_index[row][width-col-1] % 2; 

	      if (this_bit) {
		if (aux_index[row][col] >= 0) {
		  tmpvec = aux_index[row][col];
		  aux_index[row][col] = tmpvec << 1;
		} else {
		  aux_index[row][col] = 0;
		}
	      } else {
		if (aux_index[row][col] >= 0) {
		  aux_index[row][col] = aux_index[row][width-col-1];
		  aux_index[row][col] += 1;
		} else {
		  aux_index[row][col] = 1;
		}
	      }
	    }

	  } else {
	    /* partner with another row */
	    row_partner[partner]=row;
	    /* pp = pad_inhvec(partner); */
	    /* vpp = height*(pp/height) + (height-1) - pp%height; */

	    for (col=0; col<width; col++)
	      if (aux[row][col] < aux[partner][width-col-1]) {
		aux[row][col] = aux[partner][width-col-1];
		if (aux_index[row][col] >= 0) {
		  tmpvec = aux_index[partner][width-col-1];
		  aux_index[row][col] = tmpvec << 1;
		  aux_index[row][col] += 1;
		} else {
		  aux_index[row][col] = 1;
		}
	      } else {
		aux[partner][width-col-1] = aux[row][col];
		if (aux_index[row][col] >= 0) {
		  tmpvec = aux_index[row][col];
		  aux_index[row][col] = tmpvec << 1;
		} else {
		  aux_index[row][col] = 0;
		}
	      }
	  }
	}
      }
      haplo_reduce(height/2, width-1);
    }
}

/* recursively compute the vector-matrix product */
void new_reduce(int tc, int h, int w)
{
	int i, j, half;

	if (w>1) {
	  half = h/2;
	  for (i=0; i<half; ++i)
	    for (j=0; j<w; ++j)

	      if (aux[tc+i][j] < aux[tc+h-i-1][w-j-1]) {
		aux[tc+i][j] = aux[tc+h-i-1][w-j-1];
		aux_index[tc+i][j] = aux_index[tc+h-i-1][w-j-1];
	      }

	  for (i=tc+half; i<tc+h; ++i)
	    for (j=0; j<w-1; ++j) {
	      aux[i][j] = aux[i-half][j+1];
	      aux_index[i][j] = aux_index[i-half][j+1];
	    }
	  new_reduce(tc,half,w-1);
	  new_reduce(tc+half,half,w-1);
	} /* if */
} /* end new_reduce */


/* Extras from HOMOZ */

double rec_to_dist(double rec_frac)
{
    /* returns value in cM */
    double dist = -1.0;
    if (rec_frac > .4999) dist = 999.9; /* unlinked value */

    if (map_function == KOSAMBI) {
        dist = 25.0 * log ((1.0+(2.0*rec_frac)) / (1.0-(2.0*rec_frac)));
    } else if (map_function == HALDANE) {
        dist = -50.0 * log (1.0-(2.0*rec_frac));
    } else {
        error("map function undefined");
    }
    return dist;
}

double dist_to_rec(double distance)
{
    /* cM to recombination fraction */
    double rec = -1.0;
    if (map_function == KOSAMBI) {
        rec = 0.5 * tanh(.02*distance);
    } else if (map_function == HALDANE) {
        rec = 0.5 * (1.0-exp(-.02*distance));
    } else {
        error("map function undefined");
    }
    return rec;
}



/***** CONVOLUTION CODE *****/

double *temp_pval;
int saved_max;

void init_convolution(double max)
{
    /* set precision for entire set of convolutions */

    if (max > 100000.0) pval_precision=1;
    else if (max > 10000.0) pval_precision=10;
    else if (max > 1000.0) pval_precision=100;
    else if (max > 100.0) pval_precision=1000;
    else pval_precision=10000;

    saved_max= (int)(max*pval_precision)+1000;
    array(temp_pval, saved_max, double);
}



void convolve(double *pval_func, double *newpval, 
	      double *newscore, int num_scores)
{
    int i,j,this_score;

    /* convolve the two probability distributions */
    for (i=0; i<saved_max; i++) temp_pval[i]=0.0;

    for (i=0; i<saved_max; i++) {
        if (pval_func[i] == 0.0) continue;
	for (j=0; j<num_scores; j++) {
	    /* add .5 to do proper rounding */
	    this_score = (int) ((newscore[j]*pval_precision)+.5) + i;
	    temp_pval[this_score] += pval_func[i]*newpval[j];
	}
    }

    for (i=0; i<saved_max; i++) {
      pval_func[i]=temp_pval[i];
    }
}

void free_convolution(void)
{
    unarray(temp_pval, double);
}

void search_vector(void)
{
#define real_eq_100(x,y) (fabs((x)-(y))<1e-100)
#define real_eq_50(x,y)  (fabs((x)-(y))<1e-50)
#define real_eq_10(x,y)  (fabs((x)-(y))<1e-10)
#define real_eq_06(x,y)  (fabs((x)-(y))<1e-06)
#define MAX_SAMPLE   1000  
#define SVECTOR     lvector
#define DETAIL_OUT    0
	int      ibit, ilocus, ibit_padded;
	int      i, j, iclass, ips[2], ipsa, address, found;
	int      lgc[10], sum;
	double   value[2][MAX_SAMPLE], value_all[MAX_SAMPLE], temp;
	char     *binary_output(int);

	printf("\n");
	for(ilocus=0; ilocus<num_mapped_loci; ilocus++){
		printf("ilocus=%d \t",ilocus);
/*		printf("\n",ilocus);
*/		
		/* distinct values */
		for(i=0; i<MAX_SAMPLE; i++) { value_all[i]=-1; }
		ipsa = -1;
		for(i=0; i<N; i++) {
			temp = SVECTOR[ilocus][i];
			if( ipsa < MAX_SAMPLE-1 ) {
				found=FALSE;
				for(j=0; j<=ipsa; j++){
					if( real_eq_100(temp,value_all[j]) ) { 
						found = TRUE;
						break; 
					}
				}
				if( !found ) { value_all[++ipsa]=temp; }
			}
		}
		printf("(%d)\t",ipsa);
		
		/* bit pattern */
		ibit_padded=n-1;
		for(ibit=rn-1; ibit>=0; ibit--) {
			address = m_pow2[ibit];
			for(iclass=0; iclass<=1; iclass++) {
				for(i=0; i<MAX_SAMPLE; i++) { value[iclass][i]=-1; }
			}
			ips[0] = ips[1] = -1;
			for(i=0; i<N; i++) {
				temp = SVECTOR[ilocus][i];
				/* distinct by class */
				if((address & i) == address) { iclass=1;
				} else {                       iclass=0; }
				if( ips[iclass] < MAX_SAMPLE-1 ) { 
					found=FALSE;
					for(j=0; j<=ips[iclass]; j++){
						if( real_eq_100(temp,value[iclass][j]) ) { 
							found = TRUE;
							break; 
						}
					}
					if( !found ) {
						ips[iclass]++;
						j = ips[iclass];
						value[iclass][j]=temp;
					}
				}
				if( (ips[0] == MAX_SAMPLE-1) && (ips[1] == MAX_SAMPLE-1) ) { break; } 
			}

			if(DETAIL_OUT) {
				if(ibit==0) printf("\n");
				printf("    ibit=%d",ibit); 
				printf("\t ips[0/1]=%d %d",ips[0], ips[1]);
				printf("\n");

				for(iclass=0; iclass<=1; iclass++){
					printf("             value[%d][i]=(%d total)",iclass,ips[iclass]);
					for(i=0; i<5; i++) printf(" %12.4e",value[iclass][i]);
					printf("\n");
				}
			} else {
				while((m_pow2[ibit_padded] & uninf_mask) != 0) {
					if((ibit_padded & 1) == 1) printf("  ");
					printf("_");
					ibit_padded--;
				}
				if((ibit_padded & 1) == 1) printf("  ");

				lgc[0] = ips[0]==0 && real_eq_100(value[0][0],0.0);
				lgc[1] = ips[1]==0 && real_eq_100(value[1][0],0.0);
				lgc[2] = ips[0]==0 && ips[1]==0 && real_eq_100(value[0][0],value[1][0]);
				sum = lgc[0] + lgc[1] + lgc[2];
				if( sum > 1 ) {
					printf("?");
				} else if( sum==1 ) {
/*old convention */
/*					if(lgc[0])  printf("0");  
					if(lgc[1])  printf("1");
					if(lgc[2])  printf("e");
*/
/*new convention */
					if(lgc[0])  printf("1");
					if(lgc[1])  printf("0");
					if(lgc[2])  printf("e");
				} else {
					printf("x");
				}
				ibit_padded--;
			}

		}  /* ------------ end ibit loop -----*/
		printf("\n");
	}/* ------------ end ilcus loop -----*/
	printf("\n n=%d rn=%d \n", n, rn);		
	printf("uninf_mask=%x (hex),  %s (bin)\n", uninf_mask, binary_output(uninf_mask));
}
				/* eventually replace block if with an assignment outside the loop */
/*				if(       vector_flag & m_pow2(0)) { temp = lvector[ilocus][i];
				} else if(vector_flag & m_pow2(1)) { temp = pvector[ilocus][i];
				} else if(vector_flag & m_pow2(2)) { temp = rvector[ilocus][i]; }
*/


void  Load_pvct(void){
	int  i, iMarker;
	unsigned long ipvct;
	ipvct=0;
	for(iMarker=0; iMarker<num_in_map_order; iMarker++) {
		for(i=0; i<N; i++) {
			if((i & maskTest[iMarker][0]) == maskTest[iMarker][1])
				pvct[ipvct++] = pvector[iMarker][i];
		}
		if(ipvct != nz_pointer[iMarker+1]) {
			printf(
			"*** Error *** iMarker=%3d  ipvct=%10lu  nz_pointer[iMarker+1]=%10lu\n",
			iMarker, ipvct, nz_pointer[iMarker+1]);
			exit (1);
		}
	}
}

double Compute_lr_vectors(void)
{
	int i, l, p_marker;
	int n_marker = -1;          /* initialize to some invalid index */
    /* int n_pntf; */
    /* int n_bitc; */
	int found_first_informative;
	double sum, unlinked, ltheta;
	TRANSSTRUCT ts;

	/* left vector ------------------------ */
	found_first_informative = FALSE;
	for(l=0; l<num_mapped_loci; l++) {
		if(f_uninformative[l]) continue;
		if(!found_first_informative) {
			found_first_informative = TRUE;
			for(unsigned long h=nz_pointer[l]; h<nz_pointer[l+1]; h++) lvct[h] = pvct[h];
			p_marker = l;
			n_marker = l; /* just in case there is a single informative marker (compute unlinked) */
			continue;
		}
		n_marker = l;
		
		ltheta = dist_to_rec(map_position[n_marker] - map_position[p_marker]);
		
		TransSetupConvolve(ltheta, p_marker, n_marker, &lvct[0], &ts);
		TransAndMap(&ts, &lvct[0]);
		TransFree(&ts);

/* Katz, Thu Nov 20 12:25:20 2003, fix for underflow, large number of alleles */
		sum=0;
		for(unsigned long h=nz_pointer[n_marker]; h<nz_pointer[n_marker+1]; h++) sum += lvct[h];
		if ((sum < 1e-10) && (!scan_done)) {   /* K. Strauch 02/2008 */
		print("------------\n");
		print("- WARNING: -\n");
		print("------------\n");
		sf(ps,"map likelihood factor %e < 1e-10 at marker %s\n",
		   sum, locus[map_order[n_marker]].name); pr();
		sf(ps,"(marker number %d in map order) - this may indicate obligate\n",
		   n_marker+1); pr();
		print("multiple recombinations between markers separated by very small\n");
		print("recombination fractions.  Please use the option 'count recs on'\n");
		print("and specify recombination fractions not smaller than e.g. 0.005\n");
		print("around this marker in order to check for an excess number of\n");
		print("observed recombinations (in comparison to the number of expected\n");
		print("recombinations) for this pedigree.  Such an excess may be due to\n");
		print("genotyping errors or inconsistencies in the marker order.\n");
		}
		sum = 1./sum;
		for(unsigned long h=nz_pointer[n_marker]; h<nz_pointer[n_marker+1]; h++) lvct[h] *= sum;
		for(unsigned long h=nz_pointer[n_marker]; h<nz_pointer[n_marker+1]; h++) pvct[h] *= sum;
/* ---- end of underflow fix ---- */

		p_marker = n_marker;

/*	Debugging Only:
 This test does not work anymore but it is fixible
		n_pntf = nz_pointer[n_marker];
		n_bitc = nz_bits[n_marker];
		Check_results_01(p_marker, n_pntf, 
		            	 maskTest[n_marker][0], maskTest[n_marker][1],
		            	 n_bitc, rn, &lvct[n_pntf], &lvector[n_marker][0]);
*/

	}
	sum=0;
	for(unsigned long h=nz_pointer[n_marker]; h<nz_pointer[n_marker+1]; h++)
		sum += lvct[h];

	for(i=0; i<rn; i++) sum *= 0.5;
	unlinked = sum;
/*	printf(" new unlinked left = %12.4e \n", sum);
*/

	/* right vector ------------------------ */
	found_first_informative = FALSE;
	for(l=num_mapped_loci-1; l>=0; l--) {
		if(f_uninformative[l]) continue;
		if(!found_first_informative) {
			found_first_informative = TRUE;
			for(unsigned long h=nz_pointer[l]; h<nz_pointer[l+1]; h++) rvct[h] = pvct[h];
			p_marker = l;
			n_marker = l; /* just in case there is a single informative marker (compute unlinked) */
			continue;
		}
		n_marker = l;
		
		ltheta = dist_to_rec(map_position[p_marker] - map_position[n_marker]);
		TransSetupConvolve(ltheta, p_marker, n_marker, &rvct[0], &ts);
		TransAndMap(&ts, &rvct[0]);
		TransFree(&ts);

		p_marker = n_marker;

/*	Debugging Only:
 This test does not work anymore but it is fixible
		n_pntf = nz_pointer[n_marker];
		n_bitc = nz_bits[n_marker];
		Check_results_01(p_marker, n_pntf, 
		            	 maskTest[n_marker][0], maskTest[n_marker][1],
		            	 n_bitc, rn, &lvct[n_pntf], &lvector[n_marker][0]);
 */

	}
	sum=0.0;
	for(unsigned long h=nz_pointer[n_marker]; h<nz_pointer[n_marker+1]; h++)
		sum += rvct[h];
	if ((!real_eq03(sum, 1.0)) && (!scan_done)) {   /* K. Strauch 02/2008 */
		print("-----------------------\n");
		print("- !!! * WARNING * !!! -\n");
		print("-----------------------\n");
		print("Due to numerical problems with normalization (likelihood factor\n");
		sf(ps,"%e deviates from 1) for this pedigree, linkage scores are\n",
		   sum); pr();
		print("probably not accurate.  Please check the data for genotyping\n");
		print("errors or inconsistencies in the marker order before proceeding.\n");
	}
/*	for(i=0; i<rn; i++) sum *= 0.5;
	printf(" new unlinked right= %12.4e \n", sum);
*/
	return(unlinked);
}

double Compute_lr_vectors_sp(void)
{
	int i;
    /* int n_pntf; */
    /* int n_bitc; */
	volatile int l;
	volatile int p_marker;
	volatile int n_marker;
	volatile int found_first_informative;
	double sum;
	volatile double unlinked;
	volatile double ltheta_m;
	volatile double ltheta_f;
	TRANSSTRUCT ts;

	/* left vector ------------------------ */
	found_first_informative = FALSE;
	for(l=0; l<num_mapped_loci; l++) {
		if(f_uninformative[l]) continue;
		if(!found_first_informative) {
			found_first_informative = TRUE;
			for(unsigned long h=nz_pointer[l]; h<nz_pointer[l+1]; h++) lvct[h] = pvct[h];
			p_marker = l;
			n_marker = l; /* just in case there is a single informative marker (compute unlinked) */
			continue;
		}
		n_marker = l;
		
		ltheta_m = dist_to_rec(map_position_m[n_marker] - map_position_m[p_marker]);
		ltheta_f = dist_to_rec(map_position_f[n_marker] - map_position_f[p_marker]);
		
		
	        run {
		   array (m_wgt, N, int);
		   array (f_wgt, N , int); 
	        } except_when (NOMEMORY) {
		   unarray(m_wgt, int);
		   unarray(f_wgt, int); 
		   error("Can't get enough memory for this scan (local vars)\n");
	        }
				
		TransSetupConvolve_sp(ltheta_m, ltheta_f, p_marker, n_marker, &lvct[0], &ts, m_wgt, f_wgt);
		TransAndMap_sp(&ts, &lvct[0], m_wgt, f_wgt);
		TransFree(&ts);
		s_wgt_free(m_wgt, f_wgt);

/* Katz, Thu Nov 20 12:25:20 2003, fix for underflow, large number of alleles */
		sum=0;
		for(unsigned long h=nz_pointer[n_marker]; h<nz_pointer[n_marker+1]; h++) sum += lvct[h];
		if ((sum < 1e-10) && (!scan_done)) {   /* K. Strauch 02/2008 */
		  print("------------\n");
		  print("- WARNING: -\n");
		  print("------------\n");
		  sf(ps,"map likelihood factor %e < 1e-10 at marker %s\n",
		     sum, locus[f_map_order[n_marker]].name); pr();
		  sf(ps,"(marker number %d in map order) - this may indicate obligate\n",
		     n_marker+1); pr();
		  print("multiple recombinations between markers separated by very small\n");
		  print("recombination fractions.  Please use the option 'count recs on'\n");
		  print("and specify recombination fractions not smaller than e.g. 0.005\n");
		  print("around this marker in order to check for an excess number of\n");
		  print("observed recombinations (in comparison to the number of expected\n");
		  print("recombinations) for this pedigree.  Such an excess may be due to\n");
		  print("genotyping errors or inconsistencies in the marker order.\n");
		}
		sum = 1./sum;
		for(unsigned long h=nz_pointer[n_marker]; h<nz_pointer[n_marker+1]; h++) lvct[h] *= sum;
		for(unsigned long h=nz_pointer[n_marker]; h<nz_pointer[n_marker+1]; h++) pvct[h] *= sum;
/* ---- end of underflow fix ---- */

		p_marker = n_marker;

/*	Debugging Only:
 This test does not work anymore but it is fixible
		n_pntf = nz_pointer[n_marker];
		n_bitc = nz_bits[n_marker];
		Check_results_01(p_marker, n_pntf, 
		            	 maskTest[n_marker][0], maskTest[n_marker][1],
		            	 n_bitc, rn, &lvct[n_pntf], &lvector[n_marker][0]);
*/

	}
	sum=0;
	for(unsigned long h=nz_pointer[n_marker]; h<nz_pointer[n_marker+1]; h++)
		sum += lvct[h];
	for(i=0; i<rn; i++) sum *= 0.5;
	unlinked = sum;
/*	printf(" new unlinked left = %12.4e \n", sum);
*/

	/* right vector ------------------------ */
	found_first_informative = FALSE;
	for(l=num_mapped_loci-1; l>=0; l--) {
		if(f_uninformative[l]) continue;
		if(!found_first_informative) {
			found_first_informative = TRUE;
			for(unsigned long h=nz_pointer[l]; h<nz_pointer[l+1]; h++) rvct[h] = pvct[h];
			p_marker = l;
			n_marker = l; /* just in case there is a single informative marker (compute unlinked) */
			continue;
		}
		n_marker = l;
		
		ltheta_m = dist_to_rec(map_position_m[p_marker] - map_position_m[n_marker]);
		ltheta_f = dist_to_rec(map_position_f[p_marker] - map_position_f[n_marker]);
		
	        run {
		   array (m_wgt, N, int);
		   array (f_wgt, N , int); 
	        } except_when (NOMEMORY) {
		   unarray(m_wgt, int);
		   unarray(f_wgt, int); 
		   error("Can't get enough memory for this scan (local vars)\n");
	        }
		
		TransSetupConvolve_sp(ltheta_m, ltheta_f, p_marker, n_marker, &rvct[0], &ts, m_wgt, f_wgt);
		TransAndMap_sp(&ts, &rvct[0], m_wgt, f_wgt);
		TransFree(&ts);
		s_wgt_free(m_wgt, f_wgt);

		p_marker = n_marker;

/*	Debugging Only:
 This test does not work anymore but it is fixible
		n_pntf = nz_pointer[n_marker];
		n_bitc = nz_bits[n_marker];
		Check_results_01(p_marker, n_pntf, 
		            	 maskTest[n_marker][0], maskTest[n_marker][1],
		            	 n_bitc, rn, &lvct[n_pntf], &lvector[n_marker][0]);
 */

	}
	sum=0.0;
	for(unsigned long h=nz_pointer[n_marker]; h<nz_pointer[n_marker+1]; h++)
		sum += rvct[h];
	if ((!real_eq03(sum, 1.0)) && (!scan_done)) {   /* K. Strauch 02/2008 */
	  print("-----------------------\n");
	  print("- !!! * WARNING * !!! -\n");
	  print("-----------------------\n");
	  print("Due to numerical problems with normalization (likelihood factor\n");
	  sf(ps,"%e deviates from 1) for this pedigree, linkage scores are\n",
	     sum); pr();
	  print("probably not accurate.  Please check the data for genotyping\n");
	  print("errors or inconsistencies in the marker order before proceeding.\n");
	}
/*	for(i=0; i<rn; i++) sum *= 0.5;
	printf(" new unlinked right= %12.4e \n", sum);
*/
	return(unlinked);
}



/* Katz, Tue Jun 20 12:28:08 2000, Break up Transition() into
 * TransSetup(), TransAndMap(), and TransFree() so that can also use
 * between markers.
 */

/* Katz, Thu Jun 22 14:24:16 2000
 *  For use with destination between markers, set n_marker to -1 
 *  If n_marker is -1, it means no restrictions on destination.
 */
 
 void TransSetupConvolve(double theta, 
								int p_marker, 
								int n_marker, 
								double lr_vct[],
								TRANSSTRUCT *ts)
{
	int i, j, k, im;
	int p_size;
	unsigned long p_pntf;
	volatile int n_bitc;
	int nvfs, bits_on, n_legal;
	int andmask;
	int iarg3;
    /* int p_pntl; */

	ts->p_bitc = nz_bits[p_marker];


	p_size = m_pow2[ ts->p_bitc ];
	p_pntf = nz_pointer[p_marker];
	/* p_pntl = p_pntf + p_size; */
	
	if (n_marker >= 0){     /* Normal case, marker to marker transition */
	  n_bitc = nz_bits[n_marker];
	  ts->n_size = m_pow2[n_bitc];
	  ts->n_pntf = nz_pointer[n_marker];
	  ts->n_pntl = ts->n_pntf + ts->n_size;
	}
	else {            /* Transition to between markers, these are not used */
	  ts->n_size = 0; /*  (n_marker = -1) */
	  ts->n_pntf = 0;
	  ts->n_pntl = 0;
	}

	ts->l_theta[0]  = ts->l_ctheta[0] = 1.;
	ts->l_theta[1]  = theta;  ts->l_ctheta[1] = 1. - ts->l_theta[1];
	for(i=2; i<=n; i++) {
		ts->l_theta[i]  = ts->l_theta[i-1]  * ts->l_theta[1];
		ts->l_ctheta[i] = ts->l_ctheta[i-1] * ts->l_ctheta[1];
	}

	run {
		array(ts->bit_cnt, p_size, int);
		array(ts->tempvct, p_size, double);
		matrix(ts->maskShift, 64, 2, int);
	} except_when (NOMEMORY) {
		unmatrix(ts->maskShift, 64, int);
		unarray(ts->tempvct, double);
		unarray(ts->bit_cnt, int);
		error("Can't get enough memory for this scan (local vars)\n");
	}

	Setup_flip_masks(maskTest[p_marker][0], &nvfs, ts->p_v_flip_mask,
		                                    &(ts->nffs), ts->p_f_flip_mask);

	Setup_legal_vectors(ts->p_bitc, &(ts->p_mask_v_bits), p_marker, 
							  &(ts->p_legal), 
							  n_marker, &n_legal);
	if (n_marker >= 0)
	  n_legal = n_legal << n_bitc;

	/* prepare bitcount for fft */
	for(i=0; i<p_size; i++) {
		ts->bit_cnt[i] = 0;
		for(j=0; j<=mod8[ts->p_bitc]; j++) {
			ts->bit_cnt[i] += lbtc8[ (i >> (8*j)) & lbtc8mask ];
		}
		for(im=0; im<nvfs; im++) {
			andmask = i & ts->p_v_flip_mask[im];
			bits_on = 0;
			for(k=0; k<=mod8[ts->p_bitc]; k++) 
				bits_on += lbtc8[ (andmask >> (8*k)) & lbtc8mask ];
			ts->bit_cnt[i] += bits_on & 1;
		}
	}

	for(i=0; i<p_size; i++) ts->tempvct[i] = lr_vct[p_pntf+i];

	convolve_fft_2(ts->p_bitc, theta, ts->bit_cnt, ts->tempvct);

/*	printf("l=%2d  p_bitc=%2d p_size=%10d  nvfs/nffs=%2d %2d\n", 
		    p_marker, p_bitc, p_size, nvfs, nffs);
*/

   if (n_marker >= 0)
	  iarg3 = maskTest[n_marker][0];
	else
	  iarg3 = 0;

   Setup_MaskShift(rn, maskTest[p_marker][0], iarg3, n_legal, ts);

}
 
 
 void TransSetupConvolve_sp(double theta_m, double theta_f, 
								int p_marker, 
								int n_marker, 
								double lr_vct[],
								TRANSSTRUCT *ts, 
								int *m_wgt_act,
								int *f_wgt_act)	
{
	int i, im;
	int p_size;
	unsigned long p_pntf;
	volatile int n_bitc;
	int nvfs, n_legal;
    /* int bits_on; */
	int andmask;
	int iarg3;
    /* int p_pntl; */
		
	/* provisorisch */
	/* double theta_m = 0.5, theta_f = 0.5; */

	ts->p_bitc = nz_bits[p_marker];
        ts->p_bitc_m = nz_bits_m[p_marker];
	ts->p_bitc_f = nz_bits_f[p_marker];

	p_size = m_pow2[ ts->p_bitc ];
	p_pntf = nz_pointer[p_marker];
	/* p_pntl = p_pntf + p_size; */
	
	if (n_marker >= 0){     /* Normal case, marker to marker transition */
	  n_bitc = nz_bits[n_marker];
	  ts->n_size = m_pow2[n_bitc];
	  ts->n_pntf = nz_pointer[n_marker];
	  ts->n_pntl = ts->n_pntf + ts->n_size;
	}
	else {            /* Transition to between markers, these are not used */
	  ts->n_size = 0; /*  (n_marker = -1) */
	  ts->n_pntf = 0;
	  ts->n_pntl = 0;
	}


	ts->l_theta_m[0]  = ts->l_ctheta_m[0] = 1.;
	ts->l_theta_m[1]  = theta_m;  ts->l_ctheta_m[1] = 1. - ts->l_theta_m[1];
	for(i=2; i<=n; i++) {
		ts->l_theta_m[i]  = ts->l_theta_m[i-1]  * ts->l_theta_m[1];
		ts->l_ctheta_m[i] = ts->l_ctheta_m[i-1] * ts->l_ctheta_m[1];
	}
	
	ts->l_theta_f[0]  = ts->l_ctheta_f[0] = 1.;
	ts->l_theta_f[1]  = theta_f;  ts->l_ctheta_f[1] = 1. - ts->l_theta_f[1];
	for(i=2; i<=n; i++) {
		ts->l_theta_f[i]  = ts->l_theta_f[i-1]  * ts->l_theta_f[1];
		ts->l_ctheta_f[i] = ts->l_ctheta_f[i-1] * ts->l_ctheta_f[1];
	}	
	
	
	run {
		array(ts->bit_cnt_m, p_size, int);
		array(ts->bit_cnt_f, p_size, int);
		array(ts->tempvct, p_size, double);
		matrix(ts->maskShift, 64, 2, int);
	} except_when (NOMEMORY) {
		unmatrix(ts->maskShift, 64, int);
		unarray(ts->tempvct, double);
		unarray(ts->bit_cnt_m, int);
		unarray(ts->bit_cnt_f, int);
		error("Can't get enough memory for this scan (local vars)\n");
	}

	Setup_flip_masks_sp(maskTest[p_marker][0], &nvfs, ts->p_v_flip_mask,
		         &(ts->nffs), &(ts->nffs_m), &(ts->nffs_f), ts->nfbit_sex,
                         ts->p_f_flip_mask);

	Setup_legal_vectors(ts->p_bitc, &(ts->p_mask_v_bits), p_marker, 
							  &(ts->p_legal), 
							  n_marker, &n_legal);
	if (n_marker >= 0)
	  n_legal = n_legal << n_bitc;
	  
	  
	if (n_marker >= 0)
	  iarg3 = maskTest[n_marker][0];
	else
	  iarg3 = 0;

        Setup_MaskShift(rn, maskTest[p_marker][0], iarg3, n_legal, ts);  
	
        init_sex_weights(1, m_wgt_act,f_wgt_act);	
        
	for(i=0; i<p_size; i++) {
		ts->bit_cnt_m[i] = m_wgt_act[i];
		ts->bit_cnt_f[i] = f_wgt_act[i];
		for(im=0; im<nvfs; im++) {
			andmask = i & ts->p_v_flip_mask[im];
			ts->bit_cnt_m[i] += m_wgt_act[andmask] & 1;
			ts->bit_cnt_f[i] += f_wgt_act[andmask] & 1;
		}
	}
		
	
	
	/*for(i=0; i<p_size; i++) {
		ts->bit_cnt[i] = 0;
		for(int j=0; j<=mod8[ts->p_bitc]; j++) {
			ts->bit_cnt[i] += lbtc8[ (i >> (8*j)) & lbtc8mask ];
		}
		for(im=0; im<nvfs; im++) {
			andmask = i & ts->p_v_flip_mask[im];
			bits_on = 0;
			for(int k=0; k<=mod8[ts->p_bitc]; k++)
				bits_on += lbtc8[ (andmask >> (8*k)) & lbtc8mask ];
			ts->bit_cnt[i] += bits_on & 1;
		}
	}
       */


	for(i=0; i<p_size; i++) ts->tempvct[i] = lr_vct[p_pntf+i];

	convolve_fft_2_sp(ts->p_bitc, theta_m, theta_f, ts->bit_cnt_m, ts->bit_cnt_f, ts->tempvct);

/*	printf("l=%2d  p_bitc=%2d p_size=%10d  nvfs/nffs=%2d %2d\n", 
		    p_marker, p_bitc, p_size, nvfs, nffs);
*/
}
 



void TransAndMap(TRANSSTRUCT *ts, double lr_vct[])
{
  int i,j;
  unsigned long n_ptr;
  int ifs, r_size, p_address, p_varbl, p_fixed, hd_max, hd;
  int p_fix_bitc, mask, mask1, mask2, im;
  double x0;


	hd_max     = rn - ts->p_bitc + ts->nffs;
	p_fix_bitc = rn - ts->p_bitc;
	r_size     = m_pow2[ts->nffs];
	n_ptr      = ts->n_pntf;
	for(i=0; i < ts->n_size; i++) {
		p_address = ts->p_address0;
		for(j=0; j<ts->nMaskShift; j++) {
			if(ts->maskShift[j][1] >= 0) {
				p_address |= (i & ts->maskShift[j][0]) <<  ts->maskShift[j][1];
			} else {
				p_address |= (i & ts->maskShift[j][0]) >> - ts->maskShift[j][1];
			}
		}
		p_varbl = p_address & ts->p_mask_v_bits;
		p_fixed = p_address >> ts->p_bitc;

		lr_vct[n_ptr] = 0;
		for(ifs=0; ifs<r_size; ifs++) {
			hd = mask = 0;
			for(im=0; im<ts->nffs; im++) { 
				if(ifs & m_pow2[im]) {
					hd++;
					mask |= ts->p_f_flip_mask[im]; 
				}
			}
			mask1 = mask & ts->p_mask_v_bits;
			x0    = ts->tempvct[p_varbl ^ mask1];
			mask2 = (mask >> ts->p_bitc) ^ ts->p_legal; 
                                         /* contr. vector fixed bits */
			mask2 = mask2 ^ p_fixed;        /* fixed bits distance to p_address */
			for(j=0; j<=mod8[p_fix_bitc]; j++)
				hd += lbtc8[ (mask2 >> (8*j)) & lbtc8mask ];
			lr_vct[n_ptr] += x0 * ts->l_theta[hd] * ts->l_ctheta[hd_max - hd];
		}
		n_ptr++;
	}
	for(unsigned long h=ts->n_pntf; h<ts->n_pntl; h++) { lr_vct[h] *= pvct[h]; }
}


void TransAndMap_sp(TRANSSTRUCT *ts, double lr_vct[], int* m_wgt_act, int* f_wgt_act)
{
  int i,j;
  unsigned long n_ptr;
  int ifs, r_size, p_address, p_varbl, p_fixed, hd;
  int mask, mask1, mask2, im, hd_m, hd_f, hd_max_m, hd_max_f;
  double x0;


	hd_max_m     = rn_m - ts->p_bitc_m + ts->nffs_m;
	hd_max_f     = rn_f - ts->p_bitc_f + ts->nffs_f;
	r_size     = m_pow2[ts->nffs];
	n_ptr      = ts->n_pntf;
	for(i=0; i < ts->n_size; i++) {
		p_address = ts->p_address0;
		for(j=0; j<ts->nMaskShift; j++) {
			if(ts->maskShift[j][1] >= 0) {
				p_address |= (i & ts->maskShift[j][0]) <<  ts->maskShift[j][1];
			} else {
				p_address |= (i & ts->maskShift[j][0]) >> - ts->maskShift[j][1];
			}
		}
		p_varbl = p_address & ts->p_mask_v_bits;
		p_fixed = p_address >> ts->p_bitc;

		lr_vct[n_ptr] = 0;
		for(ifs=0; ifs<r_size; ifs++) {
			hd = mask = 0;
			hd_m = hd_f = 0;
			for(im=0; im<ts->nffs; im++) { 
				if(ifs & m_pow2[im]) {
					hd++;
					hd_m += ts->nfbit_sex[im];
					hd_f += (1-ts->nfbit_sex[im]); 
					mask |= ts->p_f_flip_mask[im]; 
				}
			}
			mask1 = mask & ts->p_mask_v_bits;
			x0    = ts->tempvct[p_varbl ^ mask1];
			mask2 = (mask >> ts->p_bitc) ^ ts->p_legal; 
                                         /* contr. vector fixed bits */
			mask2 = mask2 ^ p_fixed;        /* fixed bits distance to p_address */
			
		        hd_m += m_wgt_act[mask2<<ts->p_bitc];
			hd_f += f_wgt_act[mask2<<ts->p_bitc];
			
			lr_vct[n_ptr] += x0 * ts->l_theta_m[hd_m] * ts->l_ctheta_m[hd_max_m - hd_m]*
			                      ts->l_theta_f[hd_f] * ts->l_ctheta_f[hd_max_f - hd_f];
		}
		n_ptr++;
	}
	for(unsigned long h=ts->n_pntf; h<ts->n_pntl; h++) { lr_vct[h] *= pvct[h]; }
}

/* Katz, Thu Jun 22 14:21:37 2000, Used reduced FFT for between markers 
 *  This routine takes two TRANSSTRUCT's set up above for the transition
 *  from left and right to a position between markers and calculates
 *  the dot product between them, Lod Scores, etc.  Uses only the 
 *  (two) reduced space vectors.
 *
 *  (if either of these struct ptrs are NULL, it means we are off
 *   the end and should only calculate the other transition)
 *
 */

 
 void TransBetweenAndDot(TRANSSTRUCT *tsleft, TRANSSTRUCT *tsright, 
								int pos_in_scan, TRDOTSTRUCT *tdot, double *tmpdot,
								double *tmp3, double *tmp4)

/* Note: for debugging, if tmp3, tmp4 non-zero, returns the full dim
 * representation of left and right transitions in them.  In normal use
 * tmp3 and tmp4 should be 0 
 *
 * If tmpdot non-zero, return the dot product of what would be in 
 * tmp3 and tmp4 in it.  This is used for haplotyping between markers.
 *    
 */



{
  int  mask, mask1, mask2, im;
  double  myx0, froml_vct = 0.0, fromr_vct = 0.0;

  int i,j;
  int ifs, hd;

  int r_size_left, p_addr_left, p_varbl_left;
  int p_fixed_left, hd_max_left, p_fix_bitc_left;
  int r_size_right, p_addr_right, p_varbl_right;
  int p_fixed_right, hd_max_right, p_fix_bitc_right;

  double score_npl_val;
  double p, plod;
  double *this_par_score = NULL;
  if(algebra)
  this_par_score = datfile_par_score[algebra_pointer[famnum]];


  if (tsleft != NULL) {
	 hd_max_left         = rn - tsleft->p_bitc + tsleft->nffs;
	 p_fix_bitc_left = rn - tsleft->p_bitc;
	 r_size_left         = m_pow2[tsleft->nffs];
  }

  if (tsright != NULL) {
	 hd_max_right         = rn - tsright->p_bitc + tsright->nffs;
	 p_fix_bitc_right = rn - tsright->p_bitc;
	 r_size_right         = m_pow2[tsright->nffs];
  }

  tdot->like = tdot->lodlike = 0.0; tdot->statval = 0.0; tdot->varsum = 0.0;

  tdot->maxlike=0.0;
  tdot->minlike=1.0;
  tdot->un_normalized_entropy = 0;



  for(i=0; i < N; i++) {
	 if (tsleft != NULL) {
		p_addr_left = tsleft->p_address0;
		for(j=0; j<tsleft->nMaskShift; j++) {
		  if(tsleft->maskShift[j][1] >= 0) {
			 p_addr_left |= (i & tsleft->maskShift[j][0]) 
				<<  tsleft->maskShift[j][1];
		  } else {
			 p_addr_left |= (i & tsleft->maskShift[j][0]) 
				>> - tsleft->maskShift[j][1];
		  }
		}
		p_varbl_left = p_addr_left & tsleft->p_mask_v_bits;
		p_fixed_left = p_addr_left >> tsleft->p_bitc;
	 
		froml_vct = 0.0;

		for(ifs=0; ifs<r_size_left; ifs++) {
		  hd = mask = 0;
		  for(im=0; im<tsleft->nffs; im++) { 
			 if(ifs & m_pow2[im]) {
				hd++;
				mask |= tsleft->p_f_flip_mask[im]; 
			 }
		  }
		  mask1 = mask & tsleft->p_mask_v_bits;
		  myx0    = tsleft->tempvct[p_varbl_left ^ mask1];
		  mask2 = (mask >> tsleft->p_bitc) ^ tsleft->p_legal; 
		  /* contr. vector fixed bits */
		  mask2 = mask2 ^ p_fixed_left;    /* fixed bits distance to p_address */
		  for(j=0; j<=mod8[p_fix_bitc_left]; j++)
			 hd += lbtc8[ (mask2 >> (8*j)) & lbtc8mask ];
		  froml_vct += 
			 myx0 * tsleft->l_theta[hd] * tsleft->l_ctheta[hd_max_left - hd];
		}
	 }   /* if tsleft not NULL */ 

  if (tsright != NULL) {
		p_addr_right = tsright->p_address0;
		for(j=0; j<tsright->nMaskShift; j++) {
		  if(tsright->maskShift[j][1] >= 0) {
			 p_addr_right |= (i & tsright->maskShift[j][0]) 
				<<  tsright->maskShift[j][1];
		  } else {
			 p_addr_right |= (i & tsright->maskShift[j][0]) 
				>> - tsright->maskShift[j][1];
		  }
		}
		p_varbl_right = p_addr_right & tsright->p_mask_v_bits;
		p_fixed_right = p_addr_right >> tsright->p_bitc;
	 
		fromr_vct = 0.0;
		for(ifs=0; ifs<r_size_right; ifs++) {
		  hd = mask = 0;
		  for(im=0; im<tsright->nffs; im++) { 
			 if(ifs & m_pow2[im]) {
				hd++;
				mask |= tsright->p_f_flip_mask[im]; 
			 }
		  }
		  mask1 = mask & tsright->p_mask_v_bits;
		  myx0    = tsright->tempvct[p_varbl_right ^ mask1];
		  mask2 = (mask >> tsright->p_bitc) ^ tsright->p_legal; 
		  /* contr. vector fixed bits */
		  mask2 = mask2 ^ p_fixed_right;   /* fixed bits distance to p_address */
		  for(j=0; j<=mod8[p_fix_bitc_right]; j++)
			 hd += lbtc8[ (mask2 >> (8*j)) & lbtc8mask ];
		  fromr_vct += 
			 myx0 * tsright->l_theta[hd] * tsright->l_ctheta[hd_max_right - hd];
		}

  } /* if tsright not NULL */

  if ((tmp3 != (double *)0) && (tsleft != NULL))
		tmp3[i] = froml_vct;
  if ((tmp4 != (double *)0) && (tsright != NULL))
		tmp4[i] = fromr_vct;

	 /*Now that we have found the i'th bit in the full representation for
		both transitions from right and left, dot them here and calc Lod, etc
		so never have to store full vectors */

	 score_npl_val = Get_apm_score(i);
	 if (tsleft == NULL)
		p = fromr_vct;
	 else if (tsright == NULL)
		p = froml_vct;
	 else
		p = (froml_vct * fromr_vct);

/* Need this if haplotyping between markers, but it slows things down alot */
	 if (tmpdot != (double *)0)
		tmpdot[i]=p;


	 if(analysis_type != NPL_ANALYSIS)
	 {
		 if(algebra) 
			 plod = p * this_par_score[i];
		 else
			 plod = p * num_par_score[i];
	 } else {   
		 plod = p;
	 }

	 if (really_modcalc) this_p_pos[i] = p;
	 if (p > tdot->maxlike) tdot->maxlike=p;
	 if (p < tdot->minlike && p > 0.0) tdot->minlike = p;
	 tdot->lodlike += plod;
	 tdot->like += p;
	 tdot->statval += (p*score_npl_val);
	 tdot->varsum  += (p*score_npl_val*score_npl_val);

/* SUM(-(p[i]/tdot->like)*log(p[i]/tdot->like) = 
		 (1/tdot->like)*SUM(-p[i]*log(p[i])) + log(tdot->like)  
*/

	 if(p > 0) tdot->un_normalized_entropy += (-1.0) * p * log(p);
	 if(this_mod_pos==-1 && compute_sharing) 
		find_sharing_probs(pos_in_scan, n, i, p); /*** new - FIX 6/99 ***/
  }
}

 
 

void TransBetweenAndDot_sp(TRANSSTRUCT *tsleft, TRANSSTRUCT *tsright, 
								int pos_in_scan, TRDOTSTRUCT *tdot, double *tmpdot,
								double *tmp3, double *tmp4)

/* Note: for debugging, if tmp3, tmp4 non-zero, returns the full dim
 * representation of left and right transitions in them.  In normal use
 * tmp3 and tmp4 should be 0 
 *
 * If tmpdot non-zero, return the dot product of what would be in 
 * tmp3 and tmp4 in it.  This is used for haplotyping between markers.
 *    
 */
{
  int  mask, mask1, mask2, im;
  double  myx0, froml_vct = 0.0, fromr_vct = 0.0;

  int i,j;
  int ifs, hd, hd_m , hd_f, hd_max_left_m, hd_max_left_f, hd_max_right_m, hd_max_right_f;

  int r_size_left, p_addr_left, p_varbl_left;
  int p_fixed_left, p_fix_bitc_left;
  int r_size_right, p_addr_right, p_varbl_right;
  int p_fixed_right, p_fix_bitc_right;

  double score_npl_val;
  double p, plod;
  double *this_par_score = NULL;
  if(algebra)
  this_par_score = datfile_par_score[algebra_pointer[famnum]];


  if (tsleft != NULL) {
	 hd_max_left_m       = rn_m - tsleft->p_bitc_m + tsleft->nffs_m;
	 hd_max_left_f       = rn_f - tsleft->p_bitc_f + tsleft->nffs_f;
	 p_fix_bitc_left = rn - tsleft->p_bitc;
	 r_size_left         = m_pow2[tsleft->nffs];
  }

  if (tsright != NULL) {
	 hd_max_right_m         = rn_m - tsright->p_bitc_m + tsright->nffs_m;
	 hd_max_right_f         = rn_f - tsright->p_bitc_f + tsright->nffs_f;
	 p_fix_bitc_right = rn - tsright->p_bitc;
	 r_size_right         = m_pow2[tsright->nffs];
  }

  tdot->like = tdot->lodlike = 0.0; tdot->statval = 0.0; tdot->varsum = 0.0;
  
  tdot->maxlike=0.0;
  tdot->minlike=1.0;
  tdot->un_normalized_entropy = 0;



  for(i=0; i < N; i++) {
	 if (tsleft != NULL) {
		p_addr_left = tsleft->p_address0;
		for(j=0; j<tsleft->nMaskShift; j++) {
		  if(tsleft->maskShift[j][1] >= 0) {
			 p_addr_left |= (i & tsleft->maskShift[j][0]) 
				<<  tsleft->maskShift[j][1];
		  } else {
			 p_addr_left |= (i & tsleft->maskShift[j][0]) 
				>> - tsleft->maskShift[j][1];
		  }
		}
		p_varbl_left = p_addr_left & tsleft->p_mask_v_bits;
		p_fixed_left = p_addr_left >> tsleft->p_bitc;
	 
		froml_vct = 0.0;

		for(ifs=0; ifs<r_size_left; ifs++) {
		  hd = mask = 0;
		  hd_m = 0;
		  hd_f = 0;
		  for(im=0; im<tsleft->nffs; im++) { 
			 if(ifs & m_pow2[im]) {
				hd++;
				hd_m += tsleft->nfbit_sex[im];
				hd_f += (1-tsleft->nfbit_sex[im]); 
				mask |= tsleft->p_f_flip_mask[im]; 
			 }
		  }
		  mask1 = mask & tsleft->p_mask_v_bits;
		  myx0    = tsleft->tempvct[p_varbl_left ^ mask1];
		  mask2 = (mask >> tsleft->p_bitc) ^ tsleft->p_legal; 
		  /* contr. vector fixed bits */
		  mask2 = mask2 ^ p_fixed_left;    /* fixed bits distance to p_address */
		  
		  hd_m += m_wgt[mask2<<tsleft->p_bitc];
		  hd_f += f_wgt[mask2<<tsleft->p_bitc];
		  
		  froml_vct  += myx0 * tsleft->l_theta_m[hd_m] * tsleft->l_ctheta_m[hd_max_left_m - hd_m]*
			               tsleft->l_theta_f[hd_f] * tsleft->l_ctheta_f[hd_max_left_f - hd_f];
		  j = mod8[p_fix_bitc_left] + 1;	       
		}
	 }   /* if tsleft not NULL */ 

	  if (tsright != NULL) {
		p_addr_right = tsright->p_address0;
		for(j=0; j<tsright->nMaskShift; j++) {
		  if(tsright->maskShift[j][1] >= 0) {
			 p_addr_right |= (i & tsright->maskShift[j][0]) 
				<<  tsright->maskShift[j][1];
		  } else {
			 p_addr_right |= (i & tsright->maskShift[j][0]) 
				>> - tsright->maskShift[j][1];
		  }
		}
		p_varbl_right = p_addr_right & tsright->p_mask_v_bits;
		p_fixed_right = p_addr_right >> tsright->p_bitc;
		
		fromr_vct = 0.0;
		for(ifs=0; ifs<r_size_right; ifs++) {
		  hd = mask = 0;
		  hd_m = 0;
		  hd_f = 0;
		  for(im=0; im<tsright->nffs; im++) { 
			 if(ifs & m_pow2[im]) {
				hd++;
				hd_m += tsright->nfbit_sex[im];
				hd_f += (1-tsright->nfbit_sex[im]); 
				mask |= tsright->p_f_flip_mask[im]; 
			 }
		  }
		  mask1 = mask & tsright->p_mask_v_bits;
		  myx0    = tsright->tempvct[p_varbl_right ^ mask1];
		  mask2 = (mask >> tsright->p_bitc) ^ tsright->p_legal; 
		  /* contr. vector fixed bits */
		  mask2 = mask2 ^ p_fixed_right;    /* fixed bits distance to p_address */
		  
		  hd_m += m_wgt_s[mask2<<tsright->p_bitc];
		  hd_f += f_wgt_s[mask2<<tsright->p_bitc];
		  
		  fromr_vct  += myx0 * tsright->l_theta_m[hd_m] * tsright->l_ctheta_m[hd_max_right_m - hd_m]*
			               tsright->l_theta_f[hd_f] * tsright->l_ctheta_f[hd_max_right_f - hd_f];
				       
		  j = mod8[p_fix_bitc_right] + 1;		       
		}
  	} /* if tsright not NULL */

  	if ((tmp3 != (double *)0) && (tsleft != NULL))
		tmp3[i] = froml_vct;
	if ((tmp4 != (double *)0) && (tsright != NULL))
		tmp4[i] = fromr_vct;

	 /*Now that we have found the i'th bit in the full representation for
		both transitions from right and left, dot them here and calc Lod, etc
		so never have to store full vectors */

	 score_npl_val = Get_apm_score(i);
	 if (tsleft == NULL)
		p = fromr_vct;
	 else if (tsright == NULL)
		p = froml_vct;
	 else
		p = (froml_vct * fromr_vct);

/* Need this if haplotyping between markers, but it slows things down alot */
	 if (tmpdot != (double *)0)
		tmpdot[i]=p;


	 if(analysis_type != NPL_ANALYSIS)
	 {
		 if(algebra) 
			 plod = p * this_par_score[i];
		 else
			 plod = p * num_par_score[i];
	 } else {    
		 plod = p;
	 }

	 if (really_modcalc) this_p_pos[i] = p;	
	 if (p > tdot->maxlike) tdot->maxlike=p;
	 if (p < tdot->minlike && p > 0.0) tdot->minlike = p;
	 tdot->lodlike += plod;
	 tdot->like += p;
	 tdot->statval += (p*score_npl_val);
	 tdot->varsum  += (p*score_npl_val*score_npl_val);
	 	 
/* SUM(-(p[i]/tdot->like)*log(p[i]/tdot->like) = 
		 (1/tdot->like)*SUM(-p[i]*log(p[i])) + log(tdot->like)  
*/
	 if(p > 0) tdot->un_normalized_entropy += (-1.0) * p * log(p);
	 if(this_mod_pos==-1 && compute_sharing) 
		find_sharing_probs(pos_in_scan, n, i, p); /*** new - FIX 6/99 ***/
  }
}

void TransFree(TRANSSTRUCT *ts)
{
   unmatrix(ts->maskShift, 64, int);
   unarray(ts->tempvct, double);
   if (ind_sp==0){
     unarray(ts->bit_cnt, int);
   }else if (ind_sp==1){  
     unarray(ts->bit_cnt_m, int);
     unarray(ts->bit_cnt_f, int);
   }
}


void s_wgt_free(int* m_wgt_act, int* f_wgt_act){
   unarray(m_wgt_act, int);
   unarray(f_wgt_act, int); 
}


void convolve_fft_2(int nbits, double theta, int *bit_cnt, double *pv)
{
	int i, Nstates;
	double trans[100], factor;
	void fft_2(int, double *);

	Nstates = m_pow2[nbits];
	trans[0] = 1.0;
	factor = 1.0 - 2.0*theta;
	for (i=1; i<100; ++i)
		trans[i] = trans[i-1]*factor;

	fft_2(nbits, pv);
	for(i=0; i<Nstates; i++)
		pv[i] *= trans[bit_cnt[i]];
	fft_2(nbits, pv);
	for(i=0; i<Nstates; i++) {
	    pv[i] /= Nstates;
	    if (pv[i] < 0.0) pv[i] = -pv[i];
	}
}


void convolve_fft_2_sp(int nbits, double theta_m, double theta_f, int *bit_cnt_m, int *bit_cnt_f, double *pv)
{
	int i, Nstates;
	double trans_m[100], trans_f[100], factor_m, factor_f;
	void fft_2(int, double *);

        Nstates = m_pow2[nbits];
	trans_m[0] = 1.0;
	trans_f[0] = 1.0;
	factor_m = 1.0 - 2.0*theta_m;
	factor_f = 1.0 - 2.0*theta_f;
	for (i=1; i<100; ++i){
		trans_m[i] = trans_m[i-1]*factor_m;
	        trans_f[i] = trans_f[i-1]*factor_f;
        }

	fft_2(nbits, pv);
	for(i=0; i<Nstates; i++)
		pv[i] *= trans_m[bit_cnt_m[i]]*trans_f[bit_cnt_f[i]];
	fft_2(nbits, pv);
	for(i=0; i<Nstates; i++) {
	    pv[i] /= Nstates;
	    if (pv[i] < 0.0) pv[i] = -pv[i];
	}
}


void fft_2(int nbits, double *pv)
{
	int i,j,ej,flipi,Nstates;
	double temp;

	Nstates = m_pow2[nbits];
	for (j=0; j<nbits; j++) {
		ej = m_pow2[j];
		for (i=0; i<Nstates; i++) {
			if ((i & ej) == 0) {
				flipi = i ^ ej;
				temp = pv[i];
				pv[i] = temp + pv[flipi];
				pv[flipi] = temp - pv[flipi];
			}
		}
	}
}

void Setup_MaskShift(int nbits, int p_msk, int n_msk, 
                     int n_fixed, TRANSSTRUCT *ts)

{
/* -----------------------------------------------
Prepare the mask-shift operations to transform the n_representation
inheritance vector to a p_representation inh. vector.
ts                  Transition Structure (used in TransSetupConvolve_sp(), etc)
n_bits              # of bits in founder symmetry (fs) representation.
p_msk               mask of fixed bits in previous locus (fs rep.)
n_msk               mask of fixed bits in next locus (fs rep.)
n_fixed             value of fixed bits in next locus (n_representation)
ts->p_address0      value of fixed bits in next locus transformed to
                    p_representation.
ts->nMaskShift          # of mas-shift operations needed.
ts->maskShift[iop][0]   mask  for iop operation
ts->maskShift[iop][1]   shift for iop operation
k_org			contains the k_o  bit position in fs representation (original) of
                        k_p  bit position in p_representation 
   ----------------------------------------------- */
	int i, j, k_n, k_o, k_p, p_nvb, n_nvb, t_j, test, nop;
	int nop_max, sbit_flag[64];
	
	p_nvb = n_nvb = 0;
	for(i=0; i<nbits; i++) {
		if((m_pow2[i] & p_msk) == 0) p_nvb++;
		if((m_pow2[i] & n_msk) == 0) n_nvb++;
	}
	/*	printf("\t\t p_msk, n_msk = %8x  %8x \n", p_msk, n_msk);
		printf("\t\t p_nvb, n_nvb = %8d  %8d \n", p_nvb, n_nvb);
	*/

	/* define nbits mask-shift operations that map 
	   n_representation to p_representation 
	   k_n  bit position in n_representation
	   k_o  bit position in fs representation (original)
	   k_p  bit position in p_representation  
	*/	
	for(k_n=0; k_n<nbits; k_n++) {
		j=0;
		t_j  = ((k_n<n_nvb) ? k_n:(k_n-n_nvb));
		test = ((k_n<n_nvb) ? ~n_msk : n_msk );
		for(k_o=0; k_o<nbits; k_o++) {
			if(m_pow2[k_o] & test) {
				if(j==t_j) break;
				j++;
			}
		}
		
		k_p  = ((m_pow2[k_o] & p_msk) ? p_nvb:0);
		test = ((m_pow2[k_o] & p_msk) ? p_msk : ~p_msk );
		for(i=0; i<k_o; i++) {
			if(m_pow2[i] & test) k_p++;  
		}
		
		/*k_o bit position in fs representation (original) as a
		  function of k_p bit position in p_representation
		*/
		k_org[k_p] = k_o;
		
		/*printf("\t\t k_p, k_o, k_n= %2d  %2d  %2d\n", k_p, k_o, k_n); */
		ts->maskShift[k_n][0] = m_pow2[k_n];
		ts->maskShift[k_n][1] = k_p - k_n;
	}
	
	/* define p_address0 and keep transformations of variable bits only	*/	
	nop = ts->p_address0 = 0;
	for(i=0; i<nbits; i++) {
		if(i<n_nvb) {
			ts->maskShift[nop][0] = ts->maskShift[i][0];
			ts->maskShift[nop][1] = ts->maskShift[i][1];
			nop++;
		} else {
			if(ts->maskShift[i][1] >= 0) {
				ts->p_address0 |= (n_fixed & ts->maskShift[i][0]) 
				                <<  ts->maskShift[i][1];
			} else {
				ts->p_address0 |= (n_fixed & ts->maskShift[i][0]) 
				                >> - ts->maskShift[i][1];
			}
		}
	}
	
	/* group together variable bits that share the same shift value */
	nop_max = nop;
	nop     = 0;
	for(i=0; i<nop_max; i++) sbit_flag[i] = 0;
	for(i=0; i<nop_max; i++) {
		if(sbit_flag[i]) continue;
		sbit_flag[i] = 1;
		ts->maskShift[nop][0] = ts->maskShift[i][0];
		ts->maskShift[nop][1] = ts->maskShift[i][1];
		for(j=i+1; j<nop_max; j++) {
			if(ts->maskShift[nop][1] == ts->maskShift[j][1]) {
				sbit_flag[j] = 1;
				ts->maskShift[nop][0] |= ts->maskShift[j][0];
			}
		}
		nop++;
	}
	
	ts->nMaskShift = nop;
}



void Setup_flip_masks(int p_mask, int *nvfs, int p_v_flip_mask[],
                               int *nffs, int p_f_flip_mask[])
{
	int p_bitc, nv, nf, i, j, k1, k2, k3, k4, mask, test, p_flip_mask;
	int flip_m_v_bits, flip_m_t_bits;
	
/*	j=0;
	for(i=0; i<n; i++) { if((uninf_mask & m_pow2[i]) != 0) j++; }
	*tfs = j;
*/	
	p_bitc = 0;
	for(i=0; i<rn; i++) { if((p_mask & m_pow2[i]) == 0) p_bitc++; }

	nv = nf = 0;
	for (i=0; i<n; i++) {
		if ((uninf_mask & m_pow2[i]) == 0) continue;
		p_flip_mask=0;
		mask = flip_mask[i] ^ m_pow2[i]; /* all but uninformative bit */
		flip_m_v_bits = flip_m_t_bits = 0;
		for(j=0; j<n; j++) {
			if((mask & m_pow2[j]) == 0) continue;
			flip_m_t_bits++;
			test = m_pow2[j];
			/* k1   full inh. vector rep.
			   k2   founder symmetry reduced inh. vector rep.
			   k3   genotyping reduced inh. vector rep. position in variable section
			   k4   genotyping reduced inh. vector rep. position in  fixed   section
			*/
			k2 = k3 = k4 = 0;
			for(k1=0; k1<n; k1++) {
				if((uninf_mask & m_pow2[k1]) != 0) continue;
				if(test & m_pow2[k1]) break;
				if((p_mask & m_pow2[k2]) == 0) { k3++; } else { k4++; } 
				k2++;
			}
			/* if flip mask and not_fixed_bit set flip mask in 
			   genotype reduced representation */
			if(k1<n) {
				if((p_mask & m_pow2[k2]) == 0) {      /* variable bit */
					flip_m_v_bits++;
					p_flip_mask |= m_pow2[k3];
				} else {                            /*   fixed  bit */
					p_flip_mask |= m_pow2[p_bitc + k4];
				}
			}
		}
		if(p_flip_mask != 0) {   /* first check if ref. phase has no sibs */
			if(flip_m_v_bits == flip_m_t_bits) {
				p_v_flip_mask[nv] = p_flip_mask; /* full mask in variable bit set*/
				nv++;
			} else {
				p_f_flip_mask[nf] = p_flip_mask; /* mixed or all in fixed bit set */
				nf++;
			}
		}
	}
	*nvfs=nv;
	*nffs=nf;
}



void Setup_flip_masks_sp(int p_mask, int *nvfs, int p_v_flip_mask[],
                      int *nffs, int* nffs_m, int* nffs_f,
                      int *nfbit_sex, int p_f_flip_mask[])
{
	int p_bitc, nv, nf, i, j, k1, k2, k3, k4, mask, test, p_flip_mask,
            nf_m, nf_f;
	int flip_m_v_bits, flip_m_t_bits;
	
/*	j=0;
	for(i=0; i<n; i++) { if((uninf_mask & pow2[i]) != 0) j++; }
	*tfs = j;
*/	
	p_bitc = 0;
	for(i=0; i<rn; i++) { if((p_mask & m_pow2[i]) == 0) p_bitc++; }

	nv = nf = 0;
        nf_m = nf_f = 0;
	for (i=0; i<33; i++) nfbit_sex[i]=0;
	for (i=0; i<n; i++) {
		if ((uninf_mask & m_pow2[i]) == 0) continue;
		p_flip_mask=0;
		mask = flip_mask[i] ^ m_pow2[i]; /* all but uninformative bit */
		flip_m_v_bits = flip_m_t_bits = 0;
		for(j=0; j<n; j++) {
			if((mask & m_pow2[j]) == 0) continue;
			flip_m_t_bits++;
			test = m_pow2[j];
			/* k1   full inh. vector rep.
			   k2   founder symmetry reduced inh. vector rep.
			   k3   genotyping reduced inh. vector rep. position in variable section
			   k4   genotyping reduced inh. vector rep. position in  fixed   section
			*/
			k2 = k3 = k4 = 0;
			for(k1=0; k1<n; k1++) {
				if((uninf_mask & m_pow2[k1]) != 0) continue;
				if(test & m_pow2[k1]) break;
				if((p_mask & m_pow2[k2]) == 0) { k3++; } else { k4++; } 
				k2++;
			}
			/* if flip mask and not_fixed_bit set flip mask in 
			   genotype reduced representation */
			if(k1<n) {
				if((p_mask & m_pow2[k2]) == 0) {      /* variable bit */
					flip_m_v_bits++;
					p_flip_mask |= m_pow2[k3];
				} else {                            /*   fixed  bit */
					p_flip_mask |= m_pow2[p_bitc + k4];
				}
			}
		}
		if(p_flip_mask != 0) {   /* first check if ref. phase has no sibs */
			if(flip_m_v_bits == flip_m_t_bits) {
				p_v_flip_mask[nv] = p_flip_mask; /* full mask in variable bit set*/
				nv++;
			} else {
				p_f_flip_mask[nf] = p_flip_mask; /* mixed or all in fixed bit set */
				/* an der Stelle feld anlegen das sex zu i zu nf angibt, das in trans_and_map v */
				/* verwenden: im-bit */
                                if (i%2 == 1){
                                  nf_m++;
				  nfbit_sex[nf]=1;
                                }
	                        else {
                                  nf_f++;
				  nfbit_sex[nf]=0;
                                }
				nf++;
			}
		}
	}
	*nvfs=nv;
	*nffs=nf;
        *nffs_m=nf_m;
        *nffs_f=nf_f;
}

/* Katz, Thu Jun 22 14:26:31 2000
 * If n_marker = -1, are going between markers, so n_legal is 0
 * (no fixed bits in destination)
 */

void Setup_legal_vectors(int p_bitc, int *p_mask_v_bits, 
                         int p_marker, int *p_legal, 
						 int n_marker, int *n_legal)
{
	int i, j;

	*p_mask_v_bits = 0;
	for(i=0; i<p_bitc; i++)    *p_mask_v_bits |= m_pow2[i];

	*p_legal = 0;
	j       = 0;
	for(i=0; i<rn; i++) {
		if(m_pow2[i] & maskTest[p_marker][0]) {
			if(m_pow2[i] & maskTest[p_marker][1]) *p_legal |= m_pow2[j];
			j++;
		}
	}
	*n_legal = 0;

	if (n_marker < 0)  /* Going between markers, *n_legal=0, no fixed bits */
	  return;

	j       = 0;
	for(i=0; i<rn; i++) {
		if(m_pow2[i] & maskTest[n_marker][0]) {
			if(m_pow2[i] & maskTest[n_marker][1]) *n_legal |= m_pow2[j];
			j++;
		}
	}
}

void Check_results_01(int l, unsigned long n_pntf, int test0, int test1,
                      int reduced_bitc,       int full_bitc,
                   double reduced_array[], double full_array[])
{
	unsigned long dump_limit, error_limit, n_ptr, nerrors;
	int i, j, f_size;
	double x1, x2, x3;
	
	dump_limit  = 0;
	error_limit = 8;
	
	f_size = m_pow2[full_bitc];
	
	x3 = 0;
	for(unsigned long h=0; h<m_pow2[reduced_bitc]; h++) x3 += reduced_array[h];
	/* printf("  sub x3=%12.4e \n", x3); */

	n_ptr = nerrors = 0;
	for(i=0; i<f_size; i++) {
		if((i & test0) == test1) {
			x1 = reduced_array[n_ptr];
			x2 = full_array[i];
			j=!real_eq_10( (x1-x2)/x3, 0.);
			if(j) nerrors++;
			if( (j && (nerrors<error_limit)) || (n_ptr<dump_limit) ) {
				if(j) { printf("\t\t*"); }
				else  { printf("\t\t "); }
				printf(" l, n_ptr, new, old r, %2d %6lu %6lu %12.4e %12.4e",
				l, n_ptr+n_pntf, n_ptr, reduced_array[n_ptr], full_array[i]);
				if(j) {
					if(x1 != 0) {
						printf(" %16.8e \n", x2/x1);
					} else {
						printf(" %16.8e inverse\n", x1/x2);
					}
				} else {
					printf("   1.\n");
				}
			}
			n_ptr++;
		}
	}
	if(nerrors >= error_limit) {
		printf("\t\t ............. nerrors=%10lu \n", nerrors);
	} else if(nerrors != 0) {
		printf("\t\t               nerrors=%10lu \n", nerrors);
	}
}

double Get_apm_score(int new_vec)
{
	static int npl_pnt, vec_previous;

	if(new_vec != 0) {
		if( (vec_previous & score_mask_rdc) != 
	    	(new_vec & score_mask_rdc) )       npl_pnt++;
	} else {
		npl_pnt = vec_previous = 0;
	}
	vec_previous = new_vec;
	return(apm_score[npl_pnt]);
}


void get_sex_weights(int i, int size, int* f_wgt, int* m_wgt){
    int lm_wgt=0, lf_wgt=0, j;

    for (j=0; j<size; j++) {
       if (m_pow2[j]&i){       
         if (k_full[k_org[j]]%2 == 1){
	    lm_wgt++;
	  }else{
	    lf_wgt++;
	  }   
       }             
   } 
   *m_wgt = lm_wgt;                
   *f_wgt = lf_wgt;                
}

void init_sex_weights(int mod, int* m_wgt_act, int* f_wgt_act){
    int lm_wgt, lf_wgt, i, j, kf;
    
    if (mod == 1){
      
      for (j=0; j<N; j++) {
           lm_wgt=0; 
           lf_wgt=0;
           for (i=0; i<rn; i++){
              if (m_pow2[i]&j){       
 	        kf=k_full[k_org[i]]%2;
	        lm_wgt+=kf;
	        lf_wgt+=1-kf;
	      }             
           } 
           m_wgt_act[j] = lm_wgt;                
           f_wgt_act[j] = lf_wgt;
      }
	      
     }else{
      
      for (j=0; j<N; j++) {
        lm_wgt=0; 
        lf_wgt=0;
        for (i=0; i<rn; i++){
           if (m_pow2[i]&j){       
             /*if (k_full[i]%2 == 1){
	       lm_wgt++;
	     }else{
	       lf_wgt++;
	     }*/
	     kf=k_full[i]%2;
	     lm_wgt+=kf;
	     lf_wgt+=1-kf;   
           }             
        } 
        m_wgt_act[j] = lm_wgt;                
        f_wgt_act[j] = lf_wgt;
      
      }
     }  
}

